/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using System.Threading;
using gov.va.med.imaging.exchange.VixInstaller.business;

namespace gov.va.med.imaging.exchange.VixInstaller.ui
{
    public partial class PrerequisiteChecklistPage : InteriorWizardPage
    {
        private System.Diagnostics.Process externalProcess = null;
        private VixCredentialsDialog tomcatDialog = new VixCredentialsDialog();
        private LaurelBridgeInfoDialog laurelBridgeDialog = new LaurelBridgeInfoDialog();
        private FederationInfoDialog federationInfoDialog = new FederationInfoDialog();
        private ServiceAccountDialog serviceAccountDialog = new ServiceAccountDialog();
        private bool reapplyServiceAccount = (TomcatFacade.IsActiveTomcatVersionInstalled() == false) &&
            WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName);
        //private bool dcfReinstallComplete = false;

        public PrerequisiteChecklistPage()
        {
            InitializeComponent();
        }

        public PrerequisiteChecklistPage(IWizardForm wizForm, int pageIndex)
            : base(wizForm, pageIndex)
        {
            InitializeComponent();
            if (this.reapplyServiceAccount)
            {
                this.buttonTomcatServiceAccount.Text = "Assign";
            }
        }

        //protected override String Info(String infoMessage)
        //{
        //    this.InteriorPageSubHeader = infoMessage;
        //    return infoMessage;
        //}

        #region IWizardPage Members
        public override void Initialize()
        {
            this.InitializeBusinessFacadeDelegates();
            //BusinessFacade.InfoDelegate = null; // we dont want this output in the subheader
            //LaurelBridgeFacade.InfoDelegate = null; // we dont want this output in the subheader
            // restore the page header
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();

            if (config.VixRole == VixRoleType.DicomGateway)
            {
                this.InteriorPageHeader = "Install the HDIG Prerequisites.";
            }
            else
            {
                this.InteriorPageHeader = "Install the VIX Prerequisites.";
            }

            this.SetAdministratorPrerequisiteState();
            this.SetOSPrerequisiteState();
            this.SetJavaPrerequisiteState();
            this.SetTomcatPrerequisiteState();
            this.SetLaurelBridgePrerequisiteState();
            this.SetTomcatServiceAccountPrerequisiteState();
            this.SetVixCertificatePrerequisiteState();
            // set the sub header for the page
            this.SetPageSubHeader();

            // set wizard form button state
            if (this.Visible)
            {
                if (this.IsComplete())
                {
                    this.WizardForm.EnableNextButton(true);
                }
                else
                {
                    this.WizardForm.EnableNextButton(false);
                }
                this.WizardForm.EnableBackButton(true);
                this.WizardForm.EnableCancelButton(true);
            }
        }

        public override bool IsComplete()
        { 
            bool isComplete = false;
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            bool devMode = this.WizardForm.IsDeveloperMode();
            if (BusinessFacade.IsLoggedInUserAnAdministrator() && BusinessFacade.IsOperatingSystemApproved(config, devMode) &&
                JavaFacade.IsJavaInstalled(devMode) && TomcatFacade.IsActiveTomcatVersionInstalled() &&
                (LaurelBridgeFacade.IsLaurelBridgeRequired(config) == false || (LaurelBridgeFacade.IsDeprecatedLaurelBridgeInstalled() == false && LaurelBridgeFacade.IsLaurelBridgeLicensed())) && 
                WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) &&
                this.reapplyServiceAccount == false && VixFacade.IsFederationCertificateInstalledOrNotRequired(config))
            {
                isComplete = true;
            }
            return isComplete;
        }

        #endregion

        #region events

        private void buttonJava_Click(object sender, EventArgs e)
        {
            this.buttonJava.Enabled = false;
            this.InstallJava();
        }

        /// <summary>
        /// Get the tomcat administrator credentials from the user and then install tomcat.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void buttonTomcat_Click(object sender, EventArgs e)
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            if (config.TomcatAdminPassword == null)
            {
                if (this.tomcatDialog.ShowDialog() != DialogResult.OK) // ask for password associated with the admin account
                {
                    return;
                }
                config.TomcatAdminPassword = this.tomcatDialog.Password;
            }
            this.buttonTomcat.Enabled = false;
            Application.DoEvents();
            this.InstallTomcat();
        }

        private void buttonLaurelBridge_Click(object sender, EventArgs e)
        {
            //TODO: need to consider license type here. If a previous installation was removed (renamed) and the 
            // old license can be used without prompting then do so. Directory will need to be renamed to a known value
            // with the version number in it. This functionality is not needed for the initial move to 3.3.22c.

            string licensePath = null;
            bool cancel = false;
            if (this.WizardForm.IsDeveloperMode())
            {
                DialogResult result = MessageBox.Show("Are you installing a MAC based license? A MAC based license is usually a temporary license or a DCF SDK license.",
                    "Developer Mode: Specify your DCF license type", MessageBoxButtons.YesNo);
                if (result == DialogResult.Yes)
                {
                    if (this.laurelBridgeDialog.ShowDialog() == DialogResult.OK)
                    {
                        licensePath = this.laurelBridgeDialog.LicensePath;
                    }
                    else
                    {
                        cancel = true;
                    }
                }
            }
            else if (DcfLicenseType.MAC == LaurelBridgeFacade.GetActiveDcfLicenseType()) // not used now but added for completeness
            {
                if (this.laurelBridgeDialog.ShowDialog() == DialogResult.OK)
                {
                    licensePath = this.laurelBridgeDialog.LicensePath;
                }
                else
                {
                    cancel = true;
                }
            }

            if (cancel == false)
            {
                this.buttonLaurelBridge.Enabled = false;
                Application.DoEvents();
                // if licensePath is null it will indicate that this is an enterprise license install
                this.InstallAndLicenseLaurelBridgeToolkit(licensePath, LaurelBridgeFacade.GetActiveDcfRootFromManifest());
            }
        }

        private void buttonTomcatServiceAccount_Click(object sender, EventArgs e)
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            bool invokeServiceDialog = false;

            if (config.ApacheTomcatPassword == null)
            {
                invokeServiceDialog = true;
            }
            else if (this.reapplyServiceAccount)
            {
                DialogResult result = MessageBox.Show("If your service account password has not changed then answer Yes, otherwise No.",
                    "Reapply previous service account password?", MessageBoxButtons.YesNo);
                if (result == DialogResult.No)
                {
                    invokeServiceDialog = true;
                }
            }
            else if (WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) == false)
            {
                invokeServiceDialog = true;
            }

            if (invokeServiceDialog && this.serviceAccountDialog.ShowDialog() == DialogResult.OK) // ask for password associated with the admin account
            {
                config.ApacheTomcatPassword = this.serviceAccountDialog.Password;
            }

            if (config.ApacheTomcatPassword != null)
            {
                this.buttonTomcatServiceAccount.Enabled = false;
                Application.DoEvents();
                this.ConfigureServiceAccount();
            }
        }

        private void buttonVixCertificate_Click(object sender, EventArgs e)
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            if (VixFacade.IsFederationCertificateInstalledOrNotRequired(config) == false)
            {
                DialogResult result = this.federationInfoDialog.ShowDialog();

                if ( result == DialogResult.OK) // Cancel will skip this
                {
                    this.buttonVixCertificate.Enabled = false;
                    Application.DoEvents();

                    string certStoreDir = VixFacade.GetVixCertificateStoreDir();
                    if (!Directory.Exists(certStoreDir))
                    {
                        Directory.CreateDirectory(certStoreDir);
                    }
                    ZipUtilities.UnZip(this.federationInfoDialog.CertificatePath, certStoreDir);
                }
                this.Initialize();
            }
        }

        #endregion

        #region install methods
        private void ConfigureServiceAccount()
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                String username = TomcatFacade.ServiceAccountUsername;
                String password = config.ApacheTomcatPassword;

                if (WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) == false)
                {
                    this.Logger().Info(this.Info("Creating service account " + username + "."));
                    if (WindowsUserUtilities.CreateServiceAccount(username, password, "VIX service account") == true)
                    {
                        Application.DoEvents();
                        this.Logger().Info("Setting service account " + username + " privileges.");
                        LsaUtilities.SetServiceAccountPrivleges(username);
                        Application.DoEvents();
                    }
                    else
                    {
                        DialogResult result = MessageBox.Show("The password that you supplied does not meet the complexity requirements of this computer. Please try again.",
                            "Service account creation failed!", MessageBoxButtons.OK);
                        config.ApacheTomcatPassword = null;
                    }
                }
                if (WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) == true)
                {
					this.Logger().Info("Setting " + TomcatFacade.TomcatServiceName + " service to run under " + username + " account.");
					ServiceUtilities.SetServiceCredentials(TomcatFacade.TomcatServiceName, username, password, ClusterFacade.IsServerClusterNode());
					if (this.reapplyServiceAccount)
					{
						this.reapplyServiceAccount = false;
					}
					Application.DoEvents();
				}
            }
            finally
            {
                Cursor.Current = Cursors.Default;
                this.Initialize();
            }
        }

        private void InstallAndLicenseLaurelBridgeToolkit(string macBasedLicenseFilespec, string dcfPath)
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                this.WizardForm.EnableBackButton(false);
                this.WizardForm.EnableCancelButton(false);
                this.WizardForm.EnableNextButton(false);

                // uninstall Laurel Bridge if necessary - safety net for HDIG - VIX/CVIX will do this on the uninstall page
                if (LaurelBridgeFacade.IsDeprecatedLaurelBridgeInstalled())
                {
                    this.Info("Removing existing Laurel Bridge DCF toolkit version " + LaurelBridgeFacade.GetDeprecatedLaurelBridgeVersion());
                    LaurelBridgeFacade.RemoveLaurelBridgeInstallation(config);
                    LaurelBridgeFacade.SaveRenamedDeprecatedDcfRoot(config.RenamedDeprecatedDcfRoot); // save only the deprecated root to disk so we dont lose it
                }

                if (LaurelBridgeFacade.IsLaurelBridgeInstalled() == false)
                {
                    string errorMessage = "";
                    if (LaurelBridgeFacade.InstallLaurelBridgeDcfToolkit(dcfPath, ref errorMessage) == false)
                    {
                        string message = @"The Laurel Bridge DCF toolkit could not be installed due to the following error: " + errorMessage;
                        MessageBox.Show(message, "Error installing DCF toolkit", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    }
                }

                if (LaurelBridgeFacade.IsLaurelBridgeInstalled() == true && LaurelBridgeFacade.IsLaurelBridgeLicensed() == false)
                {
                    if (macBasedLicenseFilespec != null) // MAC based licensing
                    {
                        string errorMessage = "";
                        bool isLicensed = LaurelBridgeFacade.LicenseLaurelBridgeDcfToolkitWithMacBasedKeyFile(macBasedLicenseFilespec, ref errorMessage);
                        if (isLicensed == false)
                        {
                            String message = "The provided Laurel Bridge License is not valid for this computer. The dcf_info utility returned the following:\n";
                            MessageBox.Show(message + errorMessage, "Invalid Laurel Bridge License", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }
                    else
                    {
                        // License toolkit with a VA enterprise license. The UI will provide feedback about errors.
                        string macAddress = null;
                        if (config.RenamedDeprecatedDcfRoot != null)
                        {
                            macAddress = LaurelBridgeFacade.GetMacAddressFromDeprecatedKeyFile(config.RenamedDeprecatedDcfRoot);
                        }
                        // last chance at getting the mac address
                        if (macAddress == null)
                        {
                            string deprecatedDcfRoot = LaurelBridgeFacade.GetDeprecatedRenamedRootDirspec(LaurelBridgeFacade.GetActiveDcfRootFromManifest());
                            macAddress = LaurelBridgeFacade.GetMacAddressFromDeprecatedKeyFile(deprecatedDcfRoot);
                        }

                        LaurelBridgeFacade.LicenseLaurelBridgeDcfToolkitWithEnterpriseLicense(macAddress);
                    }
                }
            }
            finally
            {
                Cursor.Current = Cursors.Default;
                this.Initialize();
            }
        }

        private void InstallTomcat()
        {
                Cursor.Current = Cursors.WaitCursor;
                this.Logger().Info("Installing Apache Tomcat version " + TomcatFacade.ActiveTomcatVersion);
                this.InteriorPageHeader = "Installing Apache Tomcat version " + TomcatFacade.ActiveTomcatVersion;
                this.InteriorPageSubHeader = "This usually takes under a minute to complete.";
                this.WizardForm.EnableBackButton(false);
                this.WizardForm.EnableCancelButton(false);
                this.WizardForm.EnableNextButton(false);
                Cursor.Current = Cursors.WaitCursor;

                String exe = TomcatFacade.InstallerFilespec;

                externalProcess = new System.Diagnostics.Process();
                // Handle the Exited event that the Process class fires.
                externalProcess.Exited += new System.EventHandler(this.TomcatInstallComplete);
                externalProcess.EnableRaisingEvents = true;
                externalProcess.SynchronizingObject = this;
                externalProcess.StartInfo.FileName = exe;
                externalProcess.StartInfo.Arguments = "/S";
                externalProcess.Start();
        }

        private void InstallJava()
        {
            bool devMode = this.WizardForm.IsDeveloperMode();

            if (JavaFacade.IsJavaInstalledInWrongDirectory(devMode)) // this can happen on HDIG installs when installing for the first time over a legacy gateway
            {
                this.Logger().Warn("Java Runtime Environment version " + JavaFacade.ActiveJavaVersion + " is already installed in " + JavaFacade.GetActiveJrePathLocationIndependent());
                this.Logger().Warn("Performing uninstall so that the JRE may be re-installed in " + JavaFacade.GetActiveJavaPath(true)); // true means JRE not JDK
                JavaFacade.UninstallActiveJre();
            }

            this.Logger().Info("Installing the Java Runtime Environment version " + JavaFacade.ActiveJavaVersion);
            this.Logger().Info("Installer is located at: " + JavaFacade.GetInstallerFilespec());
            this.InteriorPageHeader = "Installing the Java Runtime Environment version " + JavaFacade.ActiveJavaVersion;
            this.InteriorPageSubHeader = "This usually takes a couple of minutes to complete.";
            this.WizardForm.EnableBackButton(false);
            this.WizardForm.EnableCancelButton(false);
            this.WizardForm.EnableNextButton(false);
            Cursor.Current = Cursors.WaitCursor;

            String exe = JavaFacade.GetInstallerFilespec();

            externalProcess = new System.Diagnostics.Process();
            // Handle the Exited event that the Process class fires.
            externalProcess.Exited += new System.EventHandler(this.JavaInstallComplete);
            externalProcess.EnableRaisingEvents = true;
            externalProcess.SynchronizingObject = this;
            externalProcess.StartInfo.FileName = exe;
            externalProcess.StartInfo.Arguments = "/s /v \"/qn REBOOT=Suppress JAVAUPDATE=0 INSTALLDIR=\\\"" + JavaFacade.GetActiveJavaPath(true) + "\\\"";
            externalProcess.Start();
        }

        private void JavaInstallComplete(object sender, System.EventArgs e)
        {
            Process p = sender as Process;
            // TODO: check for process error
            p.Close();
            p = null;


            // append to the path
            String javaPath = JavaFacade.GetActiveJavaPath(JavaFacade.IsActiveJreInstalled());
            String path = Environment.GetEnvironmentVariable("path", EnvironmentVariableTarget.Machine);
            if (path == null)
            {
                Environment.SetEnvironmentVariable("path", javaPath, EnvironmentVariableTarget.Machine);
            }
            else // prepend to path to get so that this java is found before MS java in the Windows\system32 dir
            {
                if (!path.Contains(javaPath))
                {
                    path = javaPath + @"\bin;" + path;
                    path = path.Replace(";;", ";"); // case sensisitive
                    this.Logger().Info("Adding " + path + " to the Windows PATH environment variable");
                    Environment.SetEnvironmentVariable("path", path, EnvironmentVariableTarget.Machine);
                }
            }

            Cursor.Current = Cursors.Default;
            this.Initialize();
        }

        private void TomcatInstallComplete(object sender, System.EventArgs e)
        {
            Process p = sender as Process;
            // TODO: check for process error
            p.Close();
            p = null;
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            TomcatFacade.ConfigureTomcatUsers(config);
            TomcatFacade.ConfigureTomcatService(config);
            Cursor.Current = Cursors.Default;
            this.Initialize();
        }

        #endregion

        #region initialize methods

        private bool DisableAllPageButtons()
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            return (!BusinessFacade.IsLoggedInUserAnAdministrator() || !BusinessFacade.IsOperatingSystemApproved(config, this.WizardForm.IsDeveloperMode())) ? true : false;
        }

        private void SetAdministratorPrerequisiteState()
        {
            String loggedInUser = BusinessFacade.GetLoggedInUserName();
            // report on administrator role prerequisite
            if (BusinessFacade.IsLoggedInUserAnAdministrator())
            {
                labelUserIsAdmin.Text = loggedInUser + " has the Administrator role.";
                pictureBoxUserIsAdmin.Image = global::VixInstaller.Properties.Resources.check;
            }
            else
            {
                labelUserIsAdmin.Text = BusinessFacade.GetLoggedInUserName() + " does not have Administrator role.";
                pictureBoxUserIsAdmin.Image = global::VixInstaller.Properties.Resources.error;
            }
        }

        private void SetOSPrerequisiteState()
        {
            String info = null;
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();

            if (BusinessFacade.IsOperatingSystemApproved(ref info, config, this.WizardForm.IsDeveloperMode()))
            {
                pictureBoxOS.Image = global::VixInstaller.Properties.Resources.check;
            }
            else
            {
                pictureBoxOS.Image = global::VixInstaller.Properties.Resources.error;
            }
            labelOS.Text = info;
        }

        private void SetJavaPrerequisiteState()
        {
            String info = null;
            // report on Java install status and set install button state
            if (JavaFacade.IsJavaInstalled(ref info, this.WizardForm.IsDeveloperMode()))
            {
                pictureBoxJava.Image = global::VixInstaller.Properties.Resources.check;
                buttonJava.Enabled = false;
            }
            else
            {
                pictureBoxJava.Image = global::VixInstaller.Properties.Resources.error;
                buttonJava.Enabled = this.DisableAllPageButtons() ? false : true;
                if (buttonJava.Enabled == true)
                {
                    buttonJava.Focus();
                }
            }
            labelJava.Text = info;
        }

        private void SetTomcatPrerequisiteState()
        {
            // report on Tomcat install status and set install button state
            if (TomcatFacade.IsActiveTomcatVersionInstalled())
            {
                labelTomcat.Text = "Apache Tomcat version " + TomcatFacade.ActiveTomcatVersion + " is installed.";
                pictureBoxTomcat.Image = global::VixInstaller.Properties.Resources.check;
                buttonTomcat.Enabled = false;
            }
            else
            {
                labelTomcat.Text = "Apache Tomcat version " + TomcatFacade.ActiveTomcatVersion + " is not installed.";
                pictureBoxTomcat.Image = global::VixInstaller.Properties.Resources.error;
                if (JavaFacade.IsJavaInstalled(this.WizardForm.IsDeveloperMode()))
                {
                    buttonTomcat.Enabled = this.DisableAllPageButtons() ? false : true;
                    if (buttonTomcat.Enabled == true)
                    {
                        buttonTomcat.Focus();
                    }
                }
                else
                {
                    buttonTomcat.Enabled = false; // Java must be installed first
                }
            }
        }

        private void SetLaurelBridgePrerequisiteState()
        {
            // report on Laurel Bridge status
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();

            if (LaurelBridgeFacade.IsLaurelBridgeRequired(config) == false)
            {
                labelLaurelBridge.Text = "The Laurel Bridge DICOM toolkit is not required.";
                pictureBoxLaurelBridge.Image = global::VixInstaller.Properties.Resources.check;
                buttonLaurelBridge.Enabled = false;
            }
            else if (LaurelBridgeFacade.IsDeprecatedLaurelBridgeInstalled())
            {
                labelLaurelBridge.Text = "Deprecated Laurel Bridge DICOM toolkit is installed.";
                pictureBoxLaurelBridge.Image = global::VixInstaller.Properties.Resources.error;
                buttonLaurelBridge.Enabled = true;
            }
            else if (LaurelBridgeFacade.IsLaurelBridgeLicensed()) // installed and licensed
            {
                labelLaurelBridge.Text = "The Laurel Bridge DICOM toolkit is installed.";
                pictureBoxLaurelBridge.Image = global::VixInstaller.Properties.Resources.check;
                buttonLaurelBridge.Enabled = false;
            }
            else if (LaurelBridgeFacade.IsLaurelBridgeInstalled()) // installed but not licensed
            {
                labelLaurelBridge.Text = "The Laurel Bridge DICOM toolkit is not licensed.";
                pictureBoxLaurelBridge.Image = global::VixInstaller.Properties.Resources.error;
                buttonLaurelBridge.Enabled = true;
            }
            else
            {
                labelLaurelBridge.Text = "The Laurel Bridge DICOM toolkit is not installed.";
                pictureBoxLaurelBridge.Image = global::VixInstaller.Properties.Resources.error;
                if (JavaFacade.IsJavaInstalled(this.WizardForm.IsDeveloperMode()) && TomcatFacade.IsActiveTomcatVersionInstalled())
                {
                    buttonLaurelBridge.Enabled = this.DisableAllPageButtons() ? false : true;
                    if (buttonLaurelBridge.Enabled == true)
                    {
                        buttonLaurelBridge.Focus();
                    }
                }
                else
                {
                    buttonLaurelBridge.Enabled = false;
                }
            }
        }

        private void SetTomcatServiceAccountPrerequisiteState()
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            if (WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) && (this.reapplyServiceAccount == false))
            {
                labelTomcatServiceAccount.Text = "The service account has been configured.";
                pictureBoxTomcatServiceAccount.Image = global::VixInstaller.Properties.Resources.check;
                buttonTomcatServiceAccount.Enabled = false;
            }
            else
            {
                if (reapplyServiceAccount)
                {
                    labelTomcatServiceAccount.Text = "The service account must be reassigned to the Apache Tomcat service.";
                }
                else
                {
                    labelTomcatServiceAccount.Text = "The service account has not been created yet.";
                }
                pictureBoxTomcatServiceAccount.Image = global::VixInstaller.Properties.Resources.error;
                if (JavaFacade.IsJavaInstalled(this.WizardForm.IsDeveloperMode()) && TomcatFacade.IsActiveTomcatVersionInstalled() &&
                    (LaurelBridgeFacade.IsLaurelBridgeRequired(config) == false || (LaurelBridgeFacade.IsDeprecatedLaurelBridgeInstalled() == false && LaurelBridgeFacade.IsLaurelBridgeLicensed())))
                {
                    buttonTomcatServiceAccount.Enabled = this.DisableAllPageButtons() ? false : true;
                    if (buttonTomcatServiceAccount.Enabled == true)
                    {
                        buttonTomcatServiceAccount.Focus();
                    }
                }
                else
                {
                    buttonTomcatServiceAccount.Enabled = false;
                }
            }
        }

        private void SetVixCertificatePrerequisiteState()
        {
            // report on VIX Federation certificate status
            // either the certificate not required or is already installed or we know the location of the certificate installation zip filespec
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();

            if (config.VixRole == VixRoleType.DicomGateway)
            {
                labelVixCertificate.Text = "A security certificate is not required.";
                pictureBoxVixCertificate.Image = global::VixInstaller.Properties.Resources.check;
                buttonVixCertificate.Enabled = false;
            }
            else if (VixFacade.IsFederationCertificateInstalledOrNotRequired(config))
            {
                labelVixCertificate.Text = "The VIX security certificate is installed.";
                pictureBoxVixCertificate.Image = global::VixInstaller.Properties.Resources.check;
                buttonVixCertificate.Enabled = false;
            }
            else
            {
                labelVixCertificate.Text = "The VIX security certificate is not installed.";
                pictureBoxVixCertificate.Image = global::VixInstaller.Properties.Resources.error;
                if (JavaFacade.IsJavaInstalled(this.WizardForm.IsDeveloperMode()) && TomcatFacade.IsActiveTomcatVersionInstalled() &&
                    (LaurelBridgeFacade.IsLaurelBridgeRequired(config) == false || (LaurelBridgeFacade.IsDeprecatedLaurelBridgeInstalled() == false && LaurelBridgeFacade.IsLaurelBridgeLicensed())) && 
                    WindowsUserUtilities.LocalAccountExists(TomcatFacade.TomcatServiceAccountName) && (this.reapplyServiceAccount == false))
                {
                    buttonVixCertificate.Enabled = this.DisableAllPageButtons() ? false : true;
                    if (buttonVixCertificate.Enabled == true)
                    {
                        buttonVixCertificate.Focus();
                    }
                }
                else
                {
                    buttonVixCertificate.Enabled = false;
                }
            }
        }

        private void SetPageSubHeader()
        {
            IVixConfigurationParameters config = this.WizardForm.GetVixConfigurationParameters();
            if (this.IsComplete())
            {
                if (config.VixRole == VixRoleType.DicomGateway)
                {
                    this.InteriorPageSubHeader = "All prerequisites for the HDIG installation have been met. To continue, click Next..";
                }
                else
                {
                    this.InteriorPageSubHeader = "All prerequisites for the VIX installation have been met. To continue, click Next..";
                }
            }
            else if (this.DisableAllPageButtons())
            {
                if (!BusinessFacade.IsOperatingSystemApproved(config, this.WizardForm.IsDeveloperMode()))
                {
                    if (config.VixRole == VixRoleType.DicomGateway)
                    {
                        this.InteriorPageSubHeader = "The HDIG is only approved for installation on Windows Server 2003. Click Cancel to exit the wizard.";
                    }
                    else
                    {
                        this.InteriorPageSubHeader = "The VIX is only approved for installation on Windows Server 2003. Click Cancel to exit the wizard.";
                    }
                }
                else
                {
                    if (config.VixRole == VixRoleType.DicomGateway)
                    {
                        this.InteriorPageSubHeader = "The HDIG Installer must be run by an Administrator. " + BusinessFacade.GetLoggedInUserName() + " does not have the Administrator role on this computer. Click Cancel to exit the wizard.";
                    }
                    else
                    {
                        this.InteriorPageSubHeader = "The VIX Installer must be run by an Administrator. " + BusinessFacade.GetLoggedInUserName() + " does not have the Administrator role on this computer. Click Cancel to exit the wizard.";
                    }
                }
            }
            else
            {
                //TODO: build a sub header based on next task.
                if (this.buttonJava.Enabled)
                {
                    this.InteriorPageSubHeader = "Click Install to install the Java Runtime Environment.";
                }
                else if (this.buttonTomcat.Enabled)
                {
                    this.InteriorPageSubHeader = "Click Install to install Apache Tomcat. You will be prompted to enter a password for the Tomcat admin account.";
                }
                else if (this.buttonLaurelBridge.Enabled)
                {
                    //if (this.IsDcfToolkitRefreshRequired())
                    //{
                    //    this.InteriorPageSubHeader = "Click Install to reinstall the Laurel Bridge DICOM toolkit.";
                    //}
                    //else
                    //{
                    this.InteriorPageSubHeader = "Click Install to install the Laurel Bridge DICOM toolkit. You will be prompted to activate the toolkit license.";
                    //}
                }
                else if (this.buttonTomcatServiceAccount.Enabled)
                {
                    this.InteriorPageSubHeader = "Click Create to create and configure the Tomcat service account. You will be prompted to enter a password for the service account.";
                }
                else if (this.buttonVixCertificate.Enabled)
                {
                    this.InteriorPageSubHeader = "Click Install to specify the location of the VIX security certificate.";
                }
            }
        }

        #endregion

    }
}
