/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
/**
 * 
 */
package gov.va.med.imaging.core.router.commands;

import gov.va.med.OID;
import gov.va.med.RoutingToken;
import gov.va.med.RoutingTokenImpl;
import gov.va.med.WellKnownOID;
import gov.va.med.exceptions.RoutingTokenFormatException;
import gov.va.med.imaging.artifactsource.ResolvedArtifactSource;
import gov.va.med.imaging.core.interfaces.exceptions.ConnectionException;
import gov.va.med.imaging.core.interfaces.exceptions.MethodConnectionException;
import gov.va.med.imaging.core.interfaces.exceptions.MethodException;
import gov.va.med.imaging.core.router.AbstractCommandImpl;
import gov.va.med.imaging.core.router.facade.InternalContext;
import gov.va.med.imaging.exceptions.OIDFormatException;
import gov.va.med.imaging.exchange.business.SiteNumberArtifactSourceTranslator;
import gov.va.med.imaging.transactioncontext.TransactionContext;
import gov.va.med.imaging.transactioncontext.TransactionContextFactory;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.apache.log4j.Logger;

/**
 * @author zzzzzzbeckec
 *
 */
public class GetTreatingSitesCommandImpl 
extends AbstractCommandImpl<List<ResolvedArtifactSource>> 
{
	private static final long serialVersionUID = -7099152556482314638L;
	private final RoutingToken routingToken;
	private final String patientIcn;
	private final boolean includeTrailingCharactersForSite200;
	private final SiteNumberArtifactSourceTranslator siteNumberArtifactSourceTranslator;
	
	/**
	 * @param commandContext
	 */
	public GetTreatingSitesCommandImpl(
		RoutingToken routingToken,
		String patientIcn,
		boolean includeTrailingCharactersForSite200,
		SiteNumberArtifactSourceTranslator siteNumberArtifactSourceTranslator)
	{
		super();
		
		this.routingToken = routingToken;
		this.patientIcn = patientIcn;
		this.includeTrailingCharactersForSite200 = includeTrailingCharactersForSite200;
		this.siteNumberArtifactSourceTranslator = siteNumberArtifactSourceTranslator;
	}
	
	public GetTreatingSitesCommandImpl(
			RoutingToken routingToken,
			String patientIcn,
			boolean includeTrailingCharactersForSite200)
	{
		this(routingToken, patientIcn, includeTrailingCharactersForSite200, null);
	}
	
	public GetTreatingSitesCommandImpl(
			RoutingToken routingToken,
			String patientIcn)
	{
		this(routingToken, patientIcn, false, null);
	}

	@Override
	public RoutingToken getRoutingToken()
	{
		return this.routingToken;
	}

	public String getPatientIcn()
	{
		return patientIcn;
	}

	public boolean isIncludeTrailingCharactersForSite200()
	{
		return includeTrailingCharactersForSite200;
	}

	/**
	 * @return the siteNumberArtifactSourceTranslator
	 */
	public SiteNumberArtifactSourceTranslator getSiteNumberArtifactSourceTranslator()
	{
		return siteNumberArtifactSourceTranslator;
	}

	/* (non-Javadoc)
	 * @see gov.va.med.imaging.core.router.AbstractCommandImpl#callSynchronouslyInTransactionContext()
	 */
	@Override
	public List<ResolvedArtifactSource> callSynchronouslyInTransactionContext()
	throws MethodException
	{
		TransactionContext transactionContext = TransactionContextFactory.get();
		transactionContext.setServicedSource(getRoutingToken().toRoutingTokenString());
		
		getLogger().info( "getTreatingSites - Transaction ID [" + transactionContext.getTransactionId() + "] from site [" + getRoutingToken().toString() + "] for patient [" + patientIcn + "].");
		OID routingTokenOid = null;
		
		try
		{
			routingTokenOid = OID.create( getRoutingToken().getHomeCommunityId() );
		}
		catch (OIDFormatException x)
		{
			Logger.getLogger(this.getClass()).warn("Unable to create treating facility artifact source list because routing token home community ID is not a valid OID.");
			return new ArrayList<ResolvedArtifactSource>(0);
		}
		
		if( !WellKnownOID.VA_RADIOLOGY_IMAGE.isApplicable(routingTokenOid) && !WellKnownOID.VA_DOCUMENT.isApplicable(routingTokenOid) )
		{
			Logger.getLogger(this.getClass()).warn("Unable to create treating facility artifact source list because routing token does not point to the VA home community ID.");
			return new ArrayList<ResolvedArtifactSource>(0);
		}
		
		try
		{
			List<String> siteNumbers = InternalContext.getRouter().getTreatingSitesFromDataSource(
				getRoutingToken(),
				getPatientIcn(),
				isIncludeTrailingCharactersForSite200());
			getLogger().info("Got " + ((siteNumbers == null || siteNumbers.size() == 0) ? "no" : siteNumbers.size()) + 
					" site numbers for patient '" + patientIcn + "'.");
			
			// a null Study Set indicates no studies meet the search criteria
			if(siteNumbers != null && siteNumbers.size() > 0)
			{
				List<ResolvedArtifactSource> artifactSources = new ArrayList<ResolvedArtifactSource>();
				Map<String, String> addedSiteNumbers = new HashMap<String, String>();
				for(String siteNumber : siteNumbers)
				{
					RoutingToken rt;
					try
					{						
						String originalSiteNumber = siteNumber;
						if(getSiteNumberArtifactSourceTranslator() != null)
						{
							siteNumber = getSiteNumberArtifactSourceTranslator().translateSiteNumberToArtifactSourceSiteNumber(siteNumber);
						}
						if(siteNumber == null)
						{
							getLogger().info("Site number has been translated from '" + originalSiteNumber + "' to null, indicates it should be discarded.");
						}
						else
						{
							rt = RoutingTokenImpl.createVARadiologySite(siteNumber);
							
							// JMW 7/23/2012 P124, field testing has found that patients who have been seen at both 200 and 200DOD are showing
							// duplicate DoD buttons on the remote site toolbar in the AWIV Web App. Adding a check to ensure only unique
							// sites are added to the artifactSources result since there is no reason to show the same site number twice
							if(addedSiteNumbers.containsKey(siteNumber))
							{
								getLogger().debug("Excluding duplicate site number '" + siteNumber + "' from artifact sources result.");
							}
							else
							{	
								// NOTE: do not use the internal getSite() overload here
								// we are not calling a data source
								ResolvedArtifactSource resolvedSite = getCommandContext().getResolvedArtifactSource(rt);
								if(resolvedSite != null)
								{
									artifactSources.add(resolvedSite);
									addedSiteNumbers.put(siteNumber, "");
								}
								else
									getLogger().warn("Unable to resolve site number '" + siteNumber + "'.");
							}
						}
					}
					catch (RoutingTokenFormatException x)
					{
						getLogger().error(x);
					}
				}
				
				// not sure why this was down here...
				//TransactionContextFactory.get().setDatasourceProtocol(patientUrl.getProtocol());
				StringBuilder sb = new StringBuilder();
				sb.append("Got '" + artifactSources.size() + "' artifact sources from '" + getRoutingToken().toRoutingTokenString() + "': ");
				String prefix = "";
				for(ResolvedArtifactSource resolvedArtifactSource : artifactSources)
				{
					sb.append(prefix);
					sb.append(resolvedArtifactSource.getArtifactSource().getRepositoryId());
					prefix = ", ";
				}
				transactionContext.addDebugInformation(sb.toString());
				return artifactSources;
			}
			else
			{
				return new ArrayList<ResolvedArtifactSource>(0);
			}
		}
		catch(ConnectionException cX)
		{
			throw new MethodConnectionException(cX);
		}
		
	}

	@Override
	protected String parameterToString()
	{
		StringBuilder sb = new StringBuilder();
		sb.append(getPatientIcn());
		sb.append(", ");
		sb.append(getRoutingToken().toString());
		
		return sb.toString();
	}

	@Override
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
		result = prime * result + ((this.patientIcn == null) ? 0 : this.patientIcn.hashCode());
		result = prime * result + ((this.routingToken == null) ? 0 : this.routingToken.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj)
	{
		if (this == obj)
			return true;
		if (getClass() != obj.getClass())
			return false;
		final GetTreatingSitesCommandImpl other = (GetTreatingSitesCommandImpl) obj;
		if (this.patientIcn == null)
		{
			if (other.patientIcn != null)
				return false;
		}
		else if (!this.patientIcn.equals(other.patientIcn))
			return false;
		if (this.routingToken == null)
		{
			if (other.routingToken != null)
				return false;
		}
		else if (!this.routingToken.equals(other.routingToken))
			return false;
		return true;
	}
	
	

}
