/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
package gov.va.med.imaging.exchange.enums;

/**
 * The list of mime types that the VIX understands.
 * 
 * @author ZZZZZZBECKEC
 */
public enum ImageFormat 
{
	// The order of the declared enum instances is significant.
	// More specific types should precede less specific types.
	TGA("image/x-targa", "TGA format image", null, false), 
	DOWNSAMPLEDTGA("image/x-targa", "Reduced resolution TGA format image", null, false), 
	TIFF("image/tiff", "TIFF format image", null, false), 
	BMP("image/bmp", "A bitmap image", null, false), 
	JPEG("image/jpeg", "JPEG formatted image", null, true), 
	J2K("image/j2k", "JPEG 2000 formatted image", null, true), 
	DICOM("application/dicom", "A DICOM image, the content is unknown", null, false), 
	DICOMJPEG("application/dicom", "DICOM wrapped JPEG images", "image/jpeg", true), 
	DICOMJPEG2000("application/dicom", "DICOM wrapped JPEG 2000-images", "image/j2k", true),
	DICOMPDF("application/dicom","DICOM wrapped PDF documents", "application/pdf", false),
	PDF("application/pdf", "A PDF document", null, false),
	TEXT_DICOM("text/dicom","Plain text file containing DICOM header information for an image", null, false),
	TEXT_PLAIN("text/plain","Plain text file containing patient text data (not DICOM header information)", null, false),
	IMAGE("image/*","Any image type", null, false),
	APPLICATION("application/*","Any application type", null, false),
	ANYTHING("*/*", "Any filetype available", null, false),
	DOC("application/msword", "Microsoft Word documents", null, false),
	AVI("video/x-msvideo", "AVI Video", null, false),
	RTF("text/rtf", "Rich text application document", null, false),
	WAV("audio/x-wav", "WAV audio", null, false),
	HTML("text/html","HTML document", null, false),
	MP3("audio/mpeg3", "Audio MPEG Layer 3", null, false),
	MPG("video/mpeg", "Video MPEG", null, false),
	//UNKNOWN("?/?", "Unknown file format", null),
	ORIGINAL("x/x", "Represents the original format of the image as it is stored by the owner", null, false),
	GIF("image/gif", "GIF formatted image", null, true),
	PNG("image/png", "PNG image", null, true),
	XLS("application/vnd.ms-excel", "Microsoft Excel (pre Office 2007) format", null, false),
	DOCX("application/vnd.openxmlformats-officedocument.wordprocessingml.document", "Microsoft Word 2007+ format", null, false);
			
	
	private final String mime;
	private final String type;
	private final String subType;
	private final String enclosedMime;
	private final String enclosedType;
	private final String enclosedSubType;
	private final String description;
	private final boolean compressed;
	
	private final static String mimeDelimiter = "/";
	
	/**
	 * 
	 * @param mimeType
	 * @param description
	 * @param enclosedMimeType
	 */
	ImageFormat(String mimeType, String description, String enclosedMimeType, boolean compressed)
	{
		this.mime = mimeType;
		this.compressed = compressed;
		int slashIndex = mimeType.indexOf(mimeDelimiter); 
		if(slashIndex >= 0)
		{
			this.type = mimeType.substring(0, slashIndex);
			this.subType = slashIndex < mimeType.length() ? mimeType.substring(slashIndex+1) : null;
		}
		else
		{
			this.type = mimeType;
			this.subType = null;
		}
		
		this.description = description;
		this.enclosedMime = enclosedMimeType;
		if(enclosedMime != null)
		{
			slashIndex = enclosedMime.indexOf(mimeDelimiter); 
			if(slashIndex >= 0)
			{
				this.enclosedType = enclosedMime.substring(0, slashIndex);
				this.enclosedSubType = slashIndex < enclosedMime.length() ? enclosedMime.substring(slashIndex+1) : null;
			}
			else
			{
				this.enclosedType = enclosedMime;
				this.enclosedSubType = null;
			}
		}
		else
		{
			this.enclosedType = null;
			this.enclosedSubType = null;
		}
	}

	public String getMime()
	{
		return this.mime;
	}

	public String getType()
    {
    	return type;
    }

	public String getSubType()
    {
    	return subType;
    }

	public String getDescription()
	{
		return this.description;
	}

	public String getEnclosedMime()
    {
    	return enclosedMime;
    }

	public String getEnclosedType()
    {
    	return enclosedType;
    }

	public String getEnclosedSubType()
    {
    	return enclosedSubType;
    }
	
	public String getMimeWithEnclosedMime()
	{
		if(enclosedMime != null)
		{
			return mime + mimeDelimiter + enclosedMime;
		}
		else
		{
			return mime;
		}
	}

	/**
	 * Determines if the image format is a compressed format
	 * 
	 * @return the compressed
	 */
	public boolean isCompressed() {
		return compressed;
	}

	/**
	 * 
	 * @param imageFormat
	 * @return
	 */
	public static String getContentType(ImageFormat imageFormat) 
	{
		return imageFormat.getMime();
	}
	
	/**
	 * Do a reverse lookup, find the ImageFormat enum from the mime type.
	 * Note that the mime type is specific to type and subtype and is 
	 * case sensitive.
	 * Note that this will return the first ImageFormat enum mapped to the
	 * mime type, there may be many.
	 * If the enclosed type is specified then it must match
	 * that in the ImageFormat enclosedMimeType.
	 * @param mimeType
	 * @return
	 */
	public static ImageFormat valueOfMimeType(String mimeType)
	{
		String pieces[] = mimeType.split(mimeDelimiter);
		if(pieces.length == 4)
		{
			String mime = pieces[0] + mimeDelimiter + pieces[1];
			String enclosedMime = pieces[2] + mimeDelimiter + pieces[3];
			return valueOfMimeType(mime, enclosedMime);
			
		}
		else
		{
			return valueOfMimeType(mimeType, null);
		}
	}
	
	public static ImageFormat valueOfMimeType(String mimeType, String enclosedMimeType)
	{
		for(ImageFormat imageFormat : ImageFormat.values())
			if(imageFormat.getMime().equals(mimeType))
				if(enclosedMimeType == null)
					return imageFormat;
				else if( imageFormat.getEnclosedMime() != null && 
						imageFormat.getEnclosedMime().equals(enclosedMimeType) )
					return imageFormat;
		
		return null;
	}
	
	public static boolean isDICOMFormat(ImageFormat imageFormat)
	{
		if((imageFormat == ImageFormat.DICOM) ||
			(imageFormat == ImageFormat.DICOMJPEG) ||
			(imageFormat == ImageFormat.DICOMJPEG2000) ||
			(imageFormat == ImageFormat.DICOMPDF))
			return true;
		return false;
			
	}
	
	public static boolean isDICOMFormat(String mimeType)
	{
		ImageFormat format = ImageFormat.valueOfMimeType(mimeType);
		if(format == null)
			return false;
		return isDICOMFormat(format);
	}
	
	/**
	 * 
	 * @param ordinal
	 * @return
	 */
	public static ImageFormat valueOf(int ordinal)
	{
		for(ImageFormat imageFormat : ImageFormat.values())
			if(imageFormat.ordinal() == ordinal)
				return imageFormat;
		
		return null;
	}
	
}
