/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
/**
 * 
 */
package gov.va.med.imaging.datasource;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

/**
 * The base class for service provider factories.
 * 
 * @author zzzzzzbeckec
 *
 * @param <T>
 */
abstract class ServiceProviderFactory<T extends DataSourceSpi>
{
	private final Provider parentProvider;
	private final Class<? extends DataSourceSpi> spiType;

	ServiceProviderFactory(Provider parentProvider, Class<? extends DataSourceSpi> spiType)
	{
		if(parentProvider == null || spiType == null)
			throw new IllegalArgumentException("Both parent provider and SPI type must be provided to construct a ServiceProviderFactory.");
		
		this.parentProvider = parentProvider;
		this.spiType = spiType;
	}

	public Provider getParentProvider()
	{
		return this.parentProvider;
	}

	public Class<? extends DataSourceSpi> getSpiType()
	{
		return this.spiType;
	}

	public String getProductTypeName()
	{
		return getSpiType().getSimpleName();
	}
	
	/**
	 * @param url
	 * @param site
	 * @param implementingClass
	 * @param service 
	 * @return
	 * @throws NoSuchMethodException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 */
	@SuppressWarnings("unchecked")
	protected T createInstanceUsingStaticCreateMethod(
		Class<?> implementingClass, 
		ProviderService service, 
		Class<?>[] createParameterTypes,
		Object[] createParameters) 
	throws NoSuchMethodException, IllegalAccessException, InvocationTargetException
	{
		T dataSource;
		String createLogMessage = "create(";
		if (createParameterTypes != null)
			for (Class<?> parameterType : createParameterTypes)
				createLogMessage = createLogMessage + " " + parameterType.getName();
		createLogMessage = createLogMessage + ")";

		Provider.logger.debug(this.spiType.getSimpleName() + " instance, attempting instantiation using " + createLogMessage
			+ ".");

		Method factoryMethod = implementingClass.getDeclaredMethod("create", createParameterTypes);
		int modifiers = factoryMethod.getModifiers();
		if (!Modifier.isPublic(modifiers) || !Modifier.isStatic(modifiers))
		{
			String msg = "Factory method create(URL, Site) exists but is not static and public so will not be used.";
			Provider.logger.warn(msg);
			throw new NoSuchMethodException(msg);
		}
		dataSource = (T) factoryMethod.invoke(null, createParameters);
		Provider.logger.info(this.spiType.getSimpleName() + " instance, instantiated using " + createLogMessage + ".");

		dataSource.setConfiguration(service.getProvider().getInstanceConfiguration());
		
		return dataSource;
	}

	/**
	 * 
	 * @param implementingClass
	 * @param service 
	 * @param constructorParameterTypes
	 * @param constructorParameters
	 * @return
	 * @throws NoSuchMethodException
	 * @throws IllegalAccessException
	 * @throws InvocationTargetException
	 * @throws IllegalArgumentException
	 * @throws InstantiationException
	 */
	@SuppressWarnings("unchecked")
	protected T createInstanceUsingConstructor(
		Class<?> implementingClass, 
		ProviderService service, 
		Class<?>[] constructorParameterTypes,
		Object[] constructorParameters) 
	throws NoSuchMethodException, IllegalAccessException, 
	       InvocationTargetException, IllegalArgumentException, 
	       InstantiationException
	{
		T dataSource;
		String createLogMessage = "<ctor>(";
		if (constructorParameterTypes != null)
			for (Class<?> parameterType : constructorParameterTypes)
				createLogMessage = createLogMessage + " " + parameterType.getName();
		createLogMessage = createLogMessage + ")";

		Provider.logger.debug(this.spiType.getSimpleName() + " instance, attempting instantiation using " + createLogMessage
			+ ".");
		Constructor<?> constructor = implementingClass.getConstructor(constructorParameterTypes);
		int modifiers = constructor.getModifiers();
		if (!Modifier.isPublic(modifiers))
		{
			String msg = "constructor(URL, Site) exists but is not public so will not be used.";
			Provider.logger.warn(msg);
			throw new NoSuchMethodException(msg);
		}

		dataSource = (T) constructor.newInstance(constructorParameters);
		Provider.logger.debug(this.spiType.getSimpleName() + " instance, instantiated using " + createLogMessage + ".");
		
		dataSource.setConfiguration(service.getProvider().getInstanceConfiguration());

		return dataSource;
	}
}
