/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Net;

namespace DailyTransactionLogAnalyzerEngine
{
    public class AnalysisDownloader
    {
        private LogMsgDelegate OnLogMsgEvent = null;
        public string AnalysisServerUrl {get; private set;}

        public AnalysisDownloader(LogMsgDelegate onLogMsgEvent, string analysisServerUrl)
        {
            this.OnLogMsgEvent = onLogMsgEvent;
            this.AnalysisServerUrl = analysisServerUrl;
        }

        private void LogMsg(string msg)
        {
            if (OnLogMsgEvent != null)
                OnLogMsgEvent(msg);
        }

        public void DownloadAnalysis(string analysisDirectory, DateTime startDate, DateTime endDate, DailyTransactionsType dailyTransactionType, List<string> siteIds)
        {
            DateTime date = startDate;
            while (date.CompareTo(endDate) < 1)
            {
                foreach (string siteNumber in siteIds)
                {
                    DownloadAnalysis(analysisDirectory, date, dailyTransactionType, siteNumber);
                }

                date = IncrementDate(date, dailyTransactionType);
            }
        }

        private void DownloadAnalysis(string analysisDirectory,DateTime date, DailyTransactionsType dailyTransactionType, string siteNumber)
        {
            string url = AnalysisServerUrl + "/" + GetAnalysisUri(date, dailyTransactionType, siteNumber);

            string filename = GetLocalFilename(analysisDirectory, date, dailyTransactionType, siteNumber);
            FileInfo f = new FileInfo(filename);
            if (!f.Directory.Exists)
            {
                f.Directory.Create();
            }
            WebClient client = new WebClient();
            try
            {
                LogMsg("Downloading analysis from URL '" + url + "'.");
                client.DownloadFile(url, filename);
            }
            catch (WebException webex)
            {
                LogMsg("Exception downloading from URL '" + url + "', " + webex.Message);
            }

        }

        private string GetLocalFilename(string analysisDirectory, DateTime date, DailyTransactionsType dailyTransactionType, string siteNumber)
        {
            switch (dailyTransactionType)
            {
                case DailyTransactionsType.day:
                    return DailyTransactions.GetDateFilename(analysisDirectory, siteNumber, date);
                case DailyTransactionsType.month:
                    return DailyTransactions.GetMonthFilename(analysisDirectory, siteNumber, date.Year, date.Month);
                case DailyTransactionsType.year:
                    return DailyTransactions.GetYearFilename(analysisDirectory, siteNumber, date.Year);
            }
            return null;
        }

        private string GetAnalysisUri(DateTime date, DailyTransactionsType dailyTransactionType, string siteNumber)
        {
            StringBuilder uri = new StringBuilder();
            uri.Append(siteNumber);
            uri.Append("/");
            
            switch (dailyTransactionType)
            {
                case DailyTransactionsType.day:
                    uri.Append(date.Year);
                    uri.Append("/");
                    uri.Append(date.ToString("MM"));
                    uri.Append("/");
                    uri.Append(date.ToString("MM-dd-yyyy") + ".xml");
                    break;
                case DailyTransactionsType.month:
                    uri.Append(date.Year);
                    uri.Append("/");
                    uri.Append(date.ToString("yyyy-MM") + ".xml");
                    break;
                case DailyTransactionsType.year:
                    uri.Append(date.Year + ".xml");
                    break;
            }

            return uri.ToString();
        }

        private DateTime IncrementDate(DateTime date, DailyTransactionsType dailyTransactionType)
        {
            switch (dailyTransactionType)
            {
                case DailyTransactionsType.day:
                    return date.AddDays(1);
                case DailyTransactionsType.month:
                    return date.AddMonths(1);
                case DailyTransactionsType.year:
                    return date.AddYears(1);
            }
            return date;
        }
    }
}
