/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VixLogAnalyzerEngine;
using System.IO;
using System.Xml;
using System.Xml.Serialization;

namespace DailyTransactionLogAnalyzerEngine
{
    public class DailyTransactions
    {
        public static string CurrentVersion = "1.3";

        public DateTime Date { get; private set; }
        public string SiteNumber { get; private set; }
        public TransactionCounts TransactionCounts { get; private set; }
        public Dictionary<string, FacadeCounts> FacadeCounts { get; private set; }
        public DailyTransactionsType DailyTransactionsType { get; private set; }
        public DateTime? DateCalculated { get; private set; }
        public string Version { get; private set; }    

        public DailyTransactions(string siteNumber, DateTime date, DailyTransactionsType dailyTransactionsType, 
            DateTime? dateCalculated, string version)
        {
            this.SiteNumber = siteNumber;
            this.Date = date;
            this.DailyTransactionsType = dailyTransactionsType;
            this.DateCalculated = dateCalculated;
            this.Version = version;
            TransactionCounts = new TransactionCounts();
            FacadeCounts = new Dictionary<string, FacadeCounts>();

            FacadeCounts.Add("Federation WebApp", new FacadeCounts("Federation WebApp", SiteIdentificationSource.requestingSite));

            if (siteNumber.StartsWith("2001."))
            {
                // CVIX
                //AWIV WebApp
                FacadeCounts.Add("AWIV WebApp Responses", new FacadeCounts("AWIV WebApp", "AWIV WebApp Responses", SiteIdentificationSource.respondingSite));
                FacadeCounts.Add("AWIV WebApp Requests", new FacadeCounts("AWIV WebApp", "AWIV WebApp Requests", SiteIdentificationSource.requestingSite));
                FacadeCounts.Add("Exchange WebApp", new FacadeCounts("Exchange WebApp", SiteIdentificationSource.respondingSite));
                FacadeCounts.Add("AWIV Client Web App", new FacadeCounts("AWIV Client Web App", SiteIdentificationSource.respondingSite));
                FacadeCounts.Add("XCARespondingGatewayWebApp", new FacadeCounts("XCARespondingGatewayWebApp", SiteIdentificationSource.respondingSite));
            }
            else
            {
                // VIX
                FacadeCounts.Add("VistARad WebApp", new FacadeCounts("VistARad WebApp", SiteIdentificationSource.respondingSite));
                FacadeCounts.Add("Clinical Display WebApp", new FacadeCounts("Clinical Display WebApp", SiteIdentificationSource.respondingSite));
            }            
        }

        public void AnalyzeTransactions(List<TransactionLogEntry> entries)
        {
            TransactionCounts.TotalParentTransactions += entries.Count;
            long errorCount = TransactionCounts.TransactionsWithExceptions;           

            foreach (TransactionLogEntry entry in entries)
            {
                if (entry.IsError())
                {
                    errorCount++;
                }
                foreach (string facadeName in FacadeCounts.Keys)
                {
                    FacadeCounts fc = FacadeCounts[facadeName];
                    if (entry.QueryType.StartsWith(fc.FacadeName, StringComparison.CurrentCultureIgnoreCase))
                    {
                        fc.AnalyzeTransaction(entry);
                        //break; // might be multiple FacadeCounts for this facade
                    }
                }
            }
            TransactionCounts.TransactionsWithExceptions = errorCount;            
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine(SiteNumber + " from " + Date.ToString());
            sb.AppendLine(TransactionCounts.TotalParentTransactions + " parent transactions, " + TransactionCounts.TransactionsWithExceptions + " transactions with exceptions");
            foreach (FacadeCounts fc in FacadeCounts.Values)
            {
                sb.Append(fc.ToString());
            }
            return sb.ToString();
        }
        
        public void SaveDateToXml(string outputDirectory, string siteNumber, DateTime date)
        {
            string dateString = date.ToString("MM-dd-yyyy");
            string filename = GetDateFilename(outputDirectory, siteNumber, date);// Path.Combine(outputDirectory, siteNumber, date.ToString("yyyy"), date.ToString("MM"), dateString) + ".xml";
            SaveToXml(filename, dateString, siteNumber);
        }

        public void SaveMonthToXml(string outputDirectory, string siteNumber, int year, int month)
        {
            string dateString = year + "-" + month.ToString("D2");
            //string filename = Path.Combine(outputDirectory, siteNumber, year + "", month.ToString("D2") + "", dateString) + ".xml";
            string filename = GetMonthFilename(outputDirectory, siteNumber, year, month);
            SaveToXml(filename, dateString, siteNumber);
        }

        public void SaveYearToXml(string outputDirectory, string siteNumber, int year)
        {
            string dateString = year + "";

            string filename = GetYearFilename(outputDirectory, siteNumber, year);
            SaveToXml(filename, dateString, siteNumber);
        }

        public static string GetMonthFilename(string outputDirectory, string siteNumber, int year, int month)
        {
            string dateString = year + "-" + month.ToString("D2");
            return Path.Combine(outputDirectory, siteNumber, year + "", dateString) + ".xml";
        }

        public static string GetDateFilename(string outputDirectory, string siteNumber, DateTime date)
        {
            string dateString = date.ToString("MM-dd-yyyy");
            return Path.Combine(outputDirectory, siteNumber, date.ToString("yyyy"), date.ToString("MM"), dateString) + ".xml";
        }

        public static string GetYearFilename(string outputDirectory, string siteNumber, int year)
        {
            string dateString = year + "";
            return Path.Combine(outputDirectory, siteNumber, dateString) + ".xml";
        }

        private void SaveToXml(string filename, string dateString, string siteNumber)
        {
            XmlDocument xmlDoc = new XmlDocument();

            FileInfo f = new FileInfo(filename);
            if (!f.Directory.Exists)
            {
                Directory.CreateDirectory(f.Directory.FullName);
            }

            XmlProcessingInstruction instruction =
                    xmlDoc.CreateProcessingInstruction("xml-stylesheet", "type='text/xsl' href='display.xsl'");
            xmlDoc.AppendChild(instruction);

            XmlNode rootNode = xmlDoc.CreateElement("DailyTransactions");
            xmlDoc.AppendChild(rootNode);

            XmlAttribute attr = xmlDoc.CreateAttribute("dt");
            attr.Value = dateString;
            rootNode.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("id");
            attr.Value = siteNumber;
            rootNode.Attributes.Append(attr);

            string siteName = SiteHelper.GetSiteName(siteNumber);
            if (!string.IsNullOrEmpty(siteName))
            {
                attr = xmlDoc.CreateAttribute("sn");
                attr.Value = siteName;
                rootNode.Attributes.Append(attr);
            }

            attr = xmlDoc.CreateAttribute("dc");
            if (this.DateCalculated.HasValue)
            {
                attr.Value = this.DateCalculated.Value.ToString();
            }
            else
            {
                attr.Value = "";
            }
            rootNode.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("v");
            attr.Value = Version;
            rootNode.Attributes.Append(attr);

            rootNode.AppendChild(TransactionCounts.createXmlNode(xmlDoc));

            foreach (FacadeCounts fc in FacadeCounts.Values)
            {
                rootNode.AppendChild(fc.createXmlNode(xmlDoc));
            }

            xmlDoc.Save(filename);

            //XmlSerializer serializer = new XmlSerializer(this.GetType());
            //serializer.Serialize(new FileStream(filename + ".serialized",FileMode.Create), this);
        }

        public static DailyTransactions LoadFromXml(string dailyTransactionsDirectory, string siteNumber, int year, int month)
        {
            string filename = GetMonthFilename(dailyTransactionsDirectory, siteNumber, year, month);
            return LoadFromXml(filename, DailyTransactionsType.month);
        }

        public static DailyTransactions LoadFromXml(string dailyTransactionsDirectory, string siteNumber, DateTime date)
        {
            string filename = GetDateFilename(dailyTransactionsDirectory, siteNumber, date);
            return LoadFromXml(filename, DailyTransactionsType.day);
        }

        public static DailyTransactions LoadFromXml(string dailyTransactionsDirectory, string siteNumber, int year)
        {
            string filename = GetYearFilename(dailyTransactionsDirectory, siteNumber, year);
            return LoadFromXml(filename, DailyTransactionsType.year);
        }

        public static DailyTransactions MergeIntoSingleYear(List<DailyTransactions> yearTransactions, string siteNumber, int year)
        {
            DateTime date = new DateTime(year, 1, 1);
            DailyTransactions result = new DailyTransactions(siteNumber, date, DailyTransactionsType.year, 
                DateTime.Now, CurrentVersion);
            foreach (DailyTransactions dailyTransactions in yearTransactions)
            {
                result.TransactionCounts.TotalParentTransactions += dailyTransactions.TransactionCounts.TotalParentTransactions;
                result.TransactionCounts.TransactionsWithExceptions += dailyTransactions.TransactionCounts.TransactionsWithExceptions;
                foreach (string facadeCountKey in dailyTransactions.FacadeCounts.Keys)
                {
                    FacadeCounts fc = dailyTransactions.FacadeCounts[facadeCountKey];
                    if (result.FacadeCounts.ContainsKey(facadeCountKey))
                    {
                        FacadeCounts existingFacadeCount = result.FacadeCounts[facadeCountKey];
                        existingFacadeCount.TotalTransactions += fc.TotalTransactions;
                        existingFacadeCount.TransactionsWithExceptions += fc.TransactionsWithExceptions;
                        existingFacadeCount.ImageRequestCounts += fc.ImageRequestCounts;
                        foreach (string fromSiteCountsKey in fc.SiteCounts.Keys)
                        {
                            FromSiteCounts fsc = fc.SiteCounts[fromSiteCountsKey];
                            if (existingFacadeCount.SiteCounts.ContainsKey(fromSiteCountsKey))
                            {
                                FromSiteCounts existingFromSiteCounts = existingFacadeCount.SiteCounts[fromSiteCountsKey];
                                existingFromSiteCounts.TotalTransactions += fsc.TotalTransactions;
                                existingFromSiteCounts.TransactionsWithExceptions += fsc.TransactionsWithExceptions;
                                existingFromSiteCounts.TotalImageRequests += fsc.TotalImageRequests;
                                existingFromSiteCounts.TotalImageExceptions += fsc.TotalImageExceptions;
                                existingFromSiteCounts.SuccessfulImageBytes += fsc.SuccessfulImageBytes;
                                existingFromSiteCounts.SuccessfulImageTransferTimeMs += fsc.SuccessfulImageTransferTimeMs;
                                existingFromSiteCounts.DataSourceBytesReceived += fsc.DataSourceBytesReceived;
                                existingFromSiteCounts.ImageCacheHits += fsc.ImageCacheHits;
                                existingFromSiteCounts.CachedBytesReturned += fsc.CachedBytesReturned;
                                existingFromSiteCounts.BseLogins += fsc.BseLogins;
                                existingFromSiteCounts.CapriLogins += fsc.CapriLogins;
                                foreach (string imageRequestKey in fsc.ImageRequests.Keys)
                                {
                                    ImageRequest ir = fsc.ImageRequests[imageRequestKey];
                                    if (existingFromSiteCounts.ImageRequests.ContainsKey(imageRequestKey))
                                    {
                                        ImageRequest existingImageRequest = existingFromSiteCounts.ImageRequests[imageRequestKey];
                                        existingImageRequest.RequestCount += ir.RequestCount;
                                        existingImageRequest.SuccessfulResponses += ir.SuccessfulResponses;
                                        existingImageRequest.ExceptionResponses += ir.ExceptionResponses;
                                        existingImageRequest.NotFoundResponses += ir.NotFoundResponses;
                                        existingImageRequest.InternalServerErrorResponse += ir.InternalServerErrorResponse;
                                        existingImageRequest.BytesReturned += ir.BytesReturned;
                                        existingImageRequest.TotalMs  += ir.TotalMs;
                                        existingImageRequest.DataSourceBytesReceived += ir.DataSourceBytesReceived;
                                        existingImageRequest.ImageCacheHits += ir.ImageCacheHits;
                                        existingImageRequest.CachedBytesReturned += ir.CachedBytesReturned;
                                    }
                                    else
                                    {
                                        existingFromSiteCounts.ImageRequests.Add(imageRequestKey, ir);
                                    }
                                }
                                foreach (string protocolName in fsc.ProtocolCounts.Keys)
                                {
                                    if (existingFromSiteCounts.ProtocolCounts.ContainsKey(protocolName))
                                    {
                                        existingFromSiteCounts.ProtocolCounts[protocolName] += fsc.ProtocolCounts[protocolName];
                                    }
                                    else
                                    {
                                        existingFromSiteCounts.ProtocolCounts.Add(protocolName, fsc.ProtocolCounts[protocolName]);
                                    }
                                }
                            }
                            else
                            {
                                existingFacadeCount.SiteCounts.Add(fromSiteCountsKey, fsc);
                            }
                        }
                    }
                    else
                    {
                        result.FacadeCounts.Add(facadeCountKey, fc);
                    }
                }
            }
            return result;
        }

        public static DailyTransactions LoadFromXml(string filename, DailyTransactionsType dailyTransactionsType)
        {
            XmlDocument xmlDoc = new XmlDocument();
            if (File.Exists(filename))
            {
                xmlDoc.Load(filename);

                XmlNode dailyTransactionsNode = xmlDoc.SelectSingleNode("DailyTransactions");
                string date = dailyTransactionsNode.Attributes["dt"].Value;
                string siteId = dailyTransactionsNode.Attributes["id"].Value;
                string siteName = dailyTransactionsNode.Attributes["sn"].Value;
                string version = null;
                DateTime? dateCalculated = null;
                if (dailyTransactionsNode.Attributes["dc"] != null)
                {
                    string dateCalculatedString = dailyTransactionsNode.Attributes["dc"].Value;
                    dateCalculated = DateTime.Parse(dateCalculatedString);
                }
                if (dailyTransactionsNode.Attributes["v"] != null)
                {
                    version = dailyTransactionsNode.Attributes["v"].Value;
                }

                DailyTransactions result = null;
                if (date.Length == 4)
                {
                    int year = int.Parse(date);
                    result = new DailyTransactions(siteId, new DateTime(year, 1, 1), dailyTransactionsType,
                        dateCalculated, version);
                }
                else
                {
                    result = new DailyTransactions(siteId, DateTime.Parse(date), dailyTransactionsType, dateCalculated, version);
                }

                result.TransactionCounts.loadFromXml(dailyTransactionsNode);
                XmlNodeList facadeNodes = dailyTransactionsNode.SelectNodes("Facade");
                foreach (XmlNode facadeNode in facadeNodes)
                {
                    FacadeCounts fc = DailyTransactionLogAnalyzerEngine.FacadeCounts.LoadFromXml(facadeNode);

                    if (result.FacadeCounts.ContainsKey(fc.DisplayName))
                    {
                        result.FacadeCounts[fc.DisplayName] = fc;
                    }
                    else
                    {

                        result.FacadeCounts.Add(fc.DisplayName, fc);
                    }
                }


                return result;
            }
            else
                return null;
        }

        public long BseLogins
        {
            get
            {
                long value = 0L;
                foreach (FacadeCounts fc in FacadeCounts.Values)
                {
                    value += fc.BseLogins;
                }
                return value;
            }
        }

        public long CapriLogins
        {
            get
            {
                long value = 0L;
                foreach (FacadeCounts fc in FacadeCounts.Values)
                {
                    value += fc.CapriLogins;
                }
                return value;
            }
        }
    
    
    }

    public class TransactionCounts
    {
        //public int TotalCount { get; set; }
        public long TotalParentTransactions { get; set; }
        public long TransactionsWithExceptions { get; set; }

        public TransactionCounts()
        {
            TotalParentTransactions = 0L;
            TransactionsWithExceptions = 0L;
        }

        public XmlNode createXmlNode(XmlDocument xmlDoc)
        {
            XmlNode result = xmlDoc.CreateElement("Transactions");

            XmlAttribute attr = xmlDoc.CreateAttribute("pt");
            attr.Value = TotalParentTransactions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("twe");
            attr.Value = TransactionsWithExceptions + "";
            result.Attributes.Append(attr);

            return result;
        }

        public void loadFromXml(XmlNode parentNode)
        {
            XmlNode transactionsNode = parentNode.SelectSingleNode("Transactions");
            TotalParentTransactions = long.Parse(transactionsNode.Attributes["pt"].Value);
            TransactionsWithExceptions = long.Parse(transactionsNode.Attributes["twe"].Value);
        }

    }

    /// <summary>
    /// Represents the counts that went through a specific facade
    /// </summary>
    public class FacadeCounts
    {
        public string FacadeName { get; private set; }
        public string DisplayName { get; private set; }
        public SiteIdentificationSource SiteIdentificationSource {get; private set;}
        public long TotalTransactions { get; set; }
        public long TransactionsWithExceptions { get; set; }
        public long ImageRequestCounts { get; set; }
        //public long SuccessfulImageBytes { get; set; }
        //public long SuccessfulImageTransferTimeMs { get; set; }
        public Dictionary<string, FromSiteCounts> SiteCounts { get; private set; }

        public bool IsClientFacingFacade
        {
            get
            {
                return (this.SiteIdentificationSource == DailyTransactionLogAnalyzerEngine.SiteIdentificationSource.respondingSite);
            }
        }

        public long ImageCacheResponses
        {
            get
            {
                long count = 0L;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    count += fsc.ImageCacheHits;
                }
                return count;
            }
        }

        public long SuccessfulImageResponses
        {
            get
            {
                long value = 0;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.SuccessfulImageResponses;
                }
                return value;
            }
        }

        public long SuccessfulImageBytes
        {
            get
            {
                long value = 0;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.SuccessfulImageBytes;
                }
                return value;
            }
        }

        public long DataSourceBytesReceived
        {
            get
            {
                long value = 0;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.DataSourceBytesReceived;
                }
                return value;
            }
        }

        public long ImageCacheHits
        {
            get
            {
                long value = 0L;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.ImageCacheHits;
                }
                return value;
            }
        }

        public long CachedBytesReturned
        {
            get
            {
                long value = 0L;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.CachedBytesReturned;
                }
                return value;
            }
        }


        public long RelevantBytes
        {
            get
            {
                if (this.SiteIdentificationSource == DailyTransactionLogAnalyzerEngine.SiteIdentificationSource.requestingSite)
                    return SuccessfulImageBytes;
                else
                    return DataSourceBytesReceived;
            }
        }

        public long BseLogins
        {
            get
            {
                long value = 0L;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.BseLogins;
                }
                return value;
            }
        }

        public long CapriLogins
        {
            get
            {
                long value = 0L;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.CapriLogins;
                }
                return value;
            }
        }

        public long SuccessfulImageTransferTimeMs
        {
            get
            {
                long value = 0;
                foreach (FromSiteCounts fsc in SiteCounts.Values)
                {
                    value += fsc.SuccessfulImageTransferTimeMs;
                }
                return value;
            }
        }

        public double KbSec
        {
            get
            {
                return DailyTransactionsHelper.ConvertToKbSec(SuccessfulImageBytes, SuccessfulImageTransferTimeMs);
            }
        }

        public FacadeCounts(string name, SiteIdentificationSource siteIdentificationSource)
            : this(name, null, siteIdentificationSource)
        {

        }

        public FacadeCounts(string name, string displayName, SiteIdentificationSource siteIdentificationSource)
        {
            this.FacadeName = name;            
            if (displayName == null)
                this.DisplayName = this.FacadeName;
            else
                this.DisplayName = displayName;
            this.SiteIdentificationSource = siteIdentificationSource;
            SiteCounts = new Dictionary<string, FromSiteCounts>();
            TotalTransactions = 0L;
            TransactionsWithExceptions = 0L;
            ImageRequestCounts = 0L;
        }

        //private static Dictionary<string, string> skipUsersDictionary = null;
        private static string[] skipUsers = new string[] { "WERFEL", "BUCK,DONALD K", "IMAGPROVIDERONETWOSIX,ONETWOSIX", "ayabes" };
        private static string[] skipPostImageAccessEventCommandFacades = new string[] { "Exchange WebApp" };

        private static bool SkipEntry(TransactionLogEntry entry)
        {
            /*
            if (skipUsersDictionary == null)
            {
                skipUsersDictionary = new Dictionary<string, string>();
                foreach (string name in skipUsers)
                {
                    skipUsersDictionary.Add(name, "");
                }
            }
            if (entry.User != null)
            {
                if(skipUsersDictionary.contain
            }
            return false;
             * */
            if (entry.User == null)
                return false;
            foreach (string name in skipUsers)
            {
                if (entry.User.Contains(name))
                    return true;
            }


            foreach (string skipImageAccesEventFacades in skipPostImageAccessEventCommandFacades)
            {
                if (entry.QueryType.StartsWith(skipImageAccesEventFacades))
                {
                    if(entry.CommandClassName.StartsWith("PostImageAccessEvent"))
                        return true;
                }
            }

            return false;
        }
     
        
        

        public void AnalyzeTransaction(TransactionLogEntry entry)
        {
            if (SkipEntry(entry))
                return;

            TotalTransactions++;

            string siteId = entry.RespondingSite;
            if (SiteIdentificationSource == SiteIdentificationSource.requestingSite)
            {
                // the requestingSite value comes from the DUZ(2) of the user so it might be a division number
                // recent VIX patches added a Requesting VIX Site Number when requesting data from another VIX to use
                // the actual VIX site number so we can have a better value here
                // if that value exists, use it otherwise just use the requestingSite value
                if (entry.Values.ContainsKey("Requesting VIX Site Number"))
                {
                    siteId = entry.Values["Requesting VIX Site Number"];
                    if (string.IsNullOrEmpty(siteId))
                    {
                        siteId = entry.RequestingSite;
                    }
                }
                else
                {
                    siteId = entry.RequestingSite;
                }
            }

            FromSiteCounts fsc = null;

            if (SiteCounts.ContainsKey(siteId))
            {
                fsc = SiteCounts[siteId];
            }
            else
            {
                fsc = new FromSiteCounts(siteId);
                SiteCounts.Add(siteId, fsc);
            }

            fsc.TotalTransactions++;
            object vistaLoginMethod = entry.getPropertyFromParentOrChild("VistaLoginMethod", true);
            if (vistaLoginMethod != null)
            {
                switch (vistaLoginMethod.ToString())
                {
                    case "CAPRI":
                        fsc.CapriLogins++;
                        break;
                    case "BSE":
                        fsc.BseLogins++;
                        break;
                }
            }

            object dataSource = entry.getPropertyFromParentOrChild("DataSource", true);
            if(dataSource != null)
            {
                string dataSourceString = dataSource.ToString();
                if (dataSourceString.Length > 0)
                {
                    if (fsc.ProtocolCounts.ContainsKey(dataSourceString))
                    {
                        fsc.ProtocolCounts[dataSourceString]++;
                    }
                    else
                    {
                        fsc.ProtocolCounts.Add(dataSourceString, 1);
                    }
                }
            }
            
            if (entry.IsError())
            {
                TransactionsWithExceptions++;
                fsc.TransactionsWithExceptions++;
            }

            if (entry.QueryType.Contains("image transfer") || entry.QueryType.Contains("XCARespondingGatewayWebApp Retrieve"))
            {
                ImageRequestCounts++;
                fsc.TotalImageRequests++;
                string modality = entry.Modality;
                string quality = entry.Quality;
                ImageSource imageSource = ImageSource.va;
                if (entry.Urn != null)
                {
                    if (entry.Urn.Contains("bhieimage:haims"))
                    {
                        imageSource = ImageSource.haims_radiology;
                    }
                    else if (entry.Urn.Contains("2.16.840.1.113883.3.198.1"))
                    {
                        imageSource = ImageSource.ncat;
                    }
                    else if (entry.Urn.Contains("bhieimage")
                        && (!entry.Urn.Contains("haims")))
                    {
                        imageSource = ImageSource.pacsi;
                    }
                    else if (entry.Urn.Contains("central"))
                    {
                        imageSource = ImageSource.haims_artifact;
                    }
                }

                string key = ImageRequest.CreateKey(modality, quality, imageSource);
                ImageRequest imageRequests = null;
                if (fsc.ImageRequests.ContainsKey(key))
                {
                    imageRequests = fsc.ImageRequests[key];
                }
                else
                {
                    imageRequests = new ImageRequest(entry.Modality, entry.Quality, imageSource);
                    fsc.ImageRequests.Add(key, imageRequests);
                }
                imageRequests.RequestCount++;

                if (entry.ItemInCache)
                {
                    imageRequests.ImageCacheHits++;
                    imageRequests.CachedBytesReturned += entry.FacadeBytesReturned;
                    fsc.ImageCacheHits++;
                    fsc.CachedBytesReturned += entry.FacadeBytesReturned;
                }

                long dataSourceBytesReceived = GetDataSourceBytesReceived(entry);                
                if (entry.IsError())
                {
                    imageRequests.ExceptionResponses++;
                    fsc.TotalImageExceptions++;
                    if (entry.ResponseCode == "404")
                        imageRequests.NotFoundResponses++;
                    if (entry.ResponseCode == "500")
                        imageRequests.InternalServerErrorResponse++;
                    // even with an exception we still got bytes from the datasource, so count them. Necessary for VRad since the image moves in the clientabortexception case
                    imageRequests.DataSourceBytesReceived += dataSourceBytesReceived;
                    fsc.DataSourceBytesReceived += dataSourceBytesReceived;
                }
                else
                {
                    
                    imageRequests.SuccessfulResponses++;
                    imageRequests.BytesReturned += entry.FacadeBytesReturned;                    
                    fsc.SuccessfulImageBytes += entry.FacadeBytesReturned;
                    fsc.SuccessfulImageTransferTimeMs += entry.TimeOnVix;                    
                    imageRequests.TotalMs += entry.TimeOnVix;
                    imageRequests.DataSourceBytesReceived += dataSourceBytesReceived;
                    fsc.DataSourceBytesReceived += dataSourceBytesReceived;
                }
            }
        }

        private long GetDataSourceBytesReceived(TransactionLogEntry entry)
        {
            object val = entry.getPropertyFromParentOrChild("DataSourceBytesReceived", true);
            if (val == null)
                return 0L;
            return (long)val;
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();

            sb.AppendLine(this.FacadeName + ": " + this.TotalTransactions + " total transactions, " + this.TransactionsWithExceptions + " transactions with exceptions, " + this.ImageRequestCounts + " image requests");
            foreach (FromSiteCounts fsc in SiteCounts.Values)
            {
                sb.Append("\t" + fsc.ToString());
            }

            return sb.ToString();
        }

        public XmlNode createXmlNode(XmlDocument xmlDoc)
        {
            XmlNode result = xmlDoc.CreateElement("Facade");

            XmlAttribute attr = xmlDoc.CreateAttribute("nm");
            attr.Value = this.FacadeName;
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("dn");
            attr.Value = this.DisplayName;
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("t");
            attr.Value = TotalTransactions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("twe");
            attr.Value = TransactionsWithExceptions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("irc");
            attr.Value = ImageRequestCounts + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("sBytes");
            attr.Value = SuccessfulImageBytes + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("sTimeMs");
            attr.Value = SuccessfulImageTransferTimeMs + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("kbs");
            attr.Value = KbSec.ToString("N2");
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("dsbr");
            attr.Value = DataSourceBytesReceived + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("ich");
            attr.Value = ImageCacheHits + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("cbr");
            attr.Value = CachedBytesReturned + "";
            result.Attributes.Append(attr);

            foreach (FromSiteCounts fsc in SiteCounts.Values)
            {
                result.AppendChild(fsc.createXmlNode(xmlDoc, TotalTransactions));
            }

            return result;
        }

        public static FacadeCounts LoadFromXml(XmlNode facadeNode)
        {
            string name = facadeNode.Attributes["nm"].Value;
            string displayName = name;
            if (facadeNode.Attributes["dn"] != null)
                displayName = facadeNode.Attributes["dn"].Value;

            FacadeCounts fc = new FacadeCounts(name, displayName, DailyTransactionsHelper.GetSiteIdentificationSource(name));
            fc.TotalTransactions = long.Parse(facadeNode.Attributes["t"].Value);
            fc.TransactionsWithExceptions = long.Parse(facadeNode.Attributes["twe"].Value);
            fc.ImageRequestCounts = long.Parse(facadeNode.Attributes["irc"].Value);

            XmlNodeList fromSiteNodes = facadeNode.SelectNodes("FromSite");
            foreach (XmlNode fromSiteNode in fromSiteNodes)
            {
                FromSiteCounts fsc = FromSiteCounts.LoadFromXml(fromSiteNode);
                fc.SiteCounts.Add(fsc.SiteNumber, fsc);
            }

            return fc;
        }
    }

    /// <summary>
    /// Represents the counts of requests that came from a specific site. 
    /// Depending on the SiteIdentificationSource of the facade this might count how many requests were made by a facade for data 
    /// from a site or requests made by a facade for data to a site
    /// </summary>
    public class FromSiteCounts
    {
        public string SiteNumber { get; private set; }
        public long TotalTransactions { get; set; }
        public long TransactionsWithExceptions { get; set; }
        public long TotalImageRequests { get; set; }
        public long TotalImageExceptions { get; set; }
        public long SuccessfulImageBytes { get; set; }
        public long SuccessfulImageTransferTimeMs { get; set; }
        public long DataSourceBytesReceived { get; set; }
        public long ImageCacheHits { get; set; }
        public long CachedBytesReturned { get; set; }
        public long BseLogins { get; set; }
        public long CapriLogins { get; set; }
        public Dictionary<string, ImageRequest> ImageRequests { get; private set; }
        public Dictionary<string, long> ProtocolCounts { get; private set; }

        public long ImageCacheResponses
        {
            get
            {
                long count = 0L;
                foreach (ImageRequest ir in ImageRequests.Values)
                {
                    count += ir.ImageCacheHits;
                }
                return count;
            }
        }

        public long SuccessfulImageResponses
        {
            get
            {
                long count = 0L;
                foreach (ImageRequest ir in ImageRequests.Values)
                {
                    count += ir.SuccessfulResponses;
                }
                return count;
            }
        }

        public long GetRelevantBytes(SiteIdentificationSource siteIdentificationSource)
        {
            if (siteIdentificationSource == SiteIdentificationSource.requestingSite)
                return SuccessfulImageBytes;
            else
                return DataSourceBytesReceived;
        }
    

        public double KbSec
        {
            get
            {
                return DailyTransactionsHelper.ConvertToKbSec(SuccessfulImageBytes, SuccessfulImageTransferTimeMs);
            }
        }

        public double PercentOfTotalTransactions(long totalFacadeTransactions)
        {
            if (totalFacadeTransactions >0)
                return (double)TotalTransactions / (double)totalFacadeTransactions;
            return 0f;
        }

        public FromSiteCounts(string siteNumber)
        {
            this.SiteNumber = siteNumber;
            ImageRequests = new Dictionary<string, ImageRequest>();
            TotalImageRequests = 0L;
            TotalTransactions = 0L;
            TransactionsWithExceptions = 0L;
            SuccessfulImageBytes = 0L;
            SuccessfulImageTransferTimeMs = 0L;
            TotalImageExceptions = 0L;
            DataSourceBytesReceived = 0L;
            ImageCacheHits = 0L;
            CachedBytesReturned = 0L;
            BseLogins = 0L;
            CapriLogins = 0L;
            ProtocolCounts = new Dictionary<string, long>();
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.AppendLine(SiteNumber + ": " + TotalTransactions + " total transactions, " + TransactionsWithExceptions + " transactions with exceptions, " + TotalImageRequests + " total image requests, " + SuccessfulImageBytes + " successful bytes, " + SuccessfulImageTransferTimeMs + " ms" );
            foreach (ImageRequest ir in ImageRequests.Values)
            {
                sb.Append("\t\t" + ir.ToString());
            }
            return sb.ToString();
        }

        public static FromSiteCounts LoadFromXml(XmlNode fromSiteNode)
        {
            string siteNumber = fromSiteNode.Attributes["Id"].Value;

            FromSiteCounts result = new FromSiteCounts(siteNumber);

            result.TotalTransactions = long.Parse(fromSiteNode.Attributes["t"].Value);
            result.TransactionsWithExceptions = long.Parse(fromSiteNode.Attributes["twe"].Value);
            result.TotalImageRequests = long.Parse(fromSiteNode.Attributes["tir"].Value);
            result.TotalImageExceptions = long.Parse(fromSiteNode.Attributes["tix"].Value);
            result.SuccessfulImageBytes = long.Parse(fromSiteNode.Attributes["sBytes"].Value);
            result.SuccessfulImageTransferTimeMs = long.Parse(fromSiteNode.Attributes["sTimeMs"].Value);
            if (fromSiteNode.Attributes["dsbr"] != null)
            {
                result.DataSourceBytesReceived = long.Parse(fromSiteNode.Attributes["dsbr"].Value);
            }

            if (fromSiteNode.Attributes["ich"] != null)
            {
                result.ImageCacheHits = long.Parse(fromSiteNode.Attributes["ich"].Value);
            }
            if (fromSiteNode.Attributes["cbr"] != null)
            {
                result.CachedBytesReturned = long.Parse(fromSiteNode.Attributes["cbr"].Value);
            }
            if (fromSiteNode.Attributes["bse"] != null)
            {
                result.BseLogins = long.Parse(fromSiteNode.Attributes["bse"].Value);
            }
            if (fromSiteNode.Attributes["capri"] != null)
            {
                result.CapriLogins = long.Parse(fromSiteNode.Attributes["capri"].Value);
            }

            XmlNodeList imageRequestNodes = fromSiteNode.SelectNodes("ImageRequest");
            foreach (XmlNode imageRequestNode in imageRequestNodes)
            {
                ImageRequest ir = ImageRequest.LoadFromXml(imageRequestNode);
                result.ImageRequests.Add(ir.Key, ir);
            }

            XmlNodeList protocolNodes = fromSiteNode.SelectNodes("Protocol");
            foreach (XmlNode protocolNode in protocolNodes)
            {
                string name = protocolNode.Attributes["n"].Value;
                long count = long.Parse(protocolNode.Attributes["c"].Value);
                result.ProtocolCounts.Add(name, count);
            }

            return result;
        }

        public XmlNode createXmlNode(XmlDocument xmlDoc, long totalFacadeTransactions)
        {
            XmlNode result = xmlDoc.CreateElement("FromSite");

            XmlAttribute attr = xmlDoc.CreateAttribute("Id");
            attr.Value = SiteNumber;
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("t");
            attr.Value = TotalTransactions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("twe");
            attr.Value = TransactionsWithExceptions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("tir");
            attr.Value = TotalImageRequests + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("tix");
            attr.Value = TotalImageExceptions + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("sBytes");
            attr.Value = SuccessfulImageBytes + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("sTimeMs");
            attr.Value = SuccessfulImageTransferTimeMs + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("kbs");
            attr.Value = KbSec.ToString("N2");
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("dsbr");
            attr.Value = DataSourceBytesReceived + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("tp"); // transactions percent
            double percentOfTotalTransactions = PercentOfTotalTransactions(totalFacadeTransactions);
            percentOfTotalTransactions *= 100.0f;
            attr.Value = percentOfTotalTransactions.ToString("N2");
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("ich");
            attr.Value = ImageCacheHits + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("cbr");
            attr.Value = CachedBytesReturned + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("bse");
            attr.Value = BseLogins + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("capri");
            attr.Value = CapriLogins + "";
            result.Attributes.Append(attr);

            foreach (ImageRequest ir in ImageRequests.Values)
            {
                result.AppendChild(ir.createXmlNode(xmlDoc));
            }

            foreach (string protocol in ProtocolCounts.Keys)
            {
                long value = ProtocolCounts[protocol];
                XmlNode protocolNode = xmlDoc.CreateElement("Protocol");
                result.AppendChild(protocolNode);

                attr = xmlDoc.CreateAttribute("n");
                attr.Value = protocol;
                protocolNode.Attributes.Append(attr);

                attr = xmlDoc.CreateAttribute("c");
                attr.Value = value + "";
                protocolNode.Attributes.Append(attr);
            }


            return result;
        }
    }

    public class ImageRequest
    {
        public string Modality { get; private set; }
        public string Quality { get; private set; }
        public ImageSource ImageSource { get; private set; }
        public long RequestCount { get; set; }
        public long SuccessfulResponses { get; set; }
        public long ExceptionResponses { get; set; }
        public long NotFoundResponses { get; set; } // 404 errors
        public long InternalServerErrorResponse { get; set; } // 500 errors
        public long BytesReturned { get; set; }
        public long TotalMs { get; set; }
        public long DataSourceBytesReceived { get; set; }
        public long ImageCacheHits { get; set; }
        public long CachedBytesReturned { get; set; }

        public ImageRequest(string modality, string quality, ImageSource imageSource)
        {
            this.Modality = modality;
            this.Quality = quality;
            this.ImageSource = imageSource;
            RequestCount = 0L;
            SuccessfulResponses = 0L;
            ExceptionResponses = 0L;
            NotFoundResponses = 0L;
            InternalServerErrorResponse = 0L;
            BytesReturned = 0L;
            TotalMs = 0L;
            DataSourceBytesReceived = 0L;
            ImageCacheHits = 0L;
            CachedBytesReturned = 0L;
        }

        public double KbSec
        {
            get
            {
                return DailyTransactionsHelper.ConvertToKbSec(BytesReturned, TotalMs);
            }
        }

        public string Key
        {
            get
            {
                return CreateKey(Modality, Quality, ImageSource);
            }
        }

        public static string CreateKey(string modality, string quality, ImageSource imageSource)
        {
            return modality + "_" + quality + "_" + imageSource;
        }

        public override string ToString()
        {
            return Key + ":" + RequestCount + " requests, " + SuccessfulResponses + " successful, " + ExceptionResponses + " exceptions, " + BytesReturned + " bytes, " + TotalMs + " ms" + Environment.NewLine;
        }

        public static ImageRequest LoadFromXml(XmlNode imageRequestNode)
        {
            string modality = imageRequestNode.Attributes["m"].Value;
            string quality = imageRequestNode.Attributes["q"].Value;
            ImageSource imageSource = ImageSource.va;
            if (imageRequestNode.Attributes["is"] != null)
            {
                switch (imageRequestNode.Attributes["is"].Value)
                {
                    case "haims_artifact":
                        imageSource = ImageSource.haims_artifact;
                        break;
                    case "haims_radiology":
                        imageSource = ImageSource.haims_radiology;
                        break;
                    case "pacsi":
                        imageSource = ImageSource.pacsi;
                        break;
                    case "ncat":
                        imageSource = ImageSource.ncat;
                        break;
                    case "va":
                        imageSource = ImageSource.va;
                        break;
                }
            }

            ImageRequest result = new ImageRequest(modality, quality, imageSource);

            result.RequestCount = long.Parse(imageRequestNode.Attributes["r"].Value);
            result.SuccessfulResponses = long.Parse(imageRequestNode.Attributes["sr"].Value);
            result.ExceptionResponses = long.Parse(imageRequestNode.Attributes["exResp"].Value);
            result.NotFoundResponses = long.Parse(imageRequestNode.Attributes["NFResp"].Value);
            result.InternalServerErrorResponse = long.Parse(imageRequestNode.Attributes["ISEResp"].Value);
            result.BytesReturned = long.Parse(imageRequestNode.Attributes["br"].Value);
            result.TotalMs = long.Parse(imageRequestNode.Attributes["Ms"].Value);
            if (imageRequestNode.Attributes["dsbr"] != null)
            {
                result.DataSourceBytesReceived = long.Parse(imageRequestNode.Attributes["dsbr"].Value);
            }
            if (imageRequestNode.Attributes["ich"] != null)
            {
                result.ImageCacheHits = long.Parse(imageRequestNode.Attributes["ich"].Value);
            }
            if (imageRequestNode.Attributes["cbr"] != null)
            {
                result.CachedBytesReturned = long.Parse(imageRequestNode.Attributes["cbr"].Value);
            }

            return result;
        }

        public XmlNode createXmlNode(XmlDocument xmlDoc)
        {
            XmlNode result = xmlDoc.CreateElement("ImageRequest");

            XmlAttribute attr = xmlDoc.CreateAttribute("m");
            attr.Value = Modality;
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("q");
            attr.Value = Quality;
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("is");
            attr.Value = ImageSource.ToString();
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("r");
            attr.Value = RequestCount + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("sr");
            attr.Value = SuccessfulResponses + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("exResp");
            attr.Value = ExceptionResponses + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("NFResp");
            attr.Value = NotFoundResponses + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("ISEResp");
            attr.Value = InternalServerErrorResponse + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("br");
            attr.Value = BytesReturned + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("Ms");
            attr.Value = TotalMs + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("kbs");
            attr.Value = KbSec.ToString("N2");
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("dsbr");
            attr.Value = DataSourceBytesReceived + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("ich");
            attr.Value = ImageCacheHits + "";
            result.Attributes.Append(attr);

            attr = xmlDoc.CreateAttribute("cbr");
            attr.Value = CachedBytesReturned + "";
            result.Attributes.Append(attr);

            return result;
        }
    }

    public enum SiteIdentificationSource
    {
        respondingSite, requestingSite
    }

    public enum DailyTransactionsType
    {
        day, month, year
    }

    public enum ImageSource
    {
        va, haims_radiology, pacsi, haims_artifact, ncat
    }
}
