/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Excel = Microsoft.Office.Interop.Excel;

namespace DailyTransactionLogAnalyzerEngine.reports
{
    public class GroupedReport : AbstractExcelReport
    {
        public GroupedReport()
        {
            this.Name = "Grouped Report";
            this.Description = "A report is generated where each site specified is grouped together (recommended for CVIX nodes)";
            this.ExcelOutputGroupType = ExcelOutputGroupType.allSites;
        }

        protected override void GenerateReport(string analysisDirectory, Dictionary<string, List<DailyTransactions>> sitesDailyTransactionsDictionary, ReportDateType reportDateType, 
            DateTime startDate, DateTime endDate)
        {
            throw new NotImplementedException();
        }

        protected override void GenerateReport(string analysisDirectory, DailyTransactions dailyTransactions, ReportDateType reportDateType, DateTime date)
        {
            throw new NotImplementedException();
        }

        protected override void GenerateReport(string analysisDirectory, List<DailyTransactions> dailyTransactionsList, 
            ReportDateType reportDateType, DateTime date)
        {
            switch (reportDateType)
            {
                case ReportDateType.month:
                    LogMsg("Generating Excel output for grouped sites for '" + OutputDate(date, reportDateType) + "'.");
                    OutputMultipleSitesMonthlyRollup(analysisDirectory, dailyTransactionsList, date, true);
                    break;
                case ReportDateType.year:                    
                    throw new NotImplementedException();        
                default:
                    throw new NotImplementedException();
            }
        }

        protected override ReportSiteType GetReportSiteSupportedType()
        {
            return ReportSiteType.cvix;
        }

        private void OutputMultipleSitesMonthlyRollup(string monthlyAnalysisDirectory, List<DailyTransactions> dailyTransactions,
            DateTime date, bool saveTotalsToPdf)
        {
            string filename = GetMonthlyRollupFilename(monthlyAnalysisDirectory, date.Year, date.Month);
            string dateDescription = date.ToString("MMMM yyyy");
            OutputMultipleSitesRollup(filename, dailyTransactions, dateDescription, saveTotalsToPdf);
        }

        private void OutputMultipleSitesRollup(string filename,
            List<DailyTransactions> dailyTransactionList, string dateDescription,
            bool saveTotalsToPdf)
        {
            FileInfo f = new FileInfo(filename);
            if (!f.Directory.Exists)
            {
                Directory.CreateDirectory(f.Directory.FullName);
            }
            Excel.Application oXL = new Excel.Application();
            oXL.Visible = false;

            Excel.Workbook workbook = null;
            try
            {
                List<string> siteIds = new List<string>();

                // always creating a new version from now on
                if (File.Exists(filename))
                    File.Delete(filename);

                workbook = oXL.Workbooks.Add(Type.Missing);
                // delete all existing sheets (Excel creates 3 by default)
                for (int i = workbook.Sheets.Count; i > 1; i--)
                {
                    Excel.Worksheet sheet = workbook.Sheets[i];
                    sheet.Delete();
                }
                List<Site> sites = SiteHelper.GetAllSites();
                for (int i = 0; i < sites.Count; i++)
                {
                    Site site = sites[i];
                    siteIds.Add(site.SiteId);
                }

                Excel.Worksheet totalsSheet = CreateTotalsSheet(workbook, siteIds, true);

                foreach (DailyTransactions dailyTransactions in dailyTransactionList)
                {
                    Dictionary<string, long> totalSiteCounts = new Dictionary<string, long>();
                    foreach (FacadeCounts fc in dailyTransactions.FacadeCounts.Values)
                    {
                        Dictionary<string, long> siteCounts = new Dictionary<string, long>();
                        Excel.Worksheet sheet = GetOrCreateFacadeWorksheet(workbook, fc.FacadeName, siteIds);

                        foreach (FromSiteCounts fsc in fc.SiteCounts.Values)
                        {
                            string sn = SiteHelper.CleanSiteNumber(fsc.SiteNumber);
                            if (siteCounts.ContainsKey(sn))
                            {
                                siteCounts[sn] += fsc.TotalTransactions;
                            }
                            else
                            {
                                siteCounts.Add(sn, fsc.TotalTransactions);
                            }
                            if (totalSiteCounts.ContainsKey(sn))
                            {
                                totalSiteCounts[sn] += fsc.TotalTransactions;
                            }
                            else
                            {
                                totalSiteCounts.Add(sn, fsc.TotalTransactions);
                            }
                        }

                        siteCounts = pruneToOnlyHighestCounts(siteCounts, 5);
                        int row = sheet.UsedRange.Rows.Count + 1;

                        for (int i = 0; i < siteIds.Count; i++)
                        {
                            string siteId = siteIds[i];
                            sheet.Cells[row, 1] = dailyTransactions.SiteNumber;
                            if (siteCounts.ContainsKey(siteId))
                            {
                                sheet.Cells[row, i + 2] = siteCounts[siteId];
                            }
                        }
                    }
                    totalSiteCounts = pruneToOnlyHighestCounts(totalSiteCounts, 5);
                    int totalRow = totalsSheet.UsedRange.Rows.Count + 1;
                    for (int i = 0; i < siteIds.Count; i++)
                    {
                        string siteId = siteIds[i];
                        totalsSheet.Cells[totalRow, 1] = dailyTransactions.SiteNumber;
                        if (totalSiteCounts.ContainsKey(siteId))
                        {
                            totalsSheet.Cells[totalRow, i + 2] = totalSiteCounts[siteId];
                        }
                    }
                }
                PruneEmptyColumns(workbook);
                updateSiteHeaderColumns(workbook);
                Excel.ChartObject totalsChart = CreateChartsReturnTotalsChart(workbook, dateDescription, null, true, ChartType.transactions);

                workbook.SaveAs(filename);
                if (totalsChart != null && saveTotalsToPdf)
                {
                    totalsChart.Activate();

                    totalsChart.Chart.ExportAsFixedFormat(Excel.XlFixedFormatType.xlTypePDF, filename + ".pdf",
                        Excel.XlFixedFormatQuality.xlQualityStandard, true, true, Type.Missing, Type.Missing, false, Type.Missing);
                }
            }
            finally
            {
                if (workbook != null)
                    workbook.Close();
                oXL.Quit();
                System.Runtime.InteropServices.Marshal.ReleaseComObject(workbook);
                System.Runtime.InteropServices.Marshal.ReleaseComObject(oXL);
                oXL = null;
                GC.Collect();
            }
        }

        private Excel.ChartObject CreateChartsReturnTotalsChart(Excel.Workbook workbook, string dateDescription, Site site,
            bool stacked, ChartType chartType)
        {
            Excel.ChartObject totalsChart = null;
            // create a chart for each work sheet except the sites one
            foreach (Excel.Worksheet sheet in workbook.Sheets)
            {
                if (sheet.Name == "Sites")
                    continue;
                Excel.ChartObject chart = CreateWorksheetChart(sheet, dateDescription, site, stacked, chartType);
                if (sheet.Name == "Totals")
                    totalsChart = chart;

            }
            return totalsChart;
        }

        private Excel.ChartObject CreateWorksheetChart(Excel.Worksheet sheet, string dateDescription, Site site,
            bool stacked, ChartType chartType)
        {
            return CreateWorksheetChart(sheet, dateDescription, site, stacked, chartType, false);
        }

        private Excel.Worksheet CreateTotalsSheet(Excel.Workbook workbook, List<string> siteIds, bool useExisting)
        {
            Excel.Worksheet totalsSheet = null;
            if (useExisting)
            {
                totalsSheet = workbook.ActiveSheet;
            }
            else
            {
                totalsSheet = workbook.Sheets.Add(Type.Missing, Type.Missing, Type.Missing, Type.Missing);
            }
            totalsSheet.Name = "Totals";
            OutputSitesRow(totalsSheet, 1, siteIds);
            return totalsSheet;
        }

        private Excel.Worksheet GetOrCreateFacadeWorksheet(Excel.Workbook workbook, string facadeName, List<string> siteIds)
        {
            foreach (Excel.Worksheet sheet in workbook.Sheets)
            {
                if (sheet.Name == facadeName)
                    return sheet;
            }
            int count = workbook.Sheets.Count;
            Excel.Worksheet newSheet = workbook.Sheets.Add(Type.Missing, workbook.Sheets[count], Type.Missing, Type.Missing);
            newSheet.Name = facadeName;

            OutputSitesRow(newSheet, 1, siteIds);

            return newSheet;
        }


        private void OutputSitesRow(Excel.Worksheet sheet, int row, List<string> siteIds)
        {
            for (int i = 0; i < siteIds.Count; i++)
            {
                sheet.Cells[row, i + 2] = siteIds[i];
            }
        }

        private Dictionary<string, long> pruneToOnlyHighestCounts(Dictionary<string, long> siteCounts, int maxToKeep)
        {
            if (siteCounts.Count <= maxToKeep)
                return siteCounts;
            Dictionary<string, long> result = new Dictionary<string, long>();
            List<SiteCount> siteCountsList = new List<SiteCount>();
            foreach (string siteId in siteCounts.Keys)
            {
                siteCountsList.Add(new SiteCount(siteId, siteCounts[siteId]));
            }

            siteCountsList.Sort(new SiteCountComparer());
            for (int i = 0; i < maxToKeep; i++)
            {
                SiteCount sc = siteCountsList[i];
                result.Add(sc.SiteId, sc.Count);
            }

            return result;
        }

        private void PruneEmptyColumns(Excel.Workbook workbook)
        {
            foreach (Excel.Worksheet sheet in workbook.Sheets)
            {
                PruneEmptyWorksheetColumns(sheet);
            }
        }

        private void PruneEmptyWorksheetColumns(Excel.Worksheet sheet)
        {
            for (int i = sheet.UsedRange.Columns.Count; i >= 1; i--)
            {
                bool keepColumn = false;
                // skip row 1 - it contains site numbers and is never empty
                for (int j = 2; j <= sheet.UsedRange.Rows.Count; j++)
                {
                    Excel.Range r = sheet.Cells[j, i];
                    string val = r.Value2 + "";
                    if (val != "")
                    {
                        // keep this column
                        keepColumn = true;
                        break;
                    }
                }
                if (!keepColumn)
                {
                    Excel.Range deleteRange = sheet.Cells[Type.Missing, i];
                    deleteRange.EntireColumn.Delete(Excel.XlDeleteShiftDirection.xlShiftToLeft);
                }
            }
        }
    }
}
