/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DailyTransactionLogAnalyzerEngine.reports
{
    public abstract class ReportGenerator
    {
        public string Name { get; protected set; }
        public string Description { get; protected set; }
        //public ExcelOutputOptionType ExcelOutputOptionType { get; protected set; }
        public ExcelOutputGroupType ExcelOutputGroupType { get; protected set; }
        public event LogMsgDelegate OnLogMsgEvent = null;
        
        protected void LogMsg(string msg)
        {
            if (OnLogMsgEvent != null)
                OnLogMsgEvent(msg);
        }

        protected abstract ReportSiteType GetReportSiteSupportedType();
                
        protected abstract void GenerateReport(string analysisDirectory, Dictionary<string, List<DailyTransactions>> sitesDailyTransactionsDictionary, 
            ReportDateType reportDateType, DateTime startDate, DateTime endDate);

        /// <summary>
        /// Generate a report for a specific date (year, month, day)
        /// </summary>
        /// <param name="analysisDirectory"></param>
        /// <param name="dailyTransactionsList">List of dailyTransactions, one each for a site containing the transactions for the specified date (year, month, day)</param>
        /// <param name="reportDateType"></param>
        /// <param name="date"></param>
        protected abstract void GenerateReport(string analysisDirectory, List<DailyTransactions> dailyTransactionsList, ReportDateType reportDateType, DateTime date);

        protected abstract void GenerateReport(string analysisDirectory, DailyTransactions dailyTransactions, ReportDateType reportDateType, DateTime date);

        public void GenerateReport(LogMsgDelegate onLogMsgEvent, string analysisDirectory, List<string> siteIds, DateTime startDate, 
            DateTime endDate, ReportDateType reportDateType)
        {
            DateTime date = startDate;
            this.OnLogMsgEvent = onLogMsgEvent;

            if (!ValidateSitesForSupportedSiteType(siteIds))
                return;

            LogMsg("Reading site analysis information");
            if (ExcelOutputGroupType == ExcelOutputGroupType.allDates)
            {
                
                Dictionary<string, List<DailyTransactions>> siteDailyTransactionsDictionary = new Dictionary<string, List<DailyTransactions>>();
                foreach (string siteNumber in siteIds)
                {
                    date = startDate;
                    List<DailyTransactions> siteDailyTransactions = new List<DailyTransactions>();
                    siteDailyTransactionsDictionary.Add(siteNumber, siteDailyTransactions);
                    while (date.CompareTo(endDate) < 1)
                    {
                        DailyTransactions dailyTransactions = LoadTransactions(analysisDirectory, siteNumber, date, reportDateType);
                        if (dailyTransactions != null)
                        {
                            siteDailyTransactions.Add(dailyTransactions);
                        }

                        date = IncrementDate(date, reportDateType);
                    }
                }
                GenerateReport(analysisDirectory, siteDailyTransactionsDictionary, reportDateType, startDate, endDate);
            }
            else
            {
                while (date.CompareTo(endDate) < 1)
                {
                    List<DailyTransactions> dailyTransactionsList = new List<DailyTransactions>();
                    foreach (string siteNumber in siteIds)
                    {
                        DailyTransactions dailyTransactions = LoadTransactions(analysisDirectory, siteNumber, date, reportDateType);
                        if (dailyTransactions != null)
                        {
                            if (this.ExcelOutputGroupType == ExcelOutputGroupType.none)
                            {
                                GenerateReport(analysisDirectory, dailyTransactions, reportDateType, date);
                            }
                            else
                            {
                                dailyTransactionsList.Add(dailyTransactions);
                            }
                        }
                    }
                    if (this.ExcelOutputGroupType != ExcelOutputGroupType.none)
                    {
                        GenerateReport(analysisDirectory, dailyTransactionsList, reportDateType, date);
                    }
                    date = IncrementDate(date, reportDateType);
                }
            }
        }

        private bool ValidateSitesForSupportedSiteType(List<string> siteIds)
        {
            ReportSiteType reportSiteType = GetReportSiteSupportedType();
            if (reportSiteType == ReportSiteType.vix_and_cvix)
            {
                // all good
                return true;
            }
            else if(reportSiteType == ReportSiteType.vix)
            {
                 // vix sites only
                foreach (string siteId in siteIds)
                {
                    if(!SiteHelper.IsSiteVix(siteId))
                    {
                        LogMsg("Site [" + siteId + "] is not of the supported type [" + reportSiteType + "]");
                        return false;
                    }
                }
            }
            else if (reportSiteType == ReportSiteType.cvix)
            {
                // vix sites only
                foreach (string siteId in siteIds)
                {
                    if (!SiteHelper.IsSiteCvix(siteId))
                    {
                        LogMsg("Site [" + siteId + "] is not of the supported type [" + reportSiteType + "]");
                        return false;
                    }
                }
            }
            return true;
        }

        private DailyTransactions LoadTransactions(string analysisDirectory, string siteNumber, DateTime date, ReportDateType reportDateType)
        {
            switch (reportDateType)
            {
                case ReportDateType.day:
                    return DailyTransactions.LoadFromXml(analysisDirectory, siteNumber, date);
                case ReportDateType.month:
                    return DailyTransactions.LoadFromXml(analysisDirectory, siteNumber,
                           date.Year, date.Month);
                case ReportDateType.year:
                    return DailyTransactions.LoadFromXml(analysisDirectory, siteNumber, date.Year);
            }
            return null;
        }

        protected string OutputDate(DateTime date, ReportDateType reportDateType)
        {
            switch (reportDateType)
            {
                case ReportDateType.day:
                    return date.ToString();
                case ReportDateType.month:
                    return date.ToString("MMMM yyyy");
                case ReportDateType.year:
                    return date.ToString("yyyy");
            }
            return "";
        }

        private DateTime IncrementDate(DateTime date, ReportDateType reportDateType)
        {
            switch (reportDateType)
            {
                case ReportDateType.day:
                    return date.AddDays(1);
                case ReportDateType.month:
                    return date.AddMonths(1);
                case ReportDateType.year:
                    return date.AddYears(1);
            }
            return date;
        }

        public override string ToString()
        {
            return Name;
        }

        protected Dictionary<string, Site> GetSitesDictionary()
        {
            List<Site> sites = SiteHelper.GetAllSites();
            Dictionary<string, Site> siteDictionary = new Dictionary<string, Site>();
            foreach (Site site in sites)
            {
                siteDictionary.Add(site.SiteId, site);
            }
            return siteDictionary;
        }
    }

    public enum ReportDateType
    {
        day, month, year
    }

    public class SiteDailyTransactionsList
    {
        public string SiteNumber { get; private set; }
        public string SiteName { get; private set; }
        public List<DailyTransactions> DailyTransactionsList { get; private set; }

        public SiteDailyTransactionsList(string siteNumber, string siteName, List<DailyTransactions> dailyTransactionsList)
        {
            this.SiteNumber = siteNumber;
            this.SiteName = siteName;
            this.DailyTransactionsList = dailyTransactionsList;
        }
    }

    public class SiteDailyTransactionsListComparer : IComparer<SiteDailyTransactionsList>
    {
        public int Compare(SiteDailyTransactionsList x, SiteDailyTransactionsList y)
        {
            //return (-1) * (x.SiteName.CompareTo(y.SiteName));
            return (x.SiteName.CompareTo(y.SiteName));
        }
    }

    public enum ExcelOutputOptionType
    {
        individualSites, groupedCvix, groupedTotals, nodesAndEdges, siteTotals
    }

    public enum ExcelOutputGroupType
    {
        none, allSites, allDates
    }

    public enum ReportSiteType
    {
        vix, cvix, vix_and_cvix
    }

    /*
    public class ReportOption
    {
        public string Description { get; set; }
        public bool Checked { get; set; }

        public ReportOption()
        {
            Checked = false;
        }

        public ReportOption(string description, bool isChecked)
        {
            this.Description = description;
            this.Checked = isChecked;
        }
    }*/
}
