/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Excel = Microsoft.Office.Interop.Excel;
using PowerPoint = Microsoft.Office.Interop.PowerPoint;
using Microsoft.Office.Core;

namespace DailyTransactionLogAnalyzerEngine.reports
{
    /// <summary>
    /// Creates a PDF of slides, two for each site that shows the top 5 requests and bytes transfered between sites
    /// </summary>
    public class SiteCommunicationReport : AbstractExcelReport
    {
        public SiteCommunicationReport()
        {
            this.Name = "Site Communication Report";
            this.Description = "A PDF report is generated with a page for the totals of each site";
            this.ExcelOutputGroupType = ExcelOutputGroupType.allSites;
        }

        protected override ReportSiteType GetReportSiteSupportedType()
        {
            return ReportSiteType.vix;
        }

        protected override void GenerateReport(string analysisDirectory, Dictionary<string, List<DailyTransactions>> sitesDailyTransactionsDictionary, ReportDateType reportDateType, 
            DateTime startDate, DateTime endDate)
        {
            throw new NotImplementedException();
        }

        protected override void GenerateReport(string analysisDirectory, List<DailyTransactions> dailyTransactionsList, 
            ReportDateType reportDateType, DateTime date)
        {
            switch (reportDateType)
            {
                case ReportDateType.month:
                    LogMsg("Generating PDF output for selected sites for '" + OutputDate(date, reportDateType) + "'.");
                    OutputMultipleSitesMonthlyRollupToPdf(analysisDirectory, dailyTransactionsList, date);
                    break;
                case ReportDateType.year:
                    LogMsg("Generating PDF output for selected sites for '" + OutputDate(date, reportDateType) + "'.");
                    OutputMultipleSitesYearlyRollupToPdf(analysisDirectory, dailyTransactionsList, date);
                    break;
                default:
                    throw new NotImplementedException();
            };
        }

        private void OutputMultipleSitesYearlyRollupToPdf(string monthlyAnalysisDirectory, List<DailyTransactions> dailyTransactionList,
            DateTime date)
        {
            string dateDescription = date.ToString("yyyy");
            string filename = GetYearlyRollupFilename(monthlyAnalysisDirectory, date.Year);
            OutputMultipleSitesRollupToPdf(filename, dateDescription, dailyTransactionList);
        }

        private void OutputMultipleSitesMonthlyRollupToPdf(string monthlyAnalysisDirectory, List<DailyTransactions> dailyTransactionList,
            DateTime date)
        {
            string dateDescription = date.ToString("MMMM yyyy");
            string filename = GetMonthlyRollupFilename(monthlyAnalysisDirectory, date.Year, date.Month);
            OutputMultipleSitesRollupToPdf(filename, dateDescription, dailyTransactionList);
        }        

        private string GetYearlyRollupFilename(string monthlyAnalysisDirectory,
            int year)
        {
            return monthlyAnalysisDirectory + "\\Rollup\\" + year + "\\" + year + ".xlsx";
        }

        private void OutputMultipleSitesRollupToPdf(string filename, string dateDescription, List<DailyTransactions> dailyTransactionList)
        {
            filename = filename.Replace(".xlsx", ".pdf");
            FileInfo f = new FileInfo(filename);
            if (!f.Directory.Exists)
            {
                f.Directory.Create();
            }

            if (File.Exists(filename))
                File.Delete(filename);

            Excel.Application oXL = new Excel.Application();
            oXL.Visible = false;

            PowerPoint.Application ppt = new PowerPoint.Application();

            PowerPoint.Presentation pptPres = ppt.Presentations.Add(MsoTriState.msoFalse);

            Excel.Workbook workbook = null;
            try
            {
                List<string> siteIds = new List<string>();

                workbook = oXL.Workbooks.Add(Type.Missing);
                // delete all existing sheets except first (Excel creates 3 by default, must keep at least 1)                
                for (int i = workbook.Sheets.Count; i > 1; i--)
                {
                    Excel.Worksheet sheet = workbook.Sheets[i];
                    sheet.Delete();
                }
                List<Site> sites = SiteHelper.GetAllSites();
                for (int i = 0; i < sites.Count; i++)
                {
                    Site site = sites[i];
                    siteIds.Add(site.SiteId);
                }

                Dictionary<string, Site> sitesDictionary = GetSitesDictionary();
                Dictionary<string, FacadeTotals> totalFacadeCounts = new Dictionary<string, FacadeTotals>();
                
                Dictionary<string, Dictionary<string, FacadeTotals>> facadeSiteCounts = new Dictionary<string, Dictionary<string, FacadeTotals>>();

                List<SiteDailyTansactions> siteDailyTransactionsList = SortByName(dailyTransactionList, sitesDictionary);
                foreach (SiteDailyTansactions siteDailyTransactions in siteDailyTransactionsList)
                {
                    LogMsg("Calculating totals for site '" + siteDailyTransactions.SiteName + "'");
                    DailyTransactions dailyTransactions = siteDailyTransactions.DailyTransactions;
                    Site thisSite = SiteHelper.GetSite(dailyTransactions.SiteNumber);
                    Excel.Worksheet sheet = workbook.Sheets.Add(Type.Missing, Type.Missing, Type.Missing, Type.Missing);
                    sheet.Name = thisSite.ToString(true);
                    OutputSitesRow(sheet, 1, siteIds);
                    OutputHighValuesRow(sheet, 2);

                    int count = workbook.Sheets.Count;
                    Excel.Worksheet bytesSheet = workbook.Sheets.Add(Type.Missing, workbook.Sheets[count], Type.Missing, Type.Missing);
                    bytesSheet.Name = thisSite.SiteId + "_bytes";
                    OutputSitesRow(bytesSheet, 1, siteIds);
                    OutputHighValuesRow(bytesSheet, 2);

                    count = workbook.Sheets.Count;
                    Excel.Worksheet imagesSheet = workbook.Sheets.Add(Type.Missing, workbook.Sheets[count], Type.Missing, Type.Missing);
                    imagesSheet.Name = thisSite.SiteId + "_images";
                    OutputSitesRow(imagesSheet, 1, siteIds);
                    OutputHighValuesRow(imagesSheet, 2);

                    long totalSiteBytes = 0L; // total bytes for this site to output below chart
                    long totalSuccessfulImageResponses = 0L;

                    foreach (FacadeCounts fc in dailyTransactions.FacadeCounts.Values)
                    {
                        Dictionary<string, FacadeTotals> siteCounts = new Dictionary<string, FacadeTotals>();

                        long facadeBytes = 0L;
                        if (fc.SiteIdentificationSource == SiteIdentificationSource.requestingSite)
                            facadeBytes = fc.SuccessfulImageBytes;
                        else
                            facadeBytes = fc.DataSourceBytesReceived;

                        if (!facadeSiteCounts.ContainsKey(fc.FacadeName))
                        {
                            facadeSiteCounts.Add(fc.FacadeName, new Dictionary<string, FacadeTotals>());
                        }
                        if (!facadeSiteCounts[fc.FacadeName].ContainsKey(dailyTransactions.SiteNumber))
                        {
                            facadeSiteCounts[fc.FacadeName].Add(dailyTransactions.SiteNumber, new FacadeTotals());
                        }
                        facadeSiteCounts[fc.FacadeName][dailyTransactions.SiteNumber].TotalRequests += fc.TotalTransactions;
                        facadeSiteCounts[fc.FacadeName][dailyTransactions.SiteNumber].TotalBytes += facadeBytes;
                        facadeSiteCounts[fc.FacadeName][dailyTransactions.SiteNumber].SuccessfulImageResponses += fc.SuccessfulImageResponses;

                        foreach (FromSiteCounts fsc in fc.SiteCounts.Values)
                        {
                            string sn = SiteHelper.CleanSiteNumber(fsc.SiteNumber);
                            long bytes = 0L;
                            if (fc.SiteIdentificationSource == SiteIdentificationSource.requestingSite)
                                bytes = fsc.SuccessfulImageBytes;
                            else
                                bytes = fsc.DataSourceBytesReceived;
                            totalSiteBytes += bytes;

                            if (!totalFacadeCounts.ContainsKey(fc.FacadeName))
                            {
                                totalFacadeCounts.Add(fc.FacadeName, new FacadeTotals());
                            }

                            totalFacadeCounts[fc.FacadeName].TotalBytes += bytes;
                            totalFacadeCounts[fc.FacadeName].TotalRequests += fsc.TotalTransactions;
                            totalFacadeCounts[fc.FacadeName].SuccessfulImageResponses += fsc.SuccessfulImageResponses;
                            totalSuccessfulImageResponses += fsc.SuccessfulImageResponses;
                            
                            if (sn.Length > 0)
                            {
                                if (!siteCounts.ContainsKey(sn))
                                {
                                    siteCounts.Add(sn, new FacadeTotals());
                                }
                                siteCounts[sn].TotalRequests += fsc.TotalTransactions;
                                siteCounts[sn].TotalBytes += bytes;
                                siteCounts[sn].SuccessfulImageResponses += fsc.SuccessfulImageResponses;
                            }
                        }
                        Dictionary<string, HighValuedCount> siteTransactionsHighValuesCount = markHighValueCounts(siteCounts, 5, ChartType.transactions);
                        Dictionary<string, HighValuedCount> siteBytesHighValuesCount = markHighValueCounts(siteCounts, 5, ChartType.bytes);
                        Dictionary<string, HighValuedCount> siteImagesHighValuesCount = markHighValueCounts(siteCounts, 5, ChartType.imageCounts);                        

                        int totalRow = sheet.UsedRange.Rows.Count + 1;
                        sheet.Cells[totalRow, 1] = fc.FacadeName;
                        int bytesRow = bytesSheet.UsedRange.Rows.Count + 1;
                        bytesSheet.Cells[bytesRow, 1] = fc.FacadeName;
                        int imagesRow = imagesSheet.UsedRange.Rows.Count + 1;
                        imagesSheet.Cells[imagesRow, 1] = fc.FacadeName;
                        for (int i = 0; i < siteIds.Count; i++)
                        {
                            string siteId = siteIds[i];
                            if (siteTransactionsHighValuesCount.ContainsKey(siteId))
                            {
                                HighValuedCount hvc = siteTransactionsHighValuesCount[siteId];
                                if (hvc.Count > 0)
                                {
                                    sheet.Cells[totalRow, i + 2] = hvc.Count;
                                }
                                if (hvc.HighValue)
                                {
                                    sheet.Cells[2, i + 2] = "T";
                                }
                            }
                            if (siteBytesHighValuesCount.ContainsKey(siteId))
                            {
                                HighValuedCount hvc = siteBytesHighValuesCount[siteId];
                                if (hvc.Count > 0)
                                {
                                    bytesSheet.Cells[bytesRow, i + 2] = hvc.Count;
                                }
                                if (hvc.HighValue)
                                {
                                    bytesSheet.Cells[2, i + 2] = "T";
                                }
                            }
                            if (siteImagesHighValuesCount.ContainsKey(siteId))
                            {
                                HighValuedCount hvc = siteImagesHighValuesCount[siteId];
                                if (hvc.Count > 0)
                                {
                                    imagesSheet.Cells[imagesRow, i + 2] = hvc.Count;
                                }
                                if (hvc.HighValue)
                                {
                                    imagesSheet.Cells[2, i + 2] = "T";
                                }
                            }
                        }
                    }

                    LogMsg("Generating charts for site '" + siteDailyTransactions.SiteName + "'");
                    PruneNonHighValueColumns(sheet);
                    PruneNonHighValueColumns(bytesSheet);
                    PruneNonHighValueColumns(imagesSheet);
                    ConvertBytesValuesToMB(bytesSheet);

                    updateSiteHeaderColumns(sheet, sitesDictionary);
                    updateSiteHeaderColumns(bytesSheet, sitesDictionary);
                    updateSiteHeaderColumns(imagesSheet, sitesDictionary);
                    Excel.ChartObject transactionsChart = CreateWorksheetChart(sheet, dateDescription, thisSite, false, ChartType.transactions);
                    string chartDescription = CreateDescriptionForIndividualSite(facadeSiteCounts,
                        thisSite.SiteId, ChartType.transactions) + "Total Requests: " + dailyTransactions.TransactionCounts.TotalParentTransactions.ToString("N0");
                    OutputChartToPpt(pptPres, transactionsChart, chartDescription);

                    Excel.ChartObject bytesChart = CreateWorksheetChart(bytesSheet, dateDescription, thisSite, false, ChartType.bytes);
                    chartDescription = CreateDescriptionForIndividualSite(facadeSiteCounts,
                        thisSite.SiteId, ChartType.bytes) + "Total Transferred: " + DailyTransactionsHelper.formatBytes(totalSiteBytes);
                    OutputChartToPpt(pptPres, bytesChart, chartDescription);

                    Excel.ChartObject imagesChart = CreateWorksheetChart(imagesSheet, dateDescription, thisSite, false, ChartType.imageCounts);
                    chartDescription = CreateDescriptionForIndividualSite(facadeSiteCounts,
                        thisSite.SiteId, ChartType.imageCounts) + "Total Images: " + totalSuccessfulImageResponses.ToString("N0");
                    OutputChartToPpt(pptPres, imagesChart, chartDescription);
                }

                LogMsg("Generating title and totals charts");
                PowerPointHelper.CreateSiteRelationsTitleSlide(pptPres, dateDescription); // slide 1
                CreateExplanationSlide(pptPres);
                CreateTotalsSlide(pptPres, totalFacadeCounts, 3); // slide 3

                // create a sheet/chart that shows the total transactions for all sites for each facade
                CreateTotalsSheetAndChart(workbook, siteIds, facadeSiteCounts, ChartType.transactions, dateDescription, pptPres, sitesDictionary, 4);

                // create a sheet/chart that shows the total bytes for all sites for each facade
                CreateTotalsSheetAndChart(workbook, siteIds, facadeSiteCounts, ChartType.bytes, dateDescription, pptPres, sitesDictionary, 5);

                // create a sheet/chart that shows the total successful image responses for all sites for each facade
                CreateTotalsSheetAndChart(workbook, siteIds, facadeSiteCounts, ChartType.imageCounts, dateDescription, pptPres, sitesDictionary, 6);

                if (File.Exists(filename + ".xlsx"))
                    File.Delete(filename + ".xlsx");
                workbook.SaveAs(filename + ".xlsx");

                pptPres.SaveAs(filename, PowerPoint.PpSaveAsFileType.ppSaveAsPDF, MsoTriState.msoFalse);
                string pptFilename = filename + ".pptx";
                if (File.Exists(pptFilename))
                    File.Delete(pptFilename);
                pptPres.SaveAs(pptFilename);
            }
            finally
            {
                if (workbook != null)
                    workbook.Close(false);
                if (pptPres != null)
                    pptPres.Close();
                oXL.Quit();
                ppt.Quit();
                System.Runtime.InteropServices.Marshal.ReleaseComObject(workbook);
                System.Runtime.InteropServices.Marshal.ReleaseComObject(oXL);
                System.Runtime.InteropServices.Marshal.ReleaseComObject(ppt);
                System.Runtime.InteropServices.Marshal.ReleaseComObject(pptPres);

                oXL = null;
                GC.Collect();
            }
        }

        private void OutputSitesRow(Excel.Worksheet sheet, int row, List<string> siteIds)
        {
            for (int i = 0; i < siteIds.Count; i++)
            {
                sheet.Cells[row, i + 2] = siteIds[i];
            }
        }

        private void OutputHighValuesRow(Excel.Worksheet sheet, int row)
        {
            sheet.Cells[row, 1] = "High Value";
        }    

        private void CreateTotalsSheetAndChart(Excel.Workbook workbook, List<string> siteIds,
            Dictionary<string, Dictionary<string, FacadeTotals>> facadeSiteCounts, ChartType chartType, string dateDescription,
            PowerPoint.Presentation pptPres, Dictionary<string, Site> sitesDictionary, int slideIndex)
        {
            Excel.Worksheet sheet = workbook.Sheets.Add(Type.Missing, Type.Missing, Type.Missing, Type.Missing);
            string chartDescription = "";
            switch (chartType)
            {
                case ChartType.transactions:
                    sheet.Name = "Total Requests";
                    chartDescription = "Total requests made at each site by each facade";
                    break;
                case ChartType.bytes:
                    sheet.Name = "Total Bytes";
                    chartDescription = "Total bytes transfered over the WAN on behalf of each facade at each site";
                    break;
                case ChartType.imageCounts:
                    sheet.Name = "Total Successful Images";
                    chartDescription = "Total successful images responded with by each facade at each site";
                    break;
            }
            OutputSitesRow(sheet, 1, siteIds);
            OutputHighValuesRow(sheet, 2);

            foreach (string facadeName in facadeSiteCounts.Keys)
            {
                Dictionary<string, HighValuedCount> facadeCount = markHighValueCounts(facadeSiteCounts[facadeName], 5, chartType);
                int row = sheet.UsedRange.Rows.Count + 1;
                sheet.Cells[row, 1] = facadeName;
                for (int i = 0; i < siteIds.Count; i++)
                {
                    string siteId = siteIds[i];
                    if (facadeCount.ContainsKey(siteId))
                    {
                        HighValuedCount hvc = facadeCount[siteId];
                        if (hvc.Count > 0)
                        {
                            sheet.Cells[row, i + 2] = hvc.Count;
                        }
                        if (hvc.HighValue)
                        {
                            sheet.Cells[2, i + 2] = "T";
                        }
                    }
                }
            }
            PruneNonHighValueColumns(sheet);
            if (chartType == ChartType.bytes)
                ConvertBytesValuesToMB(sheet);
            updateSiteHeaderColumns(sheet, sitesDictionary);
            Excel.ChartObject totalTransactionsChart = CreateWorksheetChart(sheet, dateDescription, null, false, chartType);
            OutputChartToPpt(pptPres, totalTransactionsChart, slideIndex, chartDescription);
        }        

        private void CreateExplanationSlide(PowerPoint.Presentation presentation)
        {
            string title = "Statistics Explanation";
            StringBuilder description = new StringBuilder();
            description.AppendLine("These statistics are based on the transactions collected from the VIX servers");
            description.AppendLine("Transactions are small actions by the VIX. A request is initiated by a client. These stats are based on requests which might generate multiple transactions");
            description.AppendLine("Federation requests are requests from another VIX or the CVIX");
            description.AppendLine("Clinical Display requests are requests made by Clinical Display");
            description.AppendLine("VistARad requests are requests made by VistARad");
            description.AppendLine("Image requests for cached data are not included in the bytes calculations");
            description.AppendLine("The totals overlap - a Clinical Display request might initiate a Federation request from another VIX. Summing the requests or bytes does not create a logical number");
            PowerPointHelper.CreateContentSlide(presentation, 2, title, description.ToString(), true, 22);
        }

        private string CreateDescriptionForIndividualSite(Dictionary<string, Dictionary<string, FacadeTotals>> allFacadeCounts, string siteId, ChartType chartType)
        {
            StringBuilder sb = new StringBuilder();

            foreach (string facadeName in allFacadeCounts.Keys)
            {
                Dictionary<string, FacadeTotals> facadeSiteCounts = allFacadeCounts[facadeName];
                if (facadeSiteCounts.ContainsKey(siteId))
                {
                    switch (chartType)
                    {
                        case ChartType.bytes:
                            sb.AppendLine(facadeName + ": " + DailyTransactionsHelper.formatBytes(facadeSiteCounts[siteId].TotalBytes));
                            break;
                        case ChartType.transactions:
                            sb.AppendLine(facadeName + ": " + facadeSiteCounts[siteId].TotalRequests.ToString("N0"));
                            break;
                        case ChartType.imageCounts:
                            sb.AppendLine(facadeName + ": " + facadeSiteCounts[siteId].SuccessfulImageResponses.ToString("N0"));
                            break;
                    }
                }
            }
            return sb.ToString();
        }

        private List<SiteDailyTansactions> SortByName(List<DailyTransactions> dailyTransactionsList, Dictionary<string, Site> sites)
        {
            List<SiteDailyTansactions> siteDailyTransactions = new List<SiteDailyTansactions>();
            foreach (DailyTransactions dailyTransaction in dailyTransactionsList)
            {
                string siteName = "";
                if (sites.ContainsKey(dailyTransaction.SiteNumber))
                {
                    siteName = sites[dailyTransaction.SiteNumber].Name;
                }
                siteDailyTransactions.Add(new SiteDailyTansactions(siteName, dailyTransaction));
            }
            siteDailyTransactions.Sort(new SiteDailyTansactionsComparer());
            return siteDailyTransactions;
        }

        /// <summary>
        /// This marks the highest value sites and returns all sites
        /// </summary>
        /// <param name="siteCounts"></param>
        /// <param name="maxToKeep"></param>
        /// <returns></returns>
        private Dictionary<string, HighValuedCount> markHighValueCounts(Dictionary<string, FacadeTotals> siteCounts, int maxToKeep, ChartType chartType)
        {
            Dictionary<string, HighValuedCount> result = new Dictionary<string, HighValuedCount>();
            List<SiteCount> siteCountsList = new List<SiteCount>();
            foreach (string siteId in siteCounts.Keys)
            {
                long value = 0L;
                switch (chartType)
                {
                    case ChartType.bytes:
                        value = siteCounts[siteId].TotalBytes;
                        break;
                    case ChartType.transactions:
                        value = siteCounts[siteId].TotalRequests;
                        break;
                    case ChartType.imageCounts:
                        value = siteCounts[siteId].SuccessfulImageResponses;
                        break;
                }

                siteCountsList.Add(new SiteCount(siteId, value));
            }

            siteCountsList.Sort(new SiteCountComparer());
            for (int i = 0; i < siteCounts.Count; i++)
            {
                SiteCount sc = siteCountsList[i];
                if (i < maxToKeep)
                {
                    result.Add(sc.SiteId, new HighValuedCount(sc.Count, true));
                }
                else
                {
                    result.Add(sc.SiteId, new HighValuedCount(sc.Count, false));
                }
                //result.Add(
                //result.Add(sc.SiteId, sc.Count);
            }

            return result;
        }

        private void PruneNonHighValueColumns(Excel.Worksheet sheet)
        {
            for (int i = sheet.UsedRange.Columns.Count; i > 1; i--)
            {
                bool keepColumn = false;

                // row 2 contains an T if the avlue should be kept
                Excel.Range r = sheet.Cells[2, i];
                string val = r.Value2 + "";
                if (val == "T")
                {
                    keepColumn = true;
                }
                if (!keepColumn)
                {
                    Excel.Range deleteRange = sheet.Cells[Type.Missing, i];
                    deleteRange.EntireColumn.Delete(Excel.XlDeleteShiftDirection.xlShiftToLeft);
                }
            }
            // now remove the second row since it is no longer useful

            Excel.Range secondRow = sheet.Cells[2, 1];//3, Type.Missing];//  [2, Type.Missing];
            secondRow.EntireRow.Delete(Excel.XlDeleteShiftDirection.xlShiftUp);
            //secondRow.EntireRow.Delete(Type.Missing);

        }

        protected override void GenerateReport(string analysisDirectory, DailyTransactions dailyTransactions, ReportDateType reportDateType, DateTime date)
        {
            throw new NotImplementedException();
        }
    }
}
