/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
//using SiteService;
//using VISACommon;
using VixLogAnalyzerEngine;

namespace TransactionEntryTracer
{
    public partial class frmMain : Form
    {
        private TransactionSearchResult SearchResults;
        private List<VASite> VaSites;

        public frmMain()
        {
            InitializeComponent();
        }

        private void frmMain_Load(object sender, EventArgs e)
        {
            tmrStart.Enabled = true;
        }

        private void tmrStart_Tick(object sender, EventArgs e)
        {
            tmrStart.Enabled = false;
            btnFindAssociatedTransactions.Enabled = false;
            LoadSites();
        }

        private void LoadSites()
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                StatusMessage("Loading Sites", false);
                SiteServiceHelper siteServiceHelper = new SiteServiceHelper();
                VaSites =
                    siteServiceHelper.GetVASites(true);
                    
                foreach (VASite site in VaSites)
                {
                    if(site.HasVix)
                    //if (site.IsVixOrCvix)
                    {
                        cboSites.Items.Add(site);
                    }                    
                }
                StatusMessage("Loaded '" + VaSites.Count + "' sites", false);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void StatusMessage(string msg, bool error)
        {
            tsStatus.Text = msg;
            Application.DoEvents();
        }

        private void btnSearchTransaction_Click(object sender, EventArgs e)
        {
            if (cboSites.SelectedItem != null)
            {
                VASite site = (VASite)cboSites.SelectedItem;
                int count = InitialSearchTransaction(site, txtTransactionId.Text.Trim(), dtDate.Value);
                if (count == 0)
                {
                    // -1 means error
                    MessageBox.Show("Did not find any transactions that matched ID");
                }
                else
                {
                    btnFindAssociatedTransactions.Enabled = true;
                    if (chkAutoFindAssociatedTransactions.Checked)
                        FindAssociatedTransactions();
                }
            }
            else
            {
                MessageBox.Show("Please select a site to retrieve the transaction from");
            }
        }

        private int InitialSearchTransaction(VASite site, string transactionId, DateTime date)
        {
            SearchResults = new TransactionSearchResult();
            transactionLogViewer1.Clear();
            
            return SearchTransaction(site, transactionId, date);
        }

        private int SearchTransaction(VASite site, string transactionId, DateTime date)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                StatusMessage("Searching for transaction '" + transactionId + "' from site '" + site.SiteName + "'.", false);
                SearchResults.SearchedSites.Add(site);
                List<TransactionLogEntry> transactions =
                    VixLogAnalyzerEngine.TransactionLogEntryDownloader.getTransaction(site.VixServer, 
                    site.VixPort, site.SiteNumber, transactionId, date);
                SearchResults.AddUnsearchedTransactions(transactions, site);

                transactionLogViewer1.DisplayTransactionLogEntries(transactionId, date, site, transactions);
                return transactions.Count;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
                return -1;
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void btnFindAssociatedTransactions_Click(object sender, EventArgs e)
        {
            FindAssociatedTransactions();
        }

        private void FindAssociatedTransactions()
        {
            StatusMessage("Searching for associated transactions", false);
            btnFindAssociatedTransactions.Enabled = false;
            List<SearchedTransaction> tempTransactions = new List<SearchedTransaction>();
            tempTransactions.AddRange(SearchResults.Transactions);

            foreach (SearchedTransaction searchTransaction in tempTransactions)
            {
                if (!searchTransaction.Searched)
                {
                    //
                    FindAssociatedTransactions(searchTransaction.Transaction);
                    searchTransaction.Searched = true;
                }
            }
        }

        private void FindAssociatedTransactions(TransactionLogEntry transaction)
        {
            /**
             * 4 possible cases (well really 6):
             * Transaction originated on a VIX to get data from another VIX
                * use Responding Site, it will start with 1.3.6.1.4.1.3768,XXX meaning VA site, just use site number portion
             * Transaction originated on a VIX to respond to another VIX
                *  use Requesting VIX Site Number to find requesting VIX
             * Transaction originated on a VIX to get data from a CVIX
                *  use Responding Site, it will contain 2.16.840.1.113883.3.42.10012.100001.207,200 which means it is DoD data so look at Data Source Response Server for specific node of CVIX
             * Transaction originated on a CVIX to response to a VIX
                *  Requesting VIX Site Number to find requesting VIX
             * 
             * (not handling requests that originated on the CVIX to get data from a VIX - probably can't trace those both directions)
             * 
             * 
             */

            VASite site = null;
            if (!String.IsNullOrEmpty(transaction.RespondingSite))
            {
                string number = transaction.RespondingSite;
                if (number.StartsWith("1.3.6.1.4.1.3768,"))
                {
                    number = number.Substring(17);
                    site = FindSiteByNumber(number);
                }
                else
                {
                    // response didn't come from VA which mean it went through the CVIX, use the Data Source Response Server for the specific node
                    if (!string.IsNullOrEmpty(transaction.DataSourceResponseServer))
                    {
                        site = FindSiteByHostname(transaction.DataSourceResponseServer);
                    }
                }
            }
            // if we didn't find a site or it was already searched, look elsewhere
            if (site == null || IsSiteSearched(site))
            {
                if (transaction.Values.ContainsKey("Requesting VIX Site Number"))
                {
                    string requestingVixSiteNumber = transaction.Values["Requesting VIX Site Number"];
                    if (!String.IsNullOrEmpty(requestingVixSiteNumber))
                    {
                        site = FindSiteByNumber(requestingVixSiteNumber);
                    }
                }
            }


            if ((site != null) && (!IsSiteSearched(site)))
            {
                SearchTransaction(site, txtTransactionId.Text.Trim(), dtDate.Value);
            }
        }

        private bool IsSiteSearched(VASite site)
        {
            return SearchResults.IsSiteSearched(site);
        }

        private VASite FindSiteByNumber(string siteNumber)
        {
            foreach (VASite site in VaSites)
            {
                if (site.SiteNumber == siteNumber)
                    return site;
            }
            return null;
        }

        private VASite FindSiteByHostname(string hostname)
        {
            foreach (VASite site in VaSites)
            {
                if (site.VixServer.ToLower().StartsWith(hostname.ToLower()))
                    return site;
            }
            return null;
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ShowAboutDialog();
        }

        private void ShowAboutDialog()
        {
            StringBuilder msg = new StringBuilder();
            msg.AppendLine("Transaction Entry Tracer");
            msg.AppendLine("Version: " + Application.ProductVersion);

            MessageBox.Show(msg.ToString(), "About", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }
    }
}
