/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VixLogAnalyzerEngine;
using TransactionLogStatisticsEngine.commands;
using System.Xml;
using System.IO;
using System.Globalization;

namespace TransactionLogStatisticsEngine
{
    public delegate void LogMsgDelegate(string msg);

    public class TransactionLogStatisticsHelper
    {

        public string LogsDirectory { get; private set; }
        public string StatsDirectory { get; private set; }

        public event LogMsgDelegate OnLogMessageEvent = null;

        public TransactionLogStatisticsHelper(string logsDirectory, string statsDirectory)
        {
            this.LogsDirectory = logsDirectory;
            this.StatsDirectory = statsDirectory;
        }

        public void searchLogs(List<string> sites, DateTime date, 
            List<TransactionLogProcessingCommand> commands)
        {
            List<TransactionLogEntry> totalEntries = new List<TransactionLogEntry>();
            foreach (string siteNumber in sites)
            {
                try
                {
                    LogMsg("Searching logs directory [" + LogsDirectory + "] for site [" + siteNumber + "] for [" + date.ToString() + "]");
                    totalEntries.AddRange(VixLogAnalyzerEngine.TransactionLogEntryDownloader.GetLogEntriesFromDirectory(LogsDirectory, siteNumber, date));
                }
                catch (Exception ex)
                {
                    LogMsg("Exception retreiving logs, " + ex.Message);
                }
            }
            foreach (TransactionLogProcessingCommand command in commands)
            {
                IEnumerable<TransactionLogEntry> matchingEntries =
                from entry in totalEntries
                where (
                    command.isEntryMatching(entry)
                    )
                orderby entry.ParentCommandId ascending
                select entry;
                command.setMatchingEntries(date, matchingEntries);
            }
        }

        public void searchLogs(string siteNumber, DateTime date, 
            List<TransactionLogProcessingCommand> commands)
        {
            List<TransactionLogEntry> entries =
                VixLogAnalyzerEngine.TransactionLogEntryDownloader.GetLogEntriesFromDirectory(LogsDirectory, siteNumber, date);
     
            foreach (TransactionLogProcessingCommand command in commands)
            {
                IEnumerable<TransactionLogEntry> matchingEntries =
                from entry in entries
                where (
                    command.isEntryMatching(entry)
                    )
                orderby entry.ParentCommandId ascending
                select entry;
                command.setMatchingEntries(date, matchingEntries);
            }
        }

        public void OutputStatistics(DateTime date, List<TransactionLogProcessingCommand> commands)
        {
            if (!Directory.Exists(StatsDirectory))
            {
                Directory.CreateDirectory(StatsDirectory);
            }

            // each command represents a test, all of them are from the same day
            string yearlyFilename = getYearlyXmlDocumentFilename(date);
            XmlDocument yearDocument = getOrCreateXmlDocument(yearlyFilename, "Yearly - " + date.Year);
            addDateStatistics(yearDocument, yearlyFilename, date, commands);

            string monthlyFilename = getMonthlyXmlDocumentFilename(date);
            XmlDocument monthDocument = getOrCreateXmlDocument(monthlyFilename, "Monthly - " + date.Month.ToString("00") + "/" + date.Year);
            addDateStatistics(monthDocument, monthlyFilename, date, commands);

            
            DateTime startofWeek = date.StartOfWeek(DayOfWeek.Sunday);
            string weeklyFilename = getWeeklyXmlDocumentFilename(startofWeek);
            XmlDocument weeklyDocument = getOrCreateXmlDocument(weeklyFilename, "Weekly - " + startofWeek.ToString("M/d/yyyy"));
            addDateStatistics(weeklyDocument, weeklyFilename, date, commands);
            
        }

        private void addDateStatistics(XmlDocument xmlDoc, string filename, DateTime date, 
            List<TransactionLogProcessingCommand> commands)
        {
            XmlNode vixStatisticsNode = xmlDoc.SelectSingleNode("/VixStatistics");
            if (vixStatisticsNode != null)
            {
                XmlAttribute lastDateAddedAttr = null;
                if (vixStatisticsNode.Attributes["lastDateAdded"] == null)
                {
                    lastDateAddedAttr = xmlDoc.CreateAttribute("lastDateAdded");
                    vixStatisticsNode.Attributes.Append(lastDateAddedAttr);
                }
                else
                {
                    lastDateAddedAttr = vixStatisticsNode.Attributes["lastDateAdded"];
                }
                lastDateAddedAttr.Value = date.ToString();
            }


            XmlNode datesNode = xmlDoc.SelectSingleNode("/VixStatistics/Dates");
            XmlNode dateNode = xmlDoc.CreateElement("Date");
            XmlAttribute attr = xmlDoc.CreateAttribute("date");
            attr.Value = date.ToString();
            dateNode.Attributes.Append(attr);
            datesNode.AppendChild(dateNode);

            XmlNode totalNode = xmlDoc.SelectSingleNode("/VixStatistics/Totals");
            XmlNode commandsDescriptionNode = xmlDoc.SelectSingleNode("/VixStatistics/CommandDescriptions");

            foreach (TransactionLogProcessingCommand command in commands)
            {
                TransactionLogEntriesStatistics stats = command.calculateStatistics();
                string name = command.GetName();// command.GetType().Name;
                dateNode.AppendChild(stats.toXml(xmlDoc, name));
                XmlNode sourceNode = findTotalSourceNode(totalNode, name);// totalNode.SelectSingleNode("/Source[@name=" + name + "]");
                if (sourceNode == null)
                {
                    totalNode.AppendChild(stats.toXml(xmlDoc, name));
                }
                else
                {
                    stats.updateXmlStats(sourceNode);
                }
                setCommandDescriptionNode(commandsDescriptionNode, command);
            }

            xmlDoc.Save(filename);
        }

        private void setCommandDescriptionNode(XmlNode commandsDescriptionNode, TransactionLogProcessingCommand command)
        {
            XmlNodeList commands = commandsDescriptionNode.SelectNodes("Command");
            foreach (XmlNode commandNode in commands)
            {
                if (commandNode.Attributes["name"].Value == command.GetName())
                {
                    // found, don't update
                    return;
                }
            }
            // not found, add and set value
            XmlNode newCommandNode = commandsDescriptionNode.OwnerDocument.CreateElement("Command");
            commandsDescriptionNode.AppendChild(newCommandNode);
            XmlAttribute attr = commandsDescriptionNode.OwnerDocument.CreateAttribute("name");
            attr.Value = command.GetName();
            newCommandNode.Attributes.Append(attr);
            attr = commandsDescriptionNode.OwnerDocument.CreateAttribute("description");
            attr.Value = command.GetDescription();
            newCommandNode.Attributes.Append(attr);

        }

        private XmlNode findTotalSourceNode(XmlNode node, string name)
        {
            XmlNodeList sourceNodes = node.SelectNodes("Source");
            foreach (XmlNode sourceNode in sourceNodes)
            {
                if (sourceNode.Attributes["name"].Value == name)
                    return sourceNode;
            }
            return null;
        }

        private string getYearlyXmlDocumentFilename(DateTime date)
        {
            return StatsDirectory + "\\Year_" + date.Year + ".xml"; 
        }

        private string getMonthlyXmlDocumentFilename(DateTime date)
        {
            return StatsDirectory + "\\Month_" + date.Year + "_" + date.Month.ToString("00") + ".xml";
        }

        
        private string getWeeklyXmlDocumentFilename(DateTime startOfWeekDate)
        {

            return StatsDirectory + "\\Weekly_" + startOfWeekDate.ToString("yyyy-MM-dd") + ".xml";
        }

        private XmlDocument getOrCreateXmlDocument(string filename, string name)
        {
            if (File.Exists(filename))
            {
                XmlDocument xmlDoc = new XmlDocument();
                xmlDoc.Load(filename);
                return xmlDoc;
            }
            else
            {
                return createBlankStatsDocument(name);
            }
        }

        private XmlDocument createBlankStatsDocument(string name)
        {
            XmlDocument xmlDoc = new XmlDocument();

            XmlProcessingInstruction instruction =
                xmlDoc.CreateProcessingInstruction("xml-stylesheet", "type='text/xsl' href='display.xsl'");
            xmlDoc.AppendChild(instruction);   

            XmlNode rootNode = xmlDoc.CreateElement("VixStatistics");
            xmlDoc.AppendChild(rootNode);

            XmlAttribute attr = xmlDoc.CreateAttribute("name");
            attr.Value = name;
            rootNode.Attributes.Append(attr);

            XmlNode datesNode = xmlDoc.CreateElement("Dates");
            rootNode.AppendChild(datesNode);

            XmlNode totalsNode = xmlDoc.CreateElement("Totals");
            rootNode.AppendChild(totalsNode);

            XmlNode commandDescriptionsNode = xmlDoc.CreateElement("CommandDescriptions");
            rootNode.AppendChild(commandDescriptionsNode);

            return xmlDoc;
        }

        /*
        private XmlNode createDateNode(DateTime date, List<TransactionLogProcessingCommand> commands)
        {
            XmlDocument xmlDoc = new XmlDocument();

            XmlNode rootNode = xmlDoc.CreateElement("VixStatistics");
            xmlDoc.AppendChild(rootNode);

            XmlNode datesNode = xmlDoc.CreateElement("Dates");
            rootNode.AppendChild(datesNode);

            XmlNode dateNode = xmlDoc.CreateElement("Date");
            XmlAttribute attr = xmlDoc.CreateAttribute("date");
            attr.Value = date.ToString();
            dateNode.Attributes.Append(attr);
            datesNode.AppendChild(dateNode);

            foreach (TransactionLogProcessingCommand command in commands)
            {
                dateNode.AppendChild(command.calculateStatistics().toXml(xmlDoc, command.GetType().Name));
            }

            xmlDoc.Save(@"C:\vix\transactionlogs\stats\test.xml");


            return null;
        }*/

        public static List<TransactionLogProcessingCommand> GetCommands()
        {
            return GetCommands(null);
        }
        public static List<TransactionLogProcessingCommand> GetCommands(string customCommandsDirectory)
        {
            List<TransactionLogProcessingCommand> commands = new List<TransactionLogProcessingCommand>();
            commands.Add(new BseRemoteLogins());
            commands.Add(new CapriRemoteLogins());
            commands.Add(new ExchangeVaImageRetrieves());
            commands.Add(new HaimsArtifactsImageRetrievesCommand());
            commands.Add(new HaimsRadiologyImageRetrievesCommand());
            commands.Add(new NcatImageRetrievesCommand());
            commands.Add(new PacsiRadiologyImageRetrievesCommand());
            commands.Add(new RadiologyImageRetrieveCommand());
            commands.Add(new XcaVaQueries());
            commands.Add(new XcaVaRetrieves());
            commands.Add(new JcifsErrors());
            commands.Add(new FederationVaImageRetreivesCommand());
            commands.Add(new MissingPatientIdException());
            commands.Add(new DailyStatsCommand());
            commands.Add(new CountStudySizes());

            if (customCommandsDirectory != null)
            {
                commands.AddRange(LoadCustomCommands(customCommandsDirectory));
            }

            return commands;
        }

        private static List<TransactionLogProcessingCommand> LoadCustomCommands(string customCommandsDirectory)
        {
            List<TransactionLogProcessingCommand> commands = new List<TransactionLogProcessingCommand>();

            FileInfo[] files = new DirectoryInfo(customCommandsDirectory).GetFiles();
            foreach (FileInfo file in files)
            {
                commands.Add(LoadCustomCommand(file));
            }

            return commands;
        }

        private static TransactionLogProcessingCommand LoadCustomCommand(FileInfo file)
        {
            GenericCommand genericCommand = GenericCommand.Load(file.FullName);
            return  GenericCommandHelper.CreateCommand(genericCommand);
        }


        public void LogMsg(string msg)
        {
            if (this.OnLogMessageEvent != null)
                OnLogMessageEvent(msg);
        }
    }

    public static class DateTimeExtensions
    {
        public static DateTime StartOfWeek(this DateTime dt, DayOfWeek startOfWeek)
        {
            System.Globalization.CultureInfo ci = System.Threading.Thread.CurrentThread.CurrentCulture;
            DayOfWeek fdow = ci.DateTimeFormat.FirstDayOfWeek;
            return dt.AddDays(-(dt.DayOfWeek - fdow));
        }
    }


}
