/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VixLogAnalyzerEngine;

namespace TransactionLogStatisticsEngine.commands
{
    public abstract class TransactionLogProcessingCommand
    {
        public List<TransactionLogEntry> MatchingEntries { get; set; }
        public DateTime Date { get; private set; }

        public TransactionLogProcessingCommand()
        {
            
        }

        public abstract bool isEntryMatching(TransactionLogEntry entry);

        public abstract string GetName();

        public abstract string GetDescription();

        /// <summary>
        /// Determines if child commands that do not have parents should be included in the result
        /// </summary>
        /// <returns></returns>
        public virtual bool includeParentlessChildCommands()
        {
            return false;
        }

        public void setMatchingEntries(DateTime date, IEnumerable<TransactionLogEntry> entries)
        {
            this.Date = date;
            Dictionary<string, TransactionLogEntryItem> entriesDictionary = new Dictionary<string, TransactionLogEntryItem>();

            // put all entries into a dictionary
            foreach (TransactionLogEntry entry in entries)
            {
                entriesDictionary.Add(entry.CommandId, new TransactionLogEntryItem(entry));
            }

            // loop through each entry in dictionary
            foreach (string commandId in entriesDictionary.Keys.ToArray())
            {
                TransactionLogEntryItem entry = entriesDictionary[commandId];
                // if the entry is a child
                if (entry.TransactionLogEntry.IsChild())
                {
                    // search to see if the parent commandId is in the dictionary
                    if (entriesDictionary.ContainsKey(entry.TransactionLogEntry.ParentCommandId))
                    {
                        // find the parent entry and add this child to it and then null this child
                        TransactionLogEntryItem parentEntry = entriesDictionary[entry.TransactionLogEntry.ParentCommandId];
                        parentEntry.TransactionLogEntry.ChildEntries.Add(entry.TransactionLogEntry);
                        entriesDictionary[commandId].MovedToParent = true;
                        //entriesDictionary[commandId] = null;
                        //entriesDictionary.Remove(commandId);
                    }
                    else
                    {
                        // didn't find the parent which means this child might have been created by a parent we are not interseted in (such as a prefetch operation)
                        // mark it as moved so it gets excluded from the result
                        
                        // determine if we want to include child commands that do not have parents in the results
                        if (!includeParentlessChildCommands())
                        {
                            entriesDictionary[commandId].MovedToParent = true;
                        }
                    }
                }
            }

            MatchingEntries = new List<TransactionLogEntry>();
            foreach (TransactionLogEntryItem entry in entriesDictionary.Values)
            {
                if((entry != null) && (!entry.MovedToParent))
                    MatchingEntries.Add(entry.TransactionLogEntry);
            }

            /*

            Dictionary<string, TransactionLogEntry> parentEntries = new Dictionary<string, TransactionLogEntry>();

            // transactions should be sorted by parent command id so parent commands should be first
            MatchingEntries = new List<TransactionLogEntry>();
            foreach (TransactionLogEntry entry in entries)
            {
                if (entry.IsChild())
                {

                }
                else
                {
                    parentEntries.Add(entry.CommandId, entry);
                }

                if (string.IsNullOrEmpty(entry.ParentCommandId))
                {
                    // this is a parent command
                    MatchingEntries.Add(entry);
                    
                    //MatchingEntries.Add(new ParentTransactionLogEntry(
                }

            }*/
        }

        public virtual TransactionLogEntriesStatistics calculateStatistics()
        {
            TransactionLogImageEntriesStatistics stats = new TransactionLogImageEntriesStatistics();
            stats.ParentTransactions = MatchingEntries.Count;
            foreach (TransactionLogEntry entry in MatchingEntries)
            {
                switch (entry.Quality)
                {
                    case "DIAGNOSTIC":
                        setImageTypeStatistics(entry, stats.DiagnosticRequests);
                        break;
                    case "THUMBNAIL":
                        setImageTypeStatistics(entry, stats.ThumbnailRequests);
                        break;
                    case "REFERENCE":
                        setImageTypeStatistics(entry, stats.ReferenceRequests);
                        break;
                    case "DIAGNOSTICUNCOMPRESSED":
                        setImageTypeStatistics(entry, stats.DiagnosticUncompressedRequests);
                        break;
                    default:
                        setImageTypeStatistics(entry, stats.ArtifactRequests);
                        break;
                }
            }
            return stats;
        }

        public void setImageTypeStatistics(TransactionLogEntry entry, ImageTypeStatistics imageTypeStats)
        {
            if (entry.IsError())
            {
                imageTypeStats.Failed += 1;
            }
            else
            {
                imageTypeStats.Successful += 1;
            }
            imageTypeStats.Requests++;
            
            imageTypeStats.BytesTransfered += (long)entry.getPropertyFromParentOrChild("FacadeBytesReturned", false);
        }

        public void OutputStatisticsToXml()
        {
            TransactionLogEntriesStatistics stats = calculateStatistics();



            // multiple stats files:
            // 1) Weekly
            // 2) Monthly
            // 3) Yearly

            // output to the main file for today's information
            // output for the 
            // get the 
        }

        public void ClearMatchingEntries()
        {
            if (MatchingEntries != null)
            {
                MatchingEntries.Clear();
                MatchingEntries = null;
            }
            //this.Date = null;
        }

        public override string ToString()
        {
            return GetName();
        }
    }

    public class TransactionLogEntryItem
    {
        public TransactionLogEntry TransactionLogEntry { get; private set; }
        public bool MovedToParent { get; set; }

        public TransactionLogEntryItem(TransactionLogEntry entry)
        {
            this.TransactionLogEntry = entry;
            this.MovedToParent = false;
        }

    }
}
