/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using GalaSoft.MvvmLight;
using GalaSoft.MvvmLight.Messaging;
using VISAHealthMonitorCommon.messages;
using VISAHealthMonitorCommon;
using VISACommon;
using System;
using System.Text;
using System.Windows.Input;

namespace VISAHealthMonitor.ViewModel
{
    /// <summary>
    /// This class contains properties that a View can data bind to.
    /// <para>
    /// Use the <strong>mvvminpc</strong> snippet to add bindable properties to this ViewModel.
    /// </para>
    /// <para>
    /// You can also use Blend to data bind with the tool's support.
    /// </para>
    /// <para>
    /// See http://www.galasoft.ch/mvvm/getstarted
    /// </para>
    /// </summary>
    public class VixDetailsViewModel : VISADetailsViewModel
    {
        public string SiteName { get; set; }
        private VisaSource visaSource = null;

        public string VixCacheDirectory { get; set; }
        public string VixCacheSize { get; set; }
        public string VixCacheDriveCapacity { get; set; }
        public string VixCacheDriveAvailableSpace { get; set; }
        public string VixCacheDrivePercentFull { get; set; }

        /*
        public string VixCacheOperationsSuccessful { get; set; }
        public string VixCacheOperationsInstanceNotFound { get; set; }
        public string VixCacheOperationsInitiated { get; set; }
        public string VixCacheOperationsError { get; set; }
         * */
        public CacheProperties CacheProperties { get; set; }
        

        /// <summary>
        /// Initializes a new instance of the VixDetailsViewModel class.
        /// </summary>
        public VixDetailsViewModel()
        {
            if (IsInDesignMode)
            {
                
                // Code runs in Blend --> create design time data.
            }
            else
            {
                Messenger.Default.Register<UpdateAndDisplayVisaHealthMessage>(this, action => ReceiveUpdateAndDisplayVisaHealthMessage(action));
                Messenger.Default.Register<DisplayVisaHealthMessage>(this, action => ReceieveDisplayVisaSourceMessage(action));
                // Code runs "for real"
            }
        }

        protected override void UpdateVisaHealth()
        {
            VisaHealth health = VisaHealthManager.GetVisaHealth(visaSource);
            Messenger.Default.Send<CursorChangeMessage>(new CursorChangeMessage(Cursors.Wait));
            health.Update();
            DisplayHealth();
            Messenger.Default.Send<CursorChangeMessage>(new CursorChangeMessage(Cursors.Arrow));
        }

        public void ReceiveUpdateAndDisplayVisaHealthMessage(UpdateAndDisplayVisaHealthMessage msg) 
        {
            visaSource = msg.VisaSource;
            UpdateVisaHealth();
            // done by the update
            //DisplayHealth();            
        }

        public void ReceieveDisplayVisaSourceMessage(DisplayVisaHealthMessage msg)
        {
            visaSource = msg.VisaSource;
            DisplayHealth();
        }

        private void DisplayHealth()
        {
            VaSite vaSite = (VaSite)visaSource;
            SiteName = vaSite.Name;
            VisaHealth health = VisaHealthManager.GetVisaHealth(visaSource);
            if (health.LoadStatus == VixHealthLoadStatus.loaded)
            {
                DateHealthLastUpdated = health.LastUpdateTime;
                VISAVersion = health.VisaVersion;
                Hostname = health.Hostname;

                long uptimeTicks = health.JVMUptime;
                TimeSpan span = ticksToTimespan(uptimeTicks);// new TimeSpan(0, 0, 0, 0, ticks);

                StringBuilder sb = new StringBuilder();
                if (span.Days > 0)
                {
                    sb.Append(span.Days);
                    sb.Append(" days, ");
                }
                sb.Append(span.Hours + " hours, " + span.Minutes + " minutes, " + span.Seconds + " seconds, " + span.Milliseconds + " ms");
                UpTime = sb.ToString();

                // time in Java starts at 1970
                DateTime seventies = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);
                DateTime d = seventies.AddMilliseconds(health.JVMStartupTime);
                StartTime = d.ToLocalTime().ToString();

                // cache properties
                SetCacheProperties(health);
            }
            else
            {
                ClearProperties();
                Hostname = vaSite.VistaHost;
            }
        }

        protected new void ClearProperties()
        {
            base.ClearProperties();
            Hostname = "";
            this.CacheProperties = new CacheProperties();
            VixCacheDirectory ="";
            VixCacheSize ="";
            VixCacheDriveCapacity ="";
            VixCacheDriveAvailableSpace ="";
            VixCacheDrivePercentFull = "";
        }

        private void SetCacheProperties(VisaHealth health)
        {
            VixCacheDirectory = health.GetPropertyValue("VixCacheDir");
            long cacheDirectorySize = health.GetPropertyValueLong("VixCacheDirSize");
            if (cacheDirectorySize >= 0)
            {
                VixCacheSize = formatBytes(cacheDirectorySize);
            }
            else
            {
                VixCacheSize = "?";
            }
            long cacheAvailableSpace = health.GetPropertyValueLong("VixCacheDirRootDriveAvailable");
            if (cacheAvailableSpace >= 0)
                VixCacheDriveAvailableSpace = formatBytes(cacheAvailableSpace);
            else
                VixCacheDriveAvailableSpace = "?";
            long cacheDriveCapacity = health.GetPropertyValueLong("VixCacheDirRootDriveTotal");
            if (cacheDriveCapacity >= 0)
                VixCacheDriveCapacity = formatBytes(cacheDriveCapacity);
            else
                VixCacheDriveCapacity = "?";
            //VixCacheDrivePercentFull

            long spaceUsed = cacheDriveCapacity - cacheAvailableSpace;
            double val = ((double)spaceUsed / (double)cacheDriveCapacity);
            VixCacheDrivePercentFull = (val * 100.0).ToString("N2") + "%";
            
            string VixCacheOperationsSuccessful = null;
            string VixCacheOperationsInstanceNotFound = null;
            string VixCacheOperationsInitiated = null;
            string VixCacheOperationsError = null;
             

            long cacheOperationsSuccessful = health.GetPropertyValueLong("VixJMXCacheOperationsSuccessful");
            if (cacheOperationsSuccessful >= 0)
                VixCacheOperationsSuccessful = cacheOperationsSuccessful.ToString(numberFormat);
            else
                VixCacheOperationsSuccessful = "?";
            long cacheOperationsInitiated = health.GetPropertyValueLong("VixJMXCacheOperationsInitiated");
            if (cacheOperationsInitiated >= 0)
                VixCacheOperationsInitiated = cacheOperationsInitiated.ToString(numberFormat);
            else
                VixCacheOperationsInitiated = "?";
            long cacheOperationsNotFound = health.GetPropertyValueLong("VixJMXCacheOperationsInstanceNotFound");
            if (cacheOperationsNotFound >= 0)
                VixCacheOperationsInstanceNotFound = cacheOperationsNotFound.ToString(numberFormat);
            else
                VixCacheOperationsInstanceNotFound = "?";
            long cacheOperationErrors = health.GetPropertyValueLong("VixJMXCacheOperationsError");
            if (cacheOperationErrors >= 0)
                VixCacheOperationsError = cacheOperationErrors.ToString(numberFormat);
            else
                VixCacheOperationsError = "?";

            CacheProperties = new CacheProperties(VixCacheOperationsInitiated, VixCacheOperationsSuccessful, VixCacheOperationsInstanceNotFound, VixCacheOperationsError);
        }

        ////public override void Cleanup()
        ////{
        ////    // Clean own resources if needed

        ////    base.Cleanup();
        ////}
    }

    public class CacheProperties
    {
        public string VixCacheOperationsSuccessful { get; set; }
        public string VixCacheOperationsInstanceNotFound { get; set; }
        public string VixCacheOperationsInitiated { get; set; }
        public string VixCacheOperationsError { get; set; }

        public CacheProperties(string initiated, string successful, string notFound, string error)
        {
            VixCacheOperationsError = error;
            VixCacheOperationsInitiated = initiated;
            VixCacheOperationsInstanceNotFound = notFound;
            VixCacheOperationsSuccessful = successful;
        }

        public CacheProperties()
        {
        }
    }
}
