/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;

namespace VixAKCommon.gov.va.med.imaging
{
    /// <summary>
    /// There are two rules for a SavableObject:
    /// 1) Must include a 0 argument constructor (so no readonly member variables)
    /// 2) Must override and implement Equals to determine if two objects are the same
    /// </summary>
    public abstract class SavableObject
    {
        public SavableObject()
        {
            // must have default constructor
        }

        public abstract override bool Equals(object obj);

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

    }

    /// <summary>
    /// This savable object contains a string with multiple lines some of which are skipped during comparison
    /// </summary>
    public class StringSkipLineSavableObject : SavableObject
    {
        private string value;
        private string[] skipLines = null;

        public string[] SkipLines
        {
            get { return skipLines; }
            set { skipLines = value; }
        }

        public string Value
        {
            get { return this.value; }
            set { this.value = value; }
        }

        public StringSkipLineSavableObject()
        {
        }

        public StringSkipLineSavableObject(string value, string [] skipLines)
        {
            this.value = value;
            this.skipLines = skipLines;
        }

        public override bool Equals(object obj)
        {
            if (obj is StringSkipLineSavableObject)
            {
                StringSkipLineSavableObject that = (StringSkipLineSavableObject)obj;
                if((this.skipLines == null && that.skipLines != null) || (this.skipLines != null && that.skipLines == null))
                {
                    // one of the skipLines is null but not both of them
                    return false;
                }
                if (this.skipLines != null && that.skipLines != null)
                {
                    if (this.skipLines.Length != that.skipLines.Length)
                        return false;
                    string[] thisLines = this.value.Split('\n');
                    string[] thatLines = that.value.Split('\n');
                    for (int i = 0; i < thisLines.Length; i++)
                    {
                        string il = thisLines[i];
                        string tl = thatLines[i];
                        bool skip = false;
                        foreach (string sl in this.skipLines)
                        {
                            if (il.StartsWith(sl))
                            {
                                skip = true;
                                // skip this line
                            }
                        }
                        if (!skip)
                        {
                            if (!il.Equals(tl))
                            {
                                return false;
                            }
                        }
                    }
                    return true;
                }
                else
                {
                    // both skipLines are null
                    return this.value.Equals(that.value);
                }
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class StringSavableObject : SavableObject
    {
        private string result;

        public string Result
        {
            get { return result; }
            set { result = value; }
        }

        public StringSavableObject()
        {
        }

        public StringSavableObject(string result)
        {
            this.result = result;
        }

        public override bool Equals(object obj)
        {
            if (obj is StringSavableObject)
            {
                StringSavableObject that = (StringSavableObject)obj;
                return this.result.Equals(that.result);
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class StringSavableObjectList : SavableObject
    {
        private List<string> results = new List<string>();

        public List<string> Results
        {
            get { return results; }
            set { results = value; }
        }

        public StringSavableObjectList()
        {
        }

        public StringSavableObjectList(List<string> results)
        {
            this.results.AddRange(results);
        }


        public override bool Equals(object obj)
        {
            if (obj is StringSavableObjectList)
            {
                StringSavableObjectList that = (StringSavableObjectList)obj;
                if (this.results.Count != that.results.Count)
                    return false;
                for (int i = 0; i < this.results.Count; i++)
                {
                    string s = this.results[i];
                    string t = that.results[i];
                    if (!s.Equals(t))
                        return false;
                }
                return true;
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class BooleanSavableObject : SavableObject
    {
        private bool result;

        public bool Result
        {
            get { return result; }
            set { result = value; }
        }

        public BooleanSavableObject()
        {
        }

        public BooleanSavableObject(bool result)
        {
            this.result = result;
        }

        public override bool Equals(object obj)
        {
            if (obj is BooleanSavableObject)
            {
                BooleanSavableObject that = (BooleanSavableObject)obj;
                return this.result == that.result;
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }
}
