/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using java.util.zip;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.clinicaldisplay;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.vistarad;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.xca;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.exchange;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.cache;
using System.Xml;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix
{

    public delegate string LogMessageDelegate(string msg);
    public delegate void IncrementProgressDelegate();
    public delegate void SetProgressBarMax(int max);
    public delegate void ActionCompleteDelegate(BaseAction action, ActionCompletedResult actionCompleteResult);
    public delegate void DoApplicationEventsDelegate();
    public delegate void LogInfoMessageDelegate(string msg);
    public delegate void LogErrorMessageDelegate(string msg);

    public class VixAKCommonUtils
    {

        public static string TrimValue(string value)
        {
            if (value.StartsWith("\""))
            {
                value = value.Substring(1);
            }
            if (value.EndsWith("\""))
            {
                value = value.Substring(0, value.Length - 1);
            }
            return value;
        }

        public static void SaveActionsAsXml(string filename, List<BaseAction> actions)
        {
            XmlDocument xmlDoc = new XmlDocument();
            XmlNode rootNode = xmlDoc.CreateElement("VixAK");
            xmlDoc.AppendChild(rootNode);
            XmlNode actionsNode = xmlDoc.CreateElement("Actions");
            rootNode.AppendChild(actionsNode);
            foreach (BaseAction action in actions)
            {
                XmlNode actionNode = action.toXml(xmlDoc);
                actionsNode.AppendChild(actionNode);
            }
            xmlDoc.Save(filename);
        }

        public static List<BaseAction> LoadActionsFromXml(string filename)
        {
            List<BaseAction> actions = new List<BaseAction>();

            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.Load(filename);

            XmlNodeList actionNodes = xmlDoc.SelectNodes("/VixAK/Actions/Action");
            foreach (XmlNode actionNode in actionNodes)
            {
                BaseAction newAction = null;
                string type = actionNode.Attributes["Type"].Value;

                bool enabled = true;
                if (actionNode.Attributes["Enabled"] != null)
                {
                    enabled = bool.Parse( actionNode.Attributes["Enabled"].Value);
                }
                string testDescription = null;
                if (actionNode.Attributes["TestDescription"] != null)
                {
                    testDescription = actionNode.Attributes["TestDescription"].Value;
                }

                List<BaseAction> allActions = GetAllActions();
                foreach (BaseAction a in allActions)
                {
                    if (type == a.GetActionName())
                    {
                        newAction = (BaseAction)Activator.CreateInstance(a.GetType());
                        break;
                    }
                }
                if (newAction != null)
                {
                    newAction.Enabled = enabled;
                    newAction.TestDescription = testDescription;
                    newAction.SetPropertiesFromXml(actionNode);
                    actions.Add(newAction);
                }
                else
                {
                    System.Diagnostics.Debug.Assert(false, "Unable to find action type in LoadActionsFromXml");
                }
            }
            return actions;
        }        

        public static String CalculateAdler32(byte[] buffer)
        {
            SByte[] sbuffer = new SByte[buffer.Length]; // checksum buffer - must be signed to make Java happy
            Adler32 adler32 = new Adler32();
            for (int i = 0; i < buffer.Length; i++) sbuffer[i] = (sbyte)buffer[i];
            adler32.update(sbuffer, 0, sbuffer.Length);
            return "{Adler32}" + ((Decimal)adler32.getValue()).ToString();
        }

        public static int CountImages(Study study)
        {
            int count = 0;
            foreach (Series series in study.Series)
            {
                count += series.Instances.Count;            
            }
            return count;
        }

        public static int CountImages(List<Study> studies)
        {
            int count = 0;
            foreach (Study study in studies)
            {
                count += CountImages(study);
            }
            return count;
        }

        public static List<ImagingPatient> GetTestPatients()
        {
            List<ImagingPatient> patients = new List<ImagingPatient>();


            return patients;
        }

        public static List<BaseAction> GetAllActions()
        {
            List<BaseAction> actions = new List<BaseAction>();
            actions.AddRange(GetDeveloperActions());
            actions.AddRange(GetPublicActions());
            actions.Sort();
            return actions;
        }

        public static List<BaseAction> GetDeveloperActions()
        {
            List<BaseAction> actions = new List<BaseAction>();
            actions.Add(new FakeTestAction());
            actions.Sort();
            return actions;
        }

        public static List<BaseAction> GetPublicActions()
        {
            List<BaseAction> actions = new List<BaseAction>();
            actions.Add(new ExchangeTestAction());
            actions.Add(new ClearCacheAction());
            actions.Add(new ClearTomcatLogsAction());
            actions.Add(new ClinicalDisplayTestAction());
            actions.Add(new FileSwapAction());
            actions.Add(new ClinicalDisplayImageAction());
            actions.Add(new ClinicalDisplayLogImageAccessAction());
            actions.Add(new ClinicalDisplayGetImageInformationAction());
            actions.Add(new ExchangeSiteServiceAction());
            actions.Add(new SiteServiceTestAction());
            actions.Add(new FileCopyAction());
            actions.Add(new FolderCopyAction());
            //actions.Add(new DocumentTestAction());
            actions.Add(new AWIVTestAction());
            actions.Add(new VistaRadFullTestAction());
            actions.Add(new VistaRadActiveExamsTestAction());
            actions.Add(new ClinicalDisplayPatientTestAction());
            actions.Add(new ContinuousAction());
            actions.Add(new XcaTestAction());
            actions.Add(new ClinicalDisplayPatientStudiesAction());             
            actions.Add(new VistaRadGetActiveExams());            
            actions.Add(new VistaRadGetExamImages());
            actions.Add(new VistaRadGetExamImage());
            actions.Add(new VistaRadGetPatientExams());
            actions.Add(new XcaGetDocument());
            actions.Add(new XcaGetDocuments());
            actions.Add(new ExchangeGetPatientStudies());
            actions.Add(new ExchangeGetImage());
            actions.Add(new VistaRadGetExamImageTextFile());
            actions.Add(new ClinicalDisplayPassthroughAction());
            actions.Add(new VistaRadPassthroughAction());
            actions.Add(new ClinicalDisplayGetImageDevFieldsAction());
            actions.Add(new ClinicalDisplayGetImageGlobalNodes());
            actions.Add(new ClinicalDisplayGetImageTextFileAction());
            actions.Add(new GetImageCachedAction());
            actions.Add(new GetDocumentCachedAction());
            actions.Sort();
            return actions;
        }

        /// <summary>
        /// This is a stupid function but i don't have a better way to do it (don't feel like
        /// dealing with reflection)
        /// </summary>
        /// <param name="action"></param>
        /// <returns></returns>
        public static BaseAction CreateNewActionInstance(BaseAction action)
        {
            List<BaseAction> actions = GetAllActions();
            foreach (BaseAction a in actions)
            {
                if (action.GetType() == a.GetType())
                {
                    return (BaseAction)Activator.CreateInstance(a.GetType());
                }
            }
            System.Diagnostics.Debug.Assert(false, "Unknown BaseAction to create in CreateNewActionInstance");        
            return null;
        }

        public static string ConvertTime(TimeSpan ts)
        {
            StringBuilder sb = new StringBuilder();
            string prefix = "";
            if (ts.Days > 0)
            {
                sb.Append(prefix);
                sb.Append(ts.Days + " days");
                prefix = ", ";
            }
            if (ts.Hours > 0)
            {
                sb.Append(prefix);
                sb.Append(ts.Hours + " hours");
                prefix = ", ";
            }
            if (ts.Minutes > 0)
            {
                sb.Append(prefix);
                sb.Append(ts.Minutes + " minutes");
                prefix = ", ";
            }
            if (ts.Seconds > 0)
            {
                sb.Append(prefix);
                sb.Append(ts.Seconds + " seconds");
                prefix = ", ";
            }
            if (ts.Milliseconds > 0)
            {
                sb.Append(prefix);
                sb.Append(ts.Milliseconds + " ms");
                prefix = ", ";
            }

            return sb.ToString();
        }

        public static String GetInstanceInfo(Study study, Series series, Instance instance, ImageQuality quality)
        {
            string imgQuality = "";
            if (quality != null)
                imgQuality = " Image Quality: " + quality.ToString("g");
            return "study: " + study.StudyDescription + " series[" + series.DicomUid + "]: " + series.Description +
                " image[" + instance.DicomUid + "]: " + instance.InstanceId + imgQuality;
        }

        public static String GetInstanceInfo(Study study, Series series, Instance instance)
        {
            return GetInstanceInfo(study, series, instance, null);
        }

        public static string getApplicationDir(string executablePath)
        {
            int loc = executablePath.LastIndexOf("\\");
            return executablePath.Substring(0, loc);
        }

        public static string getApplicationDir()
        {
            string path = System.Windows.Forms.Application.ExecutablePath;
            return getApplicationDir(path);
        }

        public static byte[] readStreamIntoBuffer(Stream stream)
        {
            MemoryStream ms = new MemoryStream();
            byte[] buffer = new byte[32768];
            int read = stream.Read(buffer, 0, buffer.Length);
            while (read > 0)
            {
                ms.Write(buffer, 0, read);
                read = stream.Read(buffer, 0, buffer.Length);
            }
            return ms.ToArray();
        }

        private static string defaultValuesFilename = null;

        public static string getDefaultValuesFile()
        {
            if (defaultValuesFilename == null)
            {
                defaultValuesFilename = getApplicationDir() + "\\defaultValues.xml";
            }
            return defaultValuesFilename;
        }

        private static string hostname = null;
        public static string GetHostname()
        {
            if (hostname == null)
            {
                hostname = System.Environment.MachineName;
            }
            return hostname;
        }
        
    }

    public enum ActionCompletedResult
    {
        /// <summary>
        /// The action completed successfully
        /// </summary>
        success, 
        /// <summary>
        /// The action failed to complete (there was an unexpected error)
        /// </summary>
        error, 
        /// <summary>
        /// The action completed however the result did not match the expected result
        /// </summary>
        resultsFailed
    }
}
