/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix.services
{
    public class AWIVServiceImpl : AbstractBaseVixInterfaceService
    {

        private readonly AWIVService_v1.ImageMetadataAWIVService metadataService;

        private readonly string mstrMetadataUrl;
        private readonly string mstrImageUrl;
        private readonly string mstrUsername = "";
        private readonly string mstrPassword = "";
        private readonly bool mbAllowJ2kCompression;
        private readonly bool mbIncludeImageFormatsSubTypes;


        public AWIVServiceImpl(string metadataUrl, string imageUrl, LogMessageDelegate OnLogMessage)
            : base(OnLogMessage)
        {
            mstrMetadataUrl = metadataUrl;
            mstrImageUrl = imageUrl;
            metadataService = new AWIVService_v1.ImageMetadataAWIVService();
            metadataService.Url = metadataUrl;
            metadataService.Credentials = new NetworkCredential(mstrUsername, mstrPassword);
            metadataService.Timeout = 500000; // 500 sec          
        }

        public Study getImageList(string siteNumber, string patientIcn, string cprsIdentifier)
        {
            string transactionId = Guid.NewGuid().ToString();
            AWIVService_v1.UserCredentialsType credentials = GetUserCredentials();
            AWIVService_v1.GroupType group = 
                metadataService.getImageList(transactionId, credentials, siteNumber, patientIcn, cprsIdentifier);

            return ServiceTranslator.TranslateGroup(group);

        }

        private AWIVService_v1.UserCredentialsType GetUserCredentials()
        {
            AWIVService_v1.UserCredentialsType credentials = new AWIVService_v1.UserCredentialsType();
            credentials.duz = "";
            credentials.fullname = "";
            credentials.siteName = "";
            credentials.siteNumber = "";
            credentials.ssn = "";
            return credentials;
        }

        public AWIVDownloadResult DownloadImage(AWIVImageRequestInput imageRequestInput)
        {
            string url = createImageUrl(imageRequestInput.Study, imageRequestInput.Instance, imageRequestInput.ImageQuality);
            imageRequestInput.RequestUrl = url;
            string requestedImageFormat = GetContentType(imageRequestInput.Instance.ImageType, imageRequestInput.ImageQuality, false);
            imageRequestInput.RequestedImageFormat = requestedImageFormat;
            WebRequest request = WebRequest.Create(url);
            request.Timeout = 500000; // 500 seconds
            request.Credentials = new NetworkCredential(mstrUsername, mstrPassword);
            AWIVService_v1.UserCredentialsType userCredentials = GetUserCredentials();
            request.Headers.Add("xxx-fullname", userCredentials.fullname);
            request.Headers.Add("xxx-sitename", userCredentials.siteName);
            request.Headers.Add("xxx-sitenumber", userCredentials.siteNumber);
            request.Headers.Add("xxx-ssn", userCredentials.ssn);
            request.Headers.Add("xxx-duz", userCredentials.duz);
            request.Headers.Add("xxx-transaction-id", imageRequestInput.TransactionId);
            string imageFormatSubTypeRequestString = "";
            if (mbIncludeImageFormatsSubTypes)
            {
                imageFormatSubTypeRequestString = GetContentType(imageRequestInput.Instance.ImageType, imageRequestInput.ImageQuality, true);
                request.Headers.Add("xxx-contentTypeWithSubType", imageFormatSubTypeRequestString);
                
            }
            imageRequestInput.ImageFormatSubTypeRequestString = imageFormatSubTypeRequestString;

            WebResponse response = null;

            try
            {
                response = request.GetResponse();
            }
            catch (System.Net.WebException webEx)
            {
                if (webEx.Status != System.Net.WebExceptionStatus.ProtocolError)
                {
                    LogMsg(webEx.InnerException + "\n\n" + webEx.StackTrace);
                }
                System.Net.HttpWebResponse resp = (System.Net.HttpWebResponse)webEx.Response;
                if (resp.StatusCode == HttpStatusCode.NotFound)
                    throw new vix.exceptions.ImageNotFoundException(webEx.Message);
                else if (resp.StatusCode == HttpStatusCode.InternalServerError)
                    throw new vix.exceptions.InternalServerErrorException(webEx.Message);
                else
                    throw webEx;
            }
            if (response == null)
                throw new Exception("Got null response from server for url '" + url + "' from transaction [" + imageRequestInput.TransactionId + "].");
            Stream stream = response.GetResponseStream();

            byte[] buffer = VixAKCommonUtils.readStreamIntoBuffer(stream);

            string calculatedAlder32 = VixAKCommonUtils.CalculateAdler32(buffer);
            string headerAlder32 = response.Headers["xxx-checksum"];

            if (calculatedAlder32 != headerAlder32)
            {
                throw new Exception("Calculated checksum [" + calculatedAlder32 + "] does not match checksum [" + headerAlder32 + "] from ViX");
            }

            string format = response.Headers["xxx-image-format"];
            string returnedImageFormat = "";
            string returnedImageQuality = "";
            if (format != null)
                returnedImageFormat = format;
            string quality = response.Headers["xxx-image-quality"];
            if (quality != null)
                returnedImageQuality = quality;

            return new AWIVDownloadResult(returnedImageQuality, returnedImageFormat, calculatedAlder32, response.Headers);
        }

        private string createImageUrl(Study study, Instance instace, ImageQuality imageQuality)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(mstrImageUrl);
            sb.Append("?imageURN=");
            sb.Append(instace.InstanceId);
            sb.Append("&imageQuality=" + imageQuality.ToString("d"));
            sb.Append("&contentType=" + GetContentType(instace.ImageType, imageQuality, false));
            return sb.ToString();
        }

        private string GetContentType(int imageType, ImageQuality imageQuality, bool includeImageSubTypes)
        {
            string contentType = "";
            string delimiter = ",";
            bool isThumbnail = (imageQuality == ImageQuality.THUMBNAIL);
            switch (imageType)
            {
                case 1: // JPEG
                case 9: // bwmed
                case 17: // color scan
                case 18: // patient photo
                case 19: // XRAY JPEG
                // document images
                case 15: // TIF
                    contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                        ImageFormat.TIFF.ToString(includeImageSubTypes) + delimiter +
                        ImageFormat.BMP.ToString(includeImageSubTypes);
                    break;
                case 21: // motion video
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "video/x-msvideo";
                    }
                    break;
                case 101: // html
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "text/html";
                    }
                    break;
                case 102: // Word document
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "application/msword";
                    }
                    break;
                case 103: // text (ASCII text)
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "text/plain";
                    }
                    break;
                case 104: // PDF
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "application/pdf";
                    }
                    break;
                case 105: // RTF
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "text/rtf";
                    }
                    break;
                case 106: // Audio
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        contentType = "audio/x-wav,audio/mpeg3";
                    }
                    break;
                case 3: // XRAY
                case 100: // DICOM
                    if (isThumbnail)
                    {
                        contentType = ImageFormat.JPG.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.TGA.ToString(includeImageSubTypes) + delimiter +
                            ImageFormat.BMP.ToString(includeImageSubTypes);
                    }
                    else
                    {
                        if (mbAllowJ2kCompression)
                        {
                            contentType = ImageFormat.DICOMJ2K.ToString(includeImageSubTypes) + delimiter +
                                ImageFormat.J2K.ToString(includeImageSubTypes) + delimiter +
                                ImageFormat.DICOM.ToString(includeImageSubTypes) + delimiter +
                                ImageFormat.TGA.ToString(includeImageSubTypes);
                        }
                        else
                        {
                            contentType = ImageFormat.DICOM.ToString(includeImageSubTypes) + delimiter +
                                ImageFormat.TGA.ToString(includeImageSubTypes);
                        }
                    }
                    break;
                default:
                    System.Diagnostics.Debug.Assert(false, "Unknown image quality type in GetContentType [" + imageQuality + "]");
                    break;
            }
            if (contentType.Length > 0)
            {
                contentType += ",*/*";
            }
            else
            {
                contentType = "*/*";
            }

            return contentType;

        }

    }

    
}
