/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
    using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Net;
using VixAKCommon.gov.va.med.imaging.vix.tester;
using System.Xml;
using VixAKCommon.gov.va.med.imaging.vix.exceptions;

namespace VixAKCommon.gov.va.med.imaging.vix.services
{
    public class XcaServiceImpl : AbstractBaseVixInterfaceService, VixInterface
    {
        private readonly string mstrQueryUrl;
        private readonly string mstrRetrieveUrl;
        private readonly VixTestUser testUser;
        private readonly string mstrUsername;
        private readonly string mstrPassword;

        public XcaServiceImpl(string queryUrl, string retrieveUrl, 
            VixTestUser testUser, string username, string password, LogMessageDelegate OnLogMessage) : base(OnLogMessage)
        {
            mstrQueryUrl = queryUrl;
            mstrRetrieveUrl = retrieveUrl;
            this.testUser = testUser;
            mstrUsername = username;
            mstrPassword = password;
            
            ServicePointManager.Expect100Continue = false;
            ServicePointManager.ServerCertificateValidationCallback += IgnoreCertificateErrorHandler;
        }

        public bool IgnoreCertificateErrorHandler(object sender, System.Security.Cryptography.X509Certificates.X509Certificate certificate,
            System.Security.Cryptography.X509Certificates.X509Chain chain, System.Net.Security.SslPolicyErrors sslPolicyErrors)
        {
            //LogMsg("Received certificate '" + certificate.Subject + "'", false);
            return true;
        }

        public List<Study> GetPatientStudies(string siteNumber, string patientIcn, string transactionId)
        {
            WebRequest request = WebRequest.Create(mstrQueryUrl);
            request.Method = "POST";
            request.ContentType = "application/soap+xml; charset=UTF-8; action=\"urn:ihe:iti:2007:CrossGatewayQuery\"";
            
            request.Headers.Add("xxx-sitename", testUser.SiteName);
            request.Headers.Add("xxx-fullname", testUser.Fullname);
            request.Headers.Add("xxx-sitenumber", testUser.SiteNumber);
            request.Headers.Add("xxx-ssn", testUser.Ssn);
            request.Headers.Add("xxx-purpose-of-use", "Routine medical care");
            request.Headers.Add("xxx-patient-id", patientIcn);
            request.Headers.Add("xxx-transaction-id", transactionId);
            request.Headers.Add("SOAPAction", "urn:ihe:iti:2007:CrossGatewayQuery");

            byte[] byteArray = Encoding.UTF8.GetBytes(createQueryRequest(patientIcn));

            request.Credentials = new NetworkCredential(mstrUsername, mstrPassword);

            Stream requestStream = request.GetRequestStream();
            requestStream.Write(byteArray, 0, byteArray.Length);
            requestStream.Close();


            WebResponse response = request.GetResponse();

            string contentType = response.Headers["Content-Type"];
            string boundary = "";
            int loc = contentType.IndexOf("boundary=");
            if (loc >= 0)
            {
                boundary = contentType.Substring(loc + 9);
                loc = boundary.IndexOf(";");
                if (loc >= 0)
                {
                    boundary = boundary.Substring(0, loc);
                }
            }
            if (boundary.StartsWith("\""))
            {
                boundary = boundary.Substring(1);
            }
            if (boundary.EndsWith("\""))
            {
                boundary = boundary.Substring(0, boundary.Length - 1);
            }

            Stream responseStream = response.GetResponseStream();

            StreamReader reader = new StreamReader(responseStream);

            string xml = extractXmlFromQueryResponse(reader, boundary);
            reader.Close();
            if (xml == null)
                throw new Exception("Could not find XML in response");
            try
            {
                return XcaServiceTranslator.createDocumentSetsFromXml(xml);    
            }
            catch (Exception ex)
            {
                string msg = ex.Message + ", xml=" + xml;
                throw new Exception(msg);
            }
        }

        public string GetStudyReport(string studyId, string transactionId)
        {
            return "";
        }

        public DownloadResult DownloadImage(ImageRequestInput imageRequestInput)
        {
            WebRequest request = WebRequest.Create(mstrRetrieveUrl);
            request.Method = "POST";
            request.ContentType = "application/soap+xml; charset=UTF-8; action=\"urn:ihe:iti:2007:CrossGatewayRetrieve\"";

            request.Headers.Add("xxx-sitename", testUser.SiteName);
            request.Headers.Add("xxx-fullname", testUser.Fullname);
            request.Headers.Add("xxx-sitenumber", testUser.SiteNumber);
            request.Headers.Add("xxx-ssn", testUser.Ssn);
            request.Headers.Add("xxx-purpose-of-use", "Routine medical care");
            request.Headers.Add("xxx-transaction-id", imageRequestInput.TransactionId);
            request.Headers.Add("SOAPAction", "urn:ihe:iti:2007:CrossGatewayQuery");

            request.Credentials = new NetworkCredential(mstrUsername, mstrPassword);

            DocumentSet documentSet = null;

            if (imageRequestInput.Study is DocumentSet)
            {
                documentSet = (DocumentSet)imageRequestInput.Study;
            }
            if (documentSet == null)
                throw new Exception("ImageRequestInput.Study is not a DocumentSet, cannot retrieve without DocumentSet");

            imageRequestInput.RequestUrl = mstrRetrieveUrl;
            byte[] byteArray = Encoding.UTF8.GetBytes(createRetrieveRequest(documentSet.HomeCommunityId, 
                documentSet.RepositoryId, documentSet.StudyId, mstrRetrieveUrl));

            Stream requestStream = request.GetRequestStream();
            requestStream.Write(byteArray, 0, byteArray.Length);
            requestStream.Close();

            WebResponse response = null;
            try
            {
                response = request.GetResponse();
            }
            catch (WebException webEx)
            {
                LogMsg("WebException, " + webEx.Message);
                // got something other than a 200 status code but we still want to parse the response

                // i think if an XCA error code comes back, the status code will still be a 200
                // so if not 200, then throw exception
                if (webEx.Status != WebExceptionStatus.Success)
                    throw webEx;

                response = webEx.Response;
            }
            if (response == null)
                throw new Exception("Null response from server");

            string boundary = "";
            string contentType = response.Headers["Content-Type"];
            if (contentType != null)
            {
                int loc = contentType.IndexOf("boundary=");
                if (loc >= 0)
                {
                    boundary = contentType.Substring(loc + 9);
                    loc = boundary.IndexOf(";");
                    if (loc >= 0)
                    {
                        boundary = boundary.Substring(0, loc);
                    }
                }
                if (boundary.StartsWith("\""))
                {
                    boundary = boundary.Substring(1);
                }
                if (boundary.EndsWith("\""))
                {
                    boundary = boundary.Substring(0, boundary.Length - 1);
                }
            }
            //LogMsg("Boundary in response is '" + boundary + "'.", false);
            Stream responseStream = response.GetResponseStream();
            StringBuilder line = new StringBuilder();

            StringBuilder fullResponse = new StringBuilder();
            MemoryStream binaryStream = new MemoryStream();
            MemoryStream tempBinaryStream = new MemoryStream();
            bool inBinarySection = false;
            bool readingBinaryData = false;
            bool inSoapSection = false;
            bool readingSoapSection = false;
            StringBuilder xml = null;

            MemoryStream imageStream = new MemoryStream();
            byte[] buffer = new byte[1024];
            int bytesRead = responseStream.Read(buffer, 0, 1024);
            while (bytesRead > 0)
            {
                imageStream.Write(buffer, 0, bytesRead);                
                bytesRead = responseStream.Read(buffer, 0, 1024);
            }
            imageStream.Seek(0, SeekOrigin.Begin);

            //int b = responseStream.ReadByte();
            int b = imageStream.ReadByte();
            string l = "";
            while (b > -1)
            {
                line.Append((char)b);
                if (readingBinaryData)
                {
                    tempBinaryStream.WriteByte((byte)b);
                }
                if (b == 10) // LF
                {
                    l = line.ToString();
                    if (readingBinaryData)
                    {
                        if (l.Contains(boundary))
                        {
                            readingBinaryData = false;
                        }
                        else
                        {
                            binaryStream.Write(tempBinaryStream.ToArray(), 0, (int)tempBinaryStream.Length);
                            tempBinaryStream = new MemoryStream();
                        }
                    }
                    else if (readingSoapSection)
                    {
                        if (l.Contains(boundary))
                        {
                            readingSoapSection = false;
                            inSoapSection = false;
                        }
                        else
                        {
                            xml.AppendLine(l);
                        }
                    }
                    else
                    {
                        fullResponse.Append(l);
                        if (l.StartsWith("Content-Type:"))
                        {
                            if (l.Contains("soap+xml"))
                            {
                                inSoapSection = true;
                                xml = new StringBuilder();
                            }
                            else
                            {
                                inBinarySection = true;
                            }
                        }
                        else if (inSoapSection)
                        {
                            if (l == "\r\n")
                            {
                                readingSoapSection = true;
                            }
                        }
                        else if (inBinarySection)
                        {
                            if (l == "\r\n")
                            {
                                readingBinaryData = true;
                            }
                        }
                    }
                    line = new StringBuilder();
                }
                b = imageStream.ReadByte();
            }
            // when we get here, there is no more data in the imageStream however we might still need to process
            // the most recent line

            l = line.ToString();
            if (readingBinaryData)
            {
                if (l.Contains(boundary))
                {
                    readingBinaryData = false;
                }
                else
                {
                    binaryStream.Write(tempBinaryStream.ToArray(), 0, (int)tempBinaryStream.Length);
                    tempBinaryStream = new MemoryStream();
                }
            }
            else if (readingSoapSection)
            {
                if (l.Contains(boundary))
                {
                    readingSoapSection = false;
                    inSoapSection = false;
                }
                else
                {
                    xml.AppendLine(l);
                }
            }
            else
            {
                fullResponse.Append(l);
                if (l.StartsWith("Content-Type:"))
                {
                    if (l.Contains("soap+xml"))
                    {
                        inSoapSection = true;
                        xml = new StringBuilder();
                    }
                    else
                    {
                        inBinarySection = true;
                    }
                }
                else if (inSoapSection)
                {
                    if (l == "\r\n")
                    {
                        readingSoapSection = true;
                    }
                }
                else if (inBinarySection)
                {
                    if (l == "\r\n")
                    {
                        readingBinaryData = true;
                    }
                }
            }


            responseStream.Close();
            ParseXmlResponse(documentSet, xml.ToString());

            byte[] imageBuffer = binaryStream.ToArray();
            string calculatedAlder32 = VixAKCommonUtils.CalculateAdler32(imageBuffer);





            DownloadResult result = new DownloadResult(imageRequestInput.ImageQuality.ImageQualityValue + "",
                "", calculatedAlder32, response.Headers, null);
            return result;


            /*
            if (xml == null)
            {
                LogMsg("Parsed XML is null, could not find XML in response", true);
            }
            else
            {
                xmlResponse = XcaRetrieveXmlResponse.Parse(xml.ToString());
            }
            XcaRetrieveResult result = new XcaRetrieveResult(xmlResponse, filename, responseHeaders.ToArray());
            return result;
            */
        }

        /// <summary>
        /// Parses the XML looking for any errors, will be thrown as exceptions
        /// </summary>
        /// <param name="documentSet"></param>
        /// <param name="xml"></param>
        private void ParseXmlResponse(DocumentSet documentSet, string xml)
        {
            XmlDocument xmlDoc = new XmlDocument();
            XmlNamespaceManager nsmgr = new XmlNamespaceManager(xmlDoc.NameTable);
            nsmgr.AddNamespace("soapenv", "http://www.w3.org/2003/05/soap-envelope");
            nsmgr.AddNamespace("ns5", "urn:ihe:iti:xds-b:2007");
            nsmgr.AddNamespace("xop", "http://www.w3.org/2004/08/xop/include");
            nsmgr.AddNamespace("ns3", "urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0");
            xmlDoc.LoadXml(xml);

            string homeCommunityId = "";
            string repositoryUniqueId = "";
            string documentUniqueId = "";
            string mimeType = "";
            string status = "";

            XmlNode retrieveDocumentSetResponse = xmlDoc.SelectSingleNode("/soapenv:Envelope/soapenv:Body/ns5:RetrieveDocumentSetResponse", nsmgr);
            if (retrieveDocumentSetResponse == null)
                return;

            XmlNode registryResponseNode = retrieveDocumentSetResponse.SelectSingleNode("ns3:RegistryResponse", nsmgr);
            if (registryResponseNode != null)
            {
                if (registryResponseNode.Attributes["status"] != null)
                {
                    status = registryResponseNode.Attributes["status"].Value;
                }
                XmlNodeList registryErrors = registryResponseNode.SelectNodes("ns3:RegistryErrorList/ns3:RegistryError", nsmgr);
                foreach (XmlNode registryErrorNode in registryErrors)
                {
                    string codeContext = "";
                    if (registryErrorNode.Attributes["codeContext"] != null)
                        codeContext = registryErrorNode.Attributes["codeContext"].Value;

                    string errorCode = "";
                    if (registryErrorNode.Attributes["errorCode"] != null)
                        errorCode = registryErrorNode.Attributes["errorCode"].Value;

                    string location = "";
                    if (registryErrorNode.Attributes["location"] != null)
                        location = registryErrorNode.Attributes["location"].Value;

                    string severity = "";
                    if (registryErrorNode.Attributes["severity"] != null)
                        severity = registryErrorNode.Attributes["severity"].Value;

                    string msg = "Error retrieving document '" + documentSet.HomeCommunityId + ":" + documentSet.RepositoryId + ":" + documentSet.StudyId + "'";

                    if (errorCode == "XDSMissingDocument")
                    {
                        throw new ImageNotFoundException(msg + ", " + errorCode);
                    }
                    else
                    {
                        throw new Exception(msg + ", " + errorCode);
                    }
                }
            }

            XmlNode documentResponseNode = retrieveDocumentSetResponse.SelectSingleNode("ns5:DocumentResponse", nsmgr);
            if (documentResponseNode != null)
            {
                XmlNode homeCommunityIdNode = documentResponseNode.SelectSingleNode("ns5:HomeCommunityId", nsmgr);
                if (homeCommunityIdNode != null)
                {
                    homeCommunityId = homeCommunityIdNode.InnerText;
                }
                XmlNode repositoryUniqueIdNode = documentResponseNode.SelectSingleNode("ns5:RepositoryUniqueId", nsmgr);
                if (repositoryUniqueIdNode != null)
                {
                    repositoryUniqueId = repositoryUniqueIdNode.InnerText;
                }
                XmlNode documentUniqueIdNode = documentResponseNode.SelectSingleNode("ns5:DocumentUniqueId", nsmgr);
                if (documentUniqueIdNode != null)
                {
                    documentUniqueId = documentUniqueIdNode.InnerText;
                }
                XmlNode mimeTypeNode = documentResponseNode.SelectSingleNode("ns5:mimeType", nsmgr);
                if (mimeTypeNode != null)
                {
                    mimeType = mimeTypeNode.InnerText;
                }
            }
        }

        public string MetadataUrl
        {
            get { return mstrQueryUrl; }
        }

        public string ImageUrl
        {
            get { return mstrRetrieveUrl ; }
        }

        private string extractXmlFromQueryResponse(StreamReader reader, string boundary)
        {
            /*
             *
             * // JMW 2/3/2011 - this was the simple code but it is insufficient
            string line = reader.ReadLine();
            while (line != null)
            {
                if (line.StartsWith("<?xml"))
                    return line;
                line = reader.ReadLine();
            }
            return null;
            */
            StringBuilder sb = new StringBuilder();
            StringBuilder xml = null;
            string line = reader.ReadLine();
            while (line != null)
            {
                sb.AppendLine(line);
                if (line.Length > 0)
                {
                    if ((string.IsNullOrEmpty(boundary) || (line.IndexOf(boundary) < 0)))

                    //if (line.IndexOf(boundary) < 0)
                    {
                        if (line.StartsWith("<?xml"))
                        {
                            xml = new StringBuilder();
                            xml.AppendLine(line);
                        }
                        else if (xml != null)
                        {
                            xml.AppendLine(line);
                        }
                    }
                }
                else
                {
                    if (xml != null)
                        return xml.ToString();
                }
                line = reader.ReadLine();
            }
            //LogMsg("Could not find XML in response: " + sb.ToString(), true);
            if (xml != null)
                return xml.ToString();
            //LogMsg("Could not find XML in response: " + sb.ToString(), true);
            return null; 
        }

        private string createQueryRequest(string patientIcn)
        {
            /*
            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");
            sb.Append("<soapenv:Body>");
            sb.Append("<urn:AdhocQueryRequest maxResults=\"1000\" id=\"http://testUri\" xmlns:urn=\"urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0\">");
            sb.Append("<urn:ResponseOption returnComposedObjects=\"true\" returnType=\"LeafClass\"/>");
            sb.Append("<urn1:AdhocQuery id=\"urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d\" xmlns:urn1=\"urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0\">");
            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryPatientId\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'");
            sb.Append(patientIcn);
            sb.Append("^^^&amp;2.16.840.1.113883.3.166&amp;ISO'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");
            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryStatus\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'urn:oasis:names:tc:ebxml-regrep:StatusType:Approved'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");
            sb.Append("</urn1:AdhocQuery>");
            sb.Append("</urn:AdhocQueryRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");
            return sb.ToString();*/
            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");

            sb.Append("<soapenv:Header xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<wsa:To>http://localhost:7010/DocSearchXcaService/RespondingGateway_Service</wsa:To>");
            sb.Append("<wsa:MessageID>urn:uuid:3C7D747C17C16CAA4A1287680303993</wsa:MessageID>");
            sb.Append("<wsa:Action>urn:ihe:iti:2007:CrossGatewayQuery</wsa:Action>");
            sb.Append("</soapenv:Header>");
            sb.Append("<soapenv:Body>");
            sb.Append("<urn:AdhocQueryRequest maxResults=\"1000\" id=\"http://testUri\" xmlns:urn=\"urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0\">");
            sb.Append("<urn:ResponseOption returnComposedObjects=\"true\" returnType=\"LeafClass\"/>");
            sb.Append("<urn1:AdhocQuery id=\"urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d\" xmlns:urn1=\"urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0\">");
            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryPatientId\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'");
            sb.Append(patientIcn);
            sb.Append("^^^&amp;2.16.840.1.113883.3.166&amp;ISO'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");

            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryStatus\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'urn:oasis:names:tc:ebxml-regrep:StatusType:Approved'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");
            sb.Append("</urn1:AdhocQuery>");
            sb.Append("</urn:AdhocQueryRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");

            return sb.ToString();
        }

        private string createRetrieveRequest(string homeCommunityId, string repositoryId, 
            string documentUniqueId, string url)
        {
            /*
            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");
            sb.Append("<soapenv:Body>");
            sb.Append("<ns5:RetrieveDocumentSetRequest xmlns:ns5=\"urn:ihe:iti:xds-b:2007\">");
            sb.Append("<ns5:DocumentRequest>");
            sb.Append("<ns5:HomeCommunityId>");
            sb.Append(homeCommunityId);
            sb.Append("</ns5:HomeCommunityId>");
            sb.Append("<ns5:RepositoryUniqueId>");
            sb.Append(repositoryId);
            sb.Append("</ns5:RepositoryUniqueId>");
            sb.Append("<ns5:DocumentUniqueId>");
            sb.Append(documentUniqueId);
            sb.Append("</ns5:DocumentUniqueId>");
            sb.Append("</ns5:DocumentRequest>");
            sb.Append("</ns5:RetrieveDocumentSetRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");
            return sb.ToString();
            */

            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");
            sb.Append("<soapenv:Header xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<wsa:To>" + url + "</wsa:To>");
            sb.Append("<wsa:MessageID>urn:uuid:3C7D747C17C16CAA4A1287680303993</wsa:MessageID>");
            sb.Append("<wsa:Action>urn:ihe:iti:2007:CrossGatewayRetrieve</wsa:Action>");
            sb.Append("</soapenv:Header>");

            sb.Append("<soapenv:Body>");
            sb.Append("<ns5:RetrieveDocumentSetRequest xmlns:ns5=\"urn:ihe:iti:xds-b:2007\">");
            sb.Append("<ns5:DocumentRequest>");
            sb.Append("<ns5:HomeCommunityId>");
            sb.Append(homeCommunityId);
            sb.Append("</ns5:HomeCommunityId>");
            sb.Append("<ns5:RepositoryUniqueId>");
            sb.Append(repositoryId);
            sb.Append("</ns5:RepositoryUniqueId>");
            sb.Append("<ns5:DocumentUniqueId>");
            sb.Append(documentUniqueId);
            sb.Append("</ns5:DocumentUniqueId>");
            sb.Append("</ns5:DocumentRequest>");
            sb.Append("</ns5:RetrieveDocumentSetRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");
            return sb.ToString();

        }
    }
}
