/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;

namespace VixAKCommon.gov.va.med.imaging.vix.tester
{
    public class VixTestResult
    {
        private List<VixPatientTestResult> patients = new List<VixPatientTestResult>();
        private DateTime startTime;
        private DateTime endTime;
        private bool completed = false;
        private readonly bool mbThreaded;
        private readonly int mintWorkerCount;
        private readonly int mintImageQualities;
        private readonly bool mbCached;
        private readonly bool mbMetadataOnly;
        private readonly bool mbStopOnNon404Error;
        private readonly int mintPatientCount;

        public int PatientCount
        {
            get { return mintPatientCount; }
        } 


        public bool StopOnNon404Error
        {
            get { return mbStopOnNon404Error; }
        } 


        public bool MetadataOnly
        {
            get { return mbMetadataOnly; }
        }

        public bool Cached
        {
            get { return mbCached; }
        }

        public int ImageQualities
        {
            get { return mintImageQualities; }
        }

        public int WorkerCount
        {
            get { return mintWorkerCount; }
        }

        public bool Threaded
        {
            get { return mbThreaded; }
        }

        public VixTestResult(bool threaded, int workerCount, int imageQualities, bool cached, 
            bool metadataOnly, bool stopOnNon404Error, int patientCount)
        {
            mbThreaded = threaded;
            mintWorkerCount = workerCount;
            mintImageQualities = imageQualities;
            mbCached = cached;
            this.mbMetadataOnly = metadataOnly;
            this.mbStopOnNon404Error = stopOnNon404Error;
            this.mintPatientCount = patientCount;
            startTime = DateTime.Now;
        }

        public void SetComplete()
        {
            if (!completed)
            {
                endTime = DateTime.Now;
            }
            completed = true;
        }

        public List<VixPatientTestResult> Patients
        {
            get { return patients; }
        }

        private int getPatientsWithImages()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                if (patientTest.TotalImages > 0)
                {
                    count++;
                }
            }
            return count;
        }

        private int getImagesDownloaded()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.ImagesDownloaded;
            }
            return count;
        }

        private int getImagesFailed()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.ImagesFailed;
            }
            return count;
        }

        private int getCountedImages()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.TotalImages;
            }
            return count;
        }

        private int get404Errors()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.Error404;
            }
            return count;
        }

        private int getChecksumErrors()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.ChecksumErrors;
            }
            return count;
        }

        private int get500Errors()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.Error500;
            }
            return count;
        }

        private int getOtherErrors()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.ErrorOther;
            }
            return count;
        }

        private int getSkippedImages()
        {
            int count = 0;
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                count += patientTest.SkippedImages;
            }
            return count;
        }

        private string getMachineNameCounts()
        {
            StringBuilder sb = new StringBuilder();
            Dictionary<string, int> machineNameCounts = new Dictionary<string, int>();
            foreach (VixPatientTestResult patientTest in this.patients)
            {
                foreach (string key in patientTest.MachineResponseCount.Keys)
                {
                    if (machineNameCounts.ContainsKey(key))
                    {
                        int count = machineNameCounts[key];
                        count += patientTest.MachineResponseCount[key];
                        machineNameCounts[key] = count;
                    }
                    else
                    {
                        machineNameCounts.Add(key, patientTest.MachineResponseCount[key]);
                    }
                }
            }
            sb.AppendLine("Received responses from '" + machineNameCounts.Count + "' machines.");
            foreach (string key in machineNameCounts.Keys)
            {
                sb.AppendLine("\t" + key + ": " + machineNameCounts[key]);
            }
            
            return sb.ToString();
        }

        public override string ToString()
        {
            SetComplete();
            StringBuilder sb = new StringBuilder();

            TimeSpan ts = endTime.Subtract(startTime);
            
            sb.AppendLine("Results for " + DateTime.Now.ToString());
            sb.AppendLine("Processing took  : " + VixAKCommonUtils.ConvertTime(ts));
            sb.AppendLine("Repeat Workers   : " + mintWorkerCount);
            sb.AppendLine("Threaded         : " + mbThreaded);
            sb.AppendLine("Images Cached    : " + mbCached);
            sb.AppendLine("Image Qualities  : " + mintImageQualities);
            sb.AppendLine("Patients with images: " + getPatientsWithImages());
            sb.AppendLine("Images downloaded: " + getImagesDownloaded());
            sb.AppendLine("Failed images    : " + getImagesFailed());
            sb.AppendLine("Stop Non 404 Errors: " + mbStopOnNon404Error);
            sb.AppendLine("Counted images   : " + getCountedImages());
            sb.AppendLine("404 Errors       : " + get404Errors());
            sb.AppendLine("500 Errors       : " + get500Errors());
            sb.AppendLine("Other Errors     : " + getOtherErrors());
            sb.AppendLine("Skipped Images   : " + getSkippedImages());
            sb.AppendLine("Checksum Errors  : " + getChecksumErrors());
            sb.Append(getMachineNameCounts());
            patients.Sort();

            foreach (VixPatientTestResult patientResult in patients)
            {
                sb.AppendLine("-----------------------");
                sb.Append(patientResult);
            }

            return sb.ToString();
        }   
    }

    public class SavableVixPatientTestResultList : SavableObject
    {
        private List<SavableVixPatientTestResult> results = new List<SavableVixPatientTestResult>();

        public List<SavableVixPatientTestResult> Results
        {
            get { return results; }
            set { results = value; }
        }

        public SavableVixPatientTestResultList()
        {
        }

        public SavableVixPatientTestResultList(List<VixPatientTestResult> patientResults)
        {
            foreach (VixPatientTestResult patientResult in patientResults)
            {
                results.Add(new SavableVixPatientTestResult(patientResult));
            }
        }

        public override bool Equals(object obj)
        {
            if (obj is SavableVixPatientTestResultList)
            {
                SavableVixPatientTestResultList that = (SavableVixPatientTestResultList)obj;
                if (this.results.Count != that.results.Count)
                    return false;
                for (int i = 0; i < this.results.Count; i++)
                {
                    SavableVixPatientTestResult s = this.results[i];
                    SavableVixPatientTestResult t = that.results[i];
                    if (!s.Equals(t))
                        return false;
                }
                return true;
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class SavableVixPatientTestResult : SavableObject
    {
        private string mstrPatientIcn;        
        private int mintStudies;
        private int mintTotalImages;
        private int mintImagesDownloaded;
        private int mintImagesFailed;
        private int mintError404;
        private int mintError500;
        private int mintErrorOther;
        private int mintSkippedImages;
        private int mintChecksumErrors;

        public string PatientIcn
        {
            get { return mstrPatientIcn; }
            set { mstrPatientIcn = value; }
        }

        public int SkippedImages
        {
            get { return mintSkippedImages; }
            set { mintSkippedImages = value; }
        }

        public int Error404
        {
            get { return mintError404; }
            set { mintError404 = value; }
        }

        public int Error500
        {
            get { return mintError500; }
            set { mintError500 = value; }
        }

        public int ErrorOther
        {
            get { return mintErrorOther; }
            set { mintErrorOther = value; }
        }

        public int ChecksumErrors
        {
            get { return mintChecksumErrors; }
            set { mintChecksumErrors = value; }
        }

        public int Studies
        {
            get { return mintStudies; }
            set { mintStudies = value; }
        }

        public int TotalImages
        {
            get { return mintTotalImages; }
            set { mintTotalImages = value; }
        }

        public int ImagesDownloaded
        {
            get { return mintImagesDownloaded; }
            set { mintImagesDownloaded = value; }
        }

        public int ImagesFailed
        {
            get { return mintImagesFailed; }
            set { mintImagesFailed = value; }
        }

        public SavableVixPatientTestResult()
        {
        }

        public SavableVixPatientTestResult(VixPatientTestResult testResult)
        {
            mstrPatientIcn = testResult.Patient.PatientIcn;
            mintChecksumErrors = testResult.ChecksumErrors;
            mintError404 = testResult.Error404;
            mintError500 = testResult.Error500;
            mintErrorOther = testResult.ErrorOther;
            mintImagesDownloaded = testResult.ImagesDownloaded;
            mintImagesFailed = testResult.ImagesFailed;
            mintSkippedImages = testResult.SkippedImages;
            mintStudies = testResult.Studies;
            mintTotalImages = testResult.TotalImages;
        }

        public override bool Equals(object obj)
        {
            if (obj is SavableVixPatientTestResult)
            {
                SavableVixPatientTestResult that = (SavableVixPatientTestResult)obj;
                return (this.mstrPatientIcn == that.mstrPatientIcn) &&
                    (this.mintChecksumErrors == that.mintChecksumErrors) &&
                    (this.mintError404 == that.mintError404) &&
                    (this.mintError500 == that.mintError500) &&
                    (this.mintErrorOther == that.mintErrorOther) &&
                    (this.mintImagesDownloaded == that.mintImagesDownloaded) &&
                    (this.mintImagesFailed == that.mintImagesFailed) &&
                    (this.mintSkippedImages == that.mintSkippedImages) &&
                    (this.mintStudies == that.mintStudies) &&
                    (this.mintTotalImages == that.mintTotalImages);
            }
            return false;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode();
        }
    }

    public class VixPatientTestResult : IComparer<VixPatientTestResult>, IComparable<VixPatientTestResult>
    {
        private readonly ImagingPatient patient;
        private int mintStudies;
        private int mintTotalImages;
        private int mintImagesDownloaded;
        private int mintImagesFailed;
        private int mintError404;
        private int mintError500;
        private int mintErrorOther;
        private int mintSkippedImages;
        private int mintChecksumErrors;
        private readonly bool mbIsExam;
        private readonly Dictionary<string, int> machineResponseCount = new Dictionary<string, int>();

        public Dictionary<string, int> MachineResponseCount
        {
            get { return machineResponseCount; }
        }

        public void incrementMachineResponse(string machineName)
        {
            incrementMachineResponse(machineName, 1);
        }

        private void incrementMachineResponse(string machineName, int incrementAmount)
        {
            if (machineName != null)
            {
                if (machineResponseCount.ContainsKey(machineName))
                {
                    int count = machineResponseCount[machineName];
                    count += incrementAmount;
                    MachineResponseCount[machineName] = count;
                }
                else
                {
                    machineResponseCount.Add(machineName, incrementAmount);
                }
            }
        }

        /*
        public void addMachineResponseCounts(Dictionary<string, int> machineResponseCount)
        {
            foreach (string key in machineResponseCount.Keys)
            {
                int count = machineResponseCount[key];
                incrementMachineResponse(key, count);
            }
        }*/

        public bool IsExam
        {
            get { return mbIsExam; }
        } 


        public int SkippedImages
        {
            get { return mintSkippedImages; }
            set { mintSkippedImages = value; }
        }

        public int Error404
        {
            get { return mintError404; }
            set { mintError404 = value; }
        }        

        public int Error500
        {
            get { return mintError500; }
            set { mintError500 = value; }
        }        

        public int ErrorOther
        {
            get { return mintErrorOther; }
            set { mintErrorOther = value; }
        }

        public int ChecksumErrors
        {
            get { return mintChecksumErrors; }
            set { mintChecksumErrors = value; }
        }

        public ImagingPatient Patient
        {
            get { return patient; }
        }

        public int Studies
        {
            get { return mintStudies; }
            set { mintStudies = value; }
        }

        public int TotalImages
        {
            get { return mintTotalImages; }
            set { mintTotalImages = value; }
        }

        public int ImagesDownloaded
        {
            get { return mintImagesDownloaded; }
            set { mintImagesDownloaded = value; }
        }

        public int ImagesFailed
        {
            get { return mintImagesFailed; }
            set { mintImagesFailed = value; }
        }

        public VixPatientTestResult(ImagingPatient patient, bool isExam)
        {
            this.patient = patient;
            mbIsExam = isExam;
            mintErrorOther = 0;
            mintError500 = 0;
            mintError404 = 0;
            mintImagesDownloaded = 0;
            mintImagesFailed = 0;
            mintStudies = 0;
            mintTotalImages = 0;
            mintSkippedImages = 0;
            mintChecksumErrors = 0;
        }

        public VixPatientTestResult(ImagingPatient patient) : this(patient, false)
        {
            
        }

        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            if (mbIsExam)
            {
                sb.AppendLine("Exam [" + patient.ToString(false) + "]");
                sb.AppendLine("[" + mintTotalImages + "] total images, [" + mintImagesDownloaded + "] images downloaded, [" + mintImagesFailed + "] failed images, [" + mintSkippedImages + "] skipped images.");                
            }
            else
            {
                sb.AppendLine("Patient [" + patient.ToString(true) + "]");
                sb.AppendLine("[" + mintStudies + "] studies, [" + mintTotalImages + "] total images, [" + mintImagesDownloaded + "] images downloaded, [" + mintImagesFailed + "] failed images, [" + mintSkippedImages + "] skipped images.");
            }
            sb.AppendLine("Error results: [" + mintError404 + "] 404 errors, [" + mintError500 + "] 500 errors, [" + mintChecksumErrors + "] checksum errors, [" + mintErrorOther + "] other errors");
            return sb.ToString();
        }

        public int Compare(VixPatientTestResult x, VixPatientTestResult y)
        {
            if (this.mbIsExam)
            {
                return x.Patient.PatientIcn.CompareTo(y.patient.PatientIcn);
            }
            else
            {
                return x.Patient.PatientName.CompareTo(y.patient.PatientName);
            }
            
        }

        public int CompareTo(VixPatientTestResult other)
        {
            return Compare(this, other);
        }
    }
}
