/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.exceptions;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class AWIVTestAction : BaseVixTestAction
    {

        public static string AWIV_ACTION_NAME = "AWIV VIX Test";

        private VixTestResult vixTestResult = null;


        public AWIVTestAction()
            : base(AWIV_ACTION_NAME)
        {
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
        }

        public override void runAction()
        {
            LogInfoMsg("Running action [" + GetActionName() + "]");
            LogInfoMsg("Client workstation [" + GetHostname() + "]");

            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            LogInfoMsg("Retrieving data from metadata url [" + metadataUrl + "], image url [" + imageUrl + "]");

            AWIVServiceImpl awivService = new AWIVServiceImpl(metadataUrl, imageUrl, LogMsg);

            List<ImagingPatient> patients = GetSelectedPatients();
            List<ImageQuality> imageQualities = GetSelectedImageQualities();

            string siteNumber = necessaryProperties[PROPERTY_SITE_NUMBER].ToString();
            LogInfoMsg("Site number [" + siteNumber + "]");
            LogInfoMsg("Selected Patients: ");
                        
            foreach (ImagingPatient patient in patients)
            {
                LogInfoMsg(patient.ToString(true));
            }
            vixTestResult = new VixTestResult(false, 1, imageQualities.Count, false, false, false, patients.Count);
            ProgressBarMax(patients.Count);
            foreach (ImagingPatient patient in patients)
            {                
                if (!mbCanRun)
                    break;
                testPatient(awivService, patient, siteNumber, imageQualities);
                IncrementProgress();
            }
            vixTestResult.SetComplete();
            LogInfoMsg("-----------------------");
            LogInfoMsg(vixTestResult.ToString());
            ActionComplete(ActionCompletedResult.success);
        }

        private bool testPatient(AWIVServiceImpl awivService, ImagingPatient patient, string siteNumber, List<ImageQuality> imageQualities)
        {
            LogInfoMsg("Testing patient '" + patient.ToString(true) + "'");
            
            VixPatientTestResult patientTestResult = new VixPatientTestResult(patient);
            vixTestResult.Patients.Add(patientTestResult);

            List<string> cprsIdentifiers = getCprsIdentifiersForPatient(patient.PatientIcn);
            LogInfoMsg("Querying for '" + cprsIdentifiers.Count + "' groups of images");
            patientTestResult.Studies = cprsIdentifiers.Count;
            foreach (string cprsIdentifier in cprsIdentifiers)
            {
                if (!mbCanRun)
                {
                    return false;
                }
                LogInfoMsg("Finding images for id '" + cprsIdentifier + "'");
                Study study = null;
                try
                {
                    study = awivService.getImageList(siteNumber, patient.PatientIcn, cprsIdentifier);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error getting image list for cprs Identifier '" + cprsIdentifier + ", " + ex.Message);
                    patientTestResult.ErrorOther++;
                    return false;
                }

                List<Instance> images = study.Series[0].Instances;
                LogInfoMsg("Group has '" + images.Count + "' images");
                patientTestResult.TotalImages += images.Count;

                foreach (Instance image in images)
                {
                    string transactionId = Guid.NewGuid().ToString();
                    try
                    {
                        AWIVDownloadResult result = 
                            awivService.DownloadImage(new AWIVImageRequestInput(study, image, ImageQuality.REFERENCE, transactionId, null));
                        patientTestResult.ImagesDownloaded++;
                        //LogInfoMsg("Downloaded image '" + image.InstanceId + "' with transaction Id '" + transactionId + "'");
                    }
                    catch (AbstractVixWebException webEx)
                    {
                        LogErrorMsg("Error downloading image '" + image.InstanceId + "' with transaction Id '" + transactionId + "', " + webEx.Message);
                        patientTestResult.ImagesFailed++;

                        
                        if (webEx.StatusCode == System.Net.HttpStatusCode.NotFound)
                        {
                            patientTestResult.Error404++;
                        }
                        else if (webEx.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                        {
                            patientTestResult.Error500++;
                        }
                        else
                        {
                            patientTestResult.ErrorOther++;
                        }
                    }
                    catch (Exception ex)
                    {
                        patientTestResult.ImagesFailed++;
                        patientTestResult.ErrorOther++;
                        LogErrorMsg("Error downloading image '" + image.InstanceId + "' with transaction Id '" + transactionId + "', " + ex.Message);
                    }
                }
            }
            LogInfoMsg("Patient '" + patient.ToString(true) + "' is complete");
            return true;
        }

        private List<string> getCprsIdentifiersForPatient(string patientIcn)
        {
            List<string> identifiers = new List<string>();
            switch (patientIcn)
            {
               
            }

            return identifiers;
        }

        public override string GetActionName()
        {
            return AWIV_ACTION_NAME;
        }

        public override bool ValidateNecessaryProperties()
        {
            return base.ValidateNecessaryProperties();
        }
    }
}
