/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using VixAKCommon.gov.va.med.imaging.vix.tester;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;
using log4net;
using System.Xml.Serialization;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public abstract class BaseAction :  IComparable<BaseAction>
    {

        protected readonly Dictionary<string, object> necessaryProperties = 
            new Dictionary<string, object>();
        protected bool mbEnabled = true;
        protected bool mbCanRun = true;
        private bool mbResultsVerified = false;
       
        protected string mstrTestDescription = null; // description of the action and its parameters - what do the parameters test with this action

        protected const string PROPERTY_RESULTS_FILENAME = "Results Filename";
        protected const string PROPERTY_OUTPUT_RESULTS_ENABLED = "Output Results Enabled";
        protected const string PROPERTY_VERIFY_RESULTS_ENABLED = "Verify Results Enabled";
        protected const string PROPERTY_EXPECTED_RESULT_EXCEPTION = "Expected Result Exception";
        protected const string PROPERTY_FAILED_RESULTS_OUTPUT_DIRECTORY = "Failed Results Output Directory";

        public BaseAction()
        {
            necessaryProperties.Add(PROPERTY_OUTPUT_RESULTS_ENABLED, "false");
            necessaryProperties.Add(PROPERTY_RESULTS_FILENAME, "");
            necessaryProperties.Add(PROPERTY_VERIFY_RESULTS_ENABLED, "false");
            necessaryProperties.Add(PROPERTY_EXPECTED_RESULT_EXCEPTION, "false");
            necessaryProperties.Add(PROPERTY_FAILED_RESULTS_OUTPUT_DIRECTORY, ""); // directory to output results that do not match expected results
        }

        public bool ResultsVerified
        {
            get { return mbResultsVerified; }
        }

        /// <summary>
        /// Determines if the expected result is an exception and that should be handled as the result (rather than reported as an error)
        /// </summary>
        /// <returns></returns>
        protected bool IsExpectedResultException()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_EXPECTED_RESULT_EXCEPTION))
            {
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_EXPECTED_RESULT_EXCEPTION].ToString()))
                    return false;
                return bool.Parse(necessaryProperties[PROPERTY_EXPECTED_RESULT_EXCEPTION].ToString());
            }
            return false;
        }

        protected bool IsOutputResultsEnabled()
        {
            if ((necessaryProperties.ContainsKey(PROPERTY_OUTPUT_RESULTS_ENABLED) && (necessaryProperties.ContainsKey(PROPERTY_RESULTS_FILENAME))))
            {
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_RESULTS_FILENAME].ToString()))
                    return false;
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_OUTPUT_RESULTS_ENABLED].ToString()))
                    return false;
                return bool.Parse(necessaryProperties[PROPERTY_OUTPUT_RESULTS_ENABLED].ToString());
            }
            return false;
        }

        protected string GetResultsFilename()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_RESULTS_FILENAME))
            {
                return necessaryProperties[PROPERTY_RESULTS_FILENAME].ToString();
            }
            return null;
        }

        public bool IsVerifyResultsEnabled()
        {
            if ((necessaryProperties.ContainsKey(PROPERTY_VERIFY_RESULTS_ENABLED) && (necessaryProperties.ContainsKey(PROPERTY_RESULTS_FILENAME))))
            {
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_RESULTS_FILENAME].ToString()))
                    return false;
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_VERIFY_RESULTS_ENABLED].ToString()))
                    return false;
                return bool.Parse(necessaryProperties[PROPERTY_VERIFY_RESULTS_ENABLED].ToString());
            }
            return false;
        }

        protected void OutputExpectedResults(SavableObject result)
        {
            string filename = GetResultsFilename();
            if ((filename != null) && (result != null))
            {
                LogInfoMsg("Outputting expected results to file '" + filename + "'.");

                XmlSerializer serializer = new XmlSerializer(result.GetType());
                TextWriter writer = new StreamWriter(filename);
                serializer.Serialize(writer, result);
                writer.Close();
            }            
        }

        protected bool VerifyExpectedResults(SavableObject result)
        {
            string filename = GetResultsFilename();
            if (filename != null)
            {
                if (result == null)
                {
                    LogErrorMsg("Received result is null, cannot compare to expected result");
                    return false;
                }
                if (!System.IO.File.Exists(filename))
                {
                    LogErrorMsg("Cannot find file '" + filename + "' to compare results to");
                    return false;
                }                
                LogInfoMsg("Comparing result to expected result from file '" + filename + "'.");
                XmlSerializer serializer = new XmlSerializer(result.GetType());
                TextReader reader = new StreamReader(filename);
                SavableObject expectedResult = (SavableObject)serializer.Deserialize(reader);
                //LogInfoMsg("Expected Result is of type: " + expectedResult.GetType().ToString());
                //LogInfoMsg("Input Result is of type: " + result.GetType().ToString());
                reader.Close();
                mbResultsVerified = true;
                bool comparisonResult = result.Equals(expectedResult);
                if (!comparisonResult)
                {
                    OutputFailedResults(result);
                }
                return comparisonResult;
            }
            return false;
        }

        private void OutputFailedResults(SavableObject result)
        {
            if (necessaryProperties.ContainsKey(PROPERTY_FAILED_RESULTS_OUTPUT_DIRECTORY))
            {

                string outputFilename = necessaryProperties[PROPERTY_FAILED_RESULTS_OUTPUT_DIRECTORY].ToString() + "\\" + this.GetActionName() + "_" + DateTime.Now.Ticks + ".xml";
                LogInfoMsg("Outputting failed results to file '" + outputFilename + "' for action '" + GetActionName() + ".");
                XmlSerializer serializer = new XmlSerializer(result.GetType());
                TextWriter writer = new StreamWriter(outputFilename);
                serializer.Serialize(writer, result);
                writer.Close();
            }
        }

        private Dictionary<string, string> defaultValues = new Dictionary<string, string>();
        private string mstrDefaultValuesFilename = "";

        protected void LoadDefaultValues(string filename)
        {
            defaultValues.Clear();
            mstrDefaultValuesFilename = filename;
            XmlDocument xmlDoc = new XmlDocument();
            if (System.IO.File.Exists(filename))
            {
                xmlDoc.Load(filename);
                XmlNodeList actionValueNodes = xmlDoc.SelectNodes("VixAK-Defaults/ActionValue");
                foreach (XmlNode actionValueNode in actionValueNodes)
                {
                    string name = actionValueNode.Attributes["Name"].Value;
                    string value = actionValueNode.Attributes["Value"].Value;
                    defaultValues.Add(name, value);
                }
            }
        }

        public void SaveDefaultValues()
        {
            XmlDocument xmlDoc = new XmlDocument();
            XmlNode rootNode = xmlDoc.CreateElement("VixAK-Defaults");
            xmlDoc.AppendChild(rootNode);
            foreach (string key in defaultValues.Keys)
            {
                XmlNode node = xmlDoc.CreateElement("ActionValue");
                rootNode.AppendChild(node);
                XmlAttribute attr = xmlDoc.CreateAttribute("Name");
                attr.Value = key;
                node.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("Value");
                attr.Value = defaultValues[key];
                node.Attributes.Append(attr);
            }
            xmlDoc.Save(mstrDefaultValuesFilename);
        }

        /// <summary>
        /// Adds the necessary property, tries to get the default value from the default value dictionary if it exists, otherwise
        /// it uses the default value supplied
        /// </summary>
        /// <param name="propertyName"></param>
        /// <param name="defaultValue"></param>
        protected void SetNecessaryPropertyWithDefault(string propertyName, string defaultValue)
        {
            string value = GetDefaultValue(propertyName);
            if (value == null)
                value = defaultValue;
            necessaryProperties.Add(propertyName, value);            
        }

        public string TestDescription
        {
            get { return mstrTestDescription; }
            set { mstrTestDescription = value; }
        }

        public bool CanRun
        {
            get { return mbCanRun; }
            set { mbCanRun = value; }
        }

        public bool Enabled
        {
            get { return mbEnabled; }
            set { mbEnabled = value; }
        }

        public Dictionary<string, object> NecessaryProperties
        {
            get { return necessaryProperties; }
        }

        protected ILog Logger()
        {
            return LogManager.GetLogger(this.GetType().Name);
        }

        public event LogMessageDelegate OnLogMessageEvent;
        public event IncrementProgressDelegate OnIncrementProgressEvent;
        public event SetProgressBarMax OnSetProgressBarMaxEvent;
        public event ActionCompleteDelegate OnActionCompleteEvent;
        public event DoApplicationEventsDelegate OnDoApplicationEventsEvent;

        protected void DoApplicationEvents()
        {
            if (OnDoApplicationEventsEvent != null)
            {
                OnDoApplicationEventsEvent();
            }
        }

        protected void ActionComplete(ActionCompletedResult actionCompleteResult)
        {
            if (OnActionCompleteEvent != null)
            {
                OnActionCompleteEvent(this, actionCompleteResult);
            }
        }

        protected void ProgressBarMax(int max)
        {
            if (OnSetProgressBarMaxEvent != null)
            {
                OnSetProgressBarMaxEvent(max);
            }
        }

        protected string LogMsg(string msg)
        {
            if (OnLogMessageEvent != null)
            {
                OnLogMessageEvent(msg);
            }
            return msg;
        }

        protected void IncrementProgress()
        {
            if (OnIncrementProgressEvent != null)
            {
                OnIncrementProgressEvent();
            }
        }

        protected void LogInfoMsg(string msg)
        {
            Logger().Info(LogMsg(msg));
            //LogMsg(msg);
        }

        protected string LogInfoMsgReturnMsg(string msg)
        {
            LogInfoMsg(msg);
            return msg;
        }

        protected void LogErrorMsg(string msg)
        {
            Logger().Error(LogMsg(msg));
        }             

        public override string ToString()
        {
            return GetActionName();
        }

        public void SetPropertiesFromXml(XmlNode actionNode)
        {
            XmlNodeList stringNodes = actionNode.SelectNodes("StringProperty");
            foreach (XmlNode stringNode in stringNodes)
            {
                string name = stringNode.Attributes["Name"].Value;
                string value = stringNode.Attributes["Value"].Value;
                this.necessaryProperties[name] = value;
            }
            XmlNodeList customOptionNodes = actionNode.SelectNodes("ActionCustomOption");
            foreach (XmlNode customOptionNode in customOptionNodes)
            {
                string name = customOptionNode.Attributes["Name"].Value;
                ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[name];
                customOption.Options.Clear();
                XmlNodeList imagingPatientNodes = customOptionNode.SelectNodes(ImagingPatient.IMAGING_PATIENT_NODE_NAME);
                foreach (XmlNode patientNode in imagingPatientNodes)
                {
                    string patientName = patientNode.Attributes["PatientName"].Value;
                    string patientIcn = patientNode.Attributes["PatientICN"].Value;
                    bool active = true;
                    if (patientNode.Attributes["active"] != null)
                    {
                        active = bool.Parse(patientNode.Attributes["active"].Value);
                    }
                    ImagingPatient patient = new ImagingPatient(patientIcn, patientName);
                    patient.Active = active;
                    if (active)
                    {
                        customOption.Options.Add(patient);

                        XmlNode sitesNode = patientNode.SelectSingleNode(ImagingPatient.IMAGING_PATIENT_SITES_NODE_NAME);
                        if (sitesNode != null)
                        {
                            XmlNodeList siteNodeList = sitesNode.SelectNodes(ImagingPatient.IMAGING_PATIENT_SITE_NODE_NAME);
                            foreach (XmlNode siteNode in siteNodeList)
                            {
                                string siteNumber = siteNode.Attributes["SiteNumber"].Value;
                                patient.SiteNumbers.Add(siteNumber);
                            }
                        }
                    }
                }
                XmlNodeList imageQualityNodes = customOptionNode.SelectNodes(ImageQuality.IMAGE_QUALITY_NODE_NAME);
                foreach (XmlNode qualityNode in imageQualityNodes)
                {
                    string qualityName = qualityNode.Attributes["QualityName"].Value;
                    int qualityValue = int.Parse( qualityNode.Attributes["QualityValue"].Value);
                    ImageQuality quality = new ImageQuality(qualityValue, qualityName);
                    customOption.Options.Add(quality);
                }
                XmlNodeList booleanNodes = customOptionNode.SelectNodes(BooleanOption.BOOLEAN_OPTION_NODE_NAME);
                foreach (XmlNode booleanNode in booleanNodes)
                {
                    bool value = bool.Parse(booleanNode.Attributes["Value"].Value);
                    BooleanOption option = new BooleanOption(value);
                    customOption.Options.Add(option);
                }
                XmlNodeList vixUserNodes = customOptionNode.SelectNodes(VixTestUser.VIX_USER_NODE_NAME);
                foreach (XmlNode vixUserNode in vixUserNodes)
                {
                    string duz = vixUserNode.Attributes["Duz"].Value;
                    string fullname = vixUserNode.Attributes["Fullname"].Value;
                    string siteName = vixUserNode.Attributes["SiteName"].Value;
                    string siteNumber = vixUserNode.Attributes["SiteNumber"].Value;
                    string ssn = vixUserNode.Attributes["Ssn"].Value;
                    string securityToken = "";
                    if (vixUserNode.Attributes["SecurityToken"] != null)
                    {
                        securityToken = vixUserNode.Attributes["SecurityToken"].Value;
                    }

                    VixTestUser user = new VixTestUser(duz, fullname, ssn, siteName, siteNumber, securityToken);
                    customOption.Options.Add(user);
                }
                //TODO: move other custom options to their own class

                customOption.Options.AddRange(SiteServiceActionType.fromXml(customOptionNode));
            }
        }

        public XmlNode toXml(XmlDocument xmlDoc)
        {
            XmlNode actionNode = xmlDoc.CreateElement("Action");
            XmlAttribute actionNameAttr = xmlDoc.CreateAttribute("Type");
            actionNameAttr.Value = GetActionName();
            actionNode.Attributes.Append(actionNameAttr);
            XmlAttribute enabledAttr = xmlDoc.CreateAttribute("Enabled");
            enabledAttr.Value = mbEnabled.ToString();
            actionNode.Attributes.Append(enabledAttr);
            if (mstrTestDescription != null)
            {
                XmlAttribute testDescriptionAttr = xmlDoc.CreateAttribute("TestDescription");
                testDescriptionAttr.Value = mstrTestDescription;
                actionNode.Attributes.Append(testDescriptionAttr);
            }
            foreach (string key in necessaryProperties.Keys)
            {
                object value = necessaryProperties[key];
                if (value is string)
                {
                    XmlNode propertyNode = xmlDoc.CreateElement("StringProperty");
                    XmlAttribute attr = xmlDoc.CreateAttribute("Name");
                    attr.Value = key;
                    propertyNode.Attributes.Append(attr);
                    attr = xmlDoc.CreateAttribute("Value");
                    attr.Value = necessaryProperties[key].ToString();
                    propertyNode.Attributes.Append(attr);
                    actionNode.AppendChild(propertyNode);
                }
                else if (value is ActionCustomOption)
                {
                    ActionCustomOption actionCustomOption = (ActionCustomOption)value;
                    XmlNode customOptionNode = xmlDoc.CreateElement("ActionCustomOption");
                    XmlAttribute attr = xmlDoc.CreateAttribute("Name");
                    attr.Value = actionCustomOption.Name;
                    customOptionNode.Attributes.Append(attr);

                    foreach (BaseCustomOption customOption in actionCustomOption.Options)
                    {
                        XmlNode optionNode = customOption.toXml(xmlDoc);
                        customOptionNode.AppendChild(optionNode);
                    }
                    actionNode.AppendChild(customOptionNode);
                }
            }
            return actionNode;
        }

        protected bool validatePropertyIncluded(string propertyName, bool canBeEmpty)
        {
            if (necessaryProperties.ContainsKey(propertyName))
            {
                if (canBeEmpty)
                    return true;
                if (necessaryProperties[propertyName] != null)
                {
                    return true;
                }
            }
            LogErrorMsg("Missing property [" + propertyName + "]");
            return false;
        }

        protected bool validatePropertyIncluded(string propertyName)
        {
            return validatePropertyIncluded(propertyName, false);
        }

        public virtual void stopAction()
        {
            mbCanRun = false;
        }

        protected string GetDefaultValue(string key)
        {
            if (defaultValues == null)
                return null;
            if (defaultValues.ContainsKey(key))
                return defaultValues[key];
            return null;
        }

        public void SetDefaultValue(string key, string value)
        {
            if (defaultValues != null)
            {
                if (defaultValues.ContainsKey(key))
                    defaultValues[key] = value;
                else
                    defaultValues.Add(key, value);
                //SaveDefaultValues();
            }
        }


        #region "Abstract Functions"
        public abstract void runAction();

        /*
        public abstract void runActionOutputResults(string resultsFilename);

        public abstract void runActionVerifyResults(string expectedResultsFilename);
        */
        public abstract string GetActionName();

        public abstract bool ValidateNecessaryProperties();        

        #endregion


        public int CompareTo(BaseAction that)
        {
            return this.GetActionName().CompareTo(that.GetActionName());
        }
        
    }
}
