/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public abstract class BaseCopyAction : BaseAction
    {

        protected const string PROPERTY_TARGET_DIRECTORY = "Target Directory";
        protected const string PROPERTY_REPEAT_COUNT = "Repeat Count";
        protected const string PROPERTY_CSV_RESULTS_FILE = "CSV Results File";
        protected const string PROPERTY_REPEAT_CONTINUOUSLY = "Repeat Continuously";

        public BaseCopyAction()
        {
            necessaryProperties.Add(PROPERTY_TARGET_DIRECTORY, @"c:\temp\target");
            necessaryProperties.Add(PROPERTY_REPEAT_COUNT, "5");
            necessaryProperties.Add(PROPERTY_CSV_RESULTS_FILE, @"c:\temp\filecopy.csv");
            necessaryProperties.Add(PROPERTY_REPEAT_CONTINUOUSLY, "false");
        }

        public abstract ObjectCopyResult ExecuteCopy(string input, string outputDir);

        protected abstract string getRemotePath();

        protected abstract string getCopyActionDescription();

        public override void runAction()
        {
            LogInfoMsg("Running action [" + GetActionName() + "]");

            string filename = getRemotePath();
            string outputDir = getTargetDirectory();
            bool repeatContinuously = getRepeatContinuously();
            int count = getRepeatCount();

            LogInfoMsg(getCopyActionDescription());
            try
            {
                List<ObjectCopyResult> results = new List<ObjectCopyResult>();
                DateTime startTime = DateTime.Now;
                if (repeatContinuously)
                {
                    if (mbCanRun)
                    {
                        results.Add(ExecuteCopy(filename, outputDir));
                    }
                }
                else
                {
                    for (int i = 0; i < count; i++)
                    {
                        results.Add(ExecuteCopy(filename, outputDir));
                    }
                }
                DateTime endTime = DateTime.Now;
                TimeSpan span = endTime.Subtract(startTime);
                LogInfoMsg(GetActionName() + " process took [" + span.TotalMilliseconds + "] ms");
                CreateCsvResults(results, getCsvOutputFilename());
                ActionComplete(ActionCompletedResult.success);
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error during " + GetActionName() + ", " + ex.Message);
                ActionComplete(ActionCompletedResult.error);
            }       
        }

        private void CreateCsvResults(List<ObjectCopyResult> fileCopyResults, string filename)
        {
            StreamWriter writer = null;
            try
            {
                writer = new StreamWriter(filename);
                WriteHeaderLine(writer);
                foreach (ObjectCopyResult result in fileCopyResults)
                {
                    WriteFileCopyResult(writer, result);
                }
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error writing CSV results to file [" + filename + "], " + ex.Message);
            }
            finally
            {
                if (writer != null)
                    writer.Close();
            }
        }

        private const string CSV_DELIMITER = ",";

        private void WriteFileCopyResult(StreamWriter writer, ObjectCopyResult result)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(result.Filename);
            sb.Append(CSV_DELIMITER);
            sb.Append(result.OutputFilename);
            sb.Append(CSV_DELIMITER);
            sb.Append(result.StartTime);
            sb.Append(CSV_DELIMITER);
            sb.Append(result.EndTime);
            sb.Append(CSV_DELIMITER);
            sb.Append(result.Ms.ToString());
            sb.Append(CSV_DELIMITER);
            sb.Append(result.FileSizeInBytes.ToString());
            sb.Append(CSV_DELIMITER);
            sb.Append(calculateKBSec(result.Ms, result.FileSizeInBytes));
            writer.WriteLine(sb.ToString());
        }

        private string calculateKBSec(long ms, long bytes)
        {
            double kb = bytes / 1024.0;
            double sec = ms / 1000.0;
            return (kb / sec) + "";
        }

        private void WriteHeaderLine(StreamWriter writer)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("Source Filename");
            sb.Append(CSV_DELIMITER);
            sb.Append("Target Filename");
            sb.Append(CSV_DELIMITER);
            sb.Append("Start Time");
            sb.Append(CSV_DELIMITER);
            sb.Append("End Time");
            sb.Append(CSV_DELIMITER);
            sb.Append("ms");
            sb.Append(CSV_DELIMITER);
            sb.Append("Bytes Transferred");
            sb.Append(CSV_DELIMITER);
            sb.Append("KB/sec");
            writer.WriteLine(sb.ToString());
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_REPEAT_COUNT))
                return false;
            if (!validatePropertyIncluded(PROPERTY_TARGET_DIRECTORY))
                return false;
            if (!validatePropertyIncluded(PROPERTY_CSV_RESULTS_FILE))
                return false;
            if (!validatePropertyIncluded(PROPERTY_REPEAT_CONTINUOUSLY))
                return false;
            string dir = getTargetDirectory();
            if (!Directory.Exists(dir))
            {
                try
                {
                    Directory.CreateDirectory(dir);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error creating directory [" + dir + "], " + ex.Message);
                    return false;
                }
            }
            try
            {
                getRepeatCount();
            }
            catch (Exception ex)
            {
                LogErrorMsg("Repeat count invalid, " + ex.Message);
                return false;
            }
            try
            {
                getRepeatContinuously();
            }
            catch (Exception ex)
            {
                LogErrorMsg("Repeat continuously invalid, " + ex.Message);
                return false;
            }

            return true;
        }

        protected string getTargetDirectory()
        {
            return necessaryProperties[PROPERTY_TARGET_DIRECTORY].ToString();
        }

        protected int getRepeatCount()
        {
            return int.Parse(necessaryProperties[PROPERTY_REPEAT_COUNT].ToString());
        }

        protected string getCsvOutputFilename()
        {
            return necessaryProperties[PROPERTY_CSV_RESULTS_FILE].ToString();
        }

        protected bool getRepeatContinuously()
        {
            return bool.Parse(necessaryProperties[PROPERTY_REPEAT_CONTINUOUSLY].ToString());
        }
    }

    public abstract class ObjectCopyResult
    {
        private readonly string mstrFilename;

        public string Filename
        {
            get { return mstrFilename; }
        }

        private readonly string mstrOutputFilename;

        public string OutputFilename
        {
            get { return mstrOutputFilename; }
        }

        private readonly long mlMs;

        public long Ms
        {
            get { return mlMs; }
        }

        private readonly string mstrStartTime;

        public string StartTime
        {
            get { return mstrStartTime; }
        }

        private readonly string mstrEndTime;

        public string EndTime
        {
            get { return mstrEndTime; }
        }

        private readonly long mintFileSizeInBytes;

        public long FileSizeInBytes
        {
            get { return mintFileSizeInBytes; }
        }

        public ObjectCopyResult(string filename, string outputFilename, long ms, string startTime, 
            string endTime, long fileSize)
        {
            mstrEndTime = endTime;
            mstrFilename = filename;
            mstrOutputFilename = outputFilename;
            mstrStartTime = startTime;
            mlMs = ms;
            mintFileSizeInBytes = fileSize;
        }
    }
}
