/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class BaseSiteServiceTestAction : BaseAction
    {
        protected string mstrActionName;

        public override string GetActionName()
        {
            return mstrActionName;
        }

        protected const string PROPERTY_SITE_SERVICE_URL = "Site Service URL";
        protected const string PROPERTY_SITE_SERVICE_ACTION = "Site Service Action";
        protected const string PROPERTY_SITE_SERVICE_QUERY = "Site Service Query";

        /// <summary>
        /// Filename to output expected results to 
        /// </summary>
        protected string mstrResultsFilename = null;
        /// <summary>
        /// Holds the expected results to compare to
        /// </summary>
        protected string mstrExpectedResultsFilename = null;

        public BaseSiteServiceTestAction(string actionName)
        {
            mstrActionName = actionName;
            this.LoadDefaultValues(VixAKCommonUtils.getDefaultValuesFile());            
            SetNecessaryPropertyWithDefault(PROPERTY_SITE_SERVICE_QUERY, "756");

            List<SiteServiceActionType> actions = SiteServiceActionType.GetSiteServiceActions();
            ActionCustomOption actionOptions = new ActionCustomOption(PROPERTY_SITE_SERVICE_ACTION, false);
            foreach (SiteServiceActionType action in actions)
            {
                actionOptions.Options.Add((BaseCustomOption)action);
            }
            this.necessaryProperties.Add(PROPERTY_SITE_SERVICE_ACTION, actionOptions);
        }

        public void runAction(SiteServiceBaseInterface siteService, string siteServiceUrl)
        {
            SiteServiceActionType action = GetSiteServiceAction();
            string actionQuery = necessaryProperties[PROPERTY_SITE_SERVICE_QUERY].ToString();

            Logger().Info(LogMsg("Running action [" + GetActionName() + "]"));
            Logger().Info(LogMsg("Client workstation [" + VixAKCommonUtils.GetHostname() + "]"));
            Logger().Info(LogMsg("Retrieving data from url [" + siteServiceUrl + "]."));
            Logger().Info(LogMsg("Site Service Action [" + action.ToString() + "]."));
            Logger().Info(LogMsg("Site Service Querty [" + actionQuery + "]."));

            ProgressBarMax(1);
            SavableObject result = null;
            try
            {
                if (action == SiteServiceActionType.GET_SITE)
                {
                    Site site = siteService.GetSite(actionQuery);
                    result = site;
                    DisplaySite(site);
                }
                else if (action == SiteServiceActionType.GET_SITES)
                {
                    List<Site> sites = siteService.GetSites(actionQuery);
                    result = new SiteArray(sites);
                    foreach (Site site in sites)
                    {
                        DisplaySite(site);
                    }
                }
                else if (action == SiteServiceActionType.GET_ALL_SITES)
                {
                    List<Site> sites = siteService.GetAllSites();
                    result = new SiteArray(sites);
                    foreach (Site site in sites)
                    {
                        DisplaySite(site);
                    }
                }
                else
                {
                    LogErrorMsg("Action [" + action.ToString() + "] not supported...");
                    ActionComplete(ActionCompletedResult.error);
                    return;
                }
            }
            catch (Exception ex)
            {
                LogErrorMsg(ex.Message);
                ActionComplete(ActionCompletedResult.error);
                return;
            }
            ActionCompletedResult actionResult = ActionCompletedResult.success;
            if (IsVerifyResultsEnabled())
            {
                if (VerifyExpectedResults(result))
                {
                    LogInfoMsg("Result matched expected result.");
                }
                else
                {
                    LogErrorMsg("Result did not match expected result.");
                    actionResult = ActionCompletedResult.resultsFailed;
                }
            }

            if (IsOutputResultsEnabled())
            {
                OutputExpectedResults(result);
            }

            ActionComplete(actionResult);
        }

        private void DisplaySite(Site site)
        {
            LogInfoMsg("Retrieved site [" + site.ToString(true) + "].");
        }

        public override void runAction()
        {
            throw new Exception("The method or operation is not implemented.");
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_SITE_SERVICE_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_SITE_SERVICE_ACTION))
                return false;
            if (!validatePropertyIncluded(PROPERTY_SITE_SERVICE_QUERY))
                return false;
            try
            {
                GetSiteServiceAction();
            }
            catch (Exception ex)
            {
                LogErrorMsg(ex.Message);
                return false;
            }
            return true;
        }

        protected SiteServiceActionType GetSiteServiceAction()
        {
            ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_SITE_SERVICE_ACTION];
            if (customOption.Options.Count != 1)
                throw new Exception("Must specify only 1 site service action");
            return (SiteServiceActionType)customOption.Options[0];
        }
    }
}
