/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using System.IO;
using VixAKCommon.gov.va.med.imaging.vix.exceptions;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    

    public abstract class BaseVistaRadAction : BaseVixTestAction
    {
        protected const string PROPERTY_TEXT_URL = "TextUrl";

        public BaseVistaRadAction(string actionName) : base(actionName)
        {
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_TEXT_URL, "");
            
            //necessaryProperties.Remove(PROPERTY_IMAGE_QUALITIES);
        }

        protected bool testExam(Exam exam, VistaRadServiceImpl vistaRadService, 
            VixPatientTestResult patientTestResult)
        {
            Exam fatExam = null;
            try
            {
                fatExam = vistaRadService.getExamDetails(exam.ExamId);
                if (exam.SiteNumber == "200")
                {
                    // not getting text files, so request count is same as actual image count
                    patientTestResult.TotalImages += fatExam.Images.Count;                    
                }
                else
                {
                    patientTestResult.TotalImages += fatExam.Images.Count * 2;
                }
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting exam details for exam '" + exam.ExamId + "', " + ex.Message);
                patientTestResult.ErrorOther++;
                return false;
            }
            LogInfoMsg("Found '" + fatExam.Images.Count + "' images for exam '" + fatExam.ExamId + "'.");
            bool writeImageChecksum = GetOutputImageChecksums();
            bool verifyImageChecksums = GetVerifyImageChecksums();
            foreach (ExamImage image in fatExam.Images)
            {
                if (!mbCanRun)
                    break;
                string transactionId = Guid.NewGuid().ToString();                

                VistaRadImageRequestInput request = null;
                try
                {
                    request =
                        new VistaRadImageRequestInput(image, transactionId);
                    VistaRadDownloadResult result = vistaRadService.DownloadImage(request);
                    if (writeImageChecksum)
                    {
                        OutputImageChecksum(image.ImageId, ImageQuality.DIAGNOSTIC, result.ReturnedImageChecksum, 
                            result.ImageBufferSize, result.ReturnedImageFormat);
                        string checksumImageDirectory = getChecksumImageDirectory();
                        if (checksumImageDirectory != null && checksumImageDirectory.Length > 0)
                        {
                            VixTestCommon.OutputImage(result.ImageBuffer, request.Image.ImageId, 
                                ImageQuality.DIAGNOSTIC, request.TransactionId, checksumImageDirectory);
                        }
                    }
                    
                    patientTestResult.incrementMachineResponse(result.HeaderMachineName);

                    if (verifyImageChecksums)
                    {
                        string expectedChecksum = GetImageChecksum(image.ImageId, ImageQuality.DIAGNOSTIC);
                        if (expectedChecksum == null)
                        {
                            throw new ChecksumNotFoundException("Did not find expected checksum for image '" + image.ImageId + "', test for this image fails without expected checksum.");
                        }
                        VixTestCommon.VerifyExpectedChecksum(result.ImageBuffer, new ChecksumTestOptions(expectedChecksum), result.ReturnedImageChecksum, request.TransactionId);
                    }


                    patientTestResult.ImagesDownloaded++;
                    OutputImageDownloadFormatResult(transactionId, image.RequestString, result.ReturnedImageQuality,
                        result.ReturnedImageFormat, "");
                    //LogInfoMsg("Downloaded image '" + image.InstanceId + "' with transaction Id '" + transactionId + "'");
                }
                catch (AbstractVixWebException webEx)
                {
                    LogErrorMsg("Error downloading image '" + image.ImageId + "' with transaction Id '" + transactionId + "', " + webEx.Message);
                    OutputImageDownloadFormatResult(transactionId, image.RequestString, "", "", webEx.Message);
                    patientTestResult.ImagesFailed++;

                    if (webEx.StatusCode == System.Net.HttpStatusCode.NotFound)
                    {
                        patientTestResult.Error404++;
                    }
                    else if (webEx.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                    {
                        patientTestResult.Error500++;
                    }
                    else
                    {
                        patientTestResult.ErrorOther++;
                    }
                }
                catch (ChecksumException cX)
                {
                    patientTestResult.ChecksumErrors++;
                    patientTestResult.ImagesFailed++;
                    OutputImageDownloadFormatResult(transactionId, image.RequestString, "", "", cX.Message);
                    LogErrorMsg("Error downloading image '" + image.ImageId + "' with transaction Id '" + transactionId + "', " + cX.Message);
                }
                catch (Exception ex)
                {
                    patientTestResult.ImagesFailed++;
                    patientTestResult.ErrorOther++;
                    OutputImageDownloadFormatResult(transactionId, image.RequestString, "", "", ex.Message);
                    LogErrorMsg("Error downloading image '" + image.ImageId + "' with transaction Id '" + transactionId + "', " + ex.Message);
                }

                // don't get the text file for station 200
                if (image.SiteNumber != "200")
                {
                    transactionId = Guid.NewGuid().ToString();
                    request = new VistaRadImageRequestInput(image, transactionId);
                    try
                    {
                        VistaRadDownloadResult result = vistaRadService.DownloadTextFile(request);
                        patientTestResult.ImagesDownloaded++;
                        patientTestResult.incrementMachineResponse(result.HeaderMachineName);
                        //LogInfoMsg("Downloaded image '" + image.InstanceId + "' with transaction Id '" + transactionId + "'");
                    }
                    catch (AbstractVixWebException webEx)
                    {
                        LogErrorMsg("Error downloading text file for image '" + image.ImageId + "' with transaction Id '" + transactionId + "', " + webEx.Message);
                        patientTestResult.ImagesFailed++;

                        if (webEx.StatusCode == System.Net.HttpStatusCode.NotFound)
                        {
                            patientTestResult.Error404++;
                        }
                        else if (webEx.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                        {
                            patientTestResult.Error500++;
                        }
                        else
                        {
                            patientTestResult.ErrorOther++;
                        }
                    }
                    catch (Exception ex)
                    {
                        patientTestResult.ImagesFailed++;
                        patientTestResult.ErrorOther++;
                        LogErrorMsg("Error downloading text file for image '" + image.ImageId + "' with transaction Id '" + transactionId + "', " + ex.Message);
                    }
                }
            }
            return true;
        }

        private void OutputImageDownloadFormatResult(string transactionId, string requestString, 
            string returnedImageQuality, string returnedImageFormat, string errorMsg)
        {
            if (formatWritter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append(transactionId);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + requestString + "\"");
                sb.Append(formatWriterDelimiter);
                sb.Append(returnedImageQuality);
                sb.Append(formatWriterDelimiter);
                sb.Append(returnedImageFormat);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + errorMsg + "\"");
                formatWritter.WriteLine(sb.ToString());
            }
        }
    }
}
