/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;
using System.Runtime.CompilerServices;
using System.IO;
using System.Xml.Serialization;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public delegate void OutputImageChecksumDelegate(string imageUrn, ImageQuality imageQuality, string checksum, 
        long imageSize, string returnedImageFormat);
    public delegate string GetImageChecksumDelegate(string imageUrn, ImageQuality imageQuality);

    public abstract class BaseVixTestAction : BaseAction
    {      
        protected string mstrActionName;

        protected const string PROPERTY_METADATA_URL = "MetadataUrl";
        protected const string PROPERTY_IMAGE_URL = "ImageUrl";

        protected const string PROPERTY_COUNT_STUDIES = "Count Studies Only";
        protected const string PROPERTY_THREADED = "Threaded";
        protected const string PROPERTY_MAX_WORKERS = "Repeat Workers";
        protected const string PROPERTY_PATIENTS = "Test Patients";
        protected const string PROPERTY_SITE_NUMBER = "Site Number";
        protected const string PROPERTY_IMAGE_QUALITIES = "Image Qualities";

        protected const string PROPERTY_ROLLUP_RESULTS = "Rollup Results";
        protected const string PROPERTY_TOMCAT_LOG_DIR = "Tomcat Log Dir";
        protected const string PROPERTY_GENERATE_PATIENT_RESULTS = "Generate Patient Results";
        protected const string PROPERTY_OUTPUT_DIRECTORY = "Results Output Directory";
        protected const string PROPERTY_CACHE_CLEAR = "Cache Clear";

        protected const string PROPERTY_OUTPUT_IMAGE_FORMAT = "Output Image Format";
        protected const string PROPERTY_VALIDATE_RESULTS_WITH_LOG = "Validate Results With Transaction Log"; // not really used anymore
        protected const string PROPERTY_FAIL_ON_NON_404_ERRORS = "Fail on non-404 errors";

        protected const string PROPERTY_RUN_FOREVER = "Run Forever";
        protected const string PROPERTY_EXCLUDE_FAILED_IMAGES = "Exclude Failed Images"; // applies to run forever tests
        protected const string PROPERTY_OUTPUT_IMAGE_CHECKSUMS = "Output Image Checksums";
        protected const string PROPERTY_VERIFY_IMAGE_CHECKSUMS = "Verify Image Checksums";
        protected const string PROPERTY_IMAGE_CHECKSUMS_FILENAME = "Image Checksums Filename"; // file containing the checksum values (to write to or read from)        

        protected const string PROPERTY_FAIL_IF_CHECKSUM_FROM_VIX_DOES_NOT_MATCH_CALCULATED = "Fail of Checksum From VIX Does Not Match Calculated";

        // directory to put images when generating checksums (for future reference)
        protected const string PROPERTY_CHECKSUM_IMAGE_DIRECTORY = "Checksum Image Directory"; 

        /// <summary>
        /// boolean Property to determine if when a checksum fails, is the image outputted for further debugging?
        /// </summary>
        protected const string PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE = "Output Checksum Failed Image";
        /// <summary>
        /// If the above property is true, this property must be set to the directory to output the failed image
        /// </summary>
        protected const string PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE_DIRECTORY = "Output Checksum Failed Directory";

        protected VixTestResult lastTestResult = null;

        private int threadCount = 0;
        private List<ThreadedVixTester> currentWorkers = new List<ThreadedVixTester>();
        private bool mbAlreadyProcessedComplete;

        protected StreamWriter formatWritter = null;
        protected TextWriter validateLogWriter = null;
        protected TextWriter imageChecksumWriter = null;
        protected string outputImageFormatsFilename = "";
        private string validateLogWriterFilename = "";

        private VixInterface mVixInterface = null;
        private int repeatCount = 0;
        private Dictionary<string, string> failedImageRequests = new Dictionary<string, string>();
        private Dictionary<string, string> imageChecksums = new Dictionary<string, string>();
        private string mstrFailedImageRequestsFilename = "";            

        protected VixTestResult LastTestResult
        {
            get { return lastTestResult; }
        }

        public BaseVixTestAction(string actionName)  :base  ()
        {
            mstrActionName = actionName;
            mstrFailedImageRequestsFilename = VixAKCommonUtils.getApplicationDir() + @"\failedImageRequests.xml";            
            this.LoadDefaultValues(VixAKCommonUtils.getDefaultValuesFile());

            SetNecessaryPropertyWithDefault(PROPERTY_COUNT_STUDIES, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_THREADED, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_MAX_WORKERS, "1");
            SetNecessaryPropertyWithDefault(PROPERTY_SITE_NUMBER, "");
            SetNecessaryPropertyWithDefault(PROPERTY_ROLLUP_RESULTS, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_TOMCAT_LOG_DIR, "c:\\Program Files\\Apache Software Foundation\\Tomcat 6.0\\logs");
            SetNecessaryPropertyWithDefault(PROPERTY_GENERATE_PATIENT_RESULTS, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_OUTPUT_DIRECTORY, "d:\\vixtests");
            SetNecessaryPropertyWithDefault(PROPERTY_CACHE_CLEAR, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_OUTPUT_IMAGE_FORMAT, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_VALIDATE_RESULTS_WITH_LOG, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_FAIL_ON_NON_404_ERRORS, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_OUTPUT_IMAGE_CHECKSUMS, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_VERIFY_IMAGE_CHECKSUMS, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_CHECKSUMS_FILENAME, "c:\\Program Files\\Vista\\Imaging\\vixak\\checksums.txt");
            SetNecessaryPropertyWithDefault(PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE, "false");
            SetNecessaryPropertyWithDefault(PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE_DIRECTORY, "c:\\Program Files\\Vista\\Imaging\\vixak\\failedImages");
            SetNecessaryPropertyWithDefault(PROPERTY_CHECKSUM_IMAGE_DIRECTORY, "");

            List<ImagingPatient> testPatients = VixAKCommonUtils.GetTestPatients();

            ActionCustomOption patientOptions = new ActionCustomOption(PROPERTY_PATIENTS);
            foreach (ImagingPatient patient in testPatients)
            {
                patientOptions.Options.Add((BaseCustomOption)patient);
            }
            this.necessaryProperties.Add(PROPERTY_PATIENTS, patientOptions);

            List<ImageQuality> imageQualities = ImageQuality.GetImageQualities();
            ActionCustomOption qualitiesOptions = new ActionCustomOption(PROPERTY_IMAGE_QUALITIES);
            foreach (ImageQuality quality in imageQualities)
            {
                qualitiesOptions.Options.Add((BaseCustomOption)quality);
            }
            this.necessaryProperties.Add(PROPERTY_IMAGE_QUALITIES, qualitiesOptions);
        }

        private void LoadFailedImageRequests()
        {
            if (File.Exists(mstrFailedImageRequestsFilename))
            {
                LogInfoMsg("Reading failed image requests from file '" + mstrFailedImageRequestsFilename + "'");
                try
                {
                    failedImageRequests.Clear();
                    XmlSerializer serializer = new XmlSerializer(typeof(List<string>));
                    TextReader reader = new StreamReader(mstrFailedImageRequestsFilename);
                    List<string> failedFiles = (List<string>)serializer.Deserialize(reader);
                    reader.Close();

                    foreach (string file in failedFiles)
                    {
                        failedImageRequests.Add(file, "");
                    }

                    LogInfoMsg("Read '" + failedImageRequests.Count + "' failed image requests from file.");
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error reading failed image requests from file '" + mstrFailedImageRequestsFilename + "', " + ex.Message);
                }
            }
        }

        private void SaveFailedImageRequests()
        {
            List<string> failedFiles = new List<string>();
            foreach (string key in failedImageRequests.Keys)
            {
                failedFiles.Add(key);
            }

            LogInfoMsg("Saving failed image requests to file '" + mstrFailedImageRequestsFilename + "'");
            try
            {
                XmlSerializer serializer = new XmlSerializer(typeof(List<string>));
                TextWriter writer = new StreamWriter(mstrFailedImageRequestsFilename);
                serializer.Serialize(writer, failedFiles);
                writer.Close();
                LogInfoMsg("Saved failed image requests to file.");
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error saving failed image requests to file '" + mstrFailedImageRequestsFilename + "', " + ex.Message);
            }
        }

        public override string GetActionName()
        {
            return mstrActionName;
        }

        public override void runAction()
        {
            throw new Exception("The method or operation is not implemented.");
        }

        private static string hostname = null;
        protected static string GetHostname()
        {
            if (hostname == null)
            {
                hostname = System.Environment.MachineName;
            }
            return hostname;
        }

        protected const string formatWriterDelimiter = ",";

        protected void CreateImageFormatHeader()
        {
            if (formatWritter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("Image Id");
                sb.Append(formatWriterDelimiter);
                sb.Append("Transaction Id");
                sb.Append(formatWriterDelimiter);
                sb.Append("Requested Image Format");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Image Format");
                sb.Append(formatWriterDelimiter);
                sb.Append("Requested Quality");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Quality");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Checksum");
                sb.Append(formatWriterDelimiter);
                sb.Append("Error Message");
                formatWritter.WriteLine(sb.ToString());
            }
        }

        protected void CreateValidateLogHeader()
        {
            if (validateLogWriter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("Image Id");
                sb.Append(formatWriterDelimiter);
                sb.Append("Transaction Id");
                sb.Append(formatWriterDelimiter);
                sb.Append("Requested Image Format");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Image Format");
                sb.Append(formatWriterDelimiter);
                sb.Append("Requested Quality");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Quality");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Code");
                sb.Append(formatWriterDelimiter);
                sb.Append("Error Message");
                validateLogWriter.WriteLine(sb.ToString());
            }
        }

        protected void runAction(VixInterface vixInterface)
        {
            repeatCount++;
            mVixInterface = vixInterface;
            threadCount = 0;
            failedStatus = false;
            mbAlreadyProcessedComplete = false;
            bool threaded = GetThreaded();
            // not sure why this was commented out
            vixInterface.OnLogMessageEvent += LogInfoMsgReturnMsg;

            
            Logger().Info(LogMsg("Running action [" + GetActionName() + "]"));
            Logger().Info(LogMsg("Client workstation [" + GetHostname() + "]"));
            Logger().Info(LogMsg("Retrieving data from metadata url [" + vixInterface.MetadataUrl + "], image url [" + vixInterface.ImageUrl + "]"));

            LoadFailedImageRequests();

            bool outputImageFormats = GetOutputImageFormat();
            bool validateResultsWithLog = GetValidateResultsWithLog();
            bool stopOnNon404Error = GetFailOnNon404Errors();
            bool runForever = GetRunForever();
            bool excludeFailedImages = GetExcludeFailedImages();
            bool outputImageChecksums = GetOutputImageChecksums();
            bool verifyImageChecksums = GetVerifyImageChecksums();
            string imageChecksumsFilename = GetImageChecksumsFilename();            
            bool outputFailedChecksumImages = getOutputChecksumFailedImage();
            string outputFailedChecksumImagesDirectory = getOutputChecksumFailedImageDirectory();
            outputImageFormatsFilename = "";
            if ((outputImageFormats) && (!threaded))
            {
                if (formatWritter == null)
                {
                    outputImageFormatsFilename = VixAKCommonUtils.getApplicationDir() + "\\imageFormats.csv";
                    LogInfoMsg("Writing output image formats to file [" + outputImageFormatsFilename + "]");
                    formatWritter = new StreamWriter(outputImageFormatsFilename, false);
                    CreateImageFormatHeader();
                }
            }
            else
            {
                LogInfoMsg("Output image formats false or threading enabled, prevents output image format");
            }
            validateLogWriterFilename = "";
            if (validateResultsWithLog)
            {
                if (validateLogWriter == null)
                {
                    validateLogWriterFilename = VixAKCommonUtils.getApplicationDir() + "\\testResults.csv";
                    LogInfoMsg("Writing image test results to file [" + validateLogWriterFilename + "]");
                    validateLogWriter = TextWriter.Synchronized(new StreamWriter(validateLogWriterFilename, false));
                    CreateValidateLogHeader();
                }
            }

            if (threaded)
            {
                outputImageChecksums = false;
            }
            if (outputImageChecksums)
            {
                if (imageChecksumWriter == null)
                {
                    if (imageChecksumsFilename == null)
                    {
                        throw new Exception("Image checksum filename is null, cannot output image checksums");
                    }
                    else
                    {
                        LogInfoMsg("Writing image checksums to file [" + imageChecksumsFilename + "]");
                        imageChecksumWriter = TextWriter.Synchronized(new StreamWriter(imageChecksumsFilename, false));
                        string checksumImageDirectory = getChecksumImageDirectory();
                        if (checksumImageDirectory != null && checksumImageDirectory.Length > 0)
                        {
                            LogInfoMsg("Writing images to directory '" + checksumImageDirectory + "'.");
                        }
                        else
                        {
                            LogInfoMsg("Not writing images locally");
                        }
                    }
                }
            }
            if (verifyImageChecksums)
            {
                if (imageChecksumsFilename == null)
                {
                    throw new Exception("Image checksum filename is null, cannot verify image checksums");
                }
                if (!LoadImageChecksums(imageChecksumsFilename))
                {
                    mbCanRun = false;
                }
            }

            if (outputFailedChecksumImages)
            {
                if (outputFailedChecksumImagesDirectory == null)
                {
                    throw new Exception("Attempting to output failed image checksum files but directory for output is null");
                }
                LogInfoMsg("Failed checksum images will be outputted to directory [" + outputFailedChecksumImagesDirectory + "].");
            }

            List<ImageQuality> qualities = GetSelectedImageQualities();
            Logger().Info(LogMsg("Image Qualities: "));
            foreach (ImageQuality quality in qualities)
            {
                Logger().Info(LogMsg(quality.ToString()));
            }
            
            string siteNumber = necessaryProperties[PROPERTY_SITE_NUMBER].ToString();
            LogInfoMsg("Site number [" + siteNumber + "]");
            bool countImages = GetCountImagesOnly();
            List<ImagingPatient> patients = GetSelectedPatients();
            Logger().Info(LogMsg("Selected Patients: "));
            foreach (ImagingPatient patient in patients)
            {
                Logger().Info(LogMsg(patient.ToString(true)));
            }
            int workers = GetWorkerCount();

            LogInfoMsg("Threaded = [" + threaded + "], workers = [" + workers + "]");
           
            if (threaded)
            {
                int maxThreads = workers * patients.Count;
                ProgressBarMax(maxThreads);
                ThreadPool.SetMaxThreads(maxThreads, 100000);
                LogInfoMsg("Setting to [" + maxThreads + "] max threads");
            }
            LogInfoMsg("Run Forever = [" + runForever + "]");
            if (runForever)
            {
                LogInfoMsg("Exclude Failed Images = [" + excludeFailedImages + "]");
            }
            else
            {
                excludeFailedImages = false;
            }
            LogInfoMsg("Output image checksums [" + outputImageChecksums + "]");
            LogInfoMsg("Verify image checksums [" + verifyImageChecksums + "]");
            currentWorkers.Clear();
            bool cacheClear = GetCacheClear();
            lastTestResult = new VixTestResult(threaded, workers, qualities.Count, 
                !cacheClear, countImages, stopOnNon404Error, patients.Count);

            if (threaded)
            {
                processPatients(workers, patients, qualities, vixInterface, siteNumber,
                    threaded, countImages, stopOnNon404Error, excludeFailedImages);
            }
            else
            {
                // always run at least once
                processPatients(workers, patients, qualities, vixInterface, siteNumber,
                    threaded, countImages, stopOnNon404Error, excludeFailedImages);
                while (runForever && mbCanRun)
                {
                    LogInfoMsg("Completed all patients, runForever is true - running patient tests again");
                    processPatients(workers, patients, qualities, vixInterface, siteNumber, 
                        threaded, countImages, stopOnNon404Error, excludeFailedImages);
                }
                ProcessingComplete();
            }                       
        }

        private void processPatients(int workers, List<ImagingPatient> patients, List<ImageQuality> qualities,
            VixInterface vixInterface, string siteNumber, bool threaded, bool countImages, bool stopOnNon404Error,
            bool excludeFailedImages)
        {
            for (int i = 0; i < workers; i++)
            {
                foreach (ImagingPatient patient in patients)
                {
                    if (!mbCanRun)
                        break;

                    ThreadedVixTester tester =
                        new ThreadedVixTester(patient, qualities, vixInterface, countImages,
                        siteNumber, validateLogWriter, stopOnNon404Error, 
                        getOutputChecksumFailedImage(), getOutputChecksumFailedImageDirectory(),
                        getFailIfChecksumFromVixDoesNotMatchCalculatedChecksum(),
                        getChecksumImageDirectory());
                    tester.OnLogMessageEvent += LogMsg;
                    tester.OnCriticalExceptionEvent += OnCriticalException;
                    if (excludeFailedImages)
                    {
                        tester.OnIsImageRequestOK += IsImageRequestOK;
                        tester.OnAddFailedImageRequest += AddFailedImageRequest;
                    }
                    currentWorkers.Add(tester);
                    if (threaded)
                    {
                        tester.OnPatientCompleteEvent += OnPatientComplete;
                        if (GetVerifyImageChecksums())
                        {
                            tester.OnGetImageChecksumEvent += GetImageChecksum;
                        }
                        try
                        {
                            threadCount++;
                            /*
                            Thread thread = new Thread(new ThreadStart(tester.ProcessPatientVoid));                            
                            thread.Start();
                            Thread.Sleep(50);
                             */
                            ThreadPool.QueueUserWorkItem(new WaitCallback(tester.ProcessPatient));
                        }
                        catch (Exception ex)
                        {
                            Logger().Error(LogMsg(ex.Message));
                        }
                    }
                    else
                    {
                        tester.FormatWriter = formatWritter;
                        tester.OnIncrementProgressEvent += IncrementProgress;
                        tester.OnSetProgressBarMaxEvent += ProgressBarMax;
                        if (GetOutputImageChecksums())
                        {
                            tester.OnOutputImageChecksum += OutputImageChecksum;
                        }
                        if (GetVerifyImageChecksums())
                        {
                            //tester.OnIsImageChecksumOK += IsImageChecksumOK;
                            tester.OnGetImageChecksumEvent += GetImageChecksum;
                        }
                        lastTestResult.Patients.Add(tester.ProcessPatient());
                    }

                } // each patient
            }     
        }

        /// <summary>
        /// Load the checksums for the images from the specfified file. 
        /// If there is a critical error in loading the checksums (such as a duplicate checksum with different values), return false to prevent processing
        /// </summary>
        /// <param name="fname"></param>
        /// <returns></returns>
        protected bool LoadImageChecksums(string fname)
        {
            imageChecksums.Clear();
            LogInfoMsg("Loading image checksums from file '" + fname + "'.");
            StreamReader reader = new StreamReader(fname);
            string line = reader.ReadLine();
            int duplicateChecksumEntries = 0; // count the number of entries that were duplicated but had same value - not critical but strange
            while (line != null)
            {
                string [] pieces = line.Split('|');
                if (pieces.Length >= 2)
                {
                    try
                    {
                        imageChecksums.Add(pieces[0], pieces[1]);
                    }
                    catch (Exception ex)
                    {
                        if (imageChecksums.ContainsKey(pieces[0]))
                        {
                            string existingChecksum = imageChecksums[pieces[0]];
                            if (existingChecksum == pieces[1])
                            {
                                // the image identifier was found more than once in the file but the checksum value is the same.
                                // not critical but weird
                                duplicateChecksumEntries++;
                                //LogInfoMsg("Checksum '" + pieces[0] + "' already exists in loaded checksums but checksum value is the same, continuing");
                            }
                            else
                            {
                                // this is a critical error - the image identifier was found more than once but the checksum was different - not sure
                                // which is correct
                                LogErrorMsg("Error adding checksum for file '" + pieces[0] + "', " + ex.Message);
                                LogErrorMsg("Checksum '" + pieces[0] + "' already exists in loaded checksums but with different checksum values, CANNOT continue!");
                                return false;
                            }
                        }
                        /*
                        LogErrorMsg("Error adding checksum for file '" + pieces[0] + "', " + ex.Message);
                        if (imageChecksums.ContainsKey(pieces[0]))
                        {
                            string existingChecksum = imageChecksums[pieces[0]];
                            if (existingChecksum == pieces[1])
                            {
                                LogInfoMsg("Checksum '" + pieces[0] + "' already exists in loaded checksums but checksum value is the same, continuing");
                            }
                            else
                            {
                                LogErrorMsg("Checksum '" + pieces[0] + "' already exists in loaded checksums but with different checksum values, CANNOT continue!");
                                return false;
                            }
                        }*/
                    }
                }

                line = reader.ReadLine();
            }
            reader.Close();
            LogInfoMsg("Done loading image checksums, loaded '" + imageChecksums.Count + "' checksum entries. Found '" + duplicateChecksumEntries + "' duplicate image entries (each duplicate had the same checksum) - this is not critical, just strange.");
            return true;
        }

        protected void OutputImageChecksum(string imageUrn, ImageQuality imageQuality, string checksum, 
            long imageSize, string returnedImageFormat)
        {
            if (imageChecksumWriter != null)
            {
                imageChecksumWriter.WriteLine(imageUrn + "_" + imageQuality.ToString("d") + "|" + checksum + "|" + imageSize + "|" + returnedImageFormat);
            }
        }

        private bool failedStatus;

        private void StopProcessing()
        {
            failedStatus = true;
            StopAllThreads();
            //ActionComplete(false);            
            OnPatientComplete(null);
        }

        private void OnCriticalException()
        {
            StopProcessing();
            return;
            // need to stop other threads?
        }

        private void StopAllThreads()
        {
            foreach (ThreadedVixTester tester in currentWorkers)
            {
                if (tester != null)
                {
                    tester.CanRun = false;
                }
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void AddFailedImageRequest(string imageUrn, ImageQuality imageQuality)
        {
            string key = imageUrn + "_" + imageQuality.ToString("d");
            if (!failedImageRequests.ContainsKey(key))
            {
                failedImageRequests.Add(key, "");
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private bool IsImageRequestOK(string imageUrn, ImageQuality imageQuality)
        {
            string key = imageUrn + "_" + imageQuality.ToString("d");
            // if the key is in the dictionary, then return false since the image request is NOT ok
            return !failedImageRequests.ContainsKey(key);
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        protected string GetImageChecksum(string imageUrn, ImageQuality imageQuality)
        {
            string key = imageUrn + "_" + imageQuality.ToString("d");

            if (imageChecksums.ContainsKey(key))
            {
                return imageChecksums[key];
            }
            else
            {
                return null;
            }      
        }

        /// <summary>
        /// Method is only called when threaded testing is being done
        /// </summary>
        /// <param name="patientResult"></param>
        [MethodImpl(MethodImplOptions.Synchronized)]
        private void OnPatientComplete(VixPatientTestResult patientResult)
        {            
            IncrementProgress();
            if (patientResult != null)
            {
                lastTestResult.Patients.Add(patientResult);
            }
            threadCount--;
            LogInfoMsg("Currently [" + threadCount + "] threads running");     

            // only do this when the count is 0, so counts less than 0 don't 
            if (threadCount == 0)
            {
                if (GetRunForever() && (mbCanRun))
                {
                    LogInfoMsg("Completed all threads, completed run #" + repeatCount);
                    LogInfoMsg(lastTestResult.ToString());

                    if (GetExcludeFailedImages())
                    {
                        SaveFailedImageRequests();
                    }

                    LogInfoMsg("Run forever is true, restarting all testing with same parameters as before.");
                    runAction(mVixInterface);
                    // running forever and not stopped, restart with the patients
                    //processPatients
                }
                else
                {
                    // either not running forever or cannot run more (stop button pressed)
                    try
                    {
                        if (formatWritter != null)
                            formatWritter.Close();
                        
                    }
                    catch (Exception ex)
                    {
                        LogErrorMsg("Error closing format writer [" + ex.Message + "]");
                    }

                    // always want to rollup results, even if stopped by user
                    ProcessingComplete();
                    /*

                    if (failedStatus)
                    {
                        ActionComplete(false);
                    }
                    else
                    {
                        ProcessingComplete();
                    }*/
                }
            }
        }

        /// <summary>
        /// Always run at the end, regardless of threaded or not
        /// </summary>
        protected void ProcessingComplete()
        {
            if (!mbAlreadyProcessedComplete)
            {
                mbAlreadyProcessedComplete = true;
                Logger().Info(LogMsg(lastTestResult.ToString()));
                if (formatWritter != null)
                {
                    formatWritter.Flush();
                    formatWritter.Close();
                    formatWritter = null;
                }
                if (validateLogWriter != null)
                {
                    validateLogWriter.Flush();
                    validateLogWriter.Close();
                    validateLogWriter = null;

                }
                if (imageChecksumWriter != null)
                {
                    imageChecksumWriter.Flush();
                    imageChecksumWriter.Close();
                    imageChecksumWriter = null;
                }

                lastTestResult.Patients.Sort();
                SavableObject result = new SavableVixPatientTestResultList(lastTestResult.Patients);
                ActionCompletedResult actionResult = ActionCompletedResult.success;
                if (IsOutputResultsEnabled())
                {
                    OutputExpectedResults(result);
                }
                if (IsVerifyResultsEnabled())
                {
                    if (!VerifyExpectedResults(result))
                    {
                        LogErrorMsg("Results do not match expected result");
                        actionResult = ActionCompletedResult.resultsFailed;
                    }
                }

                if (GetRollupResults())
                {
                    try
                    {
                        string outputDir = necessaryProperties[PROPERTY_OUTPUT_DIRECTORY].ToString();
                        string tomcatLogsDir = necessaryProperties[PROPERTY_TOMCAT_LOG_DIR].ToString();
                        Logger().Info(LogMsg("Rolling up results to [" + outputDir + "]"));
                        VixResultRollupHelper rollupHelper = new VixResultRollupHelper(LogMsg, DoApplicationEvents);
                        string compareResultsFilename = "";
                        if (GetValidateResultsWithLog())
                        {
                            compareResultsFilename = VixAKCommonUtils.getApplicationDir() + "\\compareResults.csv";
                            if (File.Exists(compareResultsFilename))
                                File.Delete(compareResultsFilename);
                            CompareResultsWithTransactionLog(compareResultsFilename);
                        }
                        rollupHelper.RollupTestResult(lastTestResult, GetGeneratePatientResults(), outputDir, tomcatLogsDir, 
                            outputImageFormatsFilename, validateLogWriterFilename, compareResultsFilename);
                    }
                    catch (Exception ex)
                    {
                        Logger().Error(LogMsg(ex.Message));
                    }
                }
                
                ActionComplete(mbCanRun ? actionResult : ActionCompletedResult.error);
            }
        }

        private void CompareResultsWithTransactionLog(string compareResults)
        {
            CompareVixResultsWithLogAction compareAction = new CompareVixResultsWithLogAction();
            compareAction.VixTestResultFilename = validateLogWriterFilename;
            compareAction.ResultsCompareOutputFilename = compareResults;
            compareAction.VixServer = extractVixServerUrl();
            compareAction.OnLogMessageEvent += LogMsg;
            compareAction.OnDoApplicationEventsEvent += this.DoApplicationEvents;
            compareAction.runAction();
        }

        private string extractVixServerUrl()
        {
            string url = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            Uri uri = new Uri(url);
            return uri.Scheme + "://" + uri.Host + ":" + uri.Port;
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_METADATA_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_COUNT_STUDIES))
                return false;
            if (!validatePropertyIncluded(PROPERTY_THREADED))
                return false;
            if (!validatePropertyIncluded(PROPERTY_MAX_WORKERS))
                return false;
            if (!validatePropertyIncluded(PROPERTY_PATIENTS))
                return false;
            if (!validatePropertyIncluded(PROPERTY_SITE_NUMBER))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_QUALITIES))
                return false;
            if (!validatePropertyIncluded(PROPERTY_ROLLUP_RESULTS))
                return false;
            if (!validatePropertyIncluded(PROPERTY_CACHE_CLEAR))
                return false;
            try
            {
                object val = necessaryProperties[PROPERTY_COUNT_STUDIES];
                bool countStudies = bool.Parse(val.ToString());
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                object val = necessaryProperties[PROPERTY_THREADED];
                bool threaded = bool.Parse(val.ToString());
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                object val = necessaryProperties[PROPERTY_MAX_WORKERS];
                int maxThreads = int.Parse(val.ToString());
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                object val = necessaryProperties[PROPERTY_PATIENTS];
                if (!(val is ActionCustomOption))
                {
                    return false;
                }
                ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_PATIENTS];
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }

            try
            {
                if (GetRollupResults())
                {
                    if (!necessaryProperties.ContainsKey(PROPERTY_OUTPUT_DIRECTORY))
                        return false;
                    if (!necessaryProperties.ContainsKey(PROPERTY_TOMCAT_LOG_DIR))
                        return false;
                    GetGeneratePatientResults();
                }
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }

            try
            {
                GetOutputImageFormat();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                GetValidateResultsWithLog();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                GetFailOnNon404Errors();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            return true;
        }

        protected List<ImagingPatient> GetSelectedPatients()
        {
            ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_PATIENTS];
            List<ImagingPatient> patients = new List<ImagingPatient>();

            foreach (BaseCustomOption option in customOption.Options)
            {
                if (option is ImagingPatient)
                {
                    ImagingPatient patient = (ImagingPatient)option;
                    if (patient.Active)
                    {
                        patients.Add(patient);
                    }
                }
            }
            return patients;
        }

        protected List<ImageQuality> GetSelectedImageQualities()
        {
            ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_IMAGE_QUALITIES];
            List<ImageQuality> qualities = new List<ImageQuality>();

            foreach (BaseCustomOption option in customOption.Options)
            {
                if (option is ImageQuality)
                {
                    ImageQuality quality = (ImageQuality)option;
                    qualities.Add(quality);
                }
            }

            return qualities;
        }

        protected bool GetCountImagesOnly()
        {
            return bool.Parse(necessaryProperties[PROPERTY_COUNT_STUDIES].ToString());
        }

        private bool GetThreaded()
        {
            return bool.Parse(necessaryProperties[PROPERTY_THREADED].ToString());
        }

        private int GetWorkerCount()
        {
            return int.Parse(necessaryProperties[PROPERTY_MAX_WORKERS].ToString());
        }

        private bool GetRollupResults()
        {
            return bool.Parse(necessaryProperties[PROPERTY_ROLLUP_RESULTS].ToString());
        }

        private bool GetGeneratePatientResults()
        {
            return bool.Parse(necessaryProperties[PROPERTY_GENERATE_PATIENT_RESULTS].ToString());
        }

        protected bool GetCacheClear()
        {
            return bool.Parse(necessaryProperties[PROPERTY_CACHE_CLEAR].ToString());
        }

        protected bool GetOutputImageFormat()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_OUTPUT_IMAGE_FORMAT))
            {
                return bool.Parse(necessaryProperties[PROPERTY_OUTPUT_IMAGE_FORMAT].ToString());
            }
            return false;
        }

        protected bool GetValidateResultsWithLog()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_VALIDATE_RESULTS_WITH_LOG))
            {
                return bool.Parse(necessaryProperties[PROPERTY_VALIDATE_RESULTS_WITH_LOG].ToString());
            }
            return false;
        }

        protected bool GetFailOnNon404Errors()
        {
            if(necessaryProperties.ContainsKey(PROPERTY_FAIL_ON_NON_404_ERRORS))
                return bool.Parse(necessaryProperties[PROPERTY_FAIL_ON_NON_404_ERRORS].ToString());
            return false;
        }

        protected bool GetRunForever()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_RUN_FOREVER))
                return bool.Parse(necessaryProperties[PROPERTY_RUN_FOREVER].ToString());
            return false;
        }

        protected bool GetExcludeFailedImages()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_EXCLUDE_FAILED_IMAGES))
                return bool.Parse(necessaryProperties[PROPERTY_EXCLUDE_FAILED_IMAGES].ToString());
            return false;
        }

        protected bool GetOutputImageChecksums()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_OUTPUT_IMAGE_CHECKSUMS))
                return bool.Parse(necessaryProperties[PROPERTY_OUTPUT_IMAGE_CHECKSUMS].ToString());
            return false;
        }

        protected bool GetVerifyImageChecksums()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_VERIFY_IMAGE_CHECKSUMS))
                return bool.Parse(necessaryProperties[PROPERTY_VERIFY_IMAGE_CHECKSUMS].ToString());
            return false;
        }

        protected string GetImageChecksumsFilename()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_IMAGE_CHECKSUMS_FILENAME))
                return necessaryProperties[PROPERTY_IMAGE_CHECKSUMS_FILENAME].ToString();
            return null;
        }

        protected bool getOutputChecksumFailedImage()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE))
            {
                return bool.Parse( necessaryProperties[PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE].ToString());
            }
            return false;
        }

        protected string getOutputChecksumFailedImageDirectory()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE_DIRECTORY))
            {
                return necessaryProperties[PROPERTY_OUTPUT_CHECKSUM_FAILED_IMAGE_DIRECTORY].ToString();
            }
            return null;
        }

        protected bool getFailIfChecksumFromVixDoesNotMatchCalculatedChecksum()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_FAIL_IF_CHECKSUM_FROM_VIX_DOES_NOT_MATCH_CALCULATED))
            {
                return bool.Parse(necessaryProperties[PROPERTY_FAIL_IF_CHECKSUM_FROM_VIX_DOES_NOT_MATCH_CALCULATED].ToString());
            }
            return false;
        }

        protected string getChecksumImageDirectory()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_CHECKSUM_IMAGE_DIRECTORY))
            {
                return necessaryProperties[PROPERTY_CHECKSUM_IMAGE_DIRECTORY].ToString();
            }
            return null;
        }

        public override void stopAction()
        {
            base.stopAction();
            StopProcessing();
        }
    }

}
