/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    class ClearCacheAction : BaseAction
    {

        private const string PROPERTY_CACHE_DIR = "Cache Directory";
        private const string PROPERTY_SERVICE_NAME = "Tomcat Service Name";
        private const string PROPERTY_MACHINE_NAME = "Machine Name";
        private const string PROPERTY_RESTART_SERVICE_AT_END = "Restart Tomcat After Clear";

        private const string dod_metadata_region = "dod-metadata-region";
        private const string dod_image_region = "dod-image-region";
        private const string va_image_region = "va-image-region";
        private const string va_metadata_region = "va-metadata-region";

        public static string CLEAR_VIX_CACHE = "Clear ViX Cache";        

        public ClearCacheAction()
        {
            this.LoadDefaultValues(VixAKCommonUtils.getDefaultValuesFile());
            SetNecessaryPropertyWithDefault(PROPERTY_CACHE_DIR, "d:\\vixcache");
            SetNecessaryPropertyWithDefault(PROPERTY_SERVICE_NAME, "tomcat6");
            SetNecessaryPropertyWithDefault(PROPERTY_MACHINE_NAME, "");
            SetNecessaryPropertyWithDefault(PROPERTY_RESTART_SERVICE_AT_END, "false");
        }

        public override string GetActionName()
        {
            return CLEAR_VIX_CACHE;
        }

        public override void runAction()
        {
            Logger().Info(LogMsg("Running action [" + GetActionName() + "]"));
            try
            {
                string serviceName = necessaryProperties[PROPERTY_SERVICE_NAME].ToString();
                string machineName = necessaryProperties[PROPERTY_MACHINE_NAME].ToString();
                ServiceControlHelper serviceHelper = new ServiceControlHelper(machineName, serviceName);
                serviceHelper.OnDoApplicationEventsEvent += DoApplicationEvents;
                serviceHelper.OnLogMessageEvent += LogMsg;
                Logger().Info(LogMsg("Using service [" + serviceName + "]"));
                if (serviceHelper.IsServiceRunning)
                {
                    if (!serviceHelper.StopService())
                    {
                        ActionComplete(ActionCompletedResult.error);
                        return;
                    }
                }                
                if (!DeleteAllCacheDirectories())
                {
                    Logger().Error(LogMsg("Error deleting cache directory"));
                    ActionComplete(ActionCompletedResult.error);
                    return;
                }
                DoApplicationEvents();
                Logger().Info(LogMsg("Vix Cache cleared"));
                if (GetRestartTomcat())
                {                    
                    serviceHelper.StartService();
                }
                DoApplicationEvents();
                Logger().Info(LogMsg("Action [" + GetActionName() + "] complete"));                
            }
            catch (Exception ex)
            {
                Logger().Error( LogMsg(ex.Message));
                ActionComplete(ActionCompletedResult.error);
                return;
            }
            ActionComplete(mbCanRun ? ActionCompletedResult.success : ActionCompletedResult.error);
        }

        private bool DeleteAllCacheDirectories()
        {
            ProgressBarMax(4);
            string baseDir = necessaryProperties[PROPERTY_CACHE_DIR].ToString();
            bool result = DeleteDirectoryCache(baseDir, dod_metadata_region);
            IncrementProgress();
            if (!result)
                return false;
            result = DeleteDirectoryCache(baseDir, dod_image_region);
            if (!result)
                return false;
            result = DeleteDirectoryCache(baseDir, va_image_region);
            if (!result)
                return false;
            result = DeleteDirectoryCache(baseDir, va_metadata_region);
            if (!result)
                return false;
            return true;
        }

        private bool DeleteDirectoryCache(string baseCacheDir, string regionDir)
        {
            string regionPath = Path.Combine(baseCacheDir, regionDir);
            string fooDir = Path.Combine(baseCacheDir, "foo");
            if (Directory.Exists(regionPath))
            {
                Logger().Info(LogMsg("Deleting directories in '" + regionPath + "'"));
                string[] dirs = Directory.GetDirectories(regionPath);
                foreach (string subDir in dirs)
                {
                    bool failed = false;
                    Logger().Info(LogMsg("Deleting directory '" + subDir + "'"));
                    try
                    {
                        Directory.Move(subDir, fooDir);
                        Directory.Delete(fooDir, true);
                    }
                    catch (Exception) {}
                    finally
                    {
                        if (Directory.Exists(subDir))
                        {
                            Logger().Error(LogMsg("Did not delete directory '" + subDir + "'"));
                            failed = true;
                        }
                    }
                    if (failed)
                        return false;
                }
            }
            return true;
        }

        private bool GetRestartTomcat()
        {
            return bool.Parse(necessaryProperties[PROPERTY_RESTART_SERVICE_AT_END].ToString());
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_CACHE_DIR))
                return false;
            if (necessaryProperties[PROPERTY_CACHE_DIR] == null)
                return false;
            string dir = necessaryProperties[PROPERTY_CACHE_DIR].ToString();
            if (!Directory.Exists(dir))
            {
                LogErrorMsg("Directory [" + dir + "] does not exist");
                return false;
            }
            if (!validatePropertyIncluded(PROPERTY_SERVICE_NAME))
                return false;
            if (!validatePropertyIncluded(PROPERTY_MACHINE_NAME))
                return false;
            if (!validatePropertyIncluded(PROPERTY_RESTART_SERVICE_AT_END))
                return false;
            return true;
        }
    }
}
