/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;


namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class ClinicalDisplayImageAction :  VerifiableTestAction
    {
        public static string CLINICAL_DISPLAY_IMAGE_ACTION_NAME = "Clinical Display Image Test";

        private const string PROPERTY_METADATA_URL = "MetadataUrl";
        private const string PROPERTY_IMAGE_URL = "ImageUrl";
        //private const string PROPERTY_SITE_NUMBER = "Site Number";
        private const string PROPERTY_IMAGE_TYPE = "Image Type";
        private const string PROPERTY_IMAGE_URN = "Image URN";
        private const string PROPERTY_ALLOW_J2K = "Allow J2K Compression";
        protected const string PROPERTY_IMAGE_QUALITIES = "Image Qualities";
        private const string PROPERTY_INCLUDE_IMAGE_SUB_TYPES = "Include Image Format Sub Types";
        private const string PROPERTY_CLINICAL_DISPLAY_INTERFACE_VERSION = "Clinical Display Interface Version";

        public ClinicalDisplayImageAction()
            : base()
        {
            this.LoadDefaultValues(VixAKCommonUtils.getDefaultValuesFile());       
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URN, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_TYPE, "1");
            SetNecessaryPropertyWithDefault(PROPERTY_ALLOW_J2K, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_INCLUDE_IMAGE_SUB_TYPES, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_CLINICAL_DISPLAY_INTERFACE_VERSION, "6");
            necessaryProperties[PROPERTY_TRANSACTION_LOG_COMMAND_NAMES] = "GetInstanceByImageUrnCommandImpl";

            List<ImageQuality> imageQualities = ImageQuality.GetImageQualities();
            ActionCustomOption qualitiesOptions = new ActionCustomOption(PROPERTY_IMAGE_QUALITIES);
            foreach (ImageQuality quality in imageQualities)
            {
                qualitiesOptions.Options.Add((BaseCustomOption)quality);
            }
            this.necessaryProperties.Add(PROPERTY_IMAGE_QUALITIES, qualitiesOptions);
        }

        public override string GetActionName()
        {
            return CLINICAL_DISPLAY_IMAGE_ACTION_NAME;
        }        

        private List<ImageQuality> GetSelectedImageQualities()
        {
            ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_IMAGE_QUALITIES];
            List<ImageQuality> qualities = new List<ImageQuality>();

            foreach (BaseCustomOption option in customOption.Options)
            {
                if (option is ImageQuality)
                {
                    ImageQuality quality = (ImageQuality)option;
                    qualities.Add(quality);
                }
            }

            return qualities;
        }

        protected override VixAKTransactionLogSavableResultHolder GetActionResult()
        {
            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            bool allowJ2K = getAllowJ2kCompression();
            bool includeSubTypes = getIncludeImageSubTypes();
            string interfaceVersion = "6";
            if (necessaryProperties.ContainsKey(PROPERTY_CLINICAL_DISPLAY_INTERFACE_VERSION))
            {
                interfaceVersion = necessaryProperties[PROPERTY_CLINICAL_DISPLAY_INTERFACE_VERSION].ToString();
            }

            VixInterface vixInterface = null;
            LogInfoMsg("Retrieving data from metadata url [" + metadataUrl + "], image url [" + imageUrl + "]");
            LogInfoMsg("Using Clinical Display interface version [" + interfaceVersion + "]");
            LogInfoMsg("Allow J2K Comprssion [" + allowJ2K + "]");
            switch (interfaceVersion)
            {
                case "2":
                    vixInterface = new ClinicalDisplayServiceImpl_v2(metadataUrl, imageUrl, false, LogMsg);
                    break;
                case "3":
                    vixInterface = new ClinicalDisplayServiceImpl_v3(metadataUrl, imageUrl, false, LogMsg);
                    break;
                case "4":
                    vixInterface = new ClinicalDisplayServiceImpl_v4(metadataUrl, imageUrl, false, false, VixTestUser.USER126, LogMsg);
                    break;
                case "6":
                    vixInterface = new ClinicalDisplayServiceImpl_v6(metadataUrl, imageUrl, false, false, VixTestUser.USER126, LogMsg);
                    break;
                default:
                    LogInfoMsg("Could not create interface for version '" + interfaceVersion + "', cannot continue");
                    break;
            }

            List<ImageQuality> qualities = GetSelectedImageQualities();
            Logger().Info(LogMsg("Image Qualities: "));
            foreach (ImageQuality quality in qualities)
            {
                Logger().Info(LogMsg(quality.ToString()));
            }
            string imageUrn = necessaryProperties[PROPERTY_IMAGE_URN].ToString();
            int imageType = int.Parse(necessaryProperties[PROPERTY_IMAGE_TYPE].ToString());
            
            /*
            string[] urnPieces = imageUrn.Split('-');
            string studyId = urnPieces[2];


            Study study = new Study(studyId);
             * */
            Instance instance = new Instance(imageUrn);
            instance.ImageType = imageType;
            int imageCount = qualities.Count;
            ProgressBarMax(imageCount);

            List<DownloadResult> results = new List<DownloadResult>();
            List<VixAKTransactionLogEntry> transactionLogEntries = new List<VixAKTransactionLogEntry>();
            List<string> imageRelevantKeys = new List<string>();
            imageRelevantKeys.Add("Item in cache?");
            
            foreach (ImageQuality imageQuality in qualities)
            {
                string transactionId = Guid.NewGuid().ToString();
                try
                {
                    DownloadResult result =
                        vixInterface.DownloadImage(new ImageRequestInput(null, instance, imageQuality, transactionId, true));
                    results.Add(result);
                    if (includeSubTypes)
                    {
                        string imageFormatSubTypeRequestString = ((ClinicalDisplayDownloadResult)result).ImageFormatSubTypeRequestString;
                        LogInfoMsg("Requested image with sub type format string [" + imageFormatSubTypeRequestString + "]");
                    }
                    GetTransactionLogEntriesForCommandsIfEnabled(new Uri(imageUrl), transactionId, 
                        transactionLogEntries, imageRelevantKeys);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error downloading image '" + imageUrn + "' with transaction Id '" + transactionId + "'.");
                    throw ex;
                }
                finally
                {
                    // only happens for 1 thread
                    IncrementProgress();
                }
            }
            return new VixAKTransactionLogSavableResultHolder(new SavableDownloadResultList(results), transactionLogEntries);
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_METADATA_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_TYPE))
                return false;            
            if (!validatePropertyIncluded(PROPERTY_ALLOW_J2K))
                return false;            
            if (!validatePropertyIncluded(PROPERTY_IMAGE_QUALITIES))
                return false;
            if (!validatePropertyIncluded(PROPERTY_INCLUDE_IMAGE_SUB_TYPES))
                return false;
            try
            {
                object val = necessaryProperties[PROPERTY_IMAGE_TYPE];
                int imageType = int.Parse(val.ToString());
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                getAllowJ2kCompression();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            try
            {
                getIncludeImageSubTypes();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
                return false;
            }
            return true;
        }

        private bool getAllowJ2kCompression()
        {
            return bool.Parse(necessaryProperties[PROPERTY_ALLOW_J2K].ToString());
        }

        private bool getIncludeImageSubTypes()
        {
            return bool.Parse(necessaryProperties[PROPERTY_INCLUDE_IMAGE_SUB_TYPES].ToString());
        }

        
    }
}
