/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class ClinicalDisplayLogImageAccessAction : BaseAction
    {
        public static string CLINICAL_DISPLAY_LOG_IMAGE_ACCESS_NAME = "Clinical Display Log Image Access";

        private const string PROPERTY_METADATA_URL = "MetadataUrl";
        private const string PROPERTY_IMAGE_URN = "Image URN";
        private const string PROPERTY_LOG_REPEAT = "Repeat Log Count";
        private const string PROPERTY_LOG_USER = "Log User";

        public ClinicalDisplayLogImageAccessAction()
        {
            this.LoadDefaultValues(VixAKCommonUtils.getDefaultValuesFile());         
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URN, "");
            SetNecessaryPropertyWithDefault(PROPERTY_LOG_REPEAT, "1");

            List<VixTestUser> users = VixTestUser.GetVixTestUsers();
            ActionCustomOption usersOptions = new ActionCustomOption(PROPERTY_LOG_USER, false);
            foreach (VixTestUser user in users)
            {
                usersOptions.Options.Add((BaseCustomOption)user);
            }
            this.necessaryProperties.Add(PROPERTY_LOG_USER, usersOptions);
            
        }

        public override string GetActionName()
        {
            return CLINICAL_DISPLAY_LOG_IMAGE_ACCESS_NAME;
        }

        public override void runAction()
        {
            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            int repeatCount = GetRepeatCount();

            ClinicalDisplayServiceImpl_v3 vixInterface = new ClinicalDisplayServiceImpl_v3(metadataUrl, "", false, LogMsg);
            VixTestUser testUser = GetSelectedUser();

            Logger().Info(LogMsg("Running action [" + GetActionName() + "]"));
            Logger().Info(LogMsg("Client workstation [" + VixAKCommonUtils.GetHostname() + "]"));
            Logger().Info(LogMsg("Retrieving data from metadata url [" + vixInterface.MetadataUrl + "]."));
            Logger().Info(LogMsg("Repeating log action [" + repeatCount + "] times."));
            LogInfoMsg("Running test as user [" + testUser + "]");
            
            string imageUrn = necessaryProperties[PROPERTY_IMAGE_URN].ToString();
            string[] urnPieces = imageUrn.Split('-');

            string studyId = urnPieces[2];
            string patientIcn = urnPieces[3];

            Study study = new Study(studyId);
            Instance instance = new Instance(imageUrn);

            ProgressBarMax(repeatCount);
            for (int i = 0; i < repeatCount; i++)
            {
                string transactionId = Guid.NewGuid().ToString();
                try
                {
                    if (vixInterface.LogImageAccess(imageUrn, patientIcn, transactionId, testUser))
                    {
                        IncrementProgress();
                    }
                    else
                    {
                        Series series = new Series("");
                        String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance);
                        Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + "False response from server for log attempt"));
                        //Logger().Error(ex.InnerException + "\n\n" + ex.StackTrace);
                        ActionComplete(ActionCompletedResult.error);
                        return;
                    }
                }
                catch (Exception ex)
                {
                    Series series = new Series("");
                    String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance);
                    Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + ex.Message));
                    //Logger().Error(ex.InnerException + "\n\n" + ex.StackTrace);
                    ActionComplete(ActionCompletedResult.error);
                    return;
                }
            }
            ActionComplete(ActionCompletedResult.success);
        }

        public override void stopAction()
        {
            base.stopAction();
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_METADATA_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_URN))
                return false;
            if (!validatePropertyIncluded(PROPERTY_LOG_REPEAT))
                return false;
            try
            {
                GetRepeatCount();
            }
            catch (Exception ex)
            {
                LogErrorMsg(ex.Message);
                return false;
            }
            

            return true;
        }

        private int GetRepeatCount()
        {
            return int.Parse(necessaryProperties[PROPERTY_LOG_REPEAT].ToString());
        }

        private VixTestUser GetSelectedUser()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_LOG_USER))
            {
                ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_LOG_USER];
                if (customOption.Options.Count != 1)
                    throw new Exception("Must specify exactly 1 user to query as");
                return (VixTestUser)customOption.Options[0];
            }
            else
            {
                return VixTestUser.USER135;
            }
        }
    }
}
