/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;
using VixAKCommon.gov.va.med.imaging.vix.services;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class ClinicalDisplayPatientTestAction : BaseVixTestAction
    {

        public static string CLINICAL_DISPLAY_PATIENT_ACTION_NAME = "Clinical Display Patient Test";

        private const string PROPERTY_IMAGE_INFO_USER = "Image Info User";
        private const string PROPERTY_ALLOW_J2K = "Allow J2K Compression";
        private const string PROPERTY_INCLUDE_IMAGE_SUB_TYPES = "Include Image Format Sub Types";
        private const string PROPERTY_DELAY_MS = "Delay ms";

        private const int maxDownloadTries = 3;

        public ClinicalDisplayPatientTestAction()
            : base(CLINICAL_DISPLAY_PATIENT_ACTION_NAME)
        {
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_ALLOW_J2K, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_INCLUDE_IMAGE_SUB_TYPES, "true");
            necessaryProperties[PROPERTY_TOMCAT_LOG_DIR] = "";
            SetNecessaryPropertyWithDefault(PROPERTY_DELAY_MS, "1000");

            List<VixTestUser> users = VixTestUser.GetVixTestUsers();
            ActionCustomOption usersOptions = new ActionCustomOption(PROPERTY_IMAGE_INFO_USER, false);
            foreach (VixTestUser user in users)
            {
                usersOptions.Options.Add((BaseCustomOption)user);
            }

            this.necessaryProperties.Add(PROPERTY_IMAGE_INFO_USER, usersOptions);
        }

        public override string GetActionName()
        {
            return CLINICAL_DISPLAY_PATIENT_ACTION_NAME;
        }

        public override void runAction()
        {
            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            bool allowJ2K = getAllowJ2kCompression();

            LogInfoMsg("Running action [" + GetActionName() + "]");
            LogInfoMsg("Client workstation [" + GetHostname() + "]");
            LogInfoMsg("Retrieving data from metadata url [" + metadataUrl + "], image url [" + imageUrl + "]");

            VixInterface vixInterface = null;
            bool includeSubTypes = getIncludeImageSubTypes();
            VixTestUser user = GetSelectedUser();
            LogInfoMsg("Running test as user '" + user.ToString() + "'.");
            vixInterface = new ClinicalDisplayServiceImpl_v4(metadataUrl, imageUrl, allowJ2K, includeSubTypes, user, LogMsg);
            int sleepTime = getDelayMs();
            LogInfoMsg("Allow J2K Comprssion [" + allowJ2K + "]");
            LogInfoMsg("Including image content type sub types [" + includeSubTypes + "]");
            LogInfoMsg("Sleeping '" + sleepTime + "'ms between image requests.");

            List<ImageQuality> qualities = GetSelectedImageQualities();
            Logger().Info(LogMsg("Image Qualities: "));
            foreach (ImageQuality quality in qualities)
            {
                Logger().Info(LogMsg(quality.ToString()));
            }

            List<ImagingPatient> patients = GetSelectedPatients();
            LogInfoMsg("Selected Patients: ");
            foreach (ImagingPatient patient in patients)
            {
                Logger().Info(LogMsg(patient.ToString(true)));
            }

            bool countImagesOnly = GetCountImagesOnly();

            lastTestResult = new VixTestResult(false, 1, qualities.Count,
                false, countImagesOnly, false, patients.Count);

            foreach (ImagingPatient patient in patients)
            {
                lastTestResult.Patients.Add(testPatient(vixInterface, patient, qualities));
            }


            ProcessingComplete();
        }

        private VixPatientTestResult testPatient(VixInterface vixInterface, ImagingPatient patient, List<ImageQuality> qualities)
        {
            VixPatientTestResult patientResult = new VixPatientTestResult(patient);
            LogInfoMsg("Testing patient '" + patient + "'.");
            LogInfoMsg("Patient has been seen at the following sites:");
            foreach (string siteNumber in patient.SiteNumbers)
            {
                LogInfoMsg(siteNumber);
            }

            foreach (string siteNumber in patient.SiteNumbers)
            {
                testPatientSite(vixInterface, patient, qualities, siteNumber, patientResult);
            }
            LogInfoMsg("Done with patient " + patient.ToString(true));
            return patientResult;
        }

        private void testPatientSite(VixInterface vixInterface, ImagingPatient patient, List<ImageQuality> qualities, 
            string siteNumber, VixPatientTestResult patientResult)
        {
            LogInfoMsg("Processing patient " + patient.ToString(true) + " from site '" + siteNumber + "'."); ;
            int sleepTime = getDelayMs();
            
            if (mbCanRun)
            {
                //lastTestResult.Patients.Add(patientResult);
                List<Study> studies = null;
                try
                {
                    String transactionId = Guid.NewGuid().ToString();
                    LogInfoMsg("Getting study graph for patient [" + patient + "] at site '" + siteNumber + "' with transaction Id '" + transactionId + ".");
                    studies = vixInterface.GetPatientStudies(siteNumber, patient.PatientIcn, transactionId);

                }
                catch (Exception ex)
                {
                    LogErrorMsg(ex.Message);
                    return ;
                }
                int qCount = qualities.Count;
                if (qCount == 0)
                    qCount = 1;
                int imageCount = VixAKCommonUtils.CountImages(studies) * qCount;
                LogInfoMsg("Found [" + studies.Count + "] studies and [" + imageCount + "] images for '" + patient.ToString(true) + "' from site '" + siteNumber + "'.");
                patientResult.Studies += studies.Count;
                patientResult.TotalImages += imageCount;

                if (mbCanRun)
                {
                    // only happens for 1 thread
                    ProgressBarMax(imageCount);
                    int downloadedImages = 0;
                    int failedImages = 0;
                    int curCount = 0;
                    foreach (Study study in studies)
                    {
                        foreach (Series series in study.Series)
                        {
                            foreach (Instance instance in series.Instances)
                            {
                                foreach (ImageQuality imageQuality in qualities)
                                {
                                    if (mbCanRun)
                                    {
                                        if (imageCount > 1000)
                                        {
                                            if ((curCount % 500) == 0)
                                            {
                                                LogInfoMsg("About to process image [" + curCount + "] of [" + imageCount + "] for patient '" + patient + "'");
                                            }
                                        }

                                        if (sleepTime > 0)
                                        {
                                            System.Threading.Thread.Sleep(sleepTime);
                                        }

                                        string transactionId = Guid.NewGuid().ToString();
                                        try
                                        {
                                            curCount++;
                                            /*
                                            if (!mbCanRun)
                                            {
                                                break;
                                            }
                                             */
                                            TryDownload(vixInterface, study, instance, imageQuality, transactionId);
                                            //vixInterface.DownloadImage(study, instance, imageQuality, transactionId);
                                            downloadedImages++;
                                        }
                                        catch (vix.exceptions.AbstractVixWebException webEx)
                                        {
                                            failedImages++;
                                            String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance, imageQuality);
                                            LogErrorMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + webEx.Message);       

                                            if (webEx.StatusCode == System.Net.HttpStatusCode.NotFound)
                                            {
                                                patientResult.Error404++;
                                            }
                                            else if (webEx.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                                            {
                                                patientResult.Error500++;
                                            }
                                            else
                                            {
                                                patientResult.ErrorOther++;
                                            }

                                        }
                                        catch (Exception ex)
                                        {
                                            failedImages++;
                                            String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance, imageQuality);
                                            LogErrorMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + ex.Message);
                                        }
                                        finally
                                        {
                                            // only happens for 1 thread
                                            IncrementProgress();
                                        }
                                    } // can run
                                } // image quality
                            } // instance
                        } // series
                    } // study
                    patientResult.ImagesFailed += failedImages;
                    patientResult.ImagesDownloaded += downloadedImages;
                } // count images
            }
            if (mbCanRun)
            {
                LogInfoMsg("Done with patient " + patient.ToString(true) + " at site '" + siteNumber + "'.");
                return;
            }
            else
            {
                Logger().Info(LogMsg("Can run is false, stopping work"));
                return;
            }
        }

        public override void stopAction()
        {
            base.stopAction();
        }

        public override bool ValidateNecessaryProperties()
        {
            try
            {
                GetSelectedUser();
            }
            catch (Exception ex)
            {
                LogErrorMsg(ex.Message);
                return false;
            }
            return true;
        }

        private VixTestUser GetSelectedUser()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_IMAGE_INFO_USER))
            {
                ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_IMAGE_INFO_USER];
                if (customOption.Options.Count != 1)
                    throw new Exception("Must specify exactly 1 user to query as");
                return (VixTestUser)customOption.Options[0];
            }
            else
            {
                return VixTestUser.USER135;
            }
        }

        private bool getAllowJ2kCompression()
        {
            return bool.Parse(necessaryProperties[PROPERTY_ALLOW_J2K].ToString());
        }

        private bool getIncludeImageSubTypes()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_INCLUDE_IMAGE_SUB_TYPES))
            {
                return bool.Parse(necessaryProperties[PROPERTY_INCLUDE_IMAGE_SUB_TYPES].ToString());
            }
            return false;
        }

        private int getDelayMs()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_DELAY_MS))
            {
                return int.Parse(necessaryProperties[PROPERTY_DELAY_MS].ToString());
            }
            return 0;
        }

        private void TryDownload(VixInterface vixInterface, Study study, Instance instance, 
            ImageQuality imageQuality, string transactionId)
        {
            int tryCount = 0;
            try
            {
                vixInterface.DownloadImage(new ImageRequestInput( study, instance, imageQuality, transactionId, true));
            }
            catch (System.Net.Sockets.SocketException socketEx)
            {
                Logger().Error(LogMsg("Socket exception: " + socketEx.Message + "occured on try '" + tryCount + "' of '" + maxDownloadTries + "'"));
                if (tryCount >= maxDownloadTries)
                {
                    throw socketEx;
                }
            }
        }
    }
}
