/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using log4net;
using System.Threading;
using System.Runtime.CompilerServices;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{

    

    public delegate void ContinuousActionCompleteDelegate(bool actionOk, ContinuousActionsTestResults testResults);

    public class ContinuousAction : BaseAction
    {
        private const string PROPERTY_METADATA_URL = "MetadataUrl";
        private const string PROPERTY_IMAGE_URL = "ImageUrl";
        //private const string PROPERTY_AWIV_IMAGE_URL = "AWIV Image URL";
        
        private const string PROPERTY_RUN_FOREVER = "Run Forever";
        private const string PROPERTY_THREAD_COUNT = "Thread Count";
        private const string PROPERTY_SITE_NUMBER = "Site Number";

        public static string PROPERTY_IMAGE_REQUEST_URN = "Image Request URN";
        public static string PROPERTY_IMAGE_INFORMATION_URN = "Image Information URN";
        public static string PROPERTY_STUDY_GRAPH_PATIENT_ICN = "Study Graph Patient ICN";
        public static string PROPERTY_STUDY_IMAGES_STUDY_URN = "Study Images Study URN";
        public static string PROPERTY_SEARCH_PATIENTS_QUERY = "Search Patients Query";
        public static string PROPERTY_RETRIEVE_PATIENT_IMAGES_PATIENT_ICN = "Retrieve Patient Images Patient ICN";
        //public static string PROPERTY_AWIV_IMAGE_REQUEST_URN ="AWIV Image Request URN";


        public static string CONTINUOUS_TEST_ACTION_NAME = "Continuous VIX Test";

        private int currentThreadCount = 0;
        private ClinicalDisplayServiceImpl_v4 service = null;
        private AWIVServiceImpl awivService = null;
        private List<ContinuousActionThread> workers = new List<ContinuousActionThread>();
        private ContinuousActionsTotalTestResults totalTestResults = new ContinuousActionsTotalTestResults();
        private ContinuousActionsTotalTestResults currentTestResults = new ContinuousActionsTotalTestResults();

        public ContinuousAction()
        {
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_RUN_FOREVER, "true");
            SetNecessaryPropertyWithDefault(PROPERTY_THREAD_COUNT, "5");
            SetNecessaryPropertyWithDefault(PROPERTY_SITE_NUMBER, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_REQUEST_URN, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_INFORMATION_URN, "");
            SetNecessaryPropertyWithDefault(PROPERTY_STUDY_GRAPH_PATIENT_ICN, "");
            SetNecessaryPropertyWithDefault(PROPERTY_STUDY_IMAGES_STUDY_URN, ""); 
            SetNecessaryPropertyWithDefault(PROPERTY_RETRIEVE_PATIENT_IMAGES_PATIENT_ICN, "");

            totalTestResults.StartTime = DateTime.Now.ToString();
        }

        public override void runAction()
        {
            LogInfoMsg("Running action '" + GetActionName() + "'");
            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            VixTestUser user = VixTestUser.USER126;
            service = new ClinicalDisplayServiceImpl_v4(metadataUrl,
                imageUrl, true, true, user, LogMsg);

            int threadCount = GetThreadCount();
            bool runForever = GetRunForever();
            string siteNumber = GetSiteNumber();

            LogInfoMsg("Metadata Url [" + metadataUrl + "], image Url [" + imageUrl + "]");
            LogInfoMsg("Thread Count [" + threadCount + "]");
            LogInfoMsg("Run Forever [" + runForever + "]");
            LogInfoMsg("Site Number [" + siteNumber + "]");

            ThreadPool.SetMaxThreads(threadCount, 100000);
            
            runTests();
        }

        private Dictionary<string, string> getTests()
        {
            Dictionary<string, string> tests = new Dictionary<string, string>();
            addTestIfIncluded(PROPERTY_IMAGE_REQUEST_URN, tests);
            addTestIfIncluded(PROPERTY_IMAGE_INFORMATION_URN, tests);
            addTestIfIncluded(PROPERTY_STUDY_GRAPH_PATIENT_ICN, tests);
            addTestIfIncluded(PROPERTY_SEARCH_PATIENTS_QUERY, tests);
            addTestIfIncluded(PROPERTY_STUDY_IMAGES_STUDY_URN, tests);
            addTestIfIncluded(PROPERTY_RETRIEVE_PATIENT_IMAGES_PATIENT_ICN, tests);

            return tests;
        }

        private void addTestIfIncluded(string testProperty, Dictionary<string, string> tests)
        {
            if (necessaryProperties.ContainsKey(testProperty))
            {
                string value = necessaryProperties[testProperty].ToString();
                if (!string.IsNullOrEmpty(value))
                {
                    tests.Add(testProperty, value);
                }
            }
        }

        [MethodImpl(MethodImplOptions.Synchronized)]
        private void ThreadComplete(bool actionOk, ContinuousActionsTestResults testResults)
        {
            currentThreadCount--;
            LogInfoMsg("action thread completed, current count [" + currentThreadCount + "]");
            currentTestResults.Results.Add(testResults);
            totalTestResults.Results.Add(testResults);
            if (currentThreadCount <= 0)
            {
                if (GetRunForever() && mbCanRun)
                {
                    currentTestResults.EndTime = DateTime.Now.ToString();
                    LogInfoMsg("All threads have completed");
                    LogInfoMsg("Results of current round of testing: " + currentTestResults.ToString());
                    //, rescheduling all of them");
                    runTests();
                    // reschedule everything
                }
                else
                {
                    LogInfoMsg("Total results: " + totalTestResults.ToString());
                    ActionComplete(mbCanRun ? ActionCompletedResult.success : ActionCompletedResult.error);
                }                
            }
        }

        private void runTests()
        {
            currentTestResults.StartTime = DateTime.Now.ToString();
            int threadCount = GetThreadCount();
            Dictionary<string, string> tests = getTests();
            LogInfoMsg("Running '" + tests.Count + "' tests per thread.");
            currentThreadCount = threadCount;
            workers.Clear();
            for (int i = 0; i < threadCount; i++)
            {
                ContinuousActionThread t = new ContinuousActionThread(service, tests, GetSiteNumber());
                workers.Add(t);
                t.OnActionCompleteEvent += ThreadComplete;
                t.OnLogMessageEvent += LogInfoMsg;
                t.OnLogErrorMessageEvent += LogErrorMsg;
                ThreadPool.QueueUserWorkItem(new WaitCallback(t.runTests));
            }
        }

        public override string GetActionName()
        {
            return CONTINUOUS_TEST_ACTION_NAME;
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_METADATA_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_THREAD_COUNT))
                return false;
            if (!validatePropertyIncluded(PROPERTY_RUN_FOREVER))
                return false;
            if (!validatePropertyIncluded(PROPERTY_SITE_NUMBER))
                return false;

            return true;
        }

        private string GetSiteNumber()
        {
            return necessaryProperties[PROPERTY_SITE_NUMBER].ToString();
        }

        private bool GetRunForever()
        {
            return bool.Parse(necessaryProperties[PROPERTY_RUN_FOREVER].ToString());
        }

        private int GetThreadCount()
        {
            return int.Parse(necessaryProperties[PROPERTY_THREAD_COUNT].ToString());
        }

        public override void stopAction()
        {

            base.stopAction();
            LogInfoMsg("Action stopped by user");
            StopAllThreads();
        }

        private void StopAllThreads()
        {
            LogInfoMsg("Stopping all threads...");
            foreach (ContinuousActionThread thread in workers)
            {
                thread.CanRun = false;
            }
        }
    }

    public class ContinuousActionThread
    {
        public event LogInfoMessageDelegate OnLogMessageEvent = null;
        public event ContinuousActionCompleteDelegate OnActionCompleteEvent = null;
        public event LogErrorMessageDelegate OnLogErrorMessageEvent = null;



        private ClinicalDisplayServiceImpl_v4 service = null;
        private AWIVServiceImpl awivService = null;
        private string siteNumber = null;
        private Dictionary<string, string> tests = null;

        private bool mbCanRun = true;

        public bool CanRun
        {
            get { return mbCanRun; }
            set { mbCanRun = value; }
        }

        private void LogErrorMsg(string msg)
        {
            if (OnLogErrorMessageEvent != null)
            {
                OnLogErrorMessageEvent(msg);
            }
        }

        private void ActionComplete(bool actionOk, ContinuousActionsTestResults results)
        {
            if (OnActionCompleteEvent != null)
            {
                OnActionCompleteEvent(actionOk, results);
            }
        }

        private string LogMsg(string msg)
        {
            if (OnLogMessageEvent != null)
            {
                OnLogMessageEvent(msg);
            }
            return msg;
        }

        protected ILog Logger()
        {
            return LogManager.GetLogger(this.GetType().Name);
        }

        public ContinuousActionThread(ClinicalDisplayServiceImpl_v4 service, Dictionary<string, string> tests, string siteNumber)
        {
            this.service = service;
            this.tests = tests;
            this.siteNumber = siteNumber;
        }

        public void runTests(Object state)
        {
            bool testOk = true;
            ContinuousActionsTestResults testResults = new ContinuousActionsTestResults();
            try
            {
                if (tests.ContainsKey(ContinuousAction.PROPERTY_IMAGE_INFORMATION_URN))
                {
                    if (mbCanRun)
                    {
                        
                        testResults.Tests.Add( testGetImageInformation(tests[ContinuousAction.PROPERTY_IMAGE_INFORMATION_URN]));
                    }
                }
                if (tests.ContainsKey(ContinuousAction.PROPERTY_IMAGE_REQUEST_URN))
                {
                    if (mbCanRun)
                    {
                        testResults.Tests.Add( testGetImage(tests[ContinuousAction.PROPERTY_IMAGE_REQUEST_URN]));
                    }
                }
                if (tests.ContainsKey(ContinuousAction.PROPERTY_SEARCH_PATIENTS_QUERY))
                {
                    if (mbCanRun)
                    {
                    }
                }
                if (tests.ContainsKey(ContinuousAction.PROPERTY_STUDY_GRAPH_PATIENT_ICN))
                {
                    if (mbCanRun)
                    {
                        testResults.Tests.Add( testGetPatientStudies(siteNumber, tests[ContinuousAction.PROPERTY_STUDY_GRAPH_PATIENT_ICN]));
                    }
                }
                /*
                if (tests.ContainsKey(ContinuousAction.PROPERTY_STUDY_IMAGES_STUDY_URN))
                {
                    if (mbCanRun)
                    {
                        testResults.Tests.Add( testGetPatientStudiesAndImages(siteNumber, 
                    }
                }*/
                if (tests.ContainsKey(ContinuousAction.PROPERTY_RETRIEVE_PATIENT_IMAGES_PATIENT_ICN))
                {
                    if (mbCanRun)
                    {
                        testResults.Tests.Add(testGetPatientStudiesAndImages(siteNumber, tests[ContinuousAction.PROPERTY_RETRIEVE_PATIENT_IMAGES_PATIENT_ICN]));
                    }
                }
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error running tests, " + ex.Message);
                testOk = false;
                
            }
            testResults.setTestsComplete();
            OnActionCompleteEvent(testOk, testResults);
        }

        private ContinuousActionTestResults testGetImageInformation(string imageUrn)
        {
            ContinuousActionTestResults test = new ContinuousActionTestResults("Get Image Information");
            string transId = Guid.NewGuid().ToString();
            try
            {
                LogMsg("Getting image information '" + imageUrn + "' with transaction Id [" + transId + "]");
                test.WSCallCount++;
                service.GetImageInformation(imageUrn, transId);
                test.WSSuccessCallCount++;
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting image information '" + imageUrn + "', " + ex.Message);
                test.OtherErrors++;
            }
            test.setTestComplete();
            return test;
        }

        private ContinuousActionTestResults testGetImage(string imageUrn)
        {
            ContinuousActionTestResults test = new ContinuousActionTestResults("Get Image");
            string transId = Guid.NewGuid().ToString();
            try
            {
                Instance instance = new Instance(imageUrn);
                instance.ImageType = 100;
                
                LogMsg("Getting image '" + imageUrn + "' with transaction Id [" + transId + "]");
                service.DownloadImage(new ImageRequestInput( null, instance, ImageQuality.REFERENCE, transId, true));
                test.ImageCount++;
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting image '" + imageUrn + "', " + ex.Message);
                test.OtherErrors++;
            }
            test.setTestComplete();
            return test;
        }

        private ContinuousActionTestResults testGetPatientStudies(string siteNumber, string patientIcn)
        {
            ContinuousActionTestResults test = new ContinuousActionTestResults("Get Patient Studies");
            try
            {
                String transactionId = Guid.NewGuid().ToString();
                LogMsg("Getting studies for patient '" + patientIcn + "' from site '" + siteNumber + "' with transactionId '" + transactionId + "'.");
                test.WSCallCount++;
                List<Study> studies = service.GetPatientStudies(siteNumber, patientIcn, transactionId);
                LogMsg("Got '" + (studies == null ? "null" : studies.Count.ToString()) + "' studies.");
                test.WSSuccessCallCount++;
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting patient studies for patient '" + patientIcn + "' from site '" + siteNumber + "', " + ex.Message);
                test.OtherErrors++;
            }
            test.setTestComplete();
            return test;
        }

        private ContinuousActionTestResults testGetPatientStudiesAndImages(string siteNumber, string patientIcn)
        {
            return testGetPatientStudiesAndImages(siteNumber, patientIcn, null);
        }

        private ContinuousActionTestResults testGetPatientStudiesAndImages(string siteNumber, string patientIcn, string specificStudyId)
        {
            ContinuousActionTestResults test = new ContinuousActionTestResults("Get Patient Studies and Images");
            try
            {
                String transactionId = Guid.NewGuid().ToString();
                LogMsg("Getting studies for patient '" + patientIcn + "' from site '" + siteNumber + "' with transactionId '" + transactionId + "'.");
                test.WSCallCount++;
                List<Study> studies = service.GetPatientStudies(siteNumber, patientIcn, transactionId);
                
                LogMsg("Got '" + (studies == null ? "null" : studies.Count.ToString()) + "' studies.");

                test.WSSuccessCallCount++;

                foreach (Study study in studies)
                {
                    if (specificStudyId == null || specificStudyId == study.StudyId)
                    {
                        if(mbCanRun)
                            DownloadStudyImages(study, test);
                    }
                }
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting patient studies for patient '" + patientIcn + "' from site '" + siteNumber + "', " + ex.Message);
                test.OtherErrors++;
            }
            test.setTestComplete();
            return test;
        }



        private void DownloadStudyImages(Study study, ContinuousActionTestResults test)
        {
            LogMsg("Retrieving images for study '" + study.StudyId + "'.");
            foreach (Series series in study.Series)
            {
                foreach (Instance image in series.Instances)
                {
                    try
                    {
                        if (mbCanRun)
                        {
                            Instance instance = new Instance(image.InstanceId);
                            instance.ImageType = image.ImageType;
                            string transactionId = Guid.NewGuid().ToString();
                            LogMsg("Getting image '" + image.InstanceId + "' with transaction Id [" + transactionId + "]");
                            service.DownloadImage(new ImageRequestInput(null, instance, ImageQuality.REFERENCE, transactionId, true));
                            test.ImageCount++;
                        }
                    }
                    catch (Exception ex)
                    {
                        LogMsg("ERROR downloading image '" + image.InstanceId + "', " + ex.Message);
                    }
                }
            }
        }
    }

    public class ContinuousActionsTotalTestResults
    {
        private string mstrStartTime;
        private string mstrEndTime;
        private readonly List<ContinuousActionsTestResults> results = new List<ContinuousActionsTestResults>();

        public string StartTime
        {
            get { return mstrStartTime; }
            set { mstrStartTime = value; }
        }        

        public string EndTime
        {
            get { return mstrEndTime; }
            set { mstrEndTime = value; }
        }

        public List<ContinuousActionsTestResults> Results
        {
            get { return results; }
        }

        public int Error404
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.Error404;
                }
                return count;
            }
        }

        public int Error500Error
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.Error500Error;
                }
                return count;
            }
        }

        public int OtherErrors
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.OtherErrors;
                }
                return count;
            }
        }

        public int ImageCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.ImageCount;
                }
                return count;
            }
        }

        public int ResultCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.ResultCount;
                }
                return count;
            }
        }

        public int WSCallCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.WSCallCount;
                }
                return count;
            }
        }

        public int WSSuccessCallCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.WSSuccessCallCount;
                }
                return count;
            }
        }

        public int TestCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionsTestResults res in results)
                {
                    count += res.Tests.Count;
                }
                return count;
            }
        }

        public override string ToString()
        {
            DateTime st = DateTime.Parse(mstrStartTime);
            DateTime et = DateTime.Parse(mstrEndTime);

            StringBuilder sb = new StringBuilder();

            sb.AppendLine("Tests started at [" + mstrStartTime + "] and ended at [" + mstrEndTime + "]");
            sb.AppendLine("Total Tests run [" + TestCount + "]");
            sb.AppendLine("WS Calls [" + WSCallCount + "]");
            sb.AppendLine("WS Successful Calls [" + WSSuccessCallCount + "]");
            sb.AppendLine("Image Count [" + ImageCount + "]");


            return sb.ToString() ;
        }
    }
   
    public class ContinuousActionsTestResults
    {
        private readonly string mstrStartTime;
        private string mstrEndTime;
        private readonly List<ContinuousActionTestResults> tests = new List<ContinuousActionTestResults>();

        public List<ContinuousActionTestResults> Tests
        {
            get { return tests; }
        } 

        public string EndTime
        {
            get { return mstrEndTime; }
        }

        public ContinuousActionsTestResults()
        {
            mstrStartTime = DateTime.Now.ToString();
        }

        public void setTestsComplete()
        {
            mstrEndTime = DateTime.Now.ToString();
        }

        public int Error404
        {
            get 
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.Error404;
                }
                return count;
            }            
        }

        public int Error500Error
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.Error500Error;
                }
                return count;
            }
        }

        public int OtherErrors
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.OtherErrors;
                }
                return count;
            }
        }

        public int ImageCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.ImageCount;
                }
                return count;
            }
        }

        public int ResultCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.ResultCount;
                }
                return count;
            }
        }

        public int WSCallCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.WSCallCount;
                }
                return count;
            }
        }

        public int WSSuccessCallCount
        {
            get
            {
                int count = 0;
                foreach (ContinuousActionTestResults res in tests)
                {
                    count += res.WSSuccessCallCount;
                }
                return count;
            }
        }
    }

    public class ContinuousActionTestResults
    {
        private readonly string mstrTestName;
        private readonly string mstrStartTime;
        private string mstrEndTime;
        private int mintError404;
        private int mintError500Error;
        private int mintOtherErrors;
        private int mintImageCount;
        private int mintResultCount;
        private int mintWSCallCount;
        private int mintWSSuccessCallCount;

        public int WSSuccessCallCount
        {
            get { return mintWSSuccessCallCount; }
            set { mintWSSuccessCallCount = value; }
        }

        public string TestName
        {
            get { return mstrTestName; }
        }

        public string StartTime
        {
            get { return mstrStartTime; }
        }

        public string EndTime
        {
            get { return mstrEndTime; }
        }

        public int Error404
        {
            get { return mintError404; }
            set { mintError404 = value; }
        }

        public int Error500Error
        {
            get { return mintError500Error; }
            set { mintError500Error = value; }
        }        

        public int OtherErrors
        {
            get { return mintOtherErrors; }
            set { mintOtherErrors = value; }
        }        

        public int ImageCount
        {
            get { return mintImageCount; }
            set { mintImageCount = value; }
        }        

        public int ResultCount
        {
            get { return mintResultCount; }
            set { mintResultCount = value; }
        }        

        public int WSCallCount
        {
            get { return mintWSCallCount; }
            set { mintWSCallCount = value; }
        }

        public ContinuousActionTestResults(string testName)
        {
            mstrTestName = testName;
            mstrStartTime = DateTime.Now.ToString();
        }

        public void setTestComplete()
        {
            mstrEndTime = DateTime.Now.ToString();
        }
    }
}
