/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using log4net;
using VixAKCommon.gov.va.med.imaging.vix.services;
using System.IO;
using VixAKCommon.gov.va.med.imaging.vix.exceptions;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class ThreadedVixTester
    {
        #region "Delegate Definitions"
        public delegate void PatientCompleteDelegate(VixPatientTestResult result);
        public delegate void CriticalExceptionDelegate();
        public delegate void AddFailedImageRequestDelegate(string imageUrn, ImageQuality imageQuality);
        public delegate bool IsImageRequestOKDelegate(string imageUrn, ImageQuality imageQuality);
        
        #endregion

        #region "Member Variables"
        private readonly ImagingPatient patient;
        private readonly List<ImageQuality> qualities;
        private readonly VixInterface vixInterface;
        private readonly bool mbCountImages;
        private readonly string mstrSiteNumber;
        private StreamWriter formatWriter = null;
        private bool mbCanRun;
        private readonly TextWriter imageResultWriter = null;
        private readonly bool mbStopOnNon404Error;
        private readonly bool mbOutputFailedChecksumImages;
        private readonly string mstrOutputFailedChecksumImagesDirectory; // this is the directory to output the image if the checksums do not match
        private readonly string mstrOutputChecksumImagesDirectory; // this is used when generated checksums to store the images for reference in the future
        private readonly bool mbFailIfChecksumFromVixDoesNotMatchCalculated;
        #endregion

        
        private const int maxDownloadTries = 3;

        public bool CanRun
        {
            get { return mbCanRun; }
            set { mbCanRun = value; }
        }

        protected ILog Logger()
        {
            return LogManager.GetLogger(this.GetType().Name);
        }

        public StreamWriter FormatWriter
        {
            get { return formatWriter; }
            set { formatWriter = value; }
        }

        public event LogMessageDelegate OnLogMessageEvent;
        public event IncrementProgressDelegate OnIncrementProgressEvent;
        public event SetProgressBarMax OnSetProgressBarMaxEvent;
        public event PatientCompleteDelegate OnPatientCompleteEvent;
        public event CriticalExceptionDelegate OnCriticalExceptionEvent;
        public event OutputImageChecksumDelegate OnOutputImageChecksum;

        public event IsImageRequestOKDelegate OnIsImageRequestOK;
        public event AddFailedImageRequestDelegate OnAddFailedImageRequest;
        public event GetImageChecksumDelegate OnGetImageChecksumEvent;


        #region "Event Handlers"

        private void AddFailedImageRequest(string imageUrn, ImageQuality imageQuality)
        {
            if (OnAddFailedImageRequest != null)
            {
                OnAddFailedImageRequest(imageUrn, imageQuality);
            }
        }

        private bool IsImageRequestOK(string imageUrn, ImageQuality imageQuality)
        {
            if (OnIsImageRequestOK != null)
            {
                return OnIsImageRequestOK(imageUrn, imageQuality);
            }
            return true;
        }

        private void CriticalException()
        {
            if (OnCriticalExceptionEvent != null)
            {
                OnCriticalExceptionEvent();
            }
        }

        private void PatientComplete(VixPatientTestResult result)
        {
            if (OnPatientCompleteEvent != null)
            {
                OnPatientCompleteEvent(result);
            }
        }

        private void ProgressBarMax(int max)
        {
            if (OnSetProgressBarMaxEvent != null)
            {
                OnSetProgressBarMaxEvent(max);
            }
        }

        private string LogMsg(string msg)
        {
            if (OnLogMessageEvent != null)
            {
                OnLogMessageEvent(msg);
            }
            return msg;
        }

        private void IncrementProgress()
        {
            if (OnIncrementProgressEvent != null)
            {
                OnIncrementProgressEvent();
            }
        }
        #endregion

        public ThreadedVixTester(ImagingPatient patient, List<ImageQuality> imageQualities, 
            VixInterface vixInterface, bool countImages, string siteNumber, 
            TextWriter imageResultWriter, bool stopNon404Error, 
            bool outputFailedChecksumImages, string outputFailedChecksumImagesDirectory,
            bool failIfChecksumFromVixDoesNotMatchCalculated,
            string outputChecksumImagesDirectory)
        {
            this.patient = patient;
            this.qualities = imageQualities;
            this.vixInterface = vixInterface;
            this.mbCountImages = countImages;
            this.mstrSiteNumber = siteNumber;
            this.imageResultWriter = imageResultWriter;
            this.mbCanRun = true;
            this.mbStopOnNon404Error = stopNon404Error;
            this.mbOutputFailedChecksumImages = outputFailedChecksumImages;
            this.mstrOutputFailedChecksumImagesDirectory = outputFailedChecksumImagesDirectory;
            this.mbFailIfChecksumFromVixDoesNotMatchCalculated = failIfChecksumFromVixDoesNotMatchCalculated;
            this.mstrOutputChecksumImagesDirectory = outputChecksumImagesDirectory;
        }

        public void ProcessPatient(Object state)
        {
            ProcessPatient();
        }

        public void ProcessPatientVoid()
        {
            ProcessPatient();
        }

        public VixPatientTestResult ProcessPatient()
        {
            Logger().Info(LogMsg("Processing patient " + patient.ToString(true)));

            VixPatientTestResult patientResult = new VixPatientTestResult(patient);
            if (mbCanRun)
            {
                //lastTestResult.Patients.Add(patientResult);
                List<Study> studies = null;
                try
                {
                    String transactionId = Guid.NewGuid().ToString();
                    Logger().Info(LogMsg("Getting study graph for patient [" + patient + "] with transactionId [" + transactionId + "]"));
                    studies = vixInterface.GetPatientStudies(mstrSiteNumber, patient.PatientIcn, transactionId);

                }
                catch (Exception ex)
                {
                    Logger().Error(LogMsg(ex.Message));
                    CriticalException();
                    return patientResult;
                }
                int imageCount = VixAKCommonUtils.CountImages(studies) * qualities.Count;
                Logger().Info(LogMsg("Found [" + studies.Count + "] studies and [" + imageCount + "] images for '" + patient.ToString(true) + "'"));
                patientResult.Studies = studies.Count;
                patientResult.TotalImages = imageCount;

                if ((!mbCountImages) && (mbCanRun))
                {
                    // only happens for 1 thread
                    ProgressBarMax(imageCount);
                    int downloadedImages = 0;
                    int failedImages = 0;
                    int curCount = 0;
                    int skippedImages = 0;
                    Dictionary<string, int> machineNameCount = new Dictionary<string, int>();
                    foreach (Study study in studies)
                    {
                        foreach (Series series in study.Series)
                        {
                            foreach (Instance instance in series.Instances)
                            {
                                foreach (ImageQuality imageQuality in qualities)
                                {
                                    if (mbCanRun)
                                    {
                                        if (imageCount > 1000)
                                        {
                                            if ((curCount % 500) == 0)
                                            {
                                                Logger().Info(LogMsg("About to process image [" + curCount + "] of [" + imageCount + "] for patient '" + this.patient + "'"));
                                            }
                                        }

                                        if (IsImageRequestOK(instance.GloballyUniqueId, imageQuality))
                                        {
                                            string transactionId = Guid.NewGuid().ToString();
                                            
                                            ImageRequestInput request = null;
                                            DownloadResult result = null;
                                            try
                                            {
                                                curCount++;
                                                //string expectedChecksum = getExpectedImageChecksum(instance, imageQuality);
                                                request = new ImageRequestInput(study, instance, imageQuality, transactionId,
                                                    mbFailIfChecksumFromVixDoesNotMatchCalculated);
                                                //request = new ImageRequestInput(study, instance, imageQuality,
                                                //    transactionId, new ChecksumTestOptions(expectedChecksum,
                                                //    mbOutputFailedChecksumImages, mstrOutputFailedChecksumImagesDirectory));

                                                result = TryDownload(request);
                                                patientResult.incrementMachineResponse(result.HeaderMachineName);

                                                string expectedChecksum = getExpectedImageChecksum(instance, imageQuality);
                                                if (expectedChecksum != null)
                                                {
                                                    VixTestCommon.VerifyExpectedChecksum(result.ImageBuffer, new ChecksumTestOptions(expectedChecksum,
                                                        mbOutputFailedChecksumImages, mstrOutputFailedChecksumImagesDirectory),
                                                        result.ReturnedImageChecksum, request.TransactionId);
                                                }

                                                downloadedImages++;
                                                OutputImageDownloadFormatResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, result.ReturnedImageFormat,
                                                    imageQuality.ToString(), result.ReturnedImageQuality, result.ReturnedImageChecksum, "");
                                                OutputImageDownloadResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, result.ReturnedImageFormat,
                                                    imageQuality.ToString(), result.ReturnedImageQuality, "", "");
                                            }
                                            catch (AbstractVixWebException webEx)
                                            {
                                                failedImages++;
                                                AddFailedImageRequest(instance.GloballyUniqueId, imageQuality);
                                                String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance, imageQuality);
                                                Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + webEx.Message));
                                                OutputImageDownloadFormatResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, "",
                                                    imageQuality.ToString(), "", "", webEx.Message);
                                                OutputImageDownloadResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, "",
                                                    imageQuality.ToString(), "",
                                                    webEx.StatusCode.ToString(), webEx.Message);

                                                if (webEx.StatusCode == System.Net.HttpStatusCode.NotFound)
                                                {
                                                    patientResult.Error404++;
                                                }
                                                else if (webEx.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                                                {
                                                    patientResult.Error500++;
                                                    checkStopNon404Exception();
                                                }
                                                else
                                                {
                                                    patientResult.ErrorOther++;
                                                    checkStopNon404Exception();
                                                }
                                            }

                                                /*
                                            catch (System.Net.WebException webEx)
                                            {
                                                failedImages++;
                                                AddFailedImageRequest(instance.GloballyUniqueId, imageQuality);
                                                String imagePrefix = VixCommon.GetInstanceInfo(study, series, instance, imageQuality);
                                                Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + webEx.Message));
                                                if (webEx.Status != System.Net.WebExceptionStatus.ProtocolError)
                                                {
                                                    Logger().Error(LogMsg(webEx.InnerException + "\n\n" + webEx.StackTrace));
                                                }
                                                System.Net.HttpWebResponse resp = (System.Net.HttpWebResponse)webEx.Response;
                                                OutputImageDownloadFormatResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, "",
                                                    imageQuality.ToString(), "", "", webEx.Message);
                                                OutputImageDownloadResult(instance.GloballyUniqueId, transactionId,
                                                    request.RequestedImageFormat, "",
                                                    imageQuality.ToString(), "",
                                                    resp.StatusCode.ToString(), webEx.Message);

                                                if (resp.StatusCode == System.Net.HttpStatusCode.NotFound)
                                                {
                                                    patientResult.Error404++;
                                                }
                                                else if (resp.StatusCode == System.Net.HttpStatusCode.InternalServerError)
                                                {
                                                    patientResult.Error500++;
                                                    checkStopNon404Exception();
                                                }
                                                else
                                                {
                                                    patientResult.ErrorOther++;
                                                    checkStopNon404Exception();
                                                }
                                            }*/
                                            catch (ChecksumException cX)
                                            {
                                                string calculatedChecksum = "";
                                                if (cX is ChecksumFailedException)
                                                {
                                                    calculatedChecksum = ((ChecksumFailedException)cX).CalculatedChecksum;
                                                }
                                                failedImages++;
                                                AddFailedImageRequest(instance.GloballyUniqueId, imageQuality);
                                                String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance, imageQuality);
                                                Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + cX.Message));
                                                // if there was an error getting the expected checksum for the image, then reqest will be null
                                                // check for that here to ensure no NPE
                                                string requestedImageFormat = (request == null ? "" : request.RequestedImageFormat);
                                                string returnedImageFormat = (result == null ? "" : result.ReturnedImageFormat);
                                                string returnedImageQuality = (result == null ? "" : result.ReturnedImageQuality);

                                                OutputImageDownloadFormatResult(instance.GloballyUniqueId, transactionId,
                                                    requestedImageFormat, returnedImageFormat,
                                                    imageQuality.ToString(), returnedImageQuality, calculatedChecksum, cX.Message);
                                                OutputImageDownloadResult(instance.GloballyUniqueId, transactionId,
                                                    requestedImageFormat, returnedImageFormat,
                                                    imageQuality.ToString(), returnedImageQuality, "", cX.Message);
                                                patientResult.ChecksumErrors++;
                                            }
                                            catch (Exception ex)
                                            {
                                                failedImages++;
                                                AddFailedImageRequest(instance.GloballyUniqueId, imageQuality);
                                                String imagePrefix = VixAKCommonUtils.GetInstanceInfo(study, series, instance, imageQuality);
                                                Logger().Error(LogMsg(imagePrefix + " Transaction Id:" + transactionId + " Error:" + ex.Message));
                                                // if there was an error getting the expected checksum for the image, then reqest will be null
                                                // check for that here to ensure no NPE
                                                string requestedImageFormat = (request == null ? "" : request.RequestedImageFormat);
                                                OutputImageDownloadFormatResult(instance.GloballyUniqueId, transactionId,
                                                    requestedImageFormat, "",
                                                    imageQuality.ToString(), "", "", ex.Message);
                                                OutputImageDownloadResult(instance.GloballyUniqueId, transactionId,
                                                    requestedImageFormat, "",
                                                    imageQuality.ToString(), "", "", ex.Message);
                                                patientResult.ErrorOther++;
                                                checkStopNon404Exception();
                                            }
                                            finally
                                            {
                                                // only happens for 1 thread
                                                IncrementProgress();
                                            }
                                        }
                                        else
                                        {
                                            skippedImages++;
                                            
                                            Logger().Info(LogMsg("Skipping image '" + instance.GloballyUniqueId + "' with quality '" + imageQuality + "' because in previous attempts this image failed."));
                                            IncrementProgress();
                                        }
                                    } // can run
                                } // image quality
                            } // instance
                        } // series
                    } // study
                    patientResult.ImagesFailed = failedImages;
                    patientResult.ImagesDownloaded = downloadedImages;
                    patientResult.SkippedImages = skippedImages;
                } // count images
            }
            if (mbCanRun)
            {
                Logger().Info(LogMsg("Done with patient " + patient.ToString(true)));
                Logger().Info(LogMsg(patientResult.ToString()));
                PatientComplete(patientResult);
            }
            else
            {
                Logger().Info(LogMsg("Can run is false, stopping work"));
                PatientComplete(patientResult);
            }
            return patientResult;
        }

        private void checkStopNon404Exception()
        {
            if (mbStopOnNon404Error)
            {
                mbCanRun = false;
                Logger().Info(LogMsg("Received non 404 exception, must fail on all non-404 errors, stopping processing..."));
                CriticalException();
            }
        }

        private const string formatWriterDelimiter = ",";

        /// <summary>
        /// Used for outputting image result to be compared to the transaction log, not really used anymore
        /// </summary>
        /// <param name="imageId"></param>
        /// <param name="transactionId"></param>
        /// <param name="requestFormat"></param>
        /// <param name="imageFormat"></param>
        /// <param name="requestedQuality"></param>
        /// <param name="returnedImageQuality"></param>
        /// <param name="responseCode"></param>
        /// <param name="errorMsg"></param>
        private void OutputImageDownloadResult(string imageId, string transactionId, string requestFormat, string imageFormat,
            string requestedQuality, string returnedImageQuality, string responseCode, string errorMsg)
        {
            if (imageResultWriter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append(imageId);
                sb.Append(formatWriterDelimiter);
                sb.Append(transactionId);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + requestFormat + "\"");
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + imageFormat + "\"");
                sb.Append(formatWriterDelimiter);
                sb.Append(requestedQuality);
                sb.Append(formatWriterDelimiter);
                sb.Append(returnedImageQuality);
                sb.Append(formatWriterDelimiter);
                sb.Append(responseCode);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + errorMsg + "\"");
                imageResultWriter.WriteLine(sb.ToString());
            }
        }

        /// <summary>
        /// Used to output the format of the each image if desired
        /// </summary>
        /// <param name="imageId"></param>
        /// <param name="transactionId"></param>
        /// <param name="requestFormat"></param>
        /// <param name="imageFormat"></param>
        /// <param name="requestedQuality"></param>
        /// <param name="returnedImageQuality"></param>
        /// <param name="receivedImageChecksum">The checksum of the image received (calculated by VIX AK)</param>
        /// <param name="errorMsg"></param>
        private void OutputImageDownloadFormatResult(string imageId, string transactionId, string requestFormat, string imageFormat, 
            string requestedQuality, string returnedImageQuality, string imageChecksum, string errorMsg)
        {
            if (formatWriter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append(imageId);
                sb.Append(formatWriterDelimiter);
                sb.Append(transactionId);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + requestFormat + "\"");
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + imageFormat + "\"");
                sb.Append(formatWriterDelimiter);
                sb.Append(requestedQuality);
                sb.Append(formatWriterDelimiter);
                sb.Append(returnedImageQuality);
                sb.Append(formatWriterDelimiter);
                sb.Append(imageChecksum);
                sb.Append(formatWriterDelimiter);
                sb.Append("\"" + errorMsg + "\"");
                formatWriter.WriteLine(sb.ToString());
            }
        }

        private string getExpectedImageChecksum(Instance instance, ImageQuality imageQuality)
        {
            string expectedChecksum = null;
            if (OnGetImageChecksumEvent != null)
            {             
                expectedChecksum = OnGetImageChecksumEvent(instance.GloballyUniqueId, imageQuality);
                
                if (expectedChecksum == null)
                {
                    throw new ChecksumNotFoundException("Did not find expected checksum for image '" + instance.GloballyUniqueId + "' of quality '" + imageQuality + "', test for this image fails without expected checksum");
                }
            }
            
            return expectedChecksum;
        }

        /// <summary>
        /// This method attempts to download the image, if there is a socket exception it is thrown after a set number of attempts
        /// (not sure why socket exceptions are so special)
        /// </summary>
        /// <param name="study"></param>
        /// <param name="instance"></param>
        /// <param name="imageQuality"></param>
        /// <param name="transactionId"></param>
        /// <returns></returns>
        private DownloadResult TryDownload(ImageRequestInput imageRequestInput)// Study study, Instance instance, ImageQuality imageQuality, string transactionId)
        {
            int tryCount= 0;
            DownloadResult result = null;
            try
            {
                result = vixInterface.DownloadImage(imageRequestInput);
                if (OnOutputImageChecksum != null)
                {
                    OnOutputImageChecksum(imageRequestInput.Instance.GloballyUniqueId, 
                        imageRequestInput.ImageQuality, result.ReturnedImageChecksum, 
                        result.ImageBufferSize, result.ReturnedImageFormat);
                    // only doing if also outputting image checksums
                    if (mstrOutputChecksumImagesDirectory != null && mstrOutputChecksumImagesDirectory.Length > 0)
                    {
                        VixTestCommon.OutputImage(result.ImageBuffer, imageRequestInput.Instance.GloballyUniqueId,
                            imageRequestInput.ImageQuality, imageRequestInput.TransactionId, 
                            mstrOutputChecksumImagesDirectory);
                    }
                }
                
                return result;
            }
            catch (System.Net.Sockets.SocketException socketEx)
            {
                Logger().Error(LogMsg("Socket exception: " + socketEx.Message + "occured on try '" + tryCount + "' of '" + maxDownloadTries + "'"));
                if (tryCount >= maxDownloadTries)
                {
                    throw socketEx;
                }
                return result;
            }
        }


    }    
}
