/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.tester.exceptions;
using System.Xml.Serialization;
using System.IO;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    /// <summary>
    /// This class extends BaseAction and is meant to be used for outputting and verifying expected results
    /// </summary>
    public abstract class VerifiableTestAction : BaseAction
    {
        protected const string PROPERTY_OUTPUT_TRANSACTION_LOG_VALUES_ENABLED = "Output Transaction Log Values Enabled";
        protected const string PROPERTY_VERIFY_TRANSACTION_LOG_VALUES_ENABLED = "Verify Transaction Log Values Enabled";
        protected const string PROPERTY_TRANSACTION_LOG_VALUES_FILENAME = "Transaction Log Values Filename";
        protected const string PROPERTY_TRANSACTION_LOG_COMMAND_NAMES = "Transaction Log Command Names";

        public VerifiableTestAction()
            : base()
        {
            necessaryProperties.Add(PROPERTY_OUTPUT_TRANSACTION_LOG_VALUES_ENABLED, "false");
            necessaryProperties.Add(PROPERTY_VERIFY_TRANSACTION_LOG_VALUES_ENABLED, "false");
            necessaryProperties.Add(PROPERTY_TRANSACTION_LOG_VALUES_FILENAME, "");
            necessaryProperties.Add(PROPERTY_TRANSACTION_LOG_COMMAND_NAMES, "");
        }

        public override void runAction()
        {
            string description = mstrTestDescription;
            string actionDetails = GetActionName();
            if (description != null && description.Length > 0)
                actionDetails += " - " + description;
            LogInfoMsg("Running action [" + actionDetails + "]");
            LogInfoMsg("Client workstation [" + VixAKCommonUtils.GetHostname() + "]");
            ActionCompletedResult actionResult = ActionCompletedResult.success;
            VixAKTransactionLogSavableResultHolder resultHolder = null;
            SavableObject savableResult = null;
            try
            {
                resultHolder = GetActionResult();
                savableResult = resultHolder.SavableObject;
            }
            catch (Exception ex)
            {                
                if (IsExpectedResultException())
                {
                    LogInfoMsg("Exception executing action [" + GetActionName() + "] but exception was expected, outputting/verifying");
                    savableResult = new ExceptionSavableObject(ex);
                    actionResult = ActionCompletedResult.success; // reset for success, will change to resultsFailed below if necessary
                }
                else
                {
                    LogErrorMsg("Error executing action [" + GetActionName() + "], " + ex.Message);
                    actionResult = ActionCompletedResult.error;
                }
            }
            if (IsOutputResultsEnabled())
            {
                OutputExpectedResults(savableResult);
            }
            if (IsVerifyResultsEnabled())
            {
                if (!VerifyExpectedResults(savableResult))
                {
                    LogErrorMsg("Results do not match expected result");
                    actionResult = ActionCompletedResult.resultsFailed;
                }
                else
                {
                    LogInfoMsg("Results match expected result!");
                }
            }

            if (IsOutputTransactionLogValuesEnabled())
            {
                OutputTransactionLogValues(resultHolder.TransactionLogEntries);
            }
            if (IsVerifyTransactionLogValuesEnabled())
            {
                if (VerifyTransactionLogValues(resultHolder.TransactionLogEntries))
                {
                    LogInfoMsg("Transaction logs match expected results");
                }
                else
                {
                    LogErrorMsg("Transaction logs do not match expected results");
                    actionResult = ActionCompletedResult.resultsFailed;
                }
            }


            ActionComplete(actionResult);
        }

        /// <summary>
        /// Implementations of this method should not trap any exceptions. This method should do the test of executing the action
        /// and returning the result. If there are any errors or configuration problems, an exception should be thrown
        /// </summary>
        /// <returns></returns>        
        protected abstract VixAKTransactionLogSavableResultHolder GetActionResult();

        protected bool IsTransactionLogValuesNecessary()
        {
            return IsOutputTransactionLogValuesEnabled() || IsVerifyTransactionLogValuesEnabled();
        }

        protected bool IsOutputTransactionLogValuesEnabled()
        {
            if ((necessaryProperties.ContainsKey(PROPERTY_OUTPUT_TRANSACTION_LOG_VALUES_ENABLED) && (necessaryProperties.ContainsKey(PROPERTY_TRANSACTION_LOG_VALUES_FILENAME))))
            {
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_TRANSACTION_LOG_VALUES_FILENAME].ToString()))
                    return false;
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_OUTPUT_TRANSACTION_LOG_VALUES_ENABLED].ToString()))
                    return false;
                string[] commandNames = GetTransactionLogEntryCommandNames();
                if (commandNames.Length <= 0)
                    return false;
                return bool.Parse(necessaryProperties[PROPERTY_OUTPUT_TRANSACTION_LOG_VALUES_ENABLED].ToString());
            }
            return false;
        }

        protected bool IsVerifyTransactionLogValuesEnabled()
        {
            if ((necessaryProperties.ContainsKey(PROPERTY_VERIFY_TRANSACTION_LOG_VALUES_ENABLED) && (necessaryProperties.ContainsKey(PROPERTY_TRANSACTION_LOG_VALUES_FILENAME))))
            {
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_TRANSACTION_LOG_VALUES_FILENAME].ToString()))
                    return false;
                if (string.IsNullOrEmpty(necessaryProperties[PROPERTY_VERIFY_TRANSACTION_LOG_VALUES_ENABLED].ToString()))
                    return false;
                string[] commandNames = GetTransactionLogEntryCommandNames();
                if (commandNames.Length <= 0)
                    return false;
                return bool.Parse(necessaryProperties[PROPERTY_VERIFY_TRANSACTION_LOG_VALUES_ENABLED].ToString());
            }
            return false;
        }

        protected string GetOutputTransactionLogValuesFilename()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_TRANSACTION_LOG_VALUES_FILENAME))
            {
                return necessaryProperties[PROPERTY_TRANSACTION_LOG_VALUES_FILENAME].ToString();
            }
            return null;
        }

        protected void OutputTransactionLogValues(List<VixAKTransactionLogEntry> entries)
        {
            string filename = GetOutputTransactionLogValuesFilename();
            if ((filename != null) && (entries != null))
            {
                LogInfoMsg("Outputting transaction log expected results to file '" + filename + "'.");
                VixAKTransactionLogEntryHolder holder = new VixAKTransactionLogEntryHolder(entries);

                XmlSerializer serializer = new XmlSerializer(holder.GetType());
                TextWriter writer = new StreamWriter(filename);
                serializer.Serialize(writer, holder);
                writer.Close();
            }   
        }

        protected string[] GetTransactionLogEntryCommandNames()
        {
            if (necessaryProperties.ContainsKey(PROPERTY_TRANSACTION_LOG_COMMAND_NAMES))
            {
                string names = necessaryProperties[PROPERTY_TRANSACTION_LOG_COMMAND_NAMES].ToString();
                return names.Split(',');
            }
            return new string[0];
        }

        protected bool VerifyTransactionLogValues(List<VixAKTransactionLogEntry> results)
        {
            string filename = GetOutputTransactionLogValuesFilename();
            if ((filename != null) && (results != null))
            {
                LogInfoMsg("Comparing result to expected result from file '" + filename + "'.");
                XmlSerializer serializer = new XmlSerializer(typeof(VixAKTransactionLogEntryHolder));
                TextReader reader = new StreamReader(filename);
                VixAKTransactionLogEntryHolder holder = (VixAKTransactionLogEntryHolder)serializer.Deserialize(reader);
                List<VixAKTransactionLogEntry> expectedResults = holder.Entries;
                reader.Close();

                if (results.Count != expectedResults.Count)
                    return false;   

                for(int i = 0; i < results.Count; i++)
                {
                    VixAKTransactionLogEntry r = results[i];
                    VixAKTransactionLogEntry t = expectedResults[i];
                    if (!r.Equals(t))
                        return false;
                }
                return true;
            }
            return false;
        }

        /// <summary>
        /// This method determines if the transaction log entries should be loaded and then retrieves those transactions and 
        /// adds them to the transactionLogEntries list.
        /// </summary>
        /// <param name="url">Server and port of VIX</param>
        /// <param name="transactionId">Transaction ID to search for</param>
        /// <param name="transactionLogEntries">List of entries to add the transaction logs from the VIX to, this list cannot be null</param>
        /// <param name="relevantKeys"></param>
        protected void GetTransactionLogEntriesForCommandsIfEnabled(Uri url, string transactionId,
            List<VixAKTransactionLogEntry> transactionLogEntries, List<string> relevantKeys)
        {
            if ((IsTransactionLogValuesNecessary()) && (transactionLogEntries != null))
            {
                string[] commandNames = GetTransactionLogEntryCommandNames();
                foreach (string commandName in commandNames)
                {
                    LogMsg("Getting transaction log entries for command name '" + commandName + "'.");
                }
                List<TransactionLog.TransactionLogEntry> entries = GetTransactionLogEntryForCommands(url,
                    transactionId, commandNames);
                if (entries != null)
                {
                    foreach (TransactionLog.TransactionLogEntry entry in entries)
                    {
                        transactionLogEntries.Add(new VixAKTransactionLogEntry(entry, relevantKeys));
                    }
                }
            }
        }

        protected TransactionLog.TransactionLogEntry GetTransactionLogEntryForCommand(Uri url, string transactionId, string commandName)
        {
            string server = url.Host + ":" + url.Port;
            return GetTransactionLogEntryForCommand(server, transactionId, commandName);
        }

        protected TransactionLog.TransactionLogEntry GetTransactionLogEntryForCommand(string server, string transactionId, string commandName)
        {
            List<TransactionLog.TransactionLogEntry> entries = GetTransactionLogEntryForCommands(server, transactionId, 
                new string[] { commandName });
            if ((entries != null) && (entries.Count > 0))
                return entries[0];
            return null;
        }

        protected List<TransactionLog.TransactionLogEntry> GetTransactionLogEntryForCommands(Uri url, string transactionId, string[] commandNames)
        {
            string server = url.Host + ":" + url.Port;
            return GetTransactionLogEntryForCommands(server, transactionId, commandNames);
        }

        /// <summary>
        /// Return multiple transaction log entries for a single transaction ID. The entries returned will match the specified command names
        /// </summary>
        /// <param name="server"></param>
        /// <param name="transactionId"></param>
        /// <param name="commandNames"></param>
        /// <returns></returns>
        protected List<TransactionLog.TransactionLogEntry> GetTransactionLogEntryForCommands(string server, string transactionId, string []commandNames)
        {
            List<TransactionLog.TransactionLogEntry> result = new List<TransactionLog.TransactionLogEntry>();
            LogInfoMsg("Searching for transaction Id [" + transactionId + "] from server [" + server + "]");
            List<TransactionLog.TransactionLogEntry> entries =
                TransactionLog.TransactionLogDownloader.DownloadTransactions(server, DateTime.Now, DateTime.Now, transactionId);
            foreach (TransactionLog.TransactionLogEntry entry in entries)
            {
                foreach (string commandName in commandNames)
                {
                    if (commandName == entry.CommandClassName)
                    {
                        result.Add(entry);
                        break;
                    }
                }
            }
            return result;
        }

        protected VixTestUser GetSelectedUser(string propertyKey)
        {
            if (necessaryProperties.ContainsKey(propertyKey))
            {
                ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[propertyKey];
                if (customOption.Options.Count != 1)
                    throw new Exception("Must specify exactly 1 user to query as");
                return (VixTestUser)customOption.Options[0];
            }
            else
            {
                return VixTestUser.USER135;
            }
        }
    }
}

