/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.services;
using System.IO;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class VistaRadActiveExamsTestAction : BaseVistaRadAction
    {

        public static string VISTARAD_ACTIVE_EXAMS_ACTION_NAME = "VistARad Active Exams VIX Test";

        private string PROPERTY_LIST_DESCRIPTOR = "ListDescriptor";
        private string PROPERTY_USER_DIVISION = "UserDivision";

        public VistaRadActiveExamsTestAction()
            : base(VISTARAD_ACTIVE_EXAMS_ACTION_NAME)
        {            
            SetNecessaryPropertyWithDefault(PROPERTY_LIST_DESCRIPTOR, "9991");
            SetNecessaryPropertyWithDefault(PROPERTY_USER_DIVISION, "");
        }

        public override string GetActionName()
        {
            return VISTARAD_ACTIVE_EXAMS_ACTION_NAME;
        }

        public override void runAction()
        {
            formatWritter = null;
            LogInfoMsg("Running action [" + GetActionName() + "]");
            LogInfoMsg("Client workstation [" + GetHostname() + "]");

            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            string txtUrl = necessaryProperties[PROPERTY_TEXT_URL].ToString();
            LogInfoMsg("Retrieving data from metadata url [" + metadataUrl + "], image url [" + imageUrl + "]");

            VistaRadServiceImpl vistaRadService = new VistaRadServiceImpl(metadataUrl, imageUrl, txtUrl, LogMsg);            
            List<ImageQuality> imageQualities = new List<ImageQuality>();
            imageQualities.Add(ImageQuality.DIAGNOSTIC);

            string siteNumber = necessaryProperties[PROPERTY_SITE_NUMBER].ToString();
            string listDescriptor = necessaryProperties[PROPERTY_LIST_DESCRIPTOR].ToString();
            string userDivision = necessaryProperties[PROPERTY_USER_DIVISION].ToString();
            bool outputImageChecksums = GetOutputImageChecksums();
            bool verifyImageChecksums = GetVerifyImageChecksums();
            bool cacheClear = GetCacheClear();
            string imageChecksumsFilename = GetImageChecksumsFilename();
            
            LogInfoMsg("Site number [" + siteNumber + "]");
            LogInfoMsg("List Descriptor [" + listDescriptor + "]");
            LogInfoMsg("User division [" + userDivision + "]");
            LogInfoMsg("Cache Clear [" + cacheClear + "]");            

            bool outputImageFormats = GetOutputImageFormat();
            outputImageFormatsFilename = "";
            if (outputImageFormats) 
            {
                if (formatWritter == null)
                {
                    outputImageFormatsFilename = VixAKCommonUtils.getApplicationDir() + "\\imageFormats.csv";
                    LogInfoMsg("Writing output image formats to file [" + outputImageFormatsFilename + "]");
                    formatWritter = new StreamWriter(outputImageFormatsFilename, false);
                    CreateVistaRadImageFormatHeader();
                }
            }
            LogInfoMsg("Output Image Format [" + outputImageFormats + "]");

            /*
            if (threaded)
                outputImageChecksums = false;
             * */
            if (outputImageChecksums)
            {
                if (imageChecksumWriter == null)
                {
                    if (imageChecksumsFilename == null)
                    {
                        throw new Exception("Image checksum filename is null, cannot output image checksums");
                    }
                    else
                    {
                        LogInfoMsg("Writing image checksums to file [" + imageChecksumsFilename + "]");
                        imageChecksumWriter = TextWriter.Synchronized(new StreamWriter(imageChecksumsFilename, false));
                        string checksumImageDirectory = getChecksumImageDirectory();
                        if (checksumImageDirectory != null && checksumImageDirectory.Length > 0)
                        {
                            LogInfoMsg("Writing images to directory '" + checksumImageDirectory + "'.");
                        }
                        else
                        {
                            LogInfoMsg("Not writing images locally");
                        }
                    }
                }
            }
            if (verifyImageChecksums)
            {
                if (imageChecksumsFilename == null)
                {
                    throw new Exception("Image checksum filename is null, cannot verify image checksums");
                }
                LoadImageChecksums(imageChecksumsFilename);
            }

            lastTestResult = new VixTestResult(false, 1, imageQualities.Count, !cacheClear, false, false, 0);

            testActiveExams(vistaRadService, siteNumber, userDivision, listDescriptor);

            lastTestResult.SetComplete();
            //LogInfoMsg("-----------------------");
            ProcessingComplete();
            //LogInfoMsg(vixTestResult.ToString());
            //ActionComplete(true);
        }

        public bool testActiveExams(VistaRadServiceImpl vistaRadService, string siteNumber, string userDivision, 
            string listDescriptor)
        {
            
            List<Exam> exams = null;
            try
            {
                LogInfoMsg("Retrieving active exams for site '" + siteNumber + "' and list descriptor '" + listDescriptor + "'.");
                exams = vistaRadService.getActiveExams(siteNumber, userDivision, listDescriptor);
                LogInfoMsg("Found '" + exams.Count + "' active exams at site '" + siteNumber + "'.");
            }
            catch (Exception ex)
            {
                LogErrorMsg("Error getting active exams, " + ex.Message);
                
                return false;
            }

            ProgressBarMax(exams.Count);
            foreach (Exam exam in exams)
            {
                if (!mbCanRun)
                    break;
                VixPatientTestResult patientTestResult = new VixPatientTestResult(new ImagingPatient(exam.ExamId), true);
                lastTestResult.Patients.Add(patientTestResult);
                testExam(exam, vistaRadService, patientTestResult);
                IncrementProgress();
            }
            return true;
        }

        protected void CreateVistaRadImageFormatHeader()
        {
            if (formatWritter != null)
            {
                StringBuilder sb = new StringBuilder();
                sb.Append("Transaction Id");
                sb.Append(formatWriterDelimiter);
                sb.Append("Request Parameters");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Image Quality");
                sb.Append(formatWriterDelimiter);
                sb.Append("Response Image Format");
                sb.Append(formatWriterDelimiter);
                sb.Append("Error Message");
                formatWritter.WriteLine(sb.ToString());
            }
        }

        public override void stopAction()
        {
            base.stopAction();
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_LIST_DESCRIPTOR))
                return false;
            if (!validatePropertyIncluded(PROPERTY_USER_DIVISION))
                return false;
            return base.ValidateNecessaryProperties();
        }
    }
}
