/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Windows.Forms;
using VixAKCommon.gov.va.med.imaging;
using log4net;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions
{
    public class VixResultRollupHelper
    {
        private const string TOMCAT_LOG_FILENAME = "ImagingExchangeWebApp.log";
        private const string VIX_AK_LOG_FILENAME = "vix-ak-log.txt";
        private const string csvDelimiter = ",";

        private event LogMessageDelegate OnLogMessage;
        private event DoApplicationEventsDelegate OnDoApplicationEventsEvent;

        private void DoApplicationEvents() 
        {
            if (OnDoApplicationEventsEvent != null)
            {
                OnDoApplicationEventsEvent();
            }
        }

        private string LogMsg(string msg)
        {
            if (OnLogMessage != null)
            {
                OnLogMessage(msg);
            }
            return msg;
        }

        protected ILog Logger()
        {
            return LogManager.GetLogger(this.GetType().Name);
        }

        public VixResultRollupHelper(LogMessageDelegate logMsgDelegate, DoApplicationEventsDelegate doApplicationEventsDelegate)
        {
            this.OnLogMessage += logMsgDelegate;
            this.OnDoApplicationEventsEvent += doApplicationEventsDelegate;
        }

        public void RollupTestResult(VixTestResult testResult, bool createPatientCSV, string outputDir, 
            string tomcatLogsDir, string formatsFile, string imageAccessResultsFilename,
            string compareResultsFilename)
        {            
            string dir = GetUniqueDirectory(outputDir, testResult);
            if (!Directory.Exists(dir))
            {
                Logger().Info(LogMsg("Creating directory [" + dir + "]"));
                Directory.CreateDirectory(dir);
            }

            CopyTomcatFile(dir, tomcatLogsDir);
            CopyImageLoaderFile(dir);
            CopyOutputFormatsFile(dir, formatsFile);
            CopyImageAccessResultsFilename(dir, imageAccessResultsFilename);
            CopyCompareResultsFilename(dir, compareResultsFilename);
            if (createPatientCSV)
            {
                CreatePatientDetailsCSV(dir, testResult);
            }        
        }

        private void CopyCompareResultsFilename(string outputDir, string compareResultsFilename)
        {
            if ((compareResultsFilename != null) && (compareResultsFilename.Length > 0))
            {
                FileInfo file = new FileInfo(compareResultsFilename);
                if (file.Exists)
                {
                    try
                    {
                        string outFile = outputDir + "\\" + file.Name;
                        Logger().Info(LogMsg("Copying compare results file [" + compareResultsFilename + "] to [" + outFile + "]"));
                        File.Copy(compareResultsFilename, outFile);
                    }
                    catch (Exception ex)
                    {
                        Logger().Error(LogMsg("Error copmare results file [" + compareResultsFilename + "], [" + ex.Message + "]"));
                    }
                }
            }
        }

        private void CopyImageAccessResultsFilename(string outputDir, string imageAccessResultsFilename)
        {
            if ((imageAccessResultsFilename != null) && (imageAccessResultsFilename.Length > 0))
            {
                FileInfo file = new FileInfo(imageAccessResultsFilename);
                if (file.Exists)
                {
                    try
                    {
                        string outFile = outputDir + "\\" + file.Name;
                        Logger().Info(LogMsg("Copying image access results file [" + imageAccessResultsFilename + "] to [" + outFile + "]"));
                        File.Copy(imageAccessResultsFilename, outFile);
                    }
                    catch (Exception ex)
                    {
                        Logger().Error(LogMsg("Error copying image access results file [" + imageAccessResultsFilename + "], [" + ex.Message + "]"));
                    }
                }
            }
        }

        private void CopyOutputFormatsFile(string outputDir, string formatsFilename)
        {
            if ((formatsFilename != null) && (formatsFilename.Length > 0))
            {
                FileInfo file = new FileInfo(formatsFilename);
                if (file.Exists)
                {
                    try
                    {
                        string outFile = outputDir + "\\" + file.Name;
                        Logger().Info(LogMsg("Copying formats file [" + formatsFilename + "] to [" + outFile + "]"));
                        File.Copy(formatsFilename, outFile);
                    }
                    catch (Exception ex)
                    {
                        Logger().Error(LogMsg("Error copying formats file [" + formatsFilename + "], [" + ex.Message +"]"));
                    }
                }
            }
        }

        private void CopyTomcatFile(string outputDir, string tomcatLogsDir)
        {
            if ((tomcatLogsDir != null) && (tomcatLogsDir.Length > 0))
            {
                string tomcatFile = tomcatLogsDir + "\\" + TOMCAT_LOG_FILENAME;
                if (File.Exists(tomcatFile))
                {
                    Logger().Info(LogMsg("Copying file [" + tomcatFile + "] to [" + outputDir + "]"));
                    File.Copy(tomcatFile, outputDir + "\\" + TOMCAT_LOG_FILENAME);
                }
                else
                {
                    Logger().Error(LogMsg("Cannot find file [" + tomcatFile + "]"));
                }
            }
        }

        private void CopyImageLoaderFile(string outputDir)
        {
            FileInfo imageLoaderFile = new FileInfo(VixAKCommonUtils.getApplicationDir(Application.ExecutablePath) + "\\" + VIX_AK_LOG_FILENAME);
            Logger().Info(LogMsg("Copying file [" + imageLoaderFile.FullName + "] to [" + outputDir + "]"));
            if (!imageLoaderFile.Exists)
            {
                Logger().Error(LogMsg("Cannot find file [" + imageLoaderFile.FullName + "]"));
            }
            File.Copy(imageLoaderFile.FullName, outputDir + "\\" + imageLoaderFile.Name);
            DoApplicationEvents();
            try
            {
                LogManager.Shutdown();
                imageLoaderFile.Delete();

                string configFilename = VixAKCommonUtils.getApplicationDir() + "\\VixAKLog.xml";
                log4net.Config.XmlConfigurator.Configure(new System.IO.FileInfo(configFilename));
                DoApplicationEvents();
            }
            catch (Exception ex)
            {
                Logger().Error(LogMsg(ex.Message));
            }
        }

        private void CreatePatientDetailsCSV(string outputDir, VixTestResult testResult)
        {
            string csv = outputDir + "\\" + getOutputDirName(testResult) + ".csv";
            StreamWriter writer = null;
            try
            {
                writer = new StreamWriter(csv);
                WriteCSVHeader(writer);
                foreach (VixPatientTestResult patientResult in testResult.Patients)
                {
                    WriteTestRunLine(writer, patientResult);
                }
            }
            finally
            {
                writer.Close();
            }
        }

        private void WriteTestRunLine(StreamWriter writer, VixPatientTestResult patientResult)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("\"" + patientResult.Patient.PatientName + "\"");
            sb.Append(csvDelimiter);
            sb.Append(patientResult.Patient.PatientIcn);
            sb.Append(csvDelimiter);
            sb.Append(patientResult.Studies.ToString());
            sb.Append(csvDelimiter);
            sb.Append(patientResult.TotalImages.ToString());
            sb.Append(csvDelimiter);
            sb.Append(patientResult.ImagesDownloaded.ToString());
            sb.Append(csvDelimiter);
            sb.Append(patientResult.ImagesFailed.ToString());
            writer.WriteLine(sb.ToString());
        }

        private void WriteCSVHeader(StreamWriter writer)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("Patient Name");
            sb.Append(csvDelimiter);
            sb.Append("Patient ICN");
            sb.Append(csvDelimiter);
            sb.Append("Study Count");
            sb.Append(csvDelimiter);
            sb.Append("Image Count");
            sb.Append(csvDelimiter);
            sb.Append("Successful Images");
            sb.Append(csvDelimiter);
            sb.Append("Failed Images");
            writer.WriteLine(sb.ToString());
        }

        private string GetUniqueDirectory(string outputDir, VixTestResult testResult)
        {
            int count = 0;
            string baseDir = GetOutputDir(outputDir, testResult);
            string curDir = baseDir;
            while (Directory.Exists(curDir))
            {
                count++;
                curDir = baseDir + "(" + count + ")";
            }
            return curDir;
        }

        private string GetOutputDir(string outputDir, VixTestResult testResult)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(outputDir);
            sb.Append("\\");
            sb.Append(getOutputDirName(testResult));
            return sb.ToString();
        }

        private string getOutputDirName(VixTestResult testResult)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(DateTime.Now.ToString("yyyy-MM-dd"));
            sb.Append("_");

            if (testResult.Cached)
            {
                sb.Append("cached");
            }
            else
            {
                sb.Append("uncached");
            }
            if (testResult != null)
            {
                sb.Append("-");
                if (testResult.Threaded)
                {
                    sb.Append(" threaded");
                }
                else
                {
                    sb.Append(" not-threaded");
                }
                //sb.Append(testResult.Threaded ThreadCount.ToString());
                //sb.Append(" threads");
                sb.Append(",");
                sb.Append(testResult.WorkerCount.ToString());
                sb.Append(" repeats");

                if (testResult.MetadataOnly)
                {
                    sb.Append("_metadata");
                }
                
            }
            return sb.ToString();
        }
    }
}
