/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions;
using VixAKCommon.gov.va.med.imaging.vix.services;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions.options;

namespace VixAKCommon.gov.va.med.imaging.vix.tester.actions.exchange
{
    public class ExchangeGetImage : VerifiableTestAction
    {
        public static string EXCHANGE_GET_IMAGE = "Exchange Get Image";

        private const string PROPERTY_METADATA_URL = "MetadataUrl";
        private const string PROPERTY_IMAGE_URL = "ImageUrl";
        private const string PROPERTY_INTERFACE_VERSION = "Interface Version";
        private const string PROPERTY_USERNAME = "Username";
        private const string PROPERTY_PASSWORD = "Password";
        private const string PROPERTY_IMAGE_ID = "Image ID";
        private const string PROPERTY_IMAGE_QUALITIES = "Image Qualities";


        public ExchangeGetImage()
            : base()
        {
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_METADATA_URL, "");
            SetNecessaryPropertyWithDefault(PROPERTY_INTERFACE_VERSION, "2");

            SetNecessaryPropertyWithDefault(PROPERTY_USERNAME, "");
            SetNecessaryPropertyWithDefault(PROPERTY_PASSWORD, "");
            SetNecessaryPropertyWithDefault(PROPERTY_IMAGE_ID, "*");

            List<ImageQuality> imageQualities = ImageQuality.GetImageQualities();
            ActionCustomOption qualitiesOptions = new ActionCustomOption(PROPERTY_IMAGE_QUALITIES);
            foreach (ImageQuality quality in imageQualities)
            {
                qualitiesOptions.Options.Add((BaseCustomOption)quality);
            }
            this.necessaryProperties.Add(PROPERTY_IMAGE_QUALITIES, qualitiesOptions);
        }

        private List<ImageQuality> GetSelectedImageQualities()
        {
            ActionCustomOption customOption = (ActionCustomOption)necessaryProperties[PROPERTY_IMAGE_QUALITIES];
            List<ImageQuality> qualities = new List<ImageQuality>();

            foreach (BaseCustomOption option in customOption.Options)
            {
                if (option is ImageQuality)
                {
                    ImageQuality quality = (ImageQuality)option;
                    qualities.Add(quality);
                }
            }

            return qualities;
        }

        protected override VixAKTransactionLogSavableResultHolder GetActionResult()
        {

            VixInterface vixInterface = null;
            string metadataUrl = necessaryProperties[PROPERTY_METADATA_URL].ToString();
            string imageUrl = necessaryProperties[PROPERTY_IMAGE_URL].ToString();
            string interfaceVersion = necessaryProperties[PROPERTY_INTERFACE_VERSION].ToString();
            string username = necessaryProperties[PROPERTY_USERNAME].ToString();
            string password = necessaryProperties[PROPERTY_PASSWORD].ToString();
            string imageUrn = necessaryProperties[PROPERTY_IMAGE_ID].ToString();
            List<ImageQuality> qualities = GetSelectedImageQualities();
            

            LogInfoMsg("Retrieving data from metadata url [" + metadataUrl + "], image url [" + imageUrl + "]");
            LogInfoMsg("Retrieving image ID [" + imageUrn + "]");

            LogInfoMsg("Image Qualities: ");
            foreach (ImageQuality quality in qualities)
            {
                LogInfoMsg(quality.ToString());
            }

            switch (interfaceVersion)
            {
                case "1":
                    vixInterface = new ExchangeServiceImpl(metadataUrl, imageUrl, username, password, LogMsg);
                    break;
                case "2":
                    vixInterface = new ExchangeServiceImpl_v2(metadataUrl, imageUrl, username, password, LogMsg);
                    break;
            }
            if (vixInterface == null)
            {
                throw new Exception("Unable to create Exchange interface for version '" + interfaceVersion + "'");
            }

            Instance instance = new Instance(imageUrn);
            int imageCount = qualities.Count;
            ProgressBarMax(imageCount);

            List<DownloadResult> results = new List<DownloadResult>();

            foreach (ImageQuality imageQuality in qualities)
            {
                string transactionId = Guid.NewGuid().ToString();
                try
                {
                    DownloadResult result =
                        vixInterface.DownloadImage(new ImageRequestInput(null, instance, imageQuality, transactionId, true));
                    results.Add(result);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error downloading image '" + imageUrn + "' with transaction Id '" + transactionId + "'.");
                    throw ex;
                }
                finally
                {
                    // only happens for 1 thread
                    IncrementProgress();
                }
            }
            return new VixAKTransactionLogSavableResultHolder(new SavableDownloadResultList(results));
        }

        public override string GetActionName()
        {
            return EXCHANGE_GET_IMAGE;
        }

        public override bool ValidateNecessaryProperties()
        {
            if (!validatePropertyIncluded(PROPERTY_METADATA_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_URL))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_ID))
                return false;
            if (!validatePropertyIncluded(PROPERTY_IMAGE_QUALITIES))
                return false;
            if (!validatePropertyIncluded(PROPERTY_INTERFACE_VERSION))
                return false;
            if (!validatePropertyIncluded(PROPERTY_USERNAME))
                return false;
            if (!validatePropertyIncluded(PROPERTY_PASSWORD))
                return false;

            return true;
        }
    }
}
