/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixAKCommon.gov.va.med.imaging.vix.tester.actions;
using VixAKCommon.gov.va.med.imaging.vix.tester;
using VixAKCommon.gov.va.med.imaging.vix;
using log4net;
using System.Windows.Forms;
using System.Configuration;
using VixCommon.email;

namespace VixAKConsole
{
    public class AKConsoleRunner
    {
        private bool mbContinueOnFailedActions = false;
        private List<BaseAction> mActions = null;
        private AKActionResults actionResults = null;
        private AKEmailConfiguration emailConfiguration = null;
        private string mstrXsltFilename = "";
        private readonly string mstrSimpleResultsOutputFilename;

        public AKConsoleRunner(string actionsFilename, string simpleResultsOutputFilename)
        {
            mstrSimpleResultsOutputFilename = simpleResultsOutputFilename;
            //string actionsFilename = @"C:\work\Objective 6\VIX AK\Functional Testing\scripts\SiteServiceVerify.xml"; // this should be provided at command line
            List<BaseAction> actions = VixAKCommon.gov.va.med.imaging.vix.VixAKCommonUtils.LoadActionsFromXml(actionsFilename);
            mActions = new List<BaseAction>();
            foreach (BaseAction action in actions)
            {
                if (action.Enabled)
                    mActions.Add(action);
            }

            mbContinueOnFailedActions = bool.Parse(ConfigurationManager.AppSettings["ContinueOnFailedActions"]);
            string AKResultsDirectory = ConfigurationManager.AppSettings["AKResultsDirectory"];
            bool emailResultsEnabled = bool.Parse(ConfigurationManager.AppSettings["EmailResultsEnabled"]);
            string emailSendTo = ConfigurationManager.AppSettings["EmailSendTo"];
            string []sendTo = emailSendTo.Split(',');

            bool useIIS = bool.Parse(ConfigurationManager.AppSettings["useIIS"]);
            string smtpServer = ConfigurationManager.AppSettings["SMTPServer"];
            string emailFrom = ConfigurationManager.AppSettings["EmailFrom"];
            if (emailResultsEnabled)
            {
                //string emailFrom = "\" VIX AK Test\" <donotreply@domain>";// "<donotreply@domain>";
                emailConfiguration = new AKEmailConfiguration(emailFrom, sendTo, useIIS, smtpServer);
            }

            DateTime now = DateTime.Now;
            actionResults = new AKActionResults(actionsFilename, 
                AKResultsDirectory + "\\" + now.ToString("yyyyMMdd_hhmmss") + ".xml", now);
            mstrXsltFilename = AKResultsDirectory + "\\display.xsl";
        }

        public bool Execute()
        {
            LogInfoMsg("Running '" + mActions.Count + "' actions");
            RunNextAction();

            return (actionResults == null ? false : actionResults.AllTestsPassed);
        }

        private void RunNextAction()
        {
            if ((mActions != null) && (mActions.Count > 0))
            {
                BaseAction action = mActions[0];
                mActions.Remove(action);
                RunAction(action);
            }
            else
            {
                AllActionsComplete();
            }
        }

        private void AllActionsComplete()
        {
            LogInfoMsg("Actions complete");
            actionResults.Save();
            LogInfoMsg(actionResults.ToString());

            if (mstrSimpleResultsOutputFilename != null)
            {
                try
                {
                    actionResults.SaveSimpleResults(mstrSimpleResultsOutputFilename);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error saving simple results to file '" + mstrSimpleResultsOutputFilename + "', " + ex.Message + ".");
                }
            }

            if (emailConfiguration != null)
            {
                LogInfoMsg("Sending email test result");
                string subject = "VIX AK Test Results " + actionResults.StartTime.ToString();
                string message = actionResults.ConvertToXslt(mstrXsltFilename);
                // this handles some strangeness with the XSLT conversion
                if (message.StartsWith("???"))
                {
                    message = message.Substring(3);
                }
                try
                {
                    EmailSender.EmailMessage(emailConfiguration.EmailServerConfiguration,
                        emailConfiguration.SendTo, emailConfiguration.SendFrom, subject, message);
                }
                catch (Exception ex)
                {
                    LogErrorMsg("Error sending email, " + ex.Message);
                }
            }
        }

        private void RunAction(BaseAction action)
        {
            if (action.ValidateNecessaryProperties())
            {
                action.OnLogMessageEvent += LogMsg;
                action.OnSetProgressBarMaxEvent += SetMaxProgress;
                action.OnIncrementProgressEvent += IncrementProgress;
                action.OnActionCompleteEvent += ActionComplete;
                action.OnDoApplicationEventsEvent += DoApplicationEvents;
                try
                {
                    action.runAction();
                }
                catch (Exception ex)
                {
                    // this shouldn't happen
                    LogErrorMsg("Error processing action '" + action.GetActionName() + "', " + ex.Message + "\n\n" + ex.StackTrace);
                    RunNextAction();
                }
            }
            else
            {
                LogMsg("Action '" + action.GetActionName() + "' failed to validate necessary properties.", true);
            }
        }

        private void ActionComplete(BaseAction action, ActionCompletedResult actionCompletedResult)
        {
            actionResults.ActionResults.Add(new ActionResult(action, actionCompletedResult));
            if (actionCompletedResult == ActionCompletedResult.success || mbContinueOnFailedActions)
            {
                RunNextAction();
            }
            else
            {
                LogMsg("Action did not complete successfully and not continuing on failed actions", true);
            }


            /*
            if (InvokeRequired)
            {
                ActionCompleteDelegate d = new ActionCompleteDelegate(this.ActionComplete);
                this.Invoke(d, new object[] { actionOk });
            }
            else
            {
                if (actionOk)
                    RunNextAction();
                else
                {
                    Logger().Error(LogMsg("The previous action failed to complete correctly, halting processing"));
                    toolStripProgressBar1.Value = toolStripProgressBar1.Minimum;
                    SetButtonsEnabled(true);
                    Cursor.Current = Cursors.Default;
                    toolStripCurActionCountLabel.Text = "";
                    curActionNumber = 0;
                    actionCount = 0;
                }
            }*/
        }

        protected ILog Logger()
        {
            return LogManager.GetLogger(this.GetType().Name);
        }

        private string LogMsg(string msg)
        {
            return LogMsg(msg, false);
        }

        /// <summary>
        /// Logs to Log4Net and console
        /// </summary>
        /// <param name="msg"></param>
        private void LogErrorMsg(string msg)
        {
            Logger().Error(msg);
            LogMsg(msg, false);
        }

        /// <summary>
        /// Logs to Log4Net and console
        /// </summary>
        /// <param name="msg"></param>
        private void LogInfoMsg(string msg)
        {
            Logger().Info(msg);
            LogMsg(msg, false);
        }

        private string LogMsg(string msg, bool isError)
        {
            /*
            if (isError)
            {
                Logger().Error(msg);
            }
            else
            {
                Logger().Info(msg);
            }*/
            Console.WriteLine(msg);
            return msg;
        }

        private void SetMaxProgress(int max)
        {
            /*
            if (InvokeRequired)
            {
                SetProgressBarMax d = new SetProgressBarMax(this.SetMaxProgress);
                this.Invoke(d, new object[] { max });
            }
            else
            {
                toolStripProgressBar1.Maximum = max;
                toolStripProgressBar1.Value = toolStripProgressBar1.Minimum;
                toolStripProgressLabel.Text = "";
                Application.DoEvents();
            }*/
        }

        private void IncrementProgress()
        {
            /*
            if (this.InvokeRequired)
            {
                IncrementProgressDelegate d = new IncrementProgressDelegate(this.IncrementProgress);
                this.Invoke(d, new object[] { });
            }
            else
            {
                if (toolStripProgressBar1.Value < toolStripProgressBar1.Maximum)
                {
                    toolStripProgressBar1.PerformStep();
                    toolStripProgressLabel.Text = toolStripProgressBar1.Value + " of " + toolStripProgressBar1.Maximum;
                    Application.DoEvents();
                }
            }*/
        }

        private void DoApplicationEvents()
        {
            /*
            if (this.InvokeRequired)
            {
                DoApplicationEventsDelegate d = new DoApplicationEventsDelegate(this.IncrementProgress);
                this.Invoke(d, new object[] { });
            }
            else
            {
                Application.DoEvents();
            }*/
            Application.DoEvents();

        }

    }

    public class AKEmailConfiguration
    {
        private readonly string mstrSmtpServer;
        private readonly bool mbUseIIS;
        private readonly string mstrSendFrom;
        private readonly string[] mstrSendTo;

        public string SmtpServer
        {
            get { return mstrSmtpServer; }
        } 

        public bool UseIIS
        {
            get { return mbUseIIS; }
        }         

        public string SendFrom
        {
            get { return mstrSendFrom; }
        } 

        public string[] SendTo
        {
            get { return mstrSendTo; }
        } 

        public AKEmailConfiguration(string sendFrom, string [] sendTo, bool useIIS, string smtpServer)
        {
            mstrSendFrom = sendFrom;
            mstrSendTo = sendTo;
            mstrSmtpServer = smtpServer;
            mbUseIIS = useIIS;
        }

        public EmailServerConfiguration EmailServerConfiguration
        {
            get
            {
                if (UseIIS)
                {
                    return EmailServerConfiguration.createIISConfiguration();
                }
                else
                {
                    return EmailServerConfiguration.createSmtpConfiguration(mstrSmtpServer);
                }
            }
        }
    }
}
