/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.IO;
using System.Net;

namespace VixCertificateTool
{
    public delegate void LogMessageDelegate(string msg);

    public class CertificateHelper
    {
        
        //public delegate void SuccessResponseDelegate(string msg);

        public static string GetJavaHome()
        {
            string javaHome = Environment.GetEnvironmentVariable("JAVA_HOME");
            return javaHome;
        }
        


        private static string GetCsrDetails(string csrFilename)
        {
            StringBuilder csr = new StringBuilder();
            StreamReader csrreader = new StreamReader(csrFilename);
            string line = csrreader.ReadLine();
            while (line != null)
            {
                csr.Append(line);
                csr.Append("%0D%0A");
                line = csrreader.ReadLine();
            }
            csrreader.Close();
            string csrString = csr.ToString();

            csrString = csrString.Replace("-----BEGIN NEW CERTIFICATE REQUEST-----", "-----BEGIN+NEW+CERTIFICATE+REQUEST-----");
            csrString = csrString.Replace("-----END NEW CERTIFICATE REQUEST-----", "-----END+NEW+CERTIFICATE+REQUEST-----");

            csrString = csrString.Replace("+", "%2B");
            csrString = csrString.Replace(" ", "%20");
            return csrString;
        }
        

        public static bool ImportCertificateResponseIntoKeystore(string keystoreFilename, string certificateResponseFilename,
            string keystorePassword, string alias, LogMessageDelegate OnErrorResponse)
        {
            if (!File.Exists(keystoreFilename))
            {
                LogMsg(OnErrorResponse, "Keystore file [" + keystoreFilename + "] does not exist");
                return false;
            }
            if (!File.Exists(certificateResponseFilename))
            {
                LogMsg(OnErrorResponse, "Certificate Response file [" + certificateResponseFilename + "] does not exist");
                return false;
            }
            string arguments = CreateImportCertificateResponseArguments(keystoreFilename, 
                certificateResponseFilename, keystorePassword, alias);
            Process proc = new Process();
            string javaHome = GetJavaHome();
            if ((javaHome == null) || (javaHome.Length <= 0))
            {
                LogMsg(OnErrorResponse, "JAVA_HOME environment variable not set");
                return false;
            }

            proc.StartInfo.Arguments = arguments;
            proc.StartInfo.WorkingDirectory = javaHome + "\\bin"; //@"c:\Program Files\Java\jdk1.6.0_03\bin";
            proc.StartInfo.FileName = "keytool.exe";

            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.RedirectStandardError = true;
            proc.StartInfo.UseShellExecute = false;
            proc.Start();
            StreamReader reader = proc.StandardOutput;
            StreamReader errorReader = proc.StandardError;

            StringBuilder errorResponse = new StringBuilder();
            string line = reader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = reader.ReadLine();
            }
            line = errorReader.ReadLine();
            while (line != null)
            {
                if (line != "Certificate reply was installed in keystore")
                {
                    errorResponse.AppendLine(line);
                }
                line = errorReader.ReadLine();
            }
            if (errorResponse.Length > 0)
            {
                LogMsg(OnErrorResponse, errorResponse.ToString());
                return false;
            }
            return true;
        }

        public static bool ImportRootCertIntoKeystore(string keystoreFilename, string rootCertificateFilename, 
            string keystorePassword, LogMessageDelegate OnErrorResponse)
        {
            if (!File.Exists(keystoreFilename))
            {
                LogMsg(OnErrorResponse, "Keystore file [" + keystoreFilename + "] does not exist");
                return false;
            }
            if (!File.Exists(rootCertificateFilename))
            {
                LogMsg(OnErrorResponse, "Root Certificate file [" + rootCertificateFilename + "] does not exist");
                return false;
            }
            string arguments = CreateImportRootCertArguments(keystoreFilename, rootCertificateFilename, keystorePassword);
            Process proc = new Process();
            string javaHome = GetJavaHome();
            if ((javaHome == null) || (javaHome.Length <= 0))
            {
                LogMsg(OnErrorResponse, "JAVA_HOME environment variable not set");
                return false;
            }

            proc.StartInfo.Arguments = arguments;
            proc.StartInfo.WorkingDirectory = javaHome + "\\bin"; //@"c:\Program Files\Java\jdk1.6.0_03\bin";
            proc.StartInfo.FileName = "keytool.exe";

            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.RedirectStandardError = true;
            proc.StartInfo.UseShellExecute = false;
            proc.Start();
            StreamReader reader = proc.StandardOutput;
            StreamReader errorReader = proc.StandardError;

            StringBuilder errorResponse = new StringBuilder();
            string line = reader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = reader.ReadLine();
            }
            line = errorReader.ReadLine();
            while (line != null)
            {
                if (line != "Certificate was added to keystore")
                {
                    errorResponse.AppendLine(line);
                }
                line = errorReader.ReadLine();
            }

            

            if (errorResponse.Length > 0)
            {
                LogMsg(OnErrorResponse, errorResponse.ToString());
                return false;
            }
            return true;
        }

        public static bool CreateCSR(string keystoreFilename, string csrFilename, string alias, string keystorePassword,
            LogMessageDelegate OnErrorResponse)
        {
            if (!File.Exists(keystoreFilename))
            {
                LogMsg(OnErrorResponse, "Keystore file [" + keystoreFilename + "] does not exist");
                return false;
            }
            if (File.Exists(csrFilename))
                File.Delete(csrFilename);

            string arguments = CreateCSRRequestArguments(keystoreFilename, csrFilename, keystorePassword, alias);
            Process proc = new Process();
            string javaHome = GetJavaHome();
            if ((javaHome == null) || (javaHome.Length <= 0))
            {
                LogMsg(OnErrorResponse, "JAVA_HOME environment variable not set");
                return false;
            }

            proc.StartInfo.Arguments = arguments;
            proc.StartInfo.WorkingDirectory = javaHome + "\\bin"; //@"c:\Program Files\Java\jdk1.6.0_03\bin";
            proc.StartInfo.FileName = "keytool.exe";

            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.RedirectStandardError = true;
            proc.StartInfo.UseShellExecute = false;
            proc.Start();
            StreamReader reader = proc.StandardOutput;
            StreamReader errorReader = proc.StandardError;

            StringBuilder errorResponse = new StringBuilder();
            string line = reader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = reader.ReadLine();
            }
            line = errorReader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = errorReader.ReadLine();
            }
            if (errorResponse.Length > 0)
            {
                LogMsg(OnErrorResponse, errorResponse.ToString());
                return false;
            }
            if (File.Exists(csrFilename))
                return true;
            return false;
        }


        public static bool CreateKeystore(string filename, KeystoreParameters parameters, 
            LogMessageDelegate OnErrorResponse)
        {
            if (File.Exists(filename))
                File.Delete(filename);

            Process proc = new Process();


            string arguments = parameters.convertToGenkeyArguments(filename);

            //txtLog.AppendText(arguments.ToString() + Environment.NewLine);

            string javaHome = GetJavaHome();
            if ((javaHome == null) || (javaHome.Length <= 0))
            {
                LogMsg(OnErrorResponse, "JAVA_HOME environment variable not set");
                return false;
            }

            proc.StartInfo.Arguments = arguments;
            proc.StartInfo.WorkingDirectory = javaHome + "\\bin"; //@"c:\Program Files\Java\jdk1.6.0_03\bin";
            proc.StartInfo.FileName = "keytool.exe";

            proc.StartInfo.RedirectStandardOutput = true;
            proc.StartInfo.RedirectStandardError = true;
            proc.StartInfo.UseShellExecute = false;
            proc.Start();
            StreamReader reader = proc.StandardOutput;
            StreamReader errorReader = proc.StandardError;



            StringBuilder errorResponse = new StringBuilder();
            string line = reader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = reader.ReadLine();
            }


            line = errorReader.ReadLine();
            while (line != null)
            {
                errorResponse.AppendLine(line);
                line = errorReader.ReadLine();                
            }
            if (errorResponse.Length > 0)
            {
                LogMsg(OnErrorResponse, errorResponse.ToString());
                return false;
            }            
            if (File.Exists(filename))
                return true;
            return false;
        }

        private static string CreateImportCertificateResponseArguments(string keystoreFilename, string certificateResponseFilename,
            string keystorePassword, string alias)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("-import ");
            sb.Append("-alias ");
            sb.Append(alias);
            sb.Append(" -file \"");
            sb.Append(certificateResponseFilename);
            sb.Append("\" -keystore \"");
            sb.Append(keystoreFilename);
            sb.Append("\"");
            sb.Append(" -storepass ");
            sb.Append(keystorePassword);
            sb.Append(" -keypass ");
            sb.Append(keystorePassword);
            sb.Append(" -noprompt");
            return sb.ToString();
        }

        private static string CreateImportRootCertArguments(string keystoreFilename, string rootCertificateFilename, 
            string keystorePassword)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("-import ");
            sb.Append("-alias ");
            sb.Append("root");
            sb.Append(" -trustcacerts");
            sb.Append(" -file \"");
            sb.Append(rootCertificateFilename);
            sb.Append("\" -keystore \"");
            sb.Append(keystoreFilename);
            sb.Append("\"");
            sb.Append(" -storepass ");
            sb.Append(keystorePassword);
            sb.Append(" -keypass ");
            sb.Append(keystorePassword);
            sb.Append(" -noprompt");
            return sb.ToString();
        }

        private static string CreateCSRRequestArguments(string keystoreFilename, string csrFilename, 
            string keystorePassword, string alias)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("-certreq ");
            sb.Append("-keyalg RSA ");
            sb.Append("-alias ");
            sb.Append(alias);
            sb.Append(" -file \"");
            sb.Append(csrFilename);
            sb.Append("\" -keystore \"");
            sb.Append(keystoreFilename);
            sb.Append("\"");
            sb.Append(" -storepass ");
            sb.Append(keystorePassword);
            return sb.ToString();
        }

        private static void LogMsg(LogMessageDelegate OnErrorResponse, string msg)
        {
            if (OnErrorResponse != null)
                OnErrorResponse(msg);
        }
    }
}
