/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using VixClientV3Common;
using System.Windows.Forms;

namespace VixClientV3
{
    public class UserConfigurations
    {
        private readonly string mstrFilename;
        private string dicomViewerPath;
        private string mstrImageDirectory;
        private bool mbShowErrorsAsOutput;
        private readonly List<UserCredentials> userCredentials = new List<UserCredentials>();
        private readonly List<VixServerConfiguration> vixServerConfigurations = new List<VixServerConfiguration>();

        private UserCredentials lastUserCredentials = null;
        private VixServerConfiguration lastVixServerConfiguration = null;
        private string lastVixInterfaceName = null;

        public string LastVixInterfaceName
        {
            get { return lastVixInterfaceName; }
            set { lastVixInterfaceName = value; }
        }

        public bool ShowErrorsAsOutput
        {
            get { return mbShowErrorsAsOutput; }
            set { mbShowErrorsAsOutput = value; }
        }

        public VixServerConfiguration LastVixServerConfiguration
        {
            get { return lastVixServerConfiguration; }
            set { lastVixServerConfiguration = value; }
        }

        public UserCredentials LastUserCredentials
        {
            get { return lastUserCredentials; }
            set { lastUserCredentials = value; }
        }

        public List<VixServerConfiguration> VixServerConfigurations
        {
            get { return vixServerConfigurations; }
        } 

        public List<UserCredentials> UserCredentials
        {
            get { return userCredentials; }
        } 

        public string DicomViewerPath
        {
            get { return dicomViewerPath; }
            set { dicomViewerPath = value; }
        }

        public string ImageDirectory
        {
            get { return mstrImageDirectory; }
            set { mstrImageDirectory = value; }
        }

        public UserConfigurations(string filename)
        {
            mstrFilename = filename;
            LoadData();
        }

        private void LoadData()
        {
            if (!System.IO.File.Exists(mstrFilename))
            {
                PopulateDefault();
                Save();
            }
            else
            {
                // populate the default values in the event there is a value not stored in the config file
                PopulateDefault(); 
            }
            userCredentials.Clear();
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.Load(mstrFilename);

            XmlNode settingsNode = xmlDoc.SelectSingleNode("/VIXClientV3Configuration/Settings");
            dicomViewerPath = settingsNode.Attributes["DICOMViewerPath"].Value;
            if (settingsNode.Attributes["ShowErrorsAsOutput"] != null)
            {
                mbShowErrorsAsOutput = bool.Parse(settingsNode.Attributes["ShowErrorsAsOutput"].Value);
            }
            else
            {
                mbShowErrorsAsOutput = false;
            }
            XmlNodeList userCredentialsNodes = xmlDoc.SelectNodes("/VIXClientV3Configuration/UserCredentials/UserCredential");
            foreach (XmlNode userCredentialNode in userCredentialsNodes)
            {
                userCredentials.Add(LoadUserCredentials(userCredentialNode));
            }
            XmlNodeList vixServerConfigurationNodes = xmlDoc.SelectNodes("/VIXClientV3Configuration/VixServerConfigurations/VixServerConfiguration");
            foreach (XmlNode vixServerConfigurationNode in vixServerConfigurationNodes)
            {
                vixServerConfigurations.Add(LoadVixServerConfiguration(vixServerConfigurationNode));
            }

            XmlNode lastVixServerConfigurationNode = xmlDoc.SelectSingleNode("/VIXClientV3Configuration/LastVixClientConfiguration/VixServerConfiguration");
            if (lastVixServerConfigurationNode != null)
            {
                this.lastVixServerConfiguration = LoadVixServerConfiguration(lastVixServerConfigurationNode);
            }
            XmlNode lastUserCredentialsNode = xmlDoc.SelectSingleNode("/VIXClientV3Configuration/LastVixClientConfiguration/UserCredential");
            if (lastUserCredentialsNode != null)
            {
                this.lastUserCredentials = LoadUserCredentials(lastUserCredentialsNode);
            }
            XmlNode lastVixInterfaceNode = xmlDoc.SelectSingleNode("/VIXClientV3Configuration/LastVixClientConfiguration/VixInterface");
            if (lastVixInterfaceNode != null)
            {
                this.lastVixInterfaceName = lastVixInterfaceNode.InnerText;
            }
        }

        private UserCredentials LoadUserCredentials(XmlNode userCredentialNode)
        {
            string duz = "";
            string name = "";
            string siteName = "";
            string siteNumber = "";
            string ssn = "";
            string cacheLocationId = "";
            foreach (XmlAttribute attr in userCredentialNode.Attributes)
            {
                switch (attr.Name)
                {
                    case "duz":
                        duz = attr.Value;
                        break;
                    case "name":
                        name = attr.Value;
                        break;
                    case "siteName":
                        siteName = attr.Value;
                        break;
                    case "siteNumber":
                        siteNumber = attr.Value;
                        break;
                    case "ssn":
                        ssn = attr.Value;
                        break;
                    case "cacheLocationId":
                        cacheLocationId = attr.Value;
                        break;
                }
            }
            return new UserCredentials(duz, name, siteName, siteNumber, ssn, cacheLocationId, "");
        }

        private VixServerConfiguration LoadVixServerConfiguration(XmlNode vixServerConfigurationNode)
        {
            string server = "";
            int port = 0;
            string name = "";
            foreach (XmlAttribute attr in vixServerConfigurationNode.Attributes)
            {
                switch (attr.Name)
                {
                    case "Server":
                        server = attr.Value;
                        break;
                    case "Port":
                        if(attr.Value.Length > 0)
                            port = int.Parse(attr.Value);
                        break;
                    case "Name":
                        name = attr.Value;
                        break;
                }
            }
            return new VixServerConfiguration(server, port, name);
        }

        private void SaveVixServerConfiguration(VixServerConfiguration vixServerConfiguration, XmlNode parentNode)
        {
            XmlDocument xmlDoc = parentNode.OwnerDocument;
            XmlNode vixServerConfigurationNode = xmlDoc.CreateElement("VixServerConfiguration");
            parentNode.AppendChild(vixServerConfigurationNode);
            XmlAttribute attr = xmlDoc.CreateAttribute("Server");
            attr.Value = vixServerConfiguration.Server;
            vixServerConfigurationNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("Port");
            attr.Value = vixServerConfiguration.Port.ToString();
            vixServerConfigurationNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("Name");
            attr.Value = vixServerConfiguration.Name;
            vixServerConfigurationNode.Attributes.Append(attr);
        }

        private void SaveUserCredentials(UserCredentials userCredential, XmlNode parentNode)
        {
            XmlDocument xmlDoc = parentNode.OwnerDocument;
            XmlNode userCredentialNode = xmlDoc.CreateElement("UserCredential");
            parentNode.AppendChild(userCredentialNode);
            XmlAttribute attr = xmlDoc.CreateAttribute("duz");
            attr.Value = userCredential.Duz;
            userCredentialNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("name");
            attr.Value = userCredential.Name;
            userCredentialNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("siteName");
            attr.Value = userCredential.SiteName;
            userCredentialNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("siteNumber");
            attr.Value = userCredential.SiteNumber;
            userCredentialNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("ssn");
            attr.Value = userCredential.Ssn;
            userCredentialNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("cacheLocationId");
            attr.Value = userCredential.CacheLocationId;
            userCredentialNode.Attributes.Append(attr);
        }

        private void PopulateDefault()
        {
            dicomViewerPath = @"C:\Program Files\Vista\Imaging\MAG_DCMView.exe";
            mbShowErrorsAsOutput = false;
            System.IO.FileInfo file = new System.IO.FileInfo(Application.ExecutablePath);
            mstrImageDirectory = file.Directory.FullName + @"\images";
            if (!System.IO.Directory.Exists(mstrImageDirectory))
            {
                System.IO.Directory.CreateDirectory(mstrImageDirectory);
            }
        }

        public void Save()
        {
            XmlDocument xmlDoc = new XmlDocument();

            XmlNode rootNode = xmlDoc.CreateElement("VIXClientV3Configuration");
            xmlDoc.AppendChild(rootNode);
            XmlNode settingsNode = xmlDoc.CreateElement("Settings");
            rootNode.AppendChild(settingsNode);
            XmlAttribute attr = xmlDoc.CreateAttribute("DICOMViewerPath");
            attr.Value = dicomViewerPath;
            settingsNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("ShowErrorsAsOutput");
            attr.Value = mbShowErrorsAsOutput.ToString();
            settingsNode.Attributes.Append(attr);

            XmlNode userConfigurationsNode = xmlDoc.CreateElement("UserCredentials");
            rootNode.AppendChild(userConfigurationsNode);
            foreach (UserCredentials userCredential in userCredentials)
            {
                SaveUserCredentials(userCredential, userConfigurationsNode);
                /*
                XmlNode userCredentialNode = xmlDoc.CreateElement("UserCredential");
                userConfigurationsNode.AppendChild(userCredentialNode);
                attr = xmlDoc.CreateAttribute("duz");
                attr.Value = userCredential.Duz;
                userCredentialNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("name");
                attr.Value = userCredential.Name;
                userCredentialNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("siteName");
                attr.Value = userCredential.SiteName;
                userCredentialNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("siteNumber");
                attr.Value = userCredential.SiteNumber;
                userCredentialNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("ssn");
                attr.Value = userCredential.Ssn;
                userCredentialNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("cacheLocationId");
                attr.Value = userCredential.CacheLocationId;
                userCredentialNode.Attributes.Append(attr);*/
            }

            XmlNode vixServerConfigurationNodes = xmlDoc.CreateElement("VixServerConfigurations");
            rootNode.AppendChild(vixServerConfigurationNodes);
            foreach (VixServerConfiguration vixServerConfiguration in vixServerConfigurations)
            {
                SaveVixServerConfiguration(vixServerConfiguration, vixServerConfigurationNodes);
                /*
                XmlNode vixServerConfigurationNode = xmlDoc.CreateElement("VixServerConfiguration");
                vixServerConfigurationNodes.AppendChild(vixServerConfigurationNode);
                attr = xmlDoc.CreateAttribute("Server");
                attr.Value = vixServerConfiguration.Server;
                vixServerConfigurationNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("Port");
                attr.Value = vixServerConfiguration.Port.ToString();
                vixServerConfigurationNode.Attributes.Append(attr);
                attr = xmlDoc.CreateAttribute("Name");
                attr.Value = vixServerConfiguration.Name;
                vixServerConfigurationNode.Attributes.Append(attr);*/
            }

            XmlNode lastVixClientConfigurationNode = xmlDoc.CreateElement("LastVixClientConfiguration");
            rootNode.AppendChild(lastVixClientConfigurationNode);

            if (lastVixServerConfiguration != null)
            {                
                SaveVixServerConfiguration(lastVixServerConfiguration, lastVixClientConfigurationNode);
            }
            if (lastUserCredentials != null)
            {
                SaveUserCredentials(lastUserCredentials, lastVixClientConfigurationNode);
            }
            if (lastVixInterfaceName != null)
            {
                XmlNode lastVixInterfaceNameNode = xmlDoc.CreateElement("VixInterface");
                lastVixClientConfigurationNode.AppendChild(lastVixInterfaceNameNode);
                XmlText n = xmlDoc.CreateTextNode(lastVixInterfaceName);
                lastVixInterfaceNameNode.AppendChild(n);
            }

            xmlDoc.Save(mstrFilename);
        }
    }
}
