/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Reflection;
using VixClientV3Common;
using VixClientV3Common.configuration;
using System.IO;

namespace VixClientV3
{
    public partial class frmMain : Form
    {
        public frmMain()
        {
            InitializeComponent();
        }

        private MethodFactoriesConfiguration factoriesConfiguration = null;
        private MethodFactoryConfiguration methodFactoryConfiguration = null;
        private VixConfiguration configuration = null;
        private UserConfigurations userConfigurations = null;
        private UserCredentials user = null;
        private IMethodFactory methodFactory = null;
        private StreamWriter writer = null;

        private void frmMain_Load(object sender, EventArgs e)
        {
            string factoriesConfigFile = getAppDir() + "/factories.xml";
            factoriesConfiguration = new MethodFactoriesConfiguration(factoriesConfigFile);
            string userConfigurationsFile = getAppDir() + "/userConfigurations.xml";
            userConfigurations = new UserConfigurations(userConfigurationsFile);
            tmrStartup.Enabled = true;
            InitializeLog();
        }

        private void InitializeLog()
        {
            string logDirectory = getAppDir() + "/logs";
            if (!Directory.Exists(logDirectory))
            {
                Directory.CreateDirectory(logDirectory);
            }
            string logFilename = logDirectory + "/log.txt";
            while (writer == null)
            {
                try
                {
                    writer = new StreamWriter(logFilename, true);
                }
                catch (Exception)
                {
                    // log file already in use (probably another instance of the client), create a new log file with a unique name
                    logFilename = logDirectory + "/log_" + DateTime.Now.Ticks + ".txt";
                }
            }
            LogMsg("Initializing Log", false);
        }

        private void CloseLog()
        {
            try
            {
                LogMsg("Closing Log", false);
                if (writer != null)
                    writer.Close();
            }
            catch (Exception) { }
        }

        private void LoadAssemblies()
        {
            foreach (MethodFactoryConfiguration configuration in factoriesConfiguration.Configurations)
            {

            }
        }

        private string getAppDir()
        {
            return VixCommon.VixCommonUtils.getApplicationDir(Application.ExecutablePath);
            //System.IO.FileInfo file = new System.IO.FileInfo(Application.ExecutablePath);
            //return file.Directory.FullName;
        }

        private void tmrStartup_Tick(object sender, EventArgs e)
        {
            tmrStartup.Enabled = false;
            if (!ShowConfiguration())
            {
                Application.Exit();
            }
        }

        private bool ShowConfiguration()
        {
            frmConfig config = new frmConfig();
            if (config.ShowConfiguration(factoriesConfiguration,
                userConfigurations) == DialogResult.OK)
            {
                user = config.User;
                configuration = config.Configuration;
                methodFactoryConfiguration = config.MethodFactoryConfiguration;

                this.Text = "VIX Client V3 - " + methodFactoryConfiguration.Name + " (" + configuration.Server + ":" + configuration.Port + ")";
                LoadMethods();
                return true;
            }
            else
            {
                return false;
            }
        }

        private void LoadMethods()
        {
            methodFactory = MethodFactoryHelper.getMethodFactory(methodFactoryConfiguration.AssemblyFilename, 
                methodFactoryConfiguration.FactoryType);
            methodFactory.setLogMsg(LogMsg);
            List<MethodFactoryParameter> methodFactoryParameters = methodFactory.getMethodFactoryParameters();
            List<MethodFactoryParameter> userMethodFactoryParameters = null;
            if ((methodFactoryParameters != null) && (methodFactoryParameters.Count > 0))
            {
                frmVixInterfaceConfiguration interfaceConfiguration = new frmVixInterfaceConfiguration();
                if (interfaceConfiguration.ShowMethodFactoryConfiguration(methodFactory) == DialogResult.OK)
                {
                    userMethodFactoryParameters = interfaceConfiguration.MethodFactoryParameters;
                }
            }
            cboMethods.Items.AddRange(methodFactory.getMethods(userMethodFactoryParameters, user, 
                configuration, LogMsg).ToArray());
        }

        private void cboMethods_SelectedIndexChanged(object sender, EventArgs e)
        {
            IMethod method = GetSelectedMethod();
            if (method != null)
                DisplayMethod(method);
        }

        private IMethod GetSelectedMethod()
        {
            if (cboMethods.SelectedItem != null)
            {
                IMethod method = (IMethod)cboMethods.SelectedItem;
                return method;
            }
            return null;
        }

        private void DisplayMethod(IMethod method)
        {
            lstProperties.Items.Clear();
            Dictionary<string, MethodParameter> parameters = method.getParameters();
            if (parameters != null)
            {
                foreach (string key in parameters.Keys)
                {
                    ListViewItem item = new ListViewItem();
                    MethodParameter param = parameters[key];
                    item.Text = param.ParameterName;
                    item.SubItems.Add(param.ParameterValue);
                    item.SubItems.Add(param.Required.ToString());
                    item.SubItems.Add(param.Saveable.ToString());
                    lstProperties.Items.Add(item);
                    item.ToolTipText = param.ParameterDescription;
                    item.Tag = param;
                }
            }
            DisplayMethodStoredParameters(method);
        }

        private void DisplayMethodStoredParameters(IMethod method)
        {
            cboSavedParameters.Items.Clear();
            List<StoredMethodParameters> parameters = methodFactory.getStoredMethodParameters(method.GetType().Name);
            cboSavedParameters.Items.AddRange(parameters.ToArray());
        }

        private void UpdateProperties(Dictionary<string, StoredMethodParameterValue> updatedParameters)
        {
            foreach (string key in updatedParameters.Keys)
            {
                StoredMethodParameterValue param = updatedParameters[key];

                foreach (ListViewItem item in lstProperties.Items)
                {
                    if (item.Text == param.ParameterName)
                    {
                        MethodParameter mp = (MethodParameter)item.Tag;
                        if (mp is AutogeneratingMethodParameter)
                        {
                            // if auto generating, don't update the value
                        }
                        else
                        {
                            mp.ParameterValue = param.ParameterValue;
                            item.SubItems[1].Text = param.ParameterValue;
                        }
                        break;
                    }
                }
            }
        }

        private void lstProperties_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (lstProperties.SelectedItems.Count > 0)
            {
                ListViewItem item = lstProperties.SelectedItems[0];
                MethodParameter parameter = (MethodParameter)item.Tag;
                txtNewValue.Text = parameter.ParameterValue;
                if (parameter.ParameterType == MethodParameterType.saveFileType)
                {
                    saveFileDialog1.FileName = parameter.ParameterValue;
                    if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                    {
                        txtNewValue.Text = saveFileDialog1.FileName;
                        item.SubItems[1].Text = saveFileDialog1.FileName;
                        parameter.ParameterValue = saveFileDialog1.FileName;
                    }
                }
                else if (parameter.ParameterType == MethodParameterType.openFileType)
                {
                    MessageBox.Show("open file type not implemented yet");
                }
                // Arianna asked for this to be disabled and for the user to manually type in the date
                // since most of our data is several years (decades) old it was difficult to adjust the date
                // ranges
                    /*
                else if (parameter.ParameterType == MethodParameterType.dateType)
                {
                    frmSelectDate selectDate = new frmSelectDate();
                    selectDate.Left = this.Left + splitContainer1.Left + lstProperties.Left + lstProperties.Width;
                    selectDate.Top = this.Top + splitContainer1.Top + lstProperties.Top + item.Bounds.Top;
                    if (selectDate.SelectDate(parameter.ParameterValue) == DialogResult.OK)
                    {
                        item.SubItems[1].Text = selectDate.SelectedDate;
                        parameter.ParameterValue = selectDate.SelectedDate;
                    }
                }*/

            }
        }

        private void btnSetValue_Click(object sender, EventArgs e)
        {
            if (lstProperties.SelectedItems.Count > 0)
            {
                ListViewItem item = lstProperties.SelectedItems[0];
                MethodParameter parameter = (MethodParameter)item.Tag;
                parameter.ParameterValue = txtNewValue.Text;
                item.SubItems[1].Text = parameter.ParameterValue;
            }
        }

        private Dictionary<string, MethodParameter> getEnteredParameters()
        {
            Dictionary<string, MethodParameter> parameters = new Dictionary<string, MethodParameter>();
            foreach (ListViewItem item in lstProperties.Items)
            {
                MethodParameter parameter = (MethodParameter)item.Tag;
                // JMW 2/2/2011 - create a new MethodParameter object to return so it is not associated with the 
                // object currently in the list. So when the transaction ID is automatically updated, it doesn't
                // impact the parameter returned by this method
                MethodParameter param = new MethodParameter(parameter.ParameterName, parameter.Required, 
                    parameter.ParameterValue, parameter.ParameterType, parameter.Saveable, 
                    parameter.ParameterDescription);
                param.ParameterValue = item.SubItems[1].Text;
                parameters.Add(param.ParameterName, param);
                //parameter.ParameterValue = item.SubItems[1].Text;
                //parameters.Add(parameter.ParameterName, parameter);

            }
            return parameters;
        }

        private void StatusMsg(string msg)
        {
            toolStripStatusLabel1.Text = msg;
            Application.DoEvents();
        }

        private void ShowSearchForSelectedOutput()
        {
            TabPage tp = tabControl1.SelectedTab;
            if (tp != null)
            {
                foreach (Control c in tp.Controls)
                {
                    if (c is MethodResultDisplay)
                    {
                        MethodResultDisplay mrd = (MethodResultDisplay)c;
                        mrd.ShowSearch();
                        break;
                    }
                }
            }
        }

        private int methodCount = 0;

        private MethodResult ExecuteMethod()
        {
            IMethod method = GetSelectedMethod();
            if (method != null)
            {
                methodCount++;
                try
                {
                    Cursor.Current = Cursors.WaitCursor;
                    Dictionary<string, MethodParameter> parameters = getEnteredParameters();
                    if (method.validateParameters(parameters))
                    {
                        
                        StatusMsg("Executing method '" + method.ToString() + "'.");
                        DateTime requestDate = DateTime.Now;
                        object result = null;
                        try
                        {
                            result = method.execute(parameters);
                        }
                        catch (Exception ex)
                        {
                            if (userConfigurations.ShowErrorsAsOutput)
                            {
                                result = ex;
                            }
                            else
                            {
                                throw ex;
                            }
                        }
                        StatusMsg("Received result from method");
                        UpdateUpdatableParameters();
                        return new MethodResult(result, method, parameters, requestDate);
                    }
                    else
                    {
                        MessageBox.Show("The input parameters are not valid. Might indicate missing required field");
                    }
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message + "\n\n" + ex.StackTrace);
                    UpdateUpdatableParameters();
                }
                finally
                {
                    Cursor.Current = Cursors.Default;
                }
            }
            return null;
        }

        private void btnExecute_Click(object sender, EventArgs e)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                MethodResult result = ExecuteMethod();
                if (result != null)
                {
                    StatusMsg("Received result from method, displaying...");

                    TabPage tp = new TabPage(methodCount + ": " + result.Method.ToString());
                    tabControl1.TabPages.Add(tp);
                    MethodResultDisplay mrd = new MethodResultDisplay();
                    mrd.OnTypeSelectedHandlerEvent += TypeSelectedHandlerDelegate;
                    mrd.OnSetParametersEvent += SetMethodParameters;

                    mrd.Parent = tp;
                    mrd.Dock = DockStyle.Fill;
                    mrd.DisplayResults(result.Parameters, result.Method, result.Result, result.RequestDate);
                    tabControl1.SelectedTab = tp;

                    UpdateUpdatableParameters();
                    StatusMsg("Result Displayed.");
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message + "\n\n" + ex.StackTrace);
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void SetMethodParameters(IMethod method, Dictionary<string, MethodParameter> parameters)
        {
            cboMethods.SelectedItem = method;
            Dictionary<string, StoredMethodParameterValue> values = new Dictionary<string,StoredMethodParameterValue>();
            foreach (string key in parameters.Keys)
            {
                MethodParameter methodParameter = parameters[key];
                values.Add(key, new StoredMethodParameterValue(methodParameter.ParameterName, 
                    methodParameter.ParameterValue));
            }
            UpdateProperties(values);
        }

        private List<VixCommon.TypeSelectedHandler> convertAssociatedMethodsToTypeSelectionHandler(List<AssociatedMethod> associatedMethods)
        {
            List<VixCommon.TypeSelectedHandler> handlers = new List<VixCommon.TypeSelectedHandler>();
            foreach (AssociatedMethod associatedMethod in associatedMethods)
            {
                handlers.Add(new VixCommon.TypeSelectedHandler(associatedMethod.MatchingType, 
                    associatedMethod.MethodToCall.Name, associatedMethod.MethodToCall, TypeSelectedHandlerDelegate));
            }
            return handlers;
        }

        private void TypeSelectedHandlerDelegate(object value, Type methodToCallType)
        {
            /*
             * if (cboMethods.SelectedItem != null)
            {
                IMethod method = (IMethod)cboMethods.SelectedItem;
                return method;
            }
             * */
            for (int i = 0; i < cboMethods.Items.Count; i++)
            {
                IMethod method = (IMethod)cboMethods.Items[i];
                if (method.GetType() == methodToCallType)
                {
                    Dictionary<string, StoredMethodParameterValue> updatedParameters = method.getParametersFromObject(value);
                    cboMethods.SelectedIndex = i;
                    UpdateProperties(updatedParameters);
                    break;
                    // need to call with 'value' to set some properties
                    //DisplayMethod(method);
                }
            }
        }

        private void UpdateUpdatableParameters()
        {
            foreach (ListViewItem item in lstProperties.Items)
            {
                MethodParameter parameter = (MethodParameter)item.Tag;
                if (parameter is AutogeneratingMethodParameter)
                {
                    AutogeneratingMethodParameter autoParam = (AutogeneratingMethodParameter)parameter;
                    autoParam.updateValue();
                    item.SubItems[1].Text = parameter.ParameterValue;
                }
            }
        }

        private void removeTabToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (tabControl1.SelectedTab != null)
            {
                tabControl1.TabPages.Remove(tabControl1.SelectedTab);
            }
        }

        private void clearTabsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            tabControl1.TabPages.Clear();
        }

        private void pasteFromClipboardToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lstProperties.SelectedItems.Count > 0)
            {
                MethodParameter parameter = (MethodParameter)lstProperties.SelectedItems[0].Tag;
                parameter.ParameterValue = Clipboard.GetText();
                lstProperties.SelectedItems[0].SubItems[1].Text = parameter.ParameterValue;
            }
        }

        private void btnSaveParameters_Click(object sender, EventArgs e)
        {
            frmSaveMethodParameters saveMethodParameters = new frmSaveMethodParameters();
            Dictionary<string, MethodParameter> methodParameters = getEnteredParameters();
            IMethod method = GetSelectedMethod();
            if (saveMethodParameters.ShowSaveMethodParameters(method, methodParameters) == DialogResult.OK)
            {
                StoredMethodParameters storedMethodParameter = saveMethodParameters.StoredMethodParameter;
                methodFactory.addStoredMethodParameter(storedMethodParameter);
                cboSavedParameters.Items.Add(storedMethodParameter);
            }
        }

        private void cboSavedParameters_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (cboSavedParameters.SelectedItem != null)
            {
                StoredMethodParameters storedMethodParameters = (StoredMethodParameters)cboSavedParameters.SelectedItem;
                UpdateProperties(storedMethodParameters.StoredParametersDictionary);
            }
        }

        private void frmMain_FormClosing(object sender, FormClosingEventArgs e)
        {
            try
            {
                if (System.IO.Directory.Exists(userConfigurations.ImageDirectory))
                {
                    System.IO.Directory.Delete(userConfigurations.ImageDirectory, true);
                }
            }
            catch (Exception) { }
            CloseLog();
        }

        private void lstProperties_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            if (lstProperties.SelectedItems.Count > 0)
            {
                MethodParameter parameter = (MethodParameter)lstProperties.SelectedItems[0].Tag;

                frmInputDialog input = new frmInputDialog();
                if (input.ShowInputDialog("Parameter - " + parameter.ParameterName, 
                    "New value for parameter '" + parameter.ParameterName + "'.", parameter.ParameterValue) == DialogResult.OK)
                {
                    parameter.ParameterValue = input.Value;
                    lstProperties.SelectedItems[0].SubItems[1].Text = parameter.ParameterValue;
                }                
            }
        }

        private void clearToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (lstProperties.SelectedItems.Count > 0)
            {
                MethodParameter parameter = (MethodParameter)lstProperties.SelectedItems[0].Tag;
                parameter.ParameterValue = "";
                lstProperties.SelectedItems[0].SubItems[1].Text = parameter.ParameterValue;
            }
        }

        private void LogMsg(string msg, bool isError)
        {
            if (writer != null)
            {
                //TODO: do something with an error message
                writer.WriteLine(DateTime.Now.ToString() + ": " + msg);
                writer.Flush();
            }
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            StringBuilder msg = new StringBuilder();
            msg.AppendLine("VIX Client V3");
            msg.AppendLine(Application.ProductVersion);
            MessageBox.Show(msg.ToString(), "About", MessageBoxButtons.OK, MessageBoxIcon.Information);
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void findToolStripMenuItem_Click(object sender, EventArgs e)
        {
            ShowSearchForSelectedOutput();
        }

        private void btnExecuteWithoutDisplaying_Click(object sender, EventArgs e)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                MethodResult result = ExecuteMethod();
                if (result != null)
                {
                    MessageBox.Show("Got result");
                }
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void tabControl1_MouseClick(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Right)
            {
                for(int i = 0; i < tabControl1.TabPages.Count; i++)
                {
                    if (tabControl1.GetTabRect(i).Contains(e.X, e.Y))
                    {
                        tabControl1.SelectedTab = tabControl1.TabPages[i];
                        break;
                    }
                }
            }
        }
    }

    class MethodResult
    {
        private readonly object mResult;
        private readonly IMethod method;
        private readonly Dictionary<string, MethodParameter> mParameters;
        private readonly DateTime mRequestDate;

        public object Result
        {
            get { return mResult; }
        }

        public IMethod Method
        {
            get { return method; }
        } 
       
        public Dictionary<string, MethodParameter> Parameters
        {
            get { return mParameters; }
        } 

        public DateTime RequestDate
        {
            get { return mRequestDate; }
        } 

        public MethodResult(object result, IMethod method, Dictionary<string, MethodParameter> Parameters,
            DateTime requestDate)
        {
            mRequestDate = requestDate;
            mResult = result;
            this.method = method;
            mParameters = Parameters;
        }
    }
}
