/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Net;

namespace VixClientV3Common
{
    public abstract class AbstractBaseGetImageMethod : AbstractBaseMethod, IMethod
    {
        private readonly bool useDuz;

        protected readonly string parameterQuality = "quality";
        protected readonly string parameterContentType = "contentType";
        protected readonly string parameterFilename = "filename";
        protected readonly string parameterOpenImageEXE = "openImageEXE";
        protected readonly string parameterImageUrn = "imageUrn";

        protected abstract string getDefaultOpenFileExecutable();
        //protected abstract string getExtraUrlParameters(Dictionary<string, MethodParameter> values);
        /// <summary>
        /// Return the full URL to get the image requesting
        /// </summary>
        /// <param name="values"></param>
        /// <returns></returns>
        protected abstract string getImageRequestUrl(Dictionary<string, MethodParameter> values);
        /// <summary>
        /// Add new parameters or override default values for existing parameters
        /// </summary>
        /// <param name="values"></param>
        protected abstract void addOrUpdateParameters(Dictionary<string, MethodParameter> values);
        protected abstract ICredentials getNetworkCredentials(Dictionary<string, MethodParameter> values);

        public AbstractBaseGetImageMethod(bool useDuz, UserCredentials userCredentials, VixConfiguration vixConfiguration)
            : base(userCredentials, vixConfiguration)
        {
            this.useDuz = useDuz;
        }

        protected abstract Dictionary<string, string> getExtraRequestHeaders(Dictionary<string, MethodParameter> values);

        public object execute(Dictionary<string, MethodParameter> values)
        {
            WebClient vix = new WebClient();
            vix.Credentials = getNetworkCredentials(values);

            UserCredentials credentials = GetUserCredentialsFromParameters(values);

            vix.Headers.Add("xxx-fullname", credentials.Name);
            vix.Headers.Add("xxx-sitename", credentials.SiteName);
            vix.Headers.Add("xxx-sitenumber", credentials.SiteNumber);
            vix.Headers.Add("xxx-ssn", credentials.Ssn);
            if (credentials.Duz != null)
            {
                vix.Headers.Add("xxx-duz", credentials.Duz);
            }
            vix.Headers.Add("xxx-cacheLocationId", credentials.CacheLocationId);
            vix.Headers.Add("xxx-transaction-id", values[parameterTransactionId].ParameterValue);
            vix.Headers.Add("xxx-securityToken", credentials.BseToken);

            Dictionary<string, string> extraRequestParameters = getExtraRequestHeaders(values);
            if (extraRequestParameters != null)
            {
                foreach (string key in extraRequestParameters.Keys)
                {
                    vix.Headers.Add(key, extraRequestParameters[key]);
                }
            }

            string filename = values[parameterFilename].ParameterValue;

            string openImageExe = null;
            if (values.ContainsKey(parameterOpenImageEXE))
            {
                openImageExe = values[parameterOpenImageEXE].ParameterValue;
            }

            string url = getImageRequestUrl(values);
            string outputFile = filename;

            /*
            Stream stream = vix.OpenRead(new Uri(url));
            byte[] buffer = new byte[1024];
            int bytesRead = stream.Read(buffer, 0, 1024);
            stream.Close();
            */

            byte[] buffer = vix.DownloadData(new Uri(url));
            BinaryWriter output = new BinaryWriter(new FileStream(outputFile, FileMode.Create));
            output.Write(buffer);
            output.Close();

            

            if (!string.IsNullOrEmpty(openImageExe))
            {
                System.Diagnostics.Process proc = new System.Diagnostics.Process();
                proc.StartInfo.FileName = openImageExe;
                proc.StartInfo.Arguments = "\"" + filename + "\"";
                proc.Start();
            }
            //return filename;
            return new ImageRequestResponse(filename, vix.ResponseHeaders);
        }

        public Dictionary<string, MethodParameter> getParameters()
        {
            Dictionary<string, MethodParameter> parameters = getCredentialsParameters(useDuz);// new Dictionary<string, MethodParameter>();

            parameters.Add(parameterTransactionId, GetTransactionIdParameter());
            parameters.Add(parameterImageUrn, new MethodParameter(parameterImageUrn, true, "", true));
            parameters.Add(parameterQuality, new MethodParameter(parameterQuality, true, "70", true));
            parameters.Add(parameterContentType, new MethodParameter(parameterContentType, true, "application/dicom,image/x-targa", true));
            //parameters.Add(parameterFilename, new MethodParameter(parameterFilename, true, "", MethodParameterType.saveFileType));
            //parameters.Add(parameterFilename, new MethodParameter(parameterFilename, true, vixConfiguration.ImageDirectory +  "\\" + DateTime.Now.Ticks, MethodParameterType.saveFileType));
            parameters.Add(parameterFilename, GetFilenameParameter(parameterFilename));
            parameters.Add(parameterOpenImageEXE, new MethodParameter(parameterOpenImageEXE,
                false, getDefaultOpenFileExecutable(), MethodParameterType.openFileType));
            addOrUpdateParameters(parameters);
            return parameters;
        }

        public bool validateParameters(Dictionary<string, MethodParameter> values)
        {
            if (values == null)
                return false;
            if (!ValidateRequiredParameters(values))
                return false;
            if (!this.ValidateCredentials(values))
                return false;
            /*
            if (!CheckParameterNotEmptyOrNull(values, parameterImageUrn))
                return false;
            if (!CheckParameterNotEmptyOrNull(values, parameterTransactionId))
                return false;
            if (!CheckParameterNotEmptyOrNull(values, parameterQuality))
                return false;
            if (!CheckParameterNotEmptyOrNull(values, parameterContentType))
                return false;
            if (!CheckParameterNotEmptyOrNull(values, parameterFilename))
                return false;
             * */
            return true;
        }

        public List<AssociatedMethod> getAssociatedMethods()
        {
            return null;
        }

        public virtual Dictionary<string, StoredMethodParameterValue> getParametersFromObject(object obj)
        {
            return new Dictionary<string, StoredMethodParameterValue>();
        }
    }

    public class ImageRequestResponse
    {
        private readonly string mstrFilename;
        private readonly WebHeaderCollection responseHeaders;
        private readonly long fileSize;

        public long FileSize
        {
            get { return fileSize; }
        }

        public string [] ResponseHeaders
        {
            get 
            {
                string[] response = new string[responseHeaders.Count];
                for (int i = 0; i < responseHeaders.Count; i++)
                {
                    response[i] = responseHeaders.Keys[i] + ":" + responseHeaders[i];
                }
                return response;
            }
        } 

        public string Filename
        {
            get { return mstrFilename; }
        }

        public ImageRequestResponse(string filename, WebHeaderCollection responseHeaders)
        {
            this.mstrFilename = filename;
            this.responseHeaders = responseHeaders;
            FileInfo file = new FileInfo(filename);
            this.fileSize = file.Length;
        }

    }
}
