/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;

namespace VixClientV3Common
{
    public abstract class AbstractBaseMethod
    {
        protected UserCredentials userCredentials = null;
        protected VixConfiguration vixConfiguration = null;

        protected readonly string parameterTransactionId = "transactionId";
        protected event LogMsgDelegate OnLogMsgEvent = null;

        public AbstractBaseMethod(UserCredentials userCredentials, VixConfiguration vixConfiguration)
            : this(userCredentials, vixConfiguration, null)
        {

        }

        public AbstractBaseMethod(UserCredentials userCredentials, VixConfiguration vixConfiguration,
            LogMsgDelegate logMsgDelegate)
        {
            this.userCredentials = userCredentials;
            this.vixConfiguration = vixConfiguration;
            this.OnLogMsgEvent = logMsgDelegate;
        }

        /// <summary>
        /// Creates a AutogeneratingMethodParameter that will automatically create random transaction IDs
        /// after every operation
        /// </summary>
        /// <returns></returns>
        protected MethodParameter GetTransactionIdParameter()
        {
            AutogeneratingMethodParameter param = new AutogeneratingMethodParameter(parameterTransactionId,
                true, GetTransactionId(), MethodParameterType.stringType, GetTransactionId);
            return param;
        }

        private string GetTransactionId()
        {
            return Guid.NewGuid().ToString();
        }

        /// <summary>
        /// Creates a AutogeneratingMethodParameter that will automatically create random filenames
        /// after every image request
        /// </summary>
        /// <param name="parameterName">The name of the parameter being stored</param>
        /// <returns></returns>
        protected MethodParameter GetFilenameParameter(string parameterName)
        {
            AutogeneratingMethodParameter param = new AutogeneratingMethodParameter(parameterName, 
                true, GetNewFilename(), MethodParameterType.saveFileType, GetNewFilename);
            return param;
        }

        protected string GetNewFilename()
        {
            return vixConfiguration.ImageDirectory + "\\" + DateTime.Now.Ticks;
        }

        protected bool ValidateCredentials(Dictionary<string, MethodParameter> values)
        {
            if (values == null)
                return false;
            if (!values.ContainsKey("credentials_fullName"))
                return false;

            // check others

            return true;
        }

        protected bool CheckParameterExists(Dictionary<string, MethodParameter> values, string parameter)
        {
            return values.ContainsKey(parameter);
        }

        protected bool CheckParameterNotEmptyOrNull(Dictionary<string, MethodParameter> values, string parameter)
        {
            if (!values.ContainsKey(parameter))
                return false;
            MethodParameter param = values[parameter];
            return !string.IsNullOrEmpty(param.ParameterValue);
        }

        protected bool ValidateRequiredParameters(Dictionary<string, MethodParameter> values)
        {
            foreach (string key in values.Keys)
            {
                MethodParameter param = values[key];
                if (param.Required)
                {
                    if (string.IsNullOrEmpty(param.ParameterValue))
                        return false;
                }
            }
            return true;
        }

        protected Dictionary<string, MethodParameter> getCredentialsParameters()
        {
            return getCredentialsParameters(true);
        }      

        protected Dictionary<string, MethodParameter> getCredentialsParameters(bool includeDuz)
        {
            Dictionary<string, MethodParameter> parameters = new Dictionary<string, MethodParameter>();
            parameters.Add("credentials_fullName", new MethodParameter("credentials_fullName", true, userCredentials.Name));
            if (includeDuz)
            {
                parameters.Add("credentials_duz", new MethodParameter("credentials_duz", true, userCredentials.Duz));
            }
            parameters.Add("credentials_ssn", new MethodParameter("credentials_ssn", true, userCredentials.Ssn));
            parameters.Add("credentials_siteName", new MethodParameter("credentials_siteName", true, userCredentials.SiteName));
            parameters.Add("credentials_siteNumber", new MethodParameter("credentials_siteNumber", true, userCredentials.SiteNumber));
            parameters.Add("credentials_securityToken", new MethodParameter("credentials_securityToken", false, userCredentials.BseToken));
            parameters.Add("credentials_cacheLocationId", new MethodParameter("credentials_cacheLocationId", false, userCredentials.CacheLocationId));

            return parameters;
        }

        protected UserCredentials GetUserCredentialsFromParameters(Dictionary<string, MethodParameter> values)
        {
            string fullname = values["credentials_fullName"].ParameterValue;
            string duz = null;
            if (values.ContainsKey("credentials_duz"))
            {
                duz = values["credentials_duz"].ParameterValue;
            }
            string ssn = values["credentials_ssn"].ParameterValue;
            string siteName = values["credentials_siteName"].ParameterValue;
            string siteNumber = values["credentials_siteNumber"].ParameterValue;
            string securityToken = values["credentials_securityToken"].ParameterValue;
            string cacheLocationId = values["credentials_cacheLocationId"].ParameterValue;

            return new UserCredentials(duz, fullname, siteName, siteNumber, ssn, cacheLocationId, securityToken);
        }

        public override string ToString()
        {
            return this.GetType().Name;
        }

        /// <summary>
        /// This method returns the string enclosed in the begining and ending delimiter.  It will always find the first instance 
        /// of each delimiter
        /// </summary>
        /// <param name="str">Input string to search, cannot be null</param>
        /// <param name="beginingDelimiter">The start delimiter or null if it should start from the begining</param>
        /// <param name="endingDelimiter">The end delimiter or null if the result should go to the end</param>
        /// <returns>The string enclosed in the delimiters or null if the piece cannot be found</returns>
        protected string extractPiece(string str, string beginingDelimiter, string endingDelimiter)
        {
            if (str == null)
                return null;
            int loc = 0;
            if(beginingDelimiter != null)
                loc = str.IndexOf(beginingDelimiter);
            if (loc < 0)
                return null;
            int endLoc = str.Length;// -1;

            if (endingDelimiter != null)
            {
                endLoc = endLoc - 1;    
                endLoc = str.IndexOf(endingDelimiter, loc + beginingDelimiter.Length);
            }
            if (endLoc < 0)
                return null;

            return str.Substring(loc + beginingDelimiter.Length, endLoc - loc - beginingDelimiter.Length);
        }

        public void setLogMsg(LogMsgDelegate logMsgDelegate)
        {
            OnLogMsgEvent = logMsgDelegate;
        }

        protected void LogMsg(string msg, bool isError)
        {
            if (OnLogMsgEvent != null)
            {
                OnLogMsgEvent(msg, isError);
            }
        }

        protected void LogMsg(string msg)
        {
            if (OnLogMsgEvent != null)
            {
                OnLogMsgEvent(msg, false);
            }
        }

        public string getWorkstationHostname()
        {
            return System.Environment.MachineName;
        }
    }
}
