/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using VixClientV3Common;
using VixClientV3_XCA.types;
using System.Diagnostics;
using System.Xml;
using System.Configuration;
using System.Reflection;

namespace VixClientV3_XCA
{
    public class XcaService
    {
        private readonly string mstrServerName;
        private readonly int mintServerPort;
        private LogMsgDelegate logMsgDelegate;

        private const string defaultProtocol = "http";

        public XcaService(string serverName, int serverPort, LogMsgDelegate logMsgDelegate)
        {
            mstrServerName = serverName;
            mintServerPort = serverPort;
            this.logMsgDelegate = logMsgDelegate;
            ServicePointManager.Expect100Continue = false;
            ServicePointManager.ServerCertificateValidationCallback += IgnoreCertificateErrorHandler;
        }

        private void LogMsg(string msg, bool error)
        {
            if (logMsgDelegate != null)
            {
                logMsgDelegate(msg, error);
            }
        }

        private string createUrl(string uri)
        {
            return createUrl(defaultProtocol, uri, mstrServerName, mintServerPort);
        }

        private string createUrl(string protocol, string uri, string host, int port)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(protocol);
            sb.Append("://");
            sb.Append(host);
            sb.Append(":");
            sb.Append(port);
            sb.Append("/");
            sb.Append(uri);

            return sb.ToString();
        }

        public XcaRetrieveResult ExecuteRetrieve(string retrieveUri, string homeCommunityId, string repositoryId,
            string documentUniqueId, string transactionId, UserCredentials userCredentials, 
            string filename, string openImageExe, string host, int port, string protocol, string username, 
            string password, bool debug)
        {
            //System.Net.ServicePointManager.Expect100Continue = false;

            string url = createUrl(protocol, retrieveUri, host, port);
            WebRequest request = WebRequest.Create(url);// "http://zzzzzzfhietis:8080/XCARespondingGateway/services/RespondingGateway_Service");
            request.Method = "POST";
            request.ContentType = "application/soap+xml; charset=UTF-8; action=\"urn:ihe:iti:2007:CrossGatewayRetrieve\"";

            request.Headers.Add("xxx-sitename", userCredentials.SiteName);
            request.Headers.Add("xxx-fullname", userCredentials.Name);
            request.Headers.Add("xxx-sitenumber", userCredentials.SiteNumber);
            request.Headers.Add("xxx-ssn", userCredentials.Ssn);
            request.Headers.Add("xxx-purpose-of-use", "Routine medical care");
            request.Headers.Add("xxx-transaction-id", transactionId);
            request.Headers.Add("SOAPAction", "urn:ihe:iti:2007:CrossGatewayRetrieve");
            //request.PreAuthenticate = true;

            //ServicePointManager.ServerCertificateValidationCallback += IgnoreCertificateErrorHandler;

            if ((!string.IsNullOrEmpty(username)) && (!string.IsNullOrEmpty(password)))
            {
                request.Credentials = new NetworkCredential(username, password);
            }

            byte[] byteArray = Encoding.UTF8.GetBytes(createRetrieveRequest(homeCommunityId, repositoryId, 
                documentUniqueId, url));

            LogMsg("Executing retrieve to url '" + url + "'.", false);

            Stream requestStream = request.GetRequestStream();
            requestStream.Write(byteArray, 0, byteArray.Length);
            requestStream.Close();

            WebResponse response = null;
            try
            {
                response = request.GetResponse();
            }
            catch (WebException webEx)
            {
                LogMsg("WebException, " + webEx.Message, true);
                // got something other than a 200 status code but we still want to parse the response

                // i think if an XCA error code comes back, the status code will still be a 200
                // so if not 200, then throw exception
                if (webEx.Status != WebExceptionStatus.Success)
                    throw webEx;

                response = webEx.Response;
            }
            if (response == null)
                throw new Exception("Null response from server");
            

            List<ResponseHeader> responseHeaders = new List<ResponseHeader>();
            foreach (string key in response.Headers.Keys)
            {
                responseHeaders.Add(new ResponseHeader(key, response.Headers[key]));
            }
            string boundary = "";
            string contentType = response.Headers["Content-Type"];
            if (contentType != null)
            {
                int loc = contentType.IndexOf("boundary=");
                if (loc >= 0)
                {
                    boundary = contentType.Substring(loc + 9);
                    loc = boundary.IndexOf(";");
                    if (loc >= 0)
                    {
                        boundary = boundary.Substring(0, loc);
                    }
                }
                if (boundary.StartsWith("\""))
                {
                    boundary = boundary.Substring(1);
                }
                if (boundary.EndsWith("\""))
                {
                    boundary = boundary.Substring(0, boundary.Length - 1);
                }
            }
            LogMsg("Boundary in response is '" + boundary + "'.", false);
            Stream responseStream = response.GetResponseStream();
            StringBuilder line = new StringBuilder();
            
            StringBuilder fullResponse = new StringBuilder();
            MemoryStream binaryStream = new MemoryStream();
            MemoryStream tempBinaryStream = new MemoryStream();
            bool inBinarySection = false;
            bool readingBinaryData = false;
            bool inSoapSection = false;
            bool readingSoapSection = false;
            StringBuilder xml = null;

            MemoryStream imageStream = new MemoryStream();
            BinaryWriter bw =   null;
            if (debug)
                bw = new BinaryWriter(new FileStream(@"C:\\temp\\out.txt", FileMode.Create));
            byte[] buffer = new byte[1024];
            int bytesRead = responseStream.Read(buffer, 0, 1024);
            while (bytesRead > 0)
            {
                imageStream.Write(buffer, 0, bytesRead);
                if(bw != null)
                    bw.Write(buffer, 0, bytesRead);
                bytesRead = responseStream.Read(buffer, 0, 1024);
            }
            if(bw != null)
                bw.Close();

            imageStream.Seek(0, SeekOrigin.Begin);

            //int b = responseStream.ReadByte();
            int b = imageStream.ReadByte();
            string l = "";
            while (b > -1)
            {
                line.Append((char)b);
                if (readingBinaryData)
                {
                    tempBinaryStream.WriteByte((byte)b);
                }
                if (b == 10) // LF
                {                   
                    l = line.ToString();
                    if (readingBinaryData)
                    {
                        if (l.Contains(boundary))
                        {
                            readingBinaryData = false;
                        }
                        else
                        {
                            binaryStream.Write(tempBinaryStream.ToArray(), 0, (int)tempBinaryStream.Length);
                            tempBinaryStream = new MemoryStream();
                        }
                    }
                    else if (readingSoapSection)
                    {
                        if (l.Contains(boundary))
                        {
                            readingSoapSection = false;
                            inSoapSection = false;
                        }
                        else
                        {
                            xml.AppendLine(l);                            
                        }
                    }
                    else
                    {
                        fullResponse.Append(l);
                        if (l.StartsWith("Content-Type:"))
                        {
                            if (l.Contains("soap+xml"))
                            {
                                inSoapSection = true;
                                xml = new StringBuilder();
                            }
                            else
                            {
                                inBinarySection = true;
                            }
                        }
                        else if (inSoapSection)
                        {
                            if (l == "\r\n")
                            {
                                readingSoapSection = true;
                            }
                        }
                        else if (inBinarySection)
                        {
                            if (l == "\r\n")
                            {
                                readingBinaryData = true;
                            }
                        }
                    }
                    line = new StringBuilder();
                }
                //b = responseStream.ReadByte();
                b = imageStream.ReadByte();
            }

            // when we get here, there is no more data in the imageStream however we might still need to process
            // the most recent line

            l = line.ToString();
            if (readingBinaryData)
            {
                if (l.Contains(boundary))
                {
                    readingBinaryData = false;
                }
                else
                {
                    if (l == "\r\n")
                    {
                        // JMW 2/21/2011 - not entirely sure about this but I think if we get here and the only data left is a CRLF
                        // then it shouldn't be written to the binary stream 
                        // I am seeing an issue when I do a retrieve to HAIMS (Silver Refresh) the files have 2 extra bytes at the end.
                        // I think this is caused by writing this line when it should not be writte..
                        // Not sure if this will cause other problems

                        // update - this did not fix the problem!
                        // maybe look if this line includes \r\n and exclude that piece of it?
                        // need to see actual output to see what is happening
                    }
                    else
                    {
                        binaryStream.Write(tempBinaryStream.ToArray(), 0, (int)tempBinaryStream.Length);
                        tempBinaryStream = new MemoryStream();
                    }
                }
            }
            else if (readingSoapSection)
            {
                if (l.Contains(boundary))
                {
                    readingSoapSection = false;
                    inSoapSection = false;
                }
                else
                {
                    xml.AppendLine(l);
                }
            }
            else
            {
                fullResponse.Append(l);
                if (l.StartsWith("Content-Type:"))
                {
                    if (l.Contains("soap+xml"))
                    {
                        inSoapSection = true;
                        xml = new StringBuilder();
                    }
                    else
                    {
                        inBinarySection = true;
                    }
                }
                else if (inSoapSection)
                {
                    if (l == "\r\n")
                    {
                        readingSoapSection = true;
                    }
                }
                else if (inBinarySection)
                {
                    if (l == "\r\n")
                    {
                        readingBinaryData = true;
                    }
                }
            }


            responseStream.Close();

            if (binaryStream.Length > 0)
            {
                BinaryWriter writer = new BinaryWriter(new FileStream(filename, FileMode.Create));
                writer.Write(binaryStream.ToArray(), 0, (int)binaryStream.Length);
                writer.Close();
            }
            else
            {
                LogMsg("Found 0 length binary data, will not save anything from this response",true);
            }

            if (File.Exists(openImageExe))
            {
                if (File.Exists(filename))
                {

                    if (!string.IsNullOrEmpty(openImageExe))
                    {
                        System.Diagnostics.Process proc = new System.Diagnostics.Process();
                        proc.StartInfo.FileName = openImageExe;
                        proc.StartInfo.Arguments = "\"" + filename + "\"";
                        proc.Start();
                    }
                }
                else
                {
                    LogMsg("File '" + filename + "' does not exist, cannot open", true);
                }
            }
            else
            {
                LogMsg("Image Viewer '" + openImageExe + "' does not exist, cannot open viewer", true);
            }

            DocumentRetrieveResponse xmlResponse = null;
            if (xml == null)
            {
                LogMsg("Parsed XML is null, could not find XML in response", true);
            }
            else
            {
                xmlResponse = DocumentRetrieveResponse.Parse(xml.ToString());
            }
            XcaRetrieveResult result = new XcaRetrieveResult(xmlResponse, filename, responseHeaders.ToArray());
            return result;
        }

        public bool IgnoreCertificateErrorHandler(object sender, System.Security.Cryptography.X509Certificates.X509Certificate certificate,
            System.Security.Cryptography.X509Certificates.X509Chain chain, System.Net.Security.SslPolicyErrors sslPolicyErrors)
        {
            LogMsg("Received certificate '" + certificate.Subject + "'", false);
            return true;
        }

        public DocumentQueryResponse ExecuteQuery(string queryUri, string patientIcn, 
            string transactionId, UserCredentials userCredentials, string fromDate,
            string toDate, string classCode, string classCodeScheme, string protocol, 
            string username, string password, bool debug)
        {
            //http://msdn.microsoft.com/en-us/library/debx8sh9.aspx

            if (string.IsNullOrEmpty(protocol))
            {
                protocol = defaultProtocol;
            }

            string url = createUrl(protocol, queryUri, mstrServerName, mintServerPort);
            HttpWebRequest request = (HttpWebRequest)WebRequest.Create(url);
            request.KeepAlive = false;
            //WebRequest request = WebRequest.Create(url);// "http://zzzzzzfhietis:8080/XCARespondingGateway/services/RespondingGateway_Service");
            
            request.Method = "POST";
            request.ContentType = "application/soap+xml; charset=UTF-8; action=\"urn:ihe:iti:2007:CrossGatewayQuery\"";

            request.Headers.Add("xxx-sitename", userCredentials.SiteName);
            request.Headers.Add("xxx-fullname", userCredentials.Name);
            request.Headers.Add("xxx-sitenumber", userCredentials.SiteNumber);
            request.Headers.Add("xxx-ssn", userCredentials.Ssn);
            request.Headers.Add("xxx-purpose-of-use", "Routine medical care");
            request.Headers.Add("xxx-patient-id", patientIcn);
            request.Headers.Add("xxx-transaction-id", transactionId);
            request.Headers.Add("SOAPAction", "\"urn:ihe:iti:2007:CrossGatewayQuery\"");
            if (SetAllowUnsafeHeaderParsing20())
            {
                LogMsg("Set allowed unsafe header parsing",false);
            }
            else
            {
                LogMsg("Could not set allow unsafe header parsing",true);
            }

            if ((!string.IsNullOrEmpty(username)) && (!string.IsNullOrEmpty(password)))
            {
                request.Credentials = new NetworkCredential(username, password);
                LogMsg("Providing credentials for user '" + username + "' as part of request", false);
            }

            //ServicePointManager.ServerCertificateValidationCallback += IgnoreCertificateErrorHandler;

            byte[] byteArray = Encoding.UTF8.GetBytes(createQueryRequest(patientIcn, 
                fromDate, toDate, classCode, classCodeScheme));

            LogMsg("Executing query to url '" + url + "' for patient '" + patientIcn + "'.", false);

            Stream requestStream =  request.GetRequestStream();
            requestStream.Write(byteArray, 0, byteArray.Length);
            requestStream.Close();

            WebResponse response = null;
            try
            {
                response = request.GetResponse();
            }
            catch (WebException webEx)
            {
                LogMsg("WebException, " + webEx.Message, true);
                // got something other than a 200 status code but we still want to parse the response

                // i think if an XCA error code comes back, the status code will still be a 200
                // so if not 200, then throw exception
                if (webEx.Status != WebExceptionStatus.Success)
                {
                    if (webEx.Response != null)
                    {
                        StreamReader r = new StreamReader(webEx.Response.GetResponseStream());
                        string s = r.ReadToEnd();
                        LogMsg("Response: " + s, true);
                    }
                    throw webEx;
                }

                response = webEx.Response;                
            }
            if (response == null)
                throw new Exception("Null response from server");

            foreach (string headerName in response.Headers.Keys)
            {
                LogMsg("Header [" + headerName + "] = [" + response.Headers[headerName] + "]", false);
            }

            string contentType = response.Headers["Content-Type"];
            string boundary = "";
            int loc = contentType.IndexOf("boundary=");
            if (loc >= 0)
            {
                boundary = contentType.Substring(loc + 9);
                loc = boundary.IndexOf(";");
                if (loc >= 0)
                {
                    boundary = boundary.Substring(0, loc);
                }
            }
            if (boundary.StartsWith("\""))
            {
                boundary = boundary.Substring(1);
            }
            if (boundary.EndsWith("\""))
            {
                boundary = boundary.Substring(0, boundary.Length - 1);
            }


            Stream responseStream = response.GetResponseStream();
            StreamReader reader = new StreamReader(responseStream);

            /*
            StringBuilder output = new StringBuilder();
            string line = reader.ReadLine();
            while (line != null)
            {
                output.AppendLine(line);
                line = reader.ReadLine();
            }
            LogMsg("Response [" + output.ToString() + "]", false);
            return null;
            */
               

            string xml = extractXmlFromQueryResponse(reader, boundary, debug);
            reader.Close();
            if (xml == null)
                throw new Exception("Could not find XML in response");
            try
            {
                //LogMsg("Xml Response [" + xml + "]", false);
                return Document.createDocumentResponseFromXml(xml);
            }
            catch (Exception ex)
            {
                LogMsg("Error parsing XML, " + ex.Message + "\n\n" + xml, true);
                throw ex;
            }
        }

        public static bool SetAllowUnsafeHeaderParsing20()
        {
            //Get the assembly that contains the internal class
            Assembly aNetAssembly = Assembly.GetAssembly(typeof(System.Net.Configuration.SettingsSection));
            if (aNetAssembly != null)
            {
                //Use the assembly in order to get the internal type for the internal class
                Type aSettingsType = aNetAssembly.GetType("System.Net.Configuration.SettingsSectionInternal");
                if (aSettingsType != null)
                {
                    //Use the internal static property to get an instance of the internal settings class.
                    //If the static instance isn't created allready the property will create it for us.
                    object anInstance = aSettingsType.InvokeMember("Section",
                      BindingFlags.Static | BindingFlags.GetProperty | BindingFlags.NonPublic, null, null, new object[] { });

                    if (anInstance != null)
                    {
                        //Locate the private bool field that tells the framework is unsafe header parsing should be allowed or not
                        FieldInfo aUseUnsafeHeaderParsing = aSettingsType.GetField("useUnsafeHeaderParsing", BindingFlags.NonPublic | BindingFlags.Instance);
                        if (aUseUnsafeHeaderParsing != null)
                        {
                            aUseUnsafeHeaderParsing.SetValue(anInstance, true);
                            return true;
                        }
                    }
                }
            }
            return false;
        }

        private string extractXmlFromQueryResponse(TextReader reader, string boundary, bool logSuccessfulOutput)
        {
            StringBuilder sb = new StringBuilder();
            StringBuilder xml = null;
            string line = reader.ReadLine();
            while (line != null)
            {
                sb.AppendLine(line);
                if (line.Length > 0)
                {
                    if((string.IsNullOrEmpty(boundary) || (line.IndexOf(boundary) < 0)))

                    //if (line.IndexOf(boundary) < 0)
                    {
                        if (line.StartsWith("<?xml"))
                        {
                            xml = new StringBuilder();
                            xml.AppendLine(line);
                        }
                        else if (xml != null)
                        {
                            xml.AppendLine(line);
                        }
                    }
                }
                else
                {
                    if (xml != null)
                        return xml.ToString();
                }
                line = reader.ReadLine();
            }
            //LogMsg("Could not find XML in response: " + sb.ToString(), true);
            if (xml != null)
            {
                if(logSuccessfulOutput)
                    LogMsg("Full Response [" + sb.ToString() + "]", false);
                return xml.ToString();
            }
            LogMsg("Could not find XML in response: " + sb.ToString(), true);
            return null; 

        }

        private string createQueryRequest(string patientIcn, string fromDate, string toDate, 
            string classCode, string classCodeScheme)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");

            /*
            // this one is failing for some reason:
            sb.Append("<S:Envelope xmlns:S=\"http://www.w3.org/2003/05/soap-envelope\">");
            sb.Append("<S:Header>");
            sb.Append("<wsa:To xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">http://vhacrbimmcvix3.vaco.domain/XCARespondingGateway/services/RespondingGateway_Service.RespondingGateway_ServiceHttpSoap12Endpoint/</wsa:To>");
            sb.Append("<wsa:Action xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">urn:ihe:iti:2007:CrossGatewayQuery</wsa:Action>");
            
            sb.Append("<ReplyTo xmlns=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<Address>http://www.w3.org/2005/08/addressing/anonymous</Address>");
            sb.Append("</ReplyTo>");
            sb.Append("<FaultTo xmlns=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<Address>http://www.w3.org/2005/08/addressing/anonymous</Address>");
            sb.Append("</FaultTo>");
             
            sb.Append("<wsa:MessageID xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">uuid:0da0b5c5-c041-444e-b5bb-0ee4879b8711</wsa:MessageID>");
            sb.Append("</S:Header>");
            sb.Append("<S:Body>");
            sb.Append("<AdhocQueryRequest maxResults=\"1000\" xmlns=\"urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0\" xmlns:ns2=\"urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0\" xmlns:ns4=\"urn:ihe:iti:xds-b:2007\" xmlns:ns3=\"urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0\" xmlns:ns5=\"urn:oasis:names:tc:ebxml-regrep:xsd:lcm:3.0\">");
            sb.Append("<ResponseOption returnComposedObjects=\"true\" />");
            sb.Append("<ns2:AdhocQuery id=\"urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d\" status=\"urn:oasis:names:tc:ebxml-regrep:StatusType:Approved\">");
            sb.Append("<ns2:Slot name=\"$XDSDocumentEntryPatientId\">");
            sb.Append("<ns2:ValueList>");
            sb.Append("<ns2:Value>'");
            sb.Append(patientIcn);
            sb.Append("^^^&amp;2.16.840.1.113883.3.166&amp;ISO'</ns2:Value>");
            sb.Append("</ns2:ValueList>");
            sb.Append("</ns2:Slot>");
            sb.Append("<ns2:Slot name=\"$XDSDocumentEntryStatus\">");
            sb.Append("<ns2:ValueList>");
            sb.Append("<ns2:Value>urn:oasis:names:tc:ebxml-regrep:StatusType:Approved</ns2:Value>");
            sb.Append("</ns2:ValueList>");
            sb.Append("</ns2:Slot>");
            
            sb.Append("<ns2:Name/>");
            sb.Append("<ns2:Description/>");
            
            //sb.Append("<ns2:QueryExpression/>");
             
            sb.Append("</ns2:AdhocQuery>");
            sb.Append("</AdhocQueryRequest>");
            sb.Append("</S:Body>");
            sb.Append("</S:Envelope>");
            */

            
             // this one works!
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");

            sb.Append("<soapenv:Header xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<wsa:To>http://localhost:7010/DocSearchXcaService/RespondingGateway_Service</wsa:To>");
            sb.Append("<wsa:MessageID>urn:uuid:3C7D747C17C16CAA4A1287680303993</wsa:MessageID>");
            sb.Append("<wsa:Action>urn:ihe:iti:2007:CrossGatewayQuery</wsa:Action>");
            sb.Append("</soapenv:Header>");
            sb.Append("<soapenv:Body>");
            sb.Append("<urn:AdhocQueryRequest maxResults=\"1000\" id=\"http://testUri\" xmlns:urn=\"urn:oasis:names:tc:ebxml-regrep:xsd:query:3.0\">");
            sb.Append("<urn:ResponseOption returnComposedObjects=\"true\" returnType=\"LeafClass\"/>");
            sb.Append("<urn1:AdhocQuery id=\"urn:uuid:14d4debf-8f97-4251-9a74-a90016b0af0d\" xmlns:urn1=\"urn:oasis:names:tc:ebxml-regrep:xsd:rim:3.0\">");
            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryPatientId\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'");
            sb.Append(patientIcn);
            sb.Append("^^^&amp;2.16.840.1.113883.3.166&amp;ISO'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");

            if (!string.IsNullOrEmpty(fromDate))
            {
                DateTime fr;
                if (DateTime.TryParse(fromDate, out fr))
                {
                    DateTime utc = fr.ToUniversalTime();
                    sb.Append("<urn1:Slot name=\"$XDSDocumentEntryCreationTimeFrom\">");
                    sb.Append("<urn1:ValueList>");
                    sb.Append("<urn1:Value>" + utc.ToString("yyyyMMddHHmm") + "</urn1:Value>");
                    sb.Append("</urn1:ValueList>");
                    sb.Append("</urn1:Slot>");
                }
            }
            if (!string.IsNullOrEmpty(toDate))
            {
                DateTime t;
                if (DateTime.TryParse(toDate, out t))
                {
                    DateTime utc = t.ToUniversalTime();
                    sb.Append("<urn1:Slot name=\"$XDSDocumentEntryCreationTimeTo\">");
                    sb.Append("<urn1:ValueList>");
                    sb.Append("<urn1:Value>" + utc.ToString("yyyyMMddHHmm") + "</urn1:Value>");
                    sb.Append("</urn1:ValueList>");
                    sb.Append("</urn1:Slot>");
                }
            }
            if (!string.IsNullOrEmpty(classCode) && (!string.IsNullOrEmpty(classCodeScheme)))
            {
                sb.Append("<urn1:Slot name=\"$XDSDocumentEntryClassCode\">");
                sb.Append("<urn1:ValueList>");
                sb.Append("<urn1:Value>" + classCode + "</urn1:Value>");
                sb.Append("</urn1:ValueList>");
                sb.Append("</urn1:Slot>");
                sb.Append("<urn1:Slot name=\"$XDSDocumentEntryClassCodeScheme\">");
                sb.Append("<urn1:ValueList>");
                sb.Append("<urn1:Value>" + classCodeScheme + "</urn1:Value>");
                sb.Append("</urn1:ValueList>");
                sb.Append("</urn1:Slot>");
            }

            sb.Append("<urn1:Slot name=\"$XDSDocumentEntryStatus\">");
            sb.Append("<urn1:ValueList>");
            sb.Append("<urn1:Value>'urn:oasis:names:tc:ebxml-regrep:StatusType:Approved'</urn1:Value>");
            sb.Append("</urn1:ValueList>");
            sb.Append("</urn1:Slot>");
            sb.Append("</urn1:AdhocQuery>");
            sb.Append("</urn:AdhocQueryRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");
            

            return sb.ToString();
        }

        private string createRetrieveRequest(string homeCommunityId, string repositoryId, 
            string documentUniqueId, string url)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
            sb.Append("<soapenv:Envelope xmlns:soapenv=\"http://www.w3.org/2003/05/soap-envelope\">");
            sb.Append("<soapenv:Header xmlns:wsa=\"http://www.w3.org/2005/08/addressing\">");
            sb.Append("<wsa:To>" + url + "</wsa:To>");
            sb.Append("<wsa:MessageID>urn:uuid:3C7D747C17C16CAA4A1287680303993</wsa:MessageID>");
            sb.Append("<wsa:Action>urn:ihe:iti:2007:CrossGatewayRetrieve</wsa:Action>");
            sb.Append("</soapenv:Header>");

            sb.Append("<soapenv:Body>");
            sb.Append("<ns5:RetrieveDocumentSetRequest xmlns:ns5=\"urn:ihe:iti:xds-b:2007\">");
            sb.Append("<ns5:DocumentRequest>");
            sb.Append("<ns5:HomeCommunityId>");
            sb.Append(homeCommunityId);
            sb.Append("</ns5:HomeCommunityId>");
            sb.Append("<ns5:RepositoryUniqueId>");
            sb.Append(repositoryId);
            sb.Append("</ns5:RepositoryUniqueId>");
            sb.Append("<ns5:DocumentUniqueId>");
            sb.Append(documentUniqueId);
            sb.Append("</ns5:DocumentUniqueId>");
            sb.Append("</ns5:DocumentRequest>");
            sb.Append("</ns5:RetrieveDocumentSetRequest>");
            sb.Append("</soapenv:Body>");
            sb.Append("</soapenv:Envelope>");
            return sb.ToString();
        }

       
        

        

        

    }

    public class IgnoreCertificatePolicy : ICertificatePolicy
    {
        

        public bool CheckValidationResult(ServicePoint srvPoint, 
            System.Security.Cryptography.X509Certificates.X509Certificate certificate, 
            WebRequest request, int certificateProblem)
        {
            return true;
        }

        
    }
}
