/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;

namespace VixClientV3_XCA.types
{
    public class DocumentResponse
    {
        private readonly DocumentResponseError [] errors;
        private readonly string status;
        private readonly DocumentResponseFault[] faults;

        public DocumentResponseFault[] Faults
        {
            get { return faults; }
        } 

        public string Status
        {
            get { return status; }
        } 

        public DocumentResponseError [] Errors
        {
            get { return errors; }
        }


        public DocumentResponse(string status, List<DocumentResponseError> errors, 
            List<DocumentResponseFault> faults)
        {
            this.errors = errors.ToArray();
            this.faults = faults.ToArray();
            this.status = status;
        }
    }

    public class XcaRetrieveResult
    {
        private readonly DocumentRetrieveResponse xmlResponse;
        private readonly string filename;
        private readonly long fileSize;
        private readonly ResponseHeader[] responseHeaders;

        public DocumentRetrieveResponse XmlResponse
        {
            get { return xmlResponse; }
        }

        public string Filename
        {
            get { return filename; }
        }

        public long FileSize
        {
            get { return fileSize; }
        }

        public ResponseHeader[] ResponseHeaders
        {
            get { return responseHeaders; }
        }

        public XcaRetrieveResult(DocumentRetrieveResponse xmlResponse, string filename, ResponseHeader[] responseHeaders)
        {
            this.xmlResponse = xmlResponse;
            this.filename = filename;
            FileInfo file = new FileInfo(filename);
            if (file.Exists)
            {
                fileSize = file.Length;
            }
            else
            {
                fileSize = -1;
            }
            this.responseHeaders = responseHeaders;
        }
    }

    /// <summary>
    /// XML portion of the XCA Retrieve response
    /// </summary>  
    public class DocumentRetrieveResponse : DocumentResponse
    {
        private string homeCommunityId;
        private string repositoryUniqueId;
        private string documentUniqueId;
        private string mimeType;

        public string HomeCommunityId
        {
            get { return homeCommunityId; }
        }

        public string RepositoryUniqueId
        {
            get { return repositoryUniqueId; }
        }

        public string DocumentUniqueId
        {
            get { return documentUniqueId; }
        }

        public string MimeType
        {
            get { return mimeType; }
        }

        private DocumentRetrieveResponse(string homeCommunityId, string repositoryUniqueId, string documentUniqueId, string mimeType,
            string status, List<DocumentResponseError> errors,
            List<DocumentResponseFault> faults) : base(status, errors, faults)
        {
            this.homeCommunityId = homeCommunityId;
            this.repositoryUniqueId = repositoryUniqueId;
            this.documentUniqueId = documentUniqueId;
            this.mimeType = mimeType;
        }


        public static DocumentRetrieveResponse Parse(string xml)
        {

            DocumentRetrieveResponse result = null;
            List<DocumentResponseError> errors = new List<DocumentResponseError>();
            List<DocumentResponseFault> faults = new List<DocumentResponseFault>();

            XmlDocument xmlDoc = new XmlDocument();
            XmlNamespaceManager nsmgr = new XmlNamespaceManager(xmlDoc.NameTable);
            nsmgr.AddNamespace("soapenv", "http://www.w3.org/2003/05/soap-envelope");
            nsmgr.AddNamespace("ns5", "urn:ihe:iti:xds-b:2007");
            nsmgr.AddNamespace("xop", "http://www.w3.org/2004/08/xop/include");
            nsmgr.AddNamespace("ns3", "urn:oasis:names:tc:ebxml-regrep:xsd:rs:3.0");
            xmlDoc.LoadXml(xml);

            string homeCommunityId = "";
            string repositoryUniqueId = "";
            string documentUniqueId = "";
            string mimeType = "";
            string status = "";

            XmlNode retrieveDocumentSetResponse = xmlDoc.SelectSingleNode("/soapenv:Envelope/soapenv:Body/ns5:RetrieveDocumentSetResponse", nsmgr);
            if (retrieveDocumentSetResponse == null)
                return result;

            XmlNode registryResponseNode = retrieveDocumentSetResponse.SelectSingleNode("ns3:RegistryResponse", nsmgr);
            if (registryResponseNode != null)
            {
                if (registryResponseNode.Attributes["status"] != null)
                {
                    status = registryResponseNode.Attributes["status"].Value;
                }
                XmlNodeList registryErrors = registryResponseNode.SelectNodes("ns3:RegistryErrorList/ns3:RegistryError", nsmgr);
                foreach (XmlNode registryErrorNode in registryErrors)
                {
                    string codeContext = "";
                    if (registryErrorNode.Attributes["codeContext"] != null)
                        codeContext = registryErrorNode.Attributes["codeContext"].Value;

                    string errorCode = "";
                    if (registryErrorNode.Attributes["errorCode"] != null)
                        errorCode = registryErrorNode.Attributes["errorCode"].Value;

                    string location = "";
                    if (registryErrorNode.Attributes["location"] != null)
                        location = registryErrorNode.Attributes["location"].Value;

                    string severity = "";
                    if (registryErrorNode.Attributes["severity"] != null)
                        severity = registryErrorNode.Attributes["severity"].Value;

                    errors.Add(new DocumentResponseError(codeContext, errorCode, location, severity));
                }
            }

            XmlNode documentResponseNode = retrieveDocumentSetResponse.SelectSingleNode("ns5:DocumentResponse", nsmgr);
            if (documentResponseNode != null)
            {
                XmlNode homeCommunityIdNode = documentResponseNode.SelectSingleNode("ns5:HomeCommunityId", nsmgr);
                if (homeCommunityIdNode != null)
                {
                    homeCommunityId = homeCommunityIdNode.InnerText;
                }
                XmlNode repositoryUniqueIdNode = documentResponseNode.SelectSingleNode("ns5:RepositoryUniqueId", nsmgr);
                if (repositoryUniqueIdNode != null)
                {
                    repositoryUniqueId = repositoryUniqueIdNode.InnerText;
                }
                XmlNode documentUniqueIdNode = documentResponseNode.SelectSingleNode("ns5:DocumentUniqueId", nsmgr);
                if (documentUniqueIdNode != null)
                {
                    documentUniqueId = documentUniqueIdNode.InnerText;
                }
                XmlNode mimeTypeNode = documentResponseNode.SelectSingleNode("ns5:mimeType", nsmgr);
                if (mimeTypeNode != null)
                {
                    mimeType = mimeTypeNode.InnerText;
                }
            }
            result = new DocumentRetrieveResponse(homeCommunityId, repositoryUniqueId, documentUniqueId, 
                mimeType,status, errors, faults);

            return result;
        }
    }

    public class ResponseHeader
    {
        private readonly string key;
        private readonly string value;

        public string Key
        {
            get { return key; }
        }

        public string Value
        {
            get { return this.value; }
        }


        public ResponseHeader(string key, string value)
        {
            this.key = key;
            this.value = value;
        }
    }

    public class DocumentQueryResponse : DocumentResponse
    {
        private readonly Document[] documents;

        public Document[] Documents
        {
            get { return documents; }
        } 

        public DocumentQueryResponse(string status, List<Document> documents,
            List<DocumentResponseError> errors, List<DocumentResponseFault> faults)
            : base(status, errors, faults)
        {
            this.documents = documents.ToArray();
        }

    }

    public class DocumentResponseError
    {
        private readonly string codeContext;
        private readonly string errorCode;
        private readonly string location;
        private readonly string severity;

        public string Location
        {
            get { return location; }
        }

        public string Severity
        {
            get { return severity; }
        }

        public string CodeContext
        {
            get { return codeContext; }
        }

        public string ErrorCode
        {
            get { return errorCode; }
        }

        public DocumentResponseError(string codeContext, string errorCode, string location, string severity)
        {
            this.codeContext = codeContext;
            this.errorCode = errorCode;
            this.severity = severity;
            this.location = location;
        }
    }

    public class DocumentResponseFault
    {
        private readonly string code;
        private readonly string reason;

        public string Code
        {
            get { return code; }
        }

        public string Reason
        {
            get { return reason; }
        }

        public DocumentResponseFault(string code, string reason)
        {
            this.code = code;
            this.reason = reason;
        }
    }

    
}
