/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using VixLogAnalyzerEngine;

namespace VixLogAnalyzer
{
    public delegate void StatusMessageDelegate(string msg);
    

    public class LogAnalyzerCommon
    {

        public event StatusMessageDelegate OnStatusMessageEvent;
        public LogAnalyzerCommon(StatusMessageDelegate onStatusMessageEvent)
        {
            this.OnStatusMessageEvent = onStatusMessageEvent;
        }

        private void StatusMsg(string msg)
        {
            if (OnStatusMessageEvent != null)
                OnStatusMessageEvent(msg);
        }

        public IEnumerable<TransactionLogEntry> Search(string logsDirectory, VASite site, DateTime date, string[] icns)
        {
            StatusMsg("Downloading transaction entries from '" + site.SiteNumber + "'.");
            string directory = logsDirectory;
            List<TransactionLogEntry> logEntries =
                TransactionLogEntryDownloader.downloadLogEntries(site.SiteNumber, site.VixServer, site.VixPort,
                date, directory);

            //logEntries.Where(entry => entry.QueryType.Contains(txtQueryType.Text));

            IEnumerable<TransactionLogEntry> results =
                from entry in logEntries
                where (
                    (icns.Contains(entry.PatientIcn))) select entry;

            return results;
        }

        /// <summary>
        /// Searchs a directory that already has the logs for the values specfied
        /// </summary>
        /// <param name="logsDirectory">The source directory where the logs already exist</param>
        /// <param name="site"></param>
        /// <param name="date"></param>
        /// <param name="queryType"></param>
        /// <param name="commandClassName"></param>
        /// <param name="transactionId"></param>
        /// <param name="errorMessage"></param>
        /// <param name="exceptionClass"></param>
        /// <param name="patientIcn"></param>
        /// <param name="urn"></param>
        /// <param name="timeOnVix"></param>
        /// <param name="respondingSite"></param>
        /// <param name="itemsReturnedMoreThan0"></param>
        /// <returns></returns>
        public IEnumerable<TransactionLogEntry> SearchDirectory(string logsDirectory, VASite site, DateTime date,
            string queryType, string commandClassName, string transactionId,
            string errorMessage, string exceptionClass, string patientIcn, string urn, SearchRange timeOnVix,
            string respondingSite, bool itemsReturnedMoreThan0, string dataSource)
        {
            StatusMsg("Loading transaction entries from site '" + site.SiteNumber + "'.");
            List<TransactionLogEntry> logEntries =
                    TransactionLogEntryDownloader.GetLogEntriesFromDirectory(logsDirectory, site.SiteNumber, date);

            //logEntries.Where(entry => entry.QueryType.Contains(txtQueryType.Text));

            IEnumerable<TransactionLogEntry> results =
                from entry in logEntries
                where (
                    (string.IsNullOrEmpty(queryType) || entry.QueryType.Contains(queryType))
                    && (string.IsNullOrEmpty(commandClassName) || entry.CommandClassName.Contains(commandClassName))
                    && (string.IsNullOrEmpty(transactionId) || entry.TransactionId.Contains(transactionId))
                    //&&  (string.IsNullOrEmpty(txtErrorMessage.Text) || ((txtErrorMessage.Text == "*" ? (entry.ErrorMessage != null && entry.ErrorMessage.Length > 0) : (entry.ErrorMessage != null &&  entry.ErrorMessage.Contains(txtErrorMessage.Text)))))
                    && (string.IsNullOrEmpty(errorMessage) || ((errorMessage == "*" ? (entry.ErrorMessage.Length > 0) : (entry.ErrorMessage.Contains(errorMessage)))))
                    && (string.IsNullOrEmpty(exceptionClass) || ((exceptionClass == "*" ? (entry.ExceptionClassName.Length > 0) : (entry.ExceptionClassName.Contains(exceptionClass)))))
                    && (string.IsNullOrEmpty(patientIcn) || entry.PatientIcn.Contains(patientIcn))
                    && (string.IsNullOrEmpty(urn) || entry.Urn.Contains(urn))
                    && ((timeOnVix == null) || (timeOnVix.Range == RangeOption.greaterThan ? (entry.TimeOnVix > timeOnVix.Value) : (entry.TimeOnVix < timeOnVix.Value)))
                    && (string.IsNullOrEmpty(respondingSite) || ((respondingSite == "*" ? (entry.RespondingSite.Length > 0) : (entry.RespondingSite.Contains(respondingSite)))))
                    //&& ((cbTimeOnVixRange.SelectedItem == null) || ((RangeOption)cbTimeOnVixRange.SelectedItem == RangeOption.greaterThan ? (entry.TimeOnVix > slTimeOnVix.Value) : (entry.TimeOnVix < slTimeOnVix.Value)))
                    && (!itemsReturnedMoreThan0 || entry.ItemsReturned > 0)
                    && (string.IsNullOrEmpty(dataSource) || (entry.DataSource.Contains(dataSource)))
                    )
                select entry;

            return results;
        }

        public IEnumerable<TransactionLogEntry> Search(string logsDirectory, VASite site, DateTime date, 
            string queryType, string commandClassName, string transactionId,
            string errorMessage, string exceptionClass, string patientIcn, string urn, SearchRange timeOnVix,
            string respondingSite, bool itemsReturnedMoreThan0, string dataSource)
        {
           
                StatusMsg("Downloading transaction entries from '" + site.SiteNumber + "'.");
                string directory = logsDirectory;
                List<TransactionLogEntry> logEntries =
                    TransactionLogEntryDownloader.downloadLogEntries(site.SiteNumber, site.VixServer, site.VixPort,
                    date, directory);

                //logEntries.Where(entry => entry.QueryType.Contains(txtQueryType.Text));

                IEnumerable<TransactionLogEntry> results =
                    from entry in logEntries
                    where (
                        (string.IsNullOrEmpty(queryType) || entry.QueryType.Contains(queryType))
                        && (string.IsNullOrEmpty(commandClassName) || entry.CommandClassName.Contains(commandClassName))
                        && (string.IsNullOrEmpty(transactionId) || entry.TransactionId.Contains(transactionId))
                        //&&  (string.IsNullOrEmpty(txtErrorMessage.Text) || ((txtErrorMessage.Text == "*" ? (entry.ErrorMessage != null && entry.ErrorMessage.Length > 0) : (entry.ErrorMessage != null &&  entry.ErrorMessage.Contains(txtErrorMessage.Text)))))
                        && (string.IsNullOrEmpty(errorMessage) || ((errorMessage == "*" ? (entry.ErrorMessage.Length > 0) : (entry.ErrorMessage.Contains(errorMessage)))))
                        && (string.IsNullOrEmpty(exceptionClass) || ((exceptionClass == "*" ? (entry.ExceptionClassName.Length > 0) : (entry.ExceptionClassName.Contains(exceptionClass)))))
                        && (string.IsNullOrEmpty(patientIcn) || entry.PatientIcn.Contains(patientIcn))
                        && (string.IsNullOrEmpty(urn) || entry.Urn.Contains(urn))
                        && ((timeOnVix == null) || (timeOnVix.Range == RangeOption.greaterThan ? (entry.TimeOnVix > timeOnVix.Value) : (entry.TimeOnVix < timeOnVix.Value)))
                        && (string.IsNullOrEmpty(respondingSite) || ((respondingSite == "*" ? (entry.RespondingSite.Length > 0) : (entry.RespondingSite.Contains(respondingSite)))))
                        //&& ((cbTimeOnVixRange.SelectedItem == null) || ((RangeOption)cbTimeOnVixRange.SelectedItem == RangeOption.greaterThan ? (entry.TimeOnVix > slTimeOnVix.Value) : (entry.TimeOnVix < slTimeOnVix.Value)))
                        && (!itemsReturnedMoreThan0 || entry.ItemsReturned > 0)
                        && (string.IsNullOrEmpty(dataSource) || (entry.DataSource.Contains(dataSource)))
                        )
                    select entry;

                return results;

            
            
        }
    }
}
