/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Web;
using System.Net;
using System.IO;

namespace VixLogAnalyzerEngine
{
    

    public class TransactionLogEntryDownloader
    {
        public static event LogMsgDelegate OnLogMsgEvent = null;

        private static void LogMsg(string msg)
        {
            if (OnLogMsgEvent != null)
                OnLogMsgEvent(msg);
        }

        /// <summary>
        /// Parse log entries from a directory on the disk in the format the VIX Log Collector does it
        /// </summary>
        /// <param name="logDirectory"></param>
        /// <param name="siteNumber"></param>
        /// <param name="date"></param>
        /// <returns></returns>
        public static List<TransactionLogEntry> GetLogEntriesFromDirectory(string logDirectory, string siteNumber, DateTime date)
        {
            //G:\VIXLogs\459\2011\07\07-04-2011.tsv

            string filename = logDirectory + "\\" + siteNumber + "\\" + date.Year + "\\" + date.Month.ToString("00") + "\\" + date.ToString("MM-dd-yyyy") + ".tsv";
            if (File.Exists(filename))
            {
                FileStream stream = new FileStream(filename, FileMode.Open);
                return ParseLogStream(siteNumber, stream, null);
            }
            // the file doesn't exist, can't do anything
            return null;
        }

        /// <summary>
        /// This method does not use a date and requires P124/P119 to work properly. This method will NOT escape the transaction ID
        /// </summary>
        /// <param name="server"></param>
        /// <param name="port"></param>
        /// <param name="siteNumber"></param>
        /// <param name="transactionId"></param>
        /// <returns></returns>
        public static List<TransactionLogEntry> getTransaction(string server, int port, 
            string siteNumber, string transactionId)
        {

            StringBuilder url = new StringBuilder();
            url.Append("http://");
            url.Append(server);
            url.Append(":");
            url.Append(port);
            url.Append("/Vix/secure/ExcelTransactionLog");
            url.Append("?transactionId=");
            url.Append(transactionId);

            // JMW 11/29/2012 - the transactionIdOnly property works correctly however it does not work if you escape the transaction ID.  The ID provided must be the real ID and not include the \{ and \} as done above
            // for now just exclude the transactionIdOnly parameter since it works fine without it (just not quite as efficient).  When we are sure all sites are 124 and 119 then it can be added back in

            // JMW 3/25/2013 - this method will always use this property to only search by transaction ID. The transaction ID should not be escaped!
            url.Append("&transactionIdOnly=true"); // specify only this transaction (new for P124 and 119)



            url.Append("&format=text%2Ftab-separated-values");
            WebRequest webRequest = WebRequest.Create(url.ToString());
            webRequest.Credentials = new NetworkCredential("", "");
            WebResponse response = webRequest.GetResponse();
            Stream transactionLogStream = transactionLogStream = response.GetResponseStream();

            return ParseLogStream(siteNumber, transactionLogStream, null);
        }

        /// <summary>
        /// This method searches for the transaction for a specific date and will work on any version of the VIX.  
        /// </summary>
        /// <param name="server"></param>
        /// <param name="port"></param>
        /// <param name="siteNumber"></param>
        /// <param name="transactionId"></param>
        /// <param name="date"></param>
        /// <returns></returns>
        public static List<TransactionLogEntry> getTransaction(string server, int port,
            string siteNumber, string transactionId, DateTime date)
        {
            //string rawTransactionId = transactionId;

            if (transactionId.StartsWith("{"))
            {
                transactionId = "\\" + transactionId;
            }
            if (transactionId.EndsWith("}"))
            {
                transactionId = transactionId.Substring(0, transactionId.Length - 1) + "\\}";
            }


            string dateFormat = "MM/dd/yyyy";
            StringBuilder url = new StringBuilder();
            url.Append("http://");
            url.Append(server);
            url.Append(":");
            url.Append(port);
            url.Append("/Vix/secure/ExcelTransactionLog");
            url.Append("?fromDate=");
            url.Append(date.ToString(dateFormat));
            url.Append("&toDate=");
            url.Append(date.ToString(dateFormat));
            url.Append("&transactionId=");
            url.Append(transactionId);

            // JMW 11/29/2012 - the transactionIdOnly property works correctly however it does not work if you escape the transaction ID.  The ID provided must be the real ID and not include the \{ and \} as done above
            // for now just exclude the transactionIdOnly parameter since it works fine without it (just not quite as efficient).  When we are sure all sites are 124 and 119 then it can be added back in
            //url.Append("&transactionIdOnly=true"); // specify only this transaction (new for P124 and 119)

            // JMW 3/25/2013
            // this doesn't really add anything since the transactionIdOnly=true works for P124/P119 the problem is if you provide an escaped value it won't work.

            // JMW 3/25/2013 - Added rawTransactionId property to P119T6, this should contain the unescaped value and can be used with transactionIdOnly, will only work with VIX/CVIX at least P119T6
            //url.Append("&transactionIdOnly=true"); // specify only this transaction (fixed in 119T6)
            // specify the raw transaction ID value with no escaping (added in 119T6)
            //url.Append("&rawTransactionId=");
            //url.Append(rawTransactionId);


            url.Append("&format=text%2Ftab-separated-values");
            WebRequest webRequest = WebRequest.Create(url.ToString());
            webRequest.Credentials = new NetworkCredential("vixlog", "tachik0ma");
            WebResponse response = webRequest.GetResponse();
            Stream transactionLogStream = transactionLogStream = response.GetResponseStream();            
            
            return ParseLogStream(siteNumber, transactionLogStream, null);
        }

        public static List<TransactionLogEntry> downloadLogEntries(string siteNumber, string server, int port, 
            DateTime date, string logTempDirectory)
        {
            //12/20/2010

            // only supporting one day at a time - not sure if that is going to be good or bad...

            // want to download each log entry and if it is today's, put into log temp directory with a .partial extension. If not today then you can assume its the full days log (if no transaction ID specified) and 
            // then store the log so you don't have to keep downloading it

            Stream transactionLogStream = null;
            bool isToday = isDateToday(date);
            string dir = logTempDirectory + "\\" + siteNumber;
            if (!Directory.Exists(dir))
            {
                Directory.CreateDirectory(dir);
            }
            string filename = dir + "\\" + date.ToString("yyyy_MM_dd") + ".csv" + (isToday ? ".partial" : "");
            if (!isToday)
            {
                // might be able to get a cached copy
                if (File.Exists(filename))
                {
                    transactionLogStream = new StreamReader(filename).BaseStream;
                }
            }
            string dateFormat = "MM/dd/yyyy";
            StreamWriter writer = null;
            if (transactionLogStream == null)
            {
                StringBuilder url = new StringBuilder();
                url.Append("http://");
                url.Append(server);
                url.Append(":");
                url.Append(port);
                url.Append("/Vix/secure/ExcelTransactionLog");
                url.Append("?fromDate=");
                url.Append(date.ToString(dateFormat));
                url.Append("&toDate=");
                url.Append(date.ToString(dateFormat));

                url.Append("&format=text%2Ftab-separated-values");
                WebRequest webRequest = WebRequest.Create(url.ToString());
                webRequest.Credentials = new NetworkCredential("vixlog", "tachik0ma");
                WebResponse response = webRequest.GetResponse();
                transactionLogStream = response.GetResponseStream();
                writer = new StreamWriter(filename);
            }
            
            return ParseLogStream(siteNumber, transactionLogStream, writer);
        }

        private static bool isDateToday(DateTime date)
        {
            DateTime now = DateTime.Now;
            if ((now.Year == date.Year) &&
                (now.Month == date.Month) &&
                (now.Day == date.Day))
                return true;
            return false;
        }

        private static List<TransactionLogEntry> ParseLogStream(string siteNumber, Stream stream, StreamWriter outputWriter)
        {
            List<TransactionLogEntry> entries = new List<TransactionLogEntry>();
            StreamReader reader = new StreamReader(stream);
            try
            {
                string headerLine = reader.ReadLine();
                if (outputWriter != null)
                    outputWriter.WriteLine(headerLine);
                string line = reader.ReadLine();
                while (line != null)
                {
                    if (outputWriter != null)
                        outputWriter.WriteLine(line);
                    TransactionLogEntry entry = TransactionLogEntry.parseTSVLine(siteNumber, headerLine, line, LogMsg);
                    if (entry != null)
                        entries.Add(entry);
                    line = reader.ReadLine();
                }
            }
            finally
            {
                reader.Close();
                if (outputWriter != null)
                    outputWriter.Close();
            }

            return entries;
        }

        /// <summary>
        /// This method groups the input entries into parent/child relationships based on the commandId and parentCommandId
        /// </summary>
        /// <param name="entries">Flat list of all entries</param>
        /// <param name="excludeParentlessTransactions">If true, child transactions without parents are excluded from the result. If true parentless child transactions are excluded</param>
        /// <returns>List of grouped transactions</returns>
        public static List<TransactionLogEntry> GroupTransactions(List<TransactionLogEntry> entries, bool excludeParentlessTransactions)
        {
            Dictionary<string, TransactionLogEntryItem> entriesDictionary = new Dictionary<string, TransactionLogEntryItem>();
            //bool foundDuplicate = false;
            // put all entries into a dictionary
            foreach (TransactionLogEntry entry in entries)
            {
                if (!string.IsNullOrEmpty(entry.CommandId))
                {
                    if (entriesDictionary.ContainsKey(entry.CommandId))
                    {
                        continue;
                        //throw new Exception("Duplicate command ID found '" + entry.CommandId + "'.");
                    }
                    entriesDictionary.Add(entry.CommandId, new TransactionLogEntryItem(entry));
                }                
            }

            // loop through each entry in dictionary
            foreach (string commandId in entriesDictionary.Keys.ToArray())
            {
                TransactionLogEntryItem entry = entriesDictionary[commandId];
                // if the entry is a child
                if (entry.TransactionLogEntry.IsChild())
                {
                    // search to see if the parent commandId is in the dictionary
                    if (entriesDictionary.ContainsKey(entry.TransactionLogEntry.ParentCommandId))
                    {
                        // find the parent entry and add this child to it and then null this child
                        TransactionLogEntryItem parentEntry = entriesDictionary[entry.TransactionLogEntry.ParentCommandId];
                        parentEntry.TransactionLogEntry.ChildEntries.Add(entry.TransactionLogEntry);
                        entriesDictionary[commandId].MovedToParent = true;
                        //entriesDictionary[commandId] = null;
                        //entriesDictionary.Remove(commandId);
                    }
                    else
                    {
                        // didn't find the parent which means this child might have been created by a parent we are not interseted in (such as a prefetch operation)
                        // mark it as moved so it gets excluded from the result

                        // determine if we want to include child commands that do not have parents in the results
                        //if (!includeParentlessChildCommands())
                        if(excludeParentlessTransactions)
                        {
                            entriesDictionary[commandId].MovedToParent = true;
                        }
                    }
                }
            }

            List<TransactionLogEntry> result = new List<TransactionLogEntry>();
            foreach (TransactionLogEntryItem entry in entriesDictionary.Values)
            {
                if ((entry != null) && (!entry.MovedToParent))
                    result.Add(entry.TransactionLogEntry);
            }
            return result;
        }

    }

    public class TransactionLogEntryItem
    {
        public TransactionLogEntry TransactionLogEntry { get; private set; }
        public bool MovedToParent { get; set; }

        public TransactionLogEntryItem(TransactionLogEntry entry)
        {
            this.TransactionLogEntry = entry;
            this.MovedToParent = false;
        }

    }
}
