/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Linq;
using System.Collections.ObjectModel;
using GalaSoft.MvvmLight;
using VixLogCollector.Client.VixLogCollectorService;
using System.Collections.Generic;
using GalaSoft.MvvmLight.Command;
using System.Windows;
using System.Collections;
using System.Windows.Controls;
using Microsoft.Practices.ServiceLocation;
using Microsoft.Practices.Unity;

namespace VixLogCollector.Client.ViewModel
{
    public class MainViewModel : ViewModelBase
    {
        #region Fields

        private IVixLogCollectorServiceClient sharedClient;

        #endregion

        #region Commands

        public RelayCommand CollectLogsCommand { get; private set; }
        public RelayCommand<IList> VixSelectionChangedCommand { get; private set; }
        public RelayCommand<IList> LogSelectionChangedCommand { get; private set; }
        public RelayCommand<SelectionChangedEventArgs> SelectedDatesChangedCommand { get; private set; }
        public RelayCommand<CalendarDateChangedEventArgs> DisplayDateChangedCommand { get; private set; }
        public RelayCommand<RoutedEventArgs> ShowOnlyProblemsChangedCommand { get; private set; }
        public RelayCommand<RoutedEventArgs> FilterByDayChangedCommand { get; private set; }
        public RelayCommand<VixLogInfo> ActivateLogCommand { get; private set; }

        #endregion

        #region Events

        public event EventHandler ErrorDaysChanged;
        public event EventHandler<EventArgs<String>> ErrorReceived;

        #endregion

        #region Properties

        // Notify Proproperty Changed event is raised by the NotifyPropertyWeaver (http://code.google.com/p/notifypropertyweaver/)
        // which is executed a a MS Build task

        public DateTime MinLogDate { get; set; }
        public DateTime MaxLogDate { get; set; }
        public DateTime SelectedDate { get; set; }
        public DateTime SelectedMonth { get; set; }
        public ObservableCollection<VixLogInfo> MonthLogs { get; set; }
        public ObservableCollection<VixLogInfo> DisplayLogs { get; set; }
        public ObservableCollection<VixLogInfo> SelectedLogs { get; set; }
        public ObservableCollection<VixInfo> Vixes { get; set; }
        public ObservableCollection<VixInfo> SelectedVixes { get; set; }
        public String[] ErrorDays { get; set; }
        public bool ShowOnlyProblems { get; set; }
        public bool FilterByDay { get; set; }

        #endregion

        #region Constructors

        public MainViewModel()
        {
            // Initialize Properties
            MinLogDate = DateTime.Today.AddMonths(-2);
            MaxLogDate = DateTime.Today;
            SelectedMonth = DateTime.Now.GetStartOfMonth();
            Vixes = new ObservableCollection<VixInfo>();
            SelectedVixes = new ObservableCollection<VixInfo>();
            MonthLogs = new ObservableCollection<VixLogInfo>();
            DisplayLogs = new ObservableCollection<VixLogInfo>();
            SelectedLogs = new ObservableCollection<VixLogInfo>();
            ErrorDays = new String[31];
            FilterByDay = true;

            // Initialize Commands
            CollectLogsCommand = new RelayCommand(() => CollectLogs(), () => SelectedLogs.Count > 0 && !IsLogCollectionRunning());
            VixSelectionChangedCommand = new RelayCommand<IList>(items => UpdateSelectedVixes(items));
            LogSelectionChangedCommand = new RelayCommand<IList>(items => UpdateSelectedLogs(items));
            SelectedDatesChangedCommand = new RelayCommand<SelectionChangedEventArgs>(args => FilterDisplayLogs());
            DisplayDateChangedCommand = new RelayCommand<CalendarDateChangedEventArgs>(args => UpdateSelectedMonth(args.RemovedDate));
            ShowOnlyProblemsChangedCommand = new RelayCommand<RoutedEventArgs>(args => FilterDisplayLogs());
            FilterByDayChangedCommand = new RelayCommand<RoutedEventArgs>(args => FilterDisplayLogs());
            ActivateLogCommand = new RelayCommand<VixLogInfo>(log => ActivateLog(log));
        }

        public MainViewModel(IVixLogCollectorServiceClient client)
            : this()
        {
            // initialize shared client
            sharedClient = client;
            sharedClient.GetVixesCompleted += new EventHandler<GetVixesCompletedEventArgs>(VixLogCollectorServiceClient_GetVixesCompleted);
            sharedClient.GetLogListCompleted += new EventHandler<GetLogListCompletedEventArgs>(VixLogCollectorServiceClient_GetLogListCompleted);
            sharedClient.CollectVixLogsCompleted += new EventHandler<CollectVixLogsCompletedEventArgs>(VixLogCollectorServiceClient_CollectVixLogsCompleted);
            sharedClient.GetErrorCompleted += new EventHandler<GetErrorCompletedEventArgs>(sharedClient_GetErrorCompleted);
            sharedClient.GetVixesAsync();
        }

        #endregion

        #region VixLogCollectorServiceClient Event Handlers

        private void sharedClient_GetErrorCompleted(object sender, GetErrorCompletedEventArgs e)
        {
            RaiseErrorReceived(e.Result);
        }

        private void VixLogCollectorServiceClient_GetVixesCompleted(object sender, GetVixesCompletedEventArgs e)
        {
            if (e.Error != null)
                throw new Exception("Error communicating with collector", e.Error);
            Vixes.Clear();
            foreach (VixInfo vix in e.Result)
                Vixes.Add(vix);
            UpdateSelectedMonth(null);
        }

        private void VixLogCollectorServiceClient_GetLogListCompleted(object sender, GetLogListCompletedEventArgs e)
        {
            if (e.Error != null)
                throw new Exception("Error communicating with collector", e.Error);
            MonthLogs.Clear();
            foreach (VixLogInfo log in e.Result)
            {
                log.Vix = Vixes.Where(var => var.Id == log.VixId).SingleOrDefault();
                MonthLogs.Add(log);
            }
            if (MonthLogs.Count() > 0)
                SelectedDate = MonthLogs.Max(m => m.LogDate);
            UpdateErrorDays();
            UpdateVixStatus();
            FilterDisplayLogs();
        }

        private void VixLogCollectorServiceClient_CollectVixLogsCompleted(object sender, CollectVixLogsCompletedEventArgs e)
        {
            if (e.Error != null)
                throw new Exception("Error communicating with collector", e.Error);
            foreach (VixLogInfo log in e.Result)
                UpdateLog(log);
            UpdateErrorDays();
            if (e.Result.Length > 0)
            {
                UpdateVixStatus(GetVix(e.Result[0].VixId));
            }
        }

        #endregion

        #region Private Methods

        private void ActivateLog(VixLogInfo log)
        {
            if (log.Status == VixLogStatus.Failed)
            {
                sharedClient.GetErrorAsync(log);
            }
        }

        private bool AreLogsMissing(VixInfo vix)
        {
            IEnumerable<VixLogInfo> missingLogs =
                from log in MonthLogs
                where log.VixId == vix.Id && log.Status == VixLogStatus.Missing
                select log;
            return missingLogs.Count() > 0;
        }

        private void CollectLogs()
        {
            IEnumerable<String> vixIds = SelectedLogs.Select(log => log.VixId).Distinct();
            IEnumerable<VixLogInfo> logs;
            foreach (String vixId in vixIds)
            {
                logs = SelectedLogs.Where(log => log.VixId == vixId);
                sharedClient.CollectVixLogsAsync(logs.ToArray());
                foreach (VixLogInfo log in logs)
                    log.Status = VixLogStatus.Collecting;
                Vixes.Where(v => v.Id == vixId).Single().Status = VixStatus.Collecting;
            }
        }

        private bool DoLogErrorsExist(DateTime targetDate)
        {
            IEnumerable<VixLogInfo> missingLogs =
                from log in MonthLogs
                where log.LogDate.Date == targetDate.Date
                      && (log.Status == VixLogStatus.Missing || log.Status == VixLogStatus.Failed)
                select log;
            return missingLogs.Count() > 0;
        }

        private bool DoLogErrorsExist(VixInfo vix)
        {
            IEnumerable<VixLogInfo> missingLogs =
                from log in MonthLogs
                where log.VixId == vix.Id && log.Status == VixLogStatus.Failed
                select log;
            return missingLogs.Count() > 0;
        }

        private void FilterDisplayLogs()
        {
            DisplayLogs.Clear();
            IEnumerable<VixLogInfo> logs = MonthLogs;
            if (FilterByDay)
                logs = from log in MonthLogs
                       where log.LogDate.Date == SelectedDate
                       select log;
            if (SelectedVixes.Count > 0)
                logs = from log in logs
                       where SelectedVixes.Contains(log.Vix)
                       select log;
            if (ShowOnlyProblems)
                logs = from log in logs
                       where log.Status == VixLogStatus.Missing || log.Status == VixLogStatus.Failed
                       select log;
            foreach (VixLogInfo log in logs)
                DisplayLogs.Add(log);
        }

        private VixInfo GetVix(String vixId)
        {
            return Vixes.Where(vix => vix.Id == vixId).SingleOrDefault();
        }

        private bool IsLogCollectionRunning()
        {
            return MonthLogs.Where(log => log.Status == VixLogStatus.Collecting).Count() > 0
                || Vixes.Where(vix => vix.Status == VixStatus.Collecting).Count() > 0;
        }

        private void RaiseErrorDaysChanged()
        {
            EventHandler handler = ErrorDaysChanged;
            if (handler != null)
                handler(this, new EventArgs());
        }

        private void RaiseErrorReceived(String error)
        {
            EventHandler<EventArgs<String>> handler = ErrorReceived;
            if (handler != null)
                handler(this,new EventArgs<String>(error));
        }

        private void UpdateErrorDays()
        {
            int year = SelectedMonth.Year;
            int month = SelectedMonth.Month;
            int daysInMonth = DateTime.DaysInMonth(year, month);
            DateTime day;
            for (int i = 0; i < ErrorDays.Length; i++)
            {
                ErrorDays[i] = null;
                if (i >= daysInMonth)
                    continue;
                day = new DateTime(year, month, i + 1);
                if (DoLogErrorsExist(day))
                    ErrorDays[i] = day.ToLongDateString();
            }
            RaiseErrorDaysChanged();
        }

        private void UpdateSelectedLogs(IList items)
        {
            SelectedLogs.Clear();
            foreach (object vixLogObject in items)
            {
                if (vixLogObject is VixLogInfo)
                    SelectedLogs.Add((VixLogInfo)vixLogObject);
            }
        }

        private void UpdateSelectedMonth(DateTime? oldDate)
        {
            if (SelectedMonth == DateTime.MinValue || oldDate.HasValue && oldDate.Value == DateTime.MinValue)
                return;
            if (oldDate == null || SelectedMonth.GetStartOfMonth() != oldDate.Value.GetStartOfMonth())
            {
                MonthLogs.Clear();
                DisplayLogs.Clear();
                sharedClient.GetLogListAsync(SelectedMonth.GetStartOfMonth(), SelectedMonth.GetEndOfMonth(), null);
            }
        }

        private void UpdateSelectedVixes(IList items)
        {
            SelectedVixes.Clear();
            foreach (object vixObject in items)
            {
                if (vixObject is VixInfo)
                    SelectedVixes.Add((VixInfo)vixObject);
            }
            FilterDisplayLogs();
        }

        private void UpdateLog(VixLogInfo log)
        {
            VixLogInfo oldLog = MonthLogs.Where(l => l.VixId == log.VixId && l.LogDate == log.LogDate).SingleOrDefault();
            if (oldLog != null)
            {
                log.Vix = oldLog.Vix;
                MonthLogs.Remove(oldLog);
                MonthLogs.Add(log);
                if (DisplayLogs.Contains(oldLog))
                {
                    DisplayLogs.Insert(DisplayLogs.IndexOf(oldLog), log);
                    DisplayLogs.Remove(oldLog);
                }
            }
        }

        private void UpdateVixStatus()
        {
            foreach (VixInfo vix in Vixes)
                UpdateVixStatus(vix);
        }

        private void UpdateVixStatus(VixInfo vix)
        {
            if (vix == null)
                return;
            if (DoLogErrorsExist(vix))
                vix.Status = VixStatus.Error;
            else if (AreLogsMissing(vix))
                vix.Status = VixStatus.Missing;
            else
                vix.Status = VixStatus.Idle;
        }

        #endregion
    }
}
