/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.Serialization;
using System.IO;
using System.Configuration;
using System.Xml;
using System.Xml.Serialization;

namespace VixLogCollector.Model
{
    [DataContract]
    public class VixInfo
    {
        public static DirectoryInfo LogDirectory;
        private const String fileName = "VixInfo.xml";
        private const String errorFileName = "error.xml";

        #region Properties

        [DataMember]
        public String Id { get; set; }

        [DataMember]
        public String Name { get; set; }

        [DataMember]
        public DateTime StartLogDate { get; set; }

        [DataMember]
        public DateTime? EndLogDate { get; set; }

        [DataMember]
        public bool IsActive { get; set; }

        [DataMember]
        public VixStatus Status
        {
            get
            {
                if (!IsActive)
                    return VixStatus.Inactive;
                FileInfo[] tempFiles = VixDirectory.GetFiles("*.tmp", SearchOption.AllDirectories);
                foreach (FileInfo file in tempFiles)
                {
                    if (file.LastWriteTime > DateTime.Now.AddMinutes(-5))
                        return VixStatus.Collecting;
                }
                return VixStatus.Idle;
            }
            set
            {
                // This is necessary in order for this to be a DataMember
                // and for the xml serialization
            }
        }

        public string Server { get; set; }

        public int Port { get; set; }

        internal DirectoryInfo VixDirectory
        {
            get
            {
                return new DirectoryInfo(LogDirectory.FullName + Path.DirectorySeparatorChar + Id);
            }
        }

        #endregion

        #region Static Methods

        public static VixInfo Load(string vixId)
        {
            FileInfo file = new FileInfo(LogDirectory.FullName + Path.DirectorySeparatorChar + vixId + Path.DirectorySeparatorChar + fileName);
            VixInfo vix = null;
            if (file.Exists)
            {
                try
                {
                    using (XmlReader reader = XmlReader.Create(file.FullName))
                    {
                        XmlSerializer serializer = new XmlSerializer(typeof(VixInfo));
                        vix = (VixInfo)serializer.Deserialize(reader);
                    }
                }
                catch (Exception e)
                {
                    VixInfo errorVix = new VixInfo(vixId);
                    errorVix.WriteException(e);
                }
            }
            return vix;
        }

        public static IEnumerable<VixInfo> LoadAll()
        {
            return LogDirectory.GetDirectories().Select(v => Load(v.Name)).Where(vix => vix != null);
        }

        #endregion

        #region Contrstructors

        static VixInfo()
        {
            if (!String.IsNullOrWhiteSpace(ConfigurationManager.AppSettings["logFolder"]))
                LogDirectory = new DirectoryInfo(ConfigurationManager.AppSettings["logFolder"]);
        }

        // required for xml serialization
        public VixInfo() { IsActive = true; }

        public VixInfo(string id)
            : this()
        {
            this.Id = id;
            StartLogDate = DateTime.Today;
        }

        #endregion

        public Uri GetLogUri(DateTime startDate, DateTime? endDate = null)
        {
            if (endDate == null)
                endDate = startDate;
            UriBuilder builder = new UriBuilder()
            {
                Scheme = "http",
                Host = Server,
                Port = Port,
                Path = "/Vix/secure/ExcelTransactionLog",
                Query = String.Format("fromDate={0}&toDate={1}&format=text/tab-separated-values", startDate.ToString("MM/dd/yyyy"), endDate.Value.ToString("MM/dd/yyyy"))
            };
            return builder.Uri;
        }

        public VixLogInfo GetLog(DateTime date)
        {
            return new VixLogInfo(this, date);
        }

        public IEnumerable<VixLogInfo> GetMissingLogs(DateTime startDate, DateTime endDate)
        {
            if (startDate < StartLogDate)
                startDate = StartLogDate;
            if (EndLogDate.HasValue && endDate > EndLogDate.Value)
                endDate = EndLogDate.Value;
            List<VixLogInfo> logs = new List<VixLogInfo>();
            VixLogInfo log;
            VixLogStatus status;
            for (DateTime i = startDate; i <= endDate; i = i.AddDays(1))
            {
                log = GetLog(i);
                status = log.Status;
                if (status == VixLogStatus.Missing || status == VixLogStatus.Failed)
                {
                    logs.Add(log);
                }
            }
            return logs;
        }

        public void Save()
        {
            if (!VixDirectory.Exists)
                VixDirectory.Create();
            FileInfo file = new FileInfo(LogDirectory.FullName + Path.DirectorySeparatorChar + Id + Path.DirectorySeparatorChar + fileName);
            XmlWriterSettings settings = new XmlWriterSettings()
            {
                Indent = true
            };
            using (XmlWriter writer = XmlWriter.Create(file.FullName, settings))
            {
                XmlSerializer serializer = new XmlSerializer(typeof(VixInfo));
                serializer.Serialize(writer, this);
            }
        }

        #region Overriden Object Methods

        public override string ToString()
        {
            return this.Name + " (" + this.Id + ")";
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is VixInfo))
                return false;
            VixInfo other = (VixInfo)obj;
            return object.Equals(this.Id, other.Id);
        }

        public override int GetHashCode()
        {
            return base.GetHashCode() ^ Id.GetHashCode();
        }
        #endregion

        #region Private Methods

        private void WriteException(Exception ex)
        {
            FileInfo errorFile = new FileInfo(VixDirectory.FullName + Path.DirectorySeparatorChar + errorFileName);
            if (errorFile.Exists)
                errorFile.Delete();
            using (StreamWriter writer = errorFile.CreateText())
            {
                writer.WriteLine(ex.ToString());
                writer.Flush();
                writer.Close();
            }
        }

        #endregion
    }
}
