/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Configuration;
using System.Runtime.Serialization;

namespace VixLogCollector.Model
{
    [DataContract]
    public class VixLogInfo
    {
        internal const string fileExtension = ".tsv";
        internal const string errorFileExtension = ".error";
        internal const string tempFileExtension = ".tmp";
        internal static readonly TimeSpan collectionDelay = new TimeSpan(2, 0, 0, 0);

        #region Fields

        private string logFileBase;

        #endregion

        #region Properties

        [DataMember]
        public String VixId { get; set; }

        [DataMember]
        public DateTime LogDate { get; set; }

        [DataMember]
        public DateTime? DownloadDate
        {
            get
            {
                return LogFileInfo.Exists ? LogFileInfo.CreationTime : (DateTime?)null;
            }
            set
            {
                // This is necessary in order for this to be a DataMember
            }
        }

        [DataMember]
        public VixLogStatus Status
        {
            get
            {
                if (LogFileInfo.Exists)
                    return VixLogStatus.Collected;
                if (LogTempFileInfo.Exists && LogTempFileInfo.LastWriteTime > DateTime.Now.AddMinutes(-5))
                    return VixLogStatus.Collecting;
                if (LogErrorFileInfo.Exists)
                    return VixLogStatus.Failed;
                if (LogDate < Vix.StartLogDate || (Vix.EndLogDate.HasValue && LogDate > Vix.EndLogDate.Value))
                    return VixLogStatus.Unscheduled;
                if (LogDate > DateTime.Now)
                    return VixLogStatus.Future;
                if (LogDate > DateTime.Now.Subtract(collectionDelay))
                    return VixLogStatus.Scheduled;
                return VixLogStatus.Missing;
            }
            set
            {
                // This is necessary in order for this to be a DataMember
            }
        }

        internal VixInfo Vix { get; set; }

        internal FileInfo LogFileInfo
        {
            get
            {
                // always return a new one so the data isn't cached
                return new FileInfo(LogFileBase + fileExtension);
            }
        }

        internal FileInfo LogErrorFileInfo
        {
            get
            {
                // always return a new one so the data isn't cached
                return new FileInfo(LogFileBase + errorFileExtension);
            }
        }

        internal FileInfo LogTempFileInfo
        {
            get
            {
                // always return a new one so the data isn't cached
                return new FileInfo(LogFileBase + tempFileExtension);
            }
        }

        internal string LogFileBase
        {
            get
            {
                if (logFileBase == null)
                    logFileBase = buildLogFileBase();
                return logFileBase;
            }
        }

        #endregion

        #region Constructor(s)

        public VixLogInfo(VixInfo vix, DateTime logDate)
        {
            this.Vix = vix;
            this.VixId = vix.Id;
            this.LogDate = logDate;
        }

        #endregion

        #region Overriden Object Methods

        public override string ToString()
        {
            return this.Vix.ToString() + " " + this.LogDate.ToShortDateString() + " " + this.Status;
        }

        public override bool Equals(object obj)
        {
            if (obj == null || !(obj is VixLogInfo))
                return false;
            VixLogInfo other = (VixLogInfo)obj;
            return object.Equals(this.Vix, other.Vix) && this.LogDate == other.LogDate;
        }

        public override int GetHashCode()
        {
            return base.GetHashCode() ^ this.Vix.GetHashCode() ^ this.LogDate.GetHashCode();
        }

        #endregion

        private string buildLogFileBase()
        {
            StringBuilder fileBuilder = new StringBuilder();
            fileBuilder.Append(VixInfo.LogDirectory.FullName);
            fileBuilder.Append(Path.DirectorySeparatorChar);
            fileBuilder.Append(VixId);
            fileBuilder.Append(Path.DirectorySeparatorChar);
            fileBuilder.Append(LogDate.Year.ToString().PadLeft(4, '0'));
            fileBuilder.Append(Path.DirectorySeparatorChar);
            fileBuilder.Append(LogDate.Month.ToString().PadLeft(2, '0'));
            fileBuilder.Append(Path.DirectorySeparatorChar);
            fileBuilder.Append(LogDate.ToString("MM-dd-yyyy"));
            return fileBuilder.ToString();
        }
    }
}
