/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.Serialization;
using System.ServiceModel;
using System.Text;
using System.IO;
using VixLogCollector.SiteService;
using System.Configuration;
using System.Net;
using System.Xml;
using System.Xml.Linq;
using VixLogCollector.Model;
using System.Security.Permissions;

namespace VixLogCollector.Services
{
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single, ConcurrencyMode = ConcurrencyMode.Multiple)]
    public class VixLogCollectorService : IVixLogCollectorService
    {
        private static readonly char pathSeperator = Path.DirectorySeparatorChar;

        #region IVixLogCollectorService Methods

        public VixLogInfo[] CollectMissingLogs(string vixId)
        {
            VixInfo vix = VixInfo.Load(vixId);
            IEnumerable<VixLogInfo> logs = vix.GetMissingLogs(DateTime.Today.AddMonths(-2), DateTime.Today);
            //VixLogInfo currentLog;
            foreach (VixLogInfo log in logs)
            {
                //currentLog = 
                CollectVixLog(log);
            }
            return logs.ToArray();
        }

        public VixLogInfo CollectVixLog(VixLogInfo log)
        {
            if (log.Vix == null)
                log.Vix = VixInfo.Load(log.VixId);
            if (!log.Vix.IsActive)
                return null;
            try
            {
                if (!log.LogFileInfo.Directory.Exists)
                    log.LogFileInfo.Directory.Create();
                if (log.LogErrorFileInfo.Exists)
                    log.LogErrorFileInfo.Delete();
                Uri vixLogUri = log.Vix.GetLogUri(log.LogDate);
                WebClient client = new WebClient()
                {
                    Credentials = new NetworkCredential()
                    {
                        UserName = ConfigurationManager.AppSettings["username"],
                        Password = ConfigurationManager.AppSettings["password"]
                    }
                };
                client.DownloadFile(vixLogUri, log.LogTempFileInfo.FullName);
                if (log.LogFileInfo.Exists)
                {
                    // only overwrite the old file if there is data in the new one
                    StreamReader reader = new StreamReader(log.LogTempFileInfo.Open(FileMode.Open));
                    reader.ReadLine();
                    bool isAtEndOfFile = reader.EndOfStream;
                    reader.Close();
                    if (!isAtEndOfFile)
                    {
                        log.LogFileInfo.Delete();
                        log.LogTempFileInfo.MoveTo(log.LogFileInfo.FullName);
                    }
                    else
                    {
                        log.LogTempFileInfo.Delete();
                    }
                }
                else
                {
                    log.LogTempFileInfo.MoveTo(log.LogFileInfo.FullName);
                }
            }
            catch (Exception e)
            {
                WriteException(log, e);
            }
            return log;
        }

        [PrincipalPermission(SecurityAction.Demand, Role = @"BUILTIN\Administrators")]
        public VixLogInfo[] CollectVixLogs(params VixLogInfo[] logs)
        {
            IEnumerable<String> vixIds = logs.Select(log => log.VixId).Distinct();
            IEnumerable<VixInfo> vixes = vixIds.Select(id => VixInfo.Load(id));
            foreach (VixLogInfo log in logs)
            {
                log.Vix = vixes.Where(vix => vix.Id == log.VixId).SingleOrDefault();
                CollectVixLog(log);
            }
            return logs;
        }

        public String GetError(VixLogInfo log)
        {
            if (!log.LogErrorFileInfo.Exists)
                return null;
            StreamReader reader = log.LogErrorFileInfo.OpenText();
            String error = reader.ReadToEnd();
            reader.Close();
            return error;
        }

        public VixLogInfo[] GetLogList(DateTime startDate, DateTime endDate, string vixId = null)
        {
            List<VixLogInfo> logs = new List<VixLogInfo>();
            IEnumerable<VixInfo> vixes;
            VixLogInfo log;
            VixLogStatus status;

            if (String.IsNullOrWhiteSpace(vixId))
                vixes = GetVixes();
            else
                vixes = new VixInfo[1] { new VixInfo(vixId) };

            if (endDate > DateTime.Today)
                endDate = DateTime.Today;
            for (DateTime i = startDate; i <= endDate; i = i.AddDays(1))
            {
                foreach (VixInfo vix in vixes)
                {
                    log = vix.GetLog(i);
                    status = log.Status;
                    if (status != VixLogStatus.Future && status != VixLogStatus.Unscheduled
                        && (vix.EndLogDate == null || vix.EndLogDate.Value >= log.LogDate))
                        logs.Add(log);
                }
            }
            return logs.ToArray();
        }

        public VixLogInfo[] GetProblems(DateTime startDate, DateTime endDate)
        {
            List<VixLogInfo> problemLogs = new List<VixLogInfo>();
            IEnumerable<VixInfo> vixes = GetVixes();
            VixLogInfo log;
            VixLogStatus status;
            for (DateTime i = startDate; i <= endDate; i = i.AddDays(1))
            {
                foreach (VixInfo vix in vixes)
                {
                    log = vix.GetLog(i);
                    status = log.Status;
                    if (status == VixLogStatus.Missing || status == VixLogStatus.Failed)
                        problemLogs.Add(log);
                }
            }
            return problemLogs.ToArray();
        }

        public Dictionary<VixInfo, VixLogInfo[]> GetProblemVixes(DateTime startDate, DateTime endDate)
        {
            Dictionary<VixInfo, VixLogInfo[]> problemVixes = new Dictionary<VixInfo,VixLogInfo[]>();
            List<VixLogInfo> problemLogs;
            VixLogInfo log;
            VixLogStatus status;
            IEnumerable<VixInfo> vixes = GetVixes();
            foreach (VixInfo vix in vixes)
            {
                problemLogs = new List<VixLogInfo>();
                for (DateTime i = startDate; i <= endDate; i = i.AddDays(1))
                {
                    log = vix.GetLog(i);
                    status = log.Status;
                    if (status == VixLogStatus.Missing || status == VixLogStatus.Failed)
                        problemLogs.Add(log);
                }
                if (problemLogs.Count() > 0)
                {
                    problemVixes.Add(vix, problemLogs.ToArray());
                }
            }
            return problemVixes;
        }

        public VixInfo[] GetVixes()
        {
            return VixInfo.LoadAll().Where(vix => vix.IsActive).ToArray();
        }

        public VixInfo[] UpdateVixes()
        {
            List<VixInfo> vixes = VixInfo.LoadAll().ToList();
            ImagingExchangeSiteServiceSoapClient client = new ImagingExchangeSiteServiceSoapClient();
            IEnumerable<ImagingExchangeSiteTO> vixSites = from site in client.getImagingExchangeSites()
                                                          where !String.IsNullOrWhiteSpace(site.acceleratorServer)
                                                          select site;
            foreach (ImagingExchangeSiteTO site in vixSites)
            {
                VixInfo vix = vixes.Where(v => v.Id == site.siteNumber).SingleOrDefault();
                if (vix == null)
                {
                    vix = new VixInfo()
                    {
                        Id = site.siteNumber,
                        StartLogDate = DateTime.Today.Date,
                    };
                    vixes.Add(vix);
                }
                vix.Name = site.siteName;
                vix.Server = site.acceleratorServer;
                vix.Port = site.acceleratorPort;
                vix.Save();
            }
            return vixes.ToArray();
        }

        #endregion

        #region Private Methods

        private void WriteException(VixLogInfo log, Exception ex)
        {
            FileInfo errorFile = log.LogErrorFileInfo;
            if (errorFile.Exists)
                errorFile.Delete();
            using (StreamWriter writer = errorFile.CreateText())
            {
                writer.WriteLine(ex.ToString());
                writer.Flush();
                writer.Close();
            }
        }

        #endregion
    }
}
