/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixLogErrorCollectorUtility;
using System.Configuration;
using System.IO;

namespace VixLogErrorConsole
{
    class Program
    {
        static void Main(string[] args)
        {
            StreamWriter writer = null;
            string smtpServer = null;
            bool useIIS = false;
            try
            {
                string errorLog = Utils.getApplicationDir() + @"\errorLog.txt";
                Uri uri = new Uri(errorLog);                
                FileInfo file = new FileInfo(uri.LocalPath);
                writer = new StreamWriter(file.FullName, true);
                Program p = new Program();
                string logDirectory = ConfigurationManager.AppSettings["LogDirectory"];
                string xmlDirectory = ConfigurationManager.AppSettings["XmlDirectory"];
                string highPriorityExceptions = ConfigurationManager.AppSettings["HighPriorityExceptions"];
                string emailTo = ConfigurationManager.AppSettings["EmailUsers"];
                useIIS = bool.Parse(ConfigurationManager.AppSettings["UseIISForEmail"]);
                smtpServer = ConfigurationManager.AppSettings["SMTPServer"];
                string webServerPath = ConfigurationManager.AppSettings["WebServerPath"];
                bool outputToXml = bool.Parse(ConfigurationManager.AppSettings["OutputToXml"]);
                string excludeSites = ConfigurationManager.AppSettings["ExcludeSites"];

                string sysLogServer = ConfigurationManager.AppSettings["SysLogServer"];
                SysLogClient.Client sysLogClient = new SysLogClient.Client(sysLogServer);

                if (args.Length == 2)
                {
                    if (args[0].Equals("test", StringComparison.CurrentCultureIgnoreCase))
                    {
                        string [] addresses = new string[1];
                        addresses[0] = args[1];
                        writer.WriteLine("Sending test email to '" + addresses[0] + "', SMTP server [" + smtpServer + "], useIIS [" + useIIS + "]");
                        TransactionLogEmailer.SendTestEmail(addresses, useIIS, smtpServer);
                        return;
                    }
                }
                Console.WriteLine("Excluding collecting logs from '" + excludeSites + "'.");
                string[] excludedSites = null;
                if ((excludeSites != null) && (excludeSites.Length > 0))
                {
                    excludedSites = excludeSites.Split(',');
                }
                List<string> failedSites = p.LoadSelectedTransactionLogEntries(writer, 
                    logDirectory, excludedSites, sysLogClient);
                if (!Directory.Exists(xmlDirectory))
                    Directory.CreateDirectory(xmlDirectory);
                if (outputToXml)
                {
                    Console.WriteLine("Outputting log data to XML");
                    TransactionLogXmlOutput.OutputToXml(p.transactionLogEntries, xmlDirectory);
                }

                
                if ((highPriorityExceptions != null) && (highPriorityExceptions.Length > 0) &&
                    (emailTo != null) && (emailTo.Length > 0))
                {
                    string[] exceptions = highPriorityExceptions.Split(',');
                    string[] emailUsers = emailTo.Split(';');
                    List<TransactionLogEntry> highPriorityEntries = p.extractHighPriorityErrors(exceptions, 
                        true, true, true);
                    if (highPriorityEntries.Count > 0)
                    {
                        DateTime endDate = DateTime.Now;
                        endDate = endDate.AddDays(-1);
                        TransactionLogEmailer.EmailTransactions(highPriorityEntries, emailUsers, endDate,
                            useIIS, smtpServer, webServerPath);
                    }
                }
                if ((failedSites != null) && (failedSites.Count > 0))
                {
                    StringBuilder exMsg = new StringBuilder();
                    exMsg.AppendLine("Processing failed for the following sites:");
                    foreach (string failedSite in failedSites)
                    {
                        exMsg.AppendLine(failedSite);
                    }
                    sendProcessingErrorEmail(writer, new Exception(exMsg.ToString()), DateTime.Now, smtpServer, useIIS);
                }
            }
            catch (Exception ex)
            {
                DateTime errorDate = DateTime.Now;
                writer.WriteLine("Error Collecting logs on " + errorDate.ToString() + ": " + ex.Message);
                writer.WriteLine(ex.StackTrace);
                writer.WriteLine("---------------------------------");
                sendProcessingErrorEmail(writer, ex, errorDate, smtpServer, useIIS);
            }
            finally
            {
                if (writer != null)
                    writer.Close();
            }
        }

        private static void sendProcessingErrorEmail(StreamWriter writer, Exception ex, DateTime errorDate,
            string smtpServer, bool useIIS)
        {
            try
            {
                string emailTo = ConfigurationManager.AppSettings["ErrorEmailUsers"];
                string[] emailUsers = emailTo.Split(';');
                TransactionLogEmailer.EmailProcessingErrorMessage(emailUsers, ex, errorDate, useIIS, smtpServer);
            }
            catch (Exception ex2)
            {
                if (writer != null)
                {
                    writer.WriteLine("Error Sending email for processing error on " + errorDate.ToString() + ": " + ex2.Message);
                    writer.WriteLine(ex2.StackTrace);
                    writer.WriteLine("---------------------------------");
                }
            }
        }

        public List<TransactionLogEntry> transactionLogEntries = new List<TransactionLogEntry>();

        /// <summary>
        /// 
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="logDir"></param>
        /// <returns>List of sites that failed, list will be empty if no sites fail</returns>
        private List<string> LoadSelectedTransactionLogEntries(StreamWriter writer, string logDir,
            string[] excludedSites, SysLogClient.Client sysLogClient)
        {           
            transactionLogEntries = new List<TransactionLogEntry>();
            DateTime startDate = DateTime.Today;
            DateTime endDate = DateTime.Today;
            startDate = clearDateTime(startDate.AddDays(-1));
            endDate = clearDateTime( endDate.AddDays(-1));
            List<VistaSite> sites = SiteServiceHelper.GetVixSites();
            List<string> failedSiteList = new List<string>();
            foreach (VistaSite site in sites)
            {
                bool skipSite = false;
                if (excludedSites != null)
                {
                    foreach (string excludedSite in excludedSites)
                    {
                        if (site.SiteNumber == excludedSite)
                        {
                            skipSite = true;
                            break;
                        }
                    }
                }
                if (!skipSite)
                {
                    Console.WriteLine("Loading logs for site [" + site + "]");
                    try
                    {
                        TransactionLogDownloader.DownloadLog(site, startDate, endDate, logDir);
                        transactionLogEntries.AddRange(TransactionLogParser.ParseLogs(logDir, site.SiteNumber, startDate, endDate));
                        SendSysLogMessage(sysLogClient, writer, site, null);
                    }
                    catch (Exception ex)
                    {
                        failedSiteList.Add(site.SiteNumber);
                        writer.WriteLine("Error loading logs for site '" + site.SiteNumber + "', " + ex.Message);
                        SendSysLogMessage(sysLogClient, writer, site, ex.Message);
                    }
                }
            }
            return failedSiteList;
        }

        private void SendSysLogMessage(SysLogClient.Client sysLogClient, StreamWriter writer, 
            VistaSite site, string errorMessage)
        {
            try
            {
                int facility = (int)SysLogClient.SysLogFacility.User;
                int level = (int)SysLogClient.SysLogLevel.Information;
                if (errorMessage != null)
                {
                    level = (int)SysLogClient.SysLogLevel.Error;
                }
                string msg = CreateSysLogXmlMessage(site, errorMessage);
                sysLogClient.Send(new SysLogClient.SysLogMessage(facility, level, msg.ToString()));
            }
            catch (Exception ex)
            {
                writer.WriteLine("Error sending syslog message for site '" + site.SiteNumber + "', " + ex.Message);
            }
        }

        private static string CreateSysLogXmlMessage(VistaSite site, string errorMessage)
        {
            bool success = true;
            if(errorMessage != null)
                success = false;
            StringBuilder xml = new StringBuilder();
            xml.Append("<msg type=\"VIXLogErrorCollector\">");
            xml.Append("<msgDetails>");
            xml.Append("<site name=\"" + site.SiteName + "\" number=\"" + site.SiteNumber + "\" />");
            xml.Append("<result success=\"" + success + "\">");
            if (!success)
            {
                xml.Append(errorMessage);
            }
            xml.Append("</result>");
            xml.Append("</msgDetails>");
            xml.Append("</msg>");
            return xml.ToString();
        }

        private List<TransactionLogEntry> extractHighPriorityErrors(string []exceptions, bool longTransactionsAsErrors, 
            bool discardClientAbortExceptions, bool discardTextFileRequestsToDoD)
        {
            // use a dictionary to ensure only a single entry per transaction ID
            Dictionary<string, TransactionLogEntry> highPriorityEntryDictionary = 
                new Dictionary<string, TransactionLogEntry>();
            
            foreach (TransactionLogEntry entry in transactionLogEntries)
            {
                bool entryLogged = false;
                if (entry.IsError())
                {
                    foreach (string exception in exceptions)
                    {
                        if (exception.Equals(entry.ExceptionClassName))
                        {
                            if ((entry.ErrorMessage.Contains("org.apache.catalina.connector.ClientAbortException")) 
                                && (discardClientAbortExceptions))
                            {
                                break;
                            }
                            else if ((discardTextFileRequestsToDoD) 
                                && (entry.ErrorMessage.Contains("Exception when requesting exam image text file: No input stream returned from data source for TXT file"))
                                && (entry.RespondingSite == "200"))
                            {
                                // exclude text file requests from VistARad to the DoD
                                break;
                            }
                            if (!highPriorityEntryDictionary.ContainsKey(entry.TransactionId))
                            {
                                highPriorityEntryDictionary.Add(entry.TransactionId, entry);
                            }
                            entryLogged = true;
                            break;
                        }

                    }
                    if (!entryLogged)
                    {
                        if ((entry.IsLongVixTransaction()) && longTransactionsAsErrors)
                        {
                            if (!highPriorityEntryDictionary.ContainsKey(entry.TransactionId))
                            {
                                highPriorityEntryDictionary.Add(entry.TransactionId, entry);
                            }
                        }
                    }
                }
            }

            List<TransactionLogEntry> highPriorityEntries = new List<TransactionLogEntry>();
            foreach (string key in highPriorityEntryDictionary.Keys)
            {
                highPriorityEntries.Add(highPriorityEntryDictionary[key]);
            }
            return highPriorityEntries;
        }

        private DateTime clearDateTime(DateTime dt)
        {
            DateTime d = new DateTime(dt.Year, dt.Month, dt.Day);
            return d;
        }

        

    }
}
