/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using VixMonitorBusiness.Vix.Monitor.Business;
using System.Diagnostics;

namespace VixMonitor
{
    public partial class frmMain : Form
    {
        public frmMain()
        {
            InitializeComponent();
        }

        private const string TEST_TIME_FORMAT = "MMM dd, yyyy hh:mm:ss tt";

        private SiteServiceHelper siteServiceHelper = null;

        private void frmMain_Load(object sender, EventArgs e)
        {
            siteServiceHelper = new SiteServiceHelper();
            try
            {
                List<VixServer> servers = siteServiceHelper.GetVixServers();
                for (int i = 0; i < servers.Count; i++)
                {
                    AddServerToTree(servers[i]);
                }
                this.Text = "ViX Monitor - " + siteServiceHelper.SiteServiceUrl;
                
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message + "\n\n" + ex.StackTrace,"Error loading site list",  MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void AddServerToTree(VixServer server)
        {
            TreeNode node = new TreeNode();
            node.Text = server.SiteName;
            node.Tag = server;
            node.ImageIndex = -1;
            node.SelectedImageIndex = node.ImageIndex;
            treServers.Nodes.Add(node);
        }

        private void TestVix(TreeNode vixNode, bool display)
        {
            VixServer vixServer = (VixServer)vixNode.Tag;
            VixTestResultGroup testResult = VixTesterHelper.TestVix(vixServer, true, 90, true);
            vixServer.ServerTests.Add(testResult);

            TreeNode resultNode = AddResultNodeToVixNode(vixNode, testResult);

            if (display)
            {
                treServers.SelectedNode = resultNode;
            }
        }

        private TreeNode AddResultNodeToVixNode(TreeNode vixNode, VixTestResultGroup testResult)
        {
            TreeNode resultNode = new TreeNode();
            resultNode.Tag = testResult;
            resultNode.Text = testResult.TestTime.ToString(TEST_TIME_FORMAT);
            resultNode.ImageIndex = getTestResultImageIndex(testResult.TestsPassed);
            resultNode.SelectedImageIndex = resultNode.ImageIndex;
            vixNode.Nodes.Add(resultNode);
            vixNode.Expand();
            vixNode.ImageIndex = resultNode.ImageIndex;
            vixNode.SelectedImageIndex = resultNode.SelectedImageIndex;
            return resultNode;
        }

        private int getTestResultImageIndex(bool result)
        {
            if (result)
            {
                return 0;
            }
            return 1;
        }

        private void treServers_NodeMouseDoubleClick(object sender, TreeNodeMouseClickEventArgs e)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                TestVix(e.Node, true);
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void treServers_AfterSelect(object sender, TreeViewEventArgs e)
        {
            if (e.Node.Tag is VixServer)
            {
                VixServer server = (VixServer)e.Node.Tag;
                DisplayVixDetails(server);
            }
            else if (e.Node.Tag is VixTestResultGroup)
            {
                VixTestResultGroup resultGroup = (VixTestResultGroup)e.Node.Tag;
                VixServer vixServer = (VixServer)e.Node.Parent.Tag;
                DisplayVixDetails(vixServer);
                DisplayTestResults(resultGroup);
            }
        }

        private void DisplayVixDetails(VixServer vixServer)
        {
            txtSiteName.Text = vixServer.SiteName;
            txtVixDetails.Text = vixServer.VixHost + ", " + vixServer.VixPort;
            ClearTestResults();
        }

        private void ClearTestResults()
        {
            txtTestTime.Text = "";
            lstTestResults.Items.Clear();
        }

        private void DisplayTestResults(VixTestResultGroup resultGroup)
        {
            ClearTestResults();
            txtTestTime.Text = resultGroup.TestTime.ToString(TEST_TIME_FORMAT);
            for (int i = 0; i < resultGroup.VixTests.Count; i++)
            {
                ListViewItem item = new ListViewItem();
                item.Text = resultGroup.VixTests[i].TestTypeString;
                item.SubItems.Add(resultGroup.VixTests[i].TestResult.ToString());
                string msg = resultGroup.VixTests[i].TestResultMessage;
                msg = msg.Replace("\n", ", ");
                item.SubItems.Add(msg);
                lstTestResults.Items.Add(item);
            }
        }

        private void btnTestAllSites_Click(object sender, EventArgs e)
        {
            try
            {
                Cursor.Current = Cursors.WaitCursor;
                for (int i = 0; i < treServers.Nodes.Count; i++)
                {
                    TestVix(treServers.Nodes[i], false);
                }
            }
            finally
            {
                Cursor.Current = Cursors.Default;
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
            
        }


        private void ExportTestResults(string filename, ExportFormat format)
        {
            List<VixServer> vixServers = new List<VixServer>();
            for (int i = 0; i < treServers.Nodes.Count; i++)
            {
                VixServer vixServer = (VixServer)treServers.Nodes[i].Tag;
                vixServers.Add(vixServer);
            }

            try
            {
                if (format == ExportFormat.XML)
                {
                    TestResultsOutputBuilder.SaveResultsAsXml(filename, vixServers);
                }
                else
                {
                    TestResultsOutputBuilder.SaveResultsAsHtml(filename, vixServers,false, true);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message + "\n\n" + ex.StackTrace, "Error saving to " + format.ToString(), MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void mnuFileExportXML_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "XML Files (*.xml)|*.xml";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportTestResults(saveFileDialog1.FileName, ExportFormat.XML);
            }
        }
        enum ExportFormat
        {
            XML, HTML
        }

        private void mnuFileExportHTML_Click(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "HTML Files (*.html)|*.html";
            if (saveFileDialog1.ShowDialog() == DialogResult.OK)
            {
                ExportTestResults(saveFileDialog1.FileName, ExportFormat.HTML);
                if (System.IO.File.Exists(saveFileDialog1.FileName))
                {
                    Process proc = new Process();
                    proc.StartInfo.FileName = saveFileDialog1.FileName;
                    proc.Start();
                }
            }
        }

        private void mnuFileImportXML_Click(object sender, EventArgs e)
        {
            openFileDialog1.Filter = "XML Files (*.xml)|*.xml";
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                List<VixServer> vixServers = TestResultsOutputBuilder.LoadResultsFromXML(openFileDialog1.FileName);
                for (int i = 0; i < vixServers.Count; i++)
                {
                    foreach (TreeNode node in treServers.Nodes)
                    {
                        if (node.Tag is VixServer)
                        {
                            VixServer nodeServer = (VixServer)node.Tag;
                            if (nodeServer.SiteNumber == vixServers[i].SiteNumber)
                            {
                                for (int j = 0; j < vixServers[i].ServerTests.Count; j++)
                                {
                                    VixTestResultGroup result = vixServers[i].ServerTests[j];
                                    AddResultNodeToVixNode(node, result);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
}
