/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using System.IO;

namespace VixMonitorBusiness.Vix.Monitor.Business
{
    public class TestResultsOutputBuilder
    {
        /// <summary>
        /// Removes all but the last test result for each ViX server 
        /// from the specified file
        /// </summary>
        /// <param name="filename">The Xml file to prune</param>
        public static void PruneXmlResults(string filename)
        {
            if (File.Exists(filename))
            {
                File.Copy(filename, filename + ".old", true);
                List<VixServer> servers = LoadResultsFromXML(filename);
                foreach (VixServer server in servers)
                {
                    VixTestResultGroup lastTestResult = server.GetMostRecentTest();
                    server.ServerTests.Clear();
                    server.ServerTests.Add(lastTestResult);
                }
                SaveResultsAsXml(filename, servers);
            }
        }


        public static List<VixServer> LoadResultsFromXML(string filename)
        {
            List<VixServer> vixServers = new List<VixServer>();
            if (!File.Exists(filename))
            {
                return vixServers;
            }
            XmlDocument xmlDoc = new XmlDocument();
            xmlDoc.Load(filename);

            XmlNodeList siteNodeList =  xmlDoc.SelectNodes("/VixMonitor/Sites/Site");
            foreach (XmlNode siteNode in siteNodeList)
            {
                string siteName = siteNode.Attributes["Name"].Value;
                string siteNumber = siteNode.Attributes["SiteNumber"].Value;
                string vixHost = siteNode.Attributes["VixHost"].Value;
                int vixPort = int.Parse(siteNode.Attributes["VixPort"].Value);
                string siteAbbreviation = siteNode.Attributes["SiteAbbreviation"].Value;
                string vistaServer = "";
                int vistaPort = 0;
                if (siteNode.Attributes["VistaHost"] != null)
                {
                    vistaServer = siteNode.Attributes["VistaHost"].Value;
                }
                if (siteNode.Attributes["VistaPort"] != null)
                {
                    vistaPort = int.Parse( siteNode.Attributes["VistaPort"].Value);
                }

                VixServer vixServer = new VixServer(siteName, siteNumber, siteAbbreviation, vistaServer, vistaPort, vixHost, vixPort);
                vixServers.Add(vixServer);
                XmlNodeList testRunList = siteNode.SelectNodes("Tests/TestRun");
                foreach (XmlNode testRun in testRunList)
                {
                    string testTime = testRun.Attributes["TestTime"].Value;
                    VixTestResultGroup resultGroup = new VixTestResultGroup(DateTime.Parse(testTime));
                    vixServer.ServerTests.Add(resultGroup);
                    XmlNodeList testNodeList = testRun.SelectNodes("Test");
                    foreach (XmlNode testNode in testNodeList)
                    {
                        VixTestResult result = null;
                        string strTestType = testNode.Attributes["TestType"].Value;

                        switch (strTestType)
                        {
                            case "PING":
                                result = new VixTestResult(VixTestType.PING, true);
                                break;
                            case "TOMCAT":
                                result = new VixTestResult(VixTestType.TOMCAT, true);
                                break;
                            case "FAILED_QUERY":
                                result = new VixTestResult(VixTestType.FAILED_QUERY, true);
                                break;
                            case "IDS_QUERY":
                                result = new VixTestResult(VixTestType.IDS_QUERY, true);
                                break;
                            case "VIX_HEALTH":
                                result = new VixTestResult(VixTestType.VIX_HEALTH, false);
                                break;
                            case "VIX_DETAILS":
                                result = new VixTestResult(VixTestType.VIX_DETAILS, false);
                                break;
                        }

                        if (result != null)
                        {
                            result.TestResult = bool.Parse(testNode.Attributes["TestResult"].Value);
                            result.TestResultMessage = testNode.Attributes["TestResultMessage"].Value;
                            resultGroup.VixTests.Add(result);
                        }
                    }
                }
            }
            return vixServers;
        }

        public static void SaveResultsAsXml(string filename, List<VixServer> vixServers)
        {
            XmlDocument xmlDoc = new XmlDocument();

            XmlNode rootNode = xmlDoc.CreateElement("VixMonitor");
            xmlDoc.AppendChild(rootNode);

            XmlNode sitesNode = xmlDoc.CreateElement("Sites");
            rootNode.AppendChild(sitesNode);

            for (int i = 0; i < vixServers.Count; i++)
            {
                XmlNode siteNode = convertSiteResultsToXml(vixServers[i], xmlDoc);
                sitesNode.AppendChild(siteNode);
            }

            xmlDoc.Save(filename);
        }

        public static void SaveResultsAsHtml(string filename, List<VixServer> vixServers, bool showAsStatusChange, bool includeVixDetails)
        {
            StringBuilder sb = new StringBuilder();
            SaveResultsAsHtml(sb, vixServers, showAsStatusChange, includeVixDetails);
            StreamWriter writer = new StreamWriter(filename);

            writer.Write(sb.ToString());
            writer.Close();
            
        }

        public static void SaveResultsAsPlainText(StringBuilder sb, List<VixServer> vixServers, bool showAsStatusChange)
        {
            foreach (VixServer server in vixServers)
            {
                //sb.AppendLine(server.SiteName + " [" + server.VixHost + ", " + server.VixPort + "]");
                sb.AppendLine(server.SiteName);
                
                foreach (VixTestResultGroup result in server.ServerTests)
                {
                    sb.AppendLine("Tests:");
                    if (showAsStatusChange)
                    {
                        sb.AppendLine("\tStatus Changed");
                    }
                    foreach (VixTestResult test in result.VixTests)
                    {
                        sb.AppendLine(test.TestTypeString + ": " + test.TestResult + " [" + test.TestResultMessage + "]");
                    }
                }
                sb.AppendLine("--------------");
            }
        }

        public static void SaveResultsAsHtml(StringBuilder sb, List<VixServer> vixServers, bool showAsStatusChange, bool includeVixDetails)
        {
            sb.AppendLine( "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">");
            sb.AppendLine("<html>");
            sb.AppendLine("<HTML><HEAD><META http-equiv=Content-Type content=\"text/html; charset=iso-8859-1\"></HEAD>");
            sb.AppendLine("<body>");

            if (includeVixDetails)
            {
                sb.AppendLine("<h2>VIX Details:</h2>");
                sb.AppendLine("<table border=1>");
                sb.AppendLine("<tr>");
                sb.AppendLine("<th>Site</th>");
                //sb.AppendLine("<th>Hostname</th>");
                sb.AppendLine("<th>VIX Version</th>");
                sb.AppendLine("<th>Startup Time</th>");
                sb.AppendLine("<th>Up Time</th>");
                sb.AppendLine("<th>Java Log Size</th>");
                sb.AppendLine("<th>Transaction Log Size</th>");
                sb.AppendLine("<th>Transactions Written</th>");
                sb.AppendLine("<th>Cache Size</th>");
                sb.AppendLine("<th>Cache % Full</th>");
                sb.AppendLine("</tr>");
                foreach (VixServer vixServer in vixServers)
                {
                    if (vixServer.ServerTests.Count > 0)
                    {
                        VixHealthMonitorCommon.VixServerHealth vixHealth = vixServer.ServerTests[0].VixHealth;
                        if (vixHealth != null)
                        {
                            sb.AppendLine("<tr>");
                            sb.AppendLine("<td>" + vixServer.SiteName + "</td>");
                            //sb.AppendLine("<td>" + vixServer.VixHost + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetVixVersion() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetVixJVMStartupTime() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetVixJVMUptimeShort() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetTomcatLogsDirSizeFormatted() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetTransactionLogsDirSizeFormatted() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetTransactionLogStatisticsTransactionsWrittenFormatted() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetVixCacheDirSizeFormatted() + "</td>");
                            sb.AppendLine("<td>" + vixHealth.GetVixCacheDirPercentFull().ToString("N2") + "%" + "</td>");
                            sb.AppendLine("</tr>");
                        }
                    }
                }
                sb.AppendLine("</table>");
            }

            for (int i = 0; i < vixServers.Count; i++)
            {
                VixServer server = vixServers[i];
                sb.AppendLine("<h2>" + server.SiteName + "</h2>");
                //sb.AppendLine("<p>" + server.VixHost + ", " + server.VixPort + "</p>");
                for (int j = 0; j < server.ServerTests.Count; j++)
                {
                    VixTestResultGroup result = server.ServerTests[j];
                    sb.AppendLine("<b>Tests:</b> " + result.TestTime.ToString());
                    if (showAsStatusChange)
                    {
                        sb.AppendLine("<br><i>Status Changed</i>");
                    }
                    sb.AppendLine("<table border=1>");
                    sb.AppendLine("<tr><th>Test Type</th><th>Test Passed</th><th>Test Result Message</th></tr>");
                    for (int k = 0; k < result.VixTests.Count; k++)
                    {
                        VixTestResult test = result.VixTests[k];
                        string msg = test.TestResultMessage;
                        if (msg.Length <= 0)
                        {
                            msg = "&nbsp;";
                        }
                        msg = msg.Replace("\n", "<br>");
                        if (test.TestResult)
                        {
                            sb.AppendLine("<tr><td>" + test.TestTypeString + "</td><td>" + test.TestResult.ToString() + "</td><td>" + msg + "</td></tr>");
                        }
                        else
                        {
                            sb.AppendLine("<tr bgcolor=red><td>" + test.TestTypeString + "</td><td>" + test.TestResult.ToString() + "</td><td>" + msg + "</td></tr>");
                        }
                    }
                    sb.AppendLine("</table>");
                }
            }

            //sb.AppendLine("<table border=1>");

            sb.AppendLine("</body>");
            sb.AppendLine("</html>");


        }

        private static XmlNode convertSiteResultsToXml(VixServer vixServer, XmlDocument xmlDoc)
        {
            XmlNode siteNode = xmlDoc.CreateElement("Site");
            XmlAttribute attr = xmlDoc.CreateAttribute("Name");
            attr.Value = vixServer.SiteName;
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("SiteNumber");
            attr.Value = vixServer.SiteNumber;
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("VixHost");
            attr.Value = vixServer.VixHost;
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("VixPort");
            attr.Value = vixServer.VixPort.ToString();
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("SiteAbbreviation");
            attr.Value = vixServer.SiteAbbreviation;
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("VistaHost");
            attr.Value = vixServer.VistaHost;
            siteNode.Attributes.Append(attr);
            attr = xmlDoc.CreateAttribute("VistaPort");
            attr.Value = vixServer.VistaPort.ToString();
            siteNode.Attributes.Append(attr);

            XmlNode testsNode = xmlDoc.CreateElement("Tests");
            siteNode.AppendChild(testsNode);

            for (int i = 0; i < vixServer.ServerTests.Count; i++)
            {
                XmlNode testRunNode = xmlDoc.CreateElement("TestRun");
                testsNode.AppendChild(testRunNode);
                attr = xmlDoc.CreateAttribute("TestTime");
                attr.Value = vixServer.ServerTests[i].TestTime.ToString();
                testRunNode.Attributes.Append(attr);
                for (int j = 0; j < vixServer.ServerTests[i].VixTests.Count; j++)
                {
                    VixTestResult testResult = vixServer.ServerTests[i].VixTests[j];
                    if (testResult.TestType != VixTestType.VIX_DETAILS)
                    {
                        XmlNode testNode = xmlDoc.CreateElement("Test");
                        testRunNode.AppendChild(testNode);

                        attr = xmlDoc.CreateAttribute("TestType");
                        attr.Value = testResult.TestType.ToString();
                        testNode.Attributes.Append(attr);
                        attr = xmlDoc.CreateAttribute("TestResult");
                        attr.Value = testResult.TestResult.ToString();
                        testNode.Attributes.Append(attr);
                        attr = xmlDoc.CreateAttribute("TestResultMessage");
                        attr.Value = testResult.TestResultMessage;
                        testNode.Attributes.Append(attr);
                    }
                }
            }

            return siteNode;
        }
    }

    /*
    public class VixSiteResultGroup
    {
        public readonly VixServer vixServer;
        public readonly List<VixTestResultGroup> testResults = new List<VixTestResultGroup>();

        public VixSiteResultGroup(VixServer server)
        {
            this.vixServer = server;
        }
    }
     */ 
}
