/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using System.Net.NetworkInformation;
using System.Net;
using System.IO;
using System.Configuration;
using System.Threading;
using System.Xml;

namespace VixMonitorBusiness.Vix.Monitor.Business
{
    public class VixTesterHelper
    {
        private static int TEST_TRIES = 3;
        private static int TEST_SLEEP_TIME = 1000 * 60; // 60 seconds

        public static VixMonitorLogger logger = null;

        public static VixTestResultGroup TestVix(VixServer vixServer, bool executePingTest,
            int vixHealthTimeout, bool testVixHealth)
        {
            VixTestResultGroup group = null;// new VixTestResultGroup(DateTime.Now);
            bool passed = false;
            int testCount = 0;
            while ((!passed) && (testCount < TEST_TRIES))
            {
                LogMsg("Testing VIX '" + vixServer + "', current test count iteration " + testCount);
                group = TestVixInternal(vixServer, executePingTest, vixHealthTimeout, testVixHealth);
                testCount++;
                
                passed = group.TestsPassed;
                if (!passed)
                {
                    if (group.RepeatTests)
                    {
                        LogMsg("At least 1 test failed of server '" + vixServer + "' on test count iteration " + testCount, true);
                        Thread.Sleep(TEST_SLEEP_TIME);
                    }
                }
            }            
            return group;
        }

        private static VixTestResultGroup TestVixInternal(VixServer vixServer,
            bool executePingTest, int vixHealthTimeout, bool testVixHealth)
        {
            VixTestResultGroup group = new VixTestResultGroup(DateTime.Now);
            if (executePingTest)
            {
                VixTestResult pingResult = TestVixPing(vixServer);
                group.VixTests.Add(pingResult);
            }
            else
            {
                group.VixTests.Add(new VixTestResult(VixTestType.PING, true, "Test Skipped", false));
            }

            VixTestResult tomcatResult = TestTomcatQuery(vixServer);
            group.VixTests.Add(tomcatResult);

            /*
            VixTestResult vixResult = TestFailedVixQuery(vixServer);
            group.VixTests.Add(vixResult);
             */
            VixTestResult vixResult = TestIDSQuery(vixServer);
            group.VixTests.Add(vixResult);

            if (group.TestsPassed)
            {
                if (testVixHealth)
                {
                    // only do health test if all other tests passed
                    VixHealthTestResult healthResults = TestVixLongRunningThreads(vixServer,
                        vixHealthTimeout);
                    group.VixHealth = healthResults.VixHealth;
                    group.VixTests.Add(healthResults);
                }
                else
                {
                    group.VixTests.Add(new VixTestResult(VixTestType.VIX_HEALTH, true, "Test Skipped", false));
                    group.VixHealth = null;
                }
            }

            return group;
        }

        private static VixTestResult TestTomcatQuery(VixServer vixServer)
        {
            VixTestResult result = new VixTestResult(VixTestType.TOMCAT, true);
            LogTestStarted(vixServer, VixTestType.TOMCAT);
            try
            {
                //WebClient client = new WebClient();
                string vixServerBaseUrl = "http://" + vixServer.VixHost + ":" + vixServer.VixPort + "/";
                //Stream readStream = client.OpenRead(vixServerBaseUrl);
                WebRequest request = WebRequest.Create(vixServerBaseUrl);
                request.Timeout = 1000 * 30; // 30 seconds
                Stream readStream = request.GetResponse().GetResponseStream();
                if (readStream.CanRead)
                {
                    readStream.Close();
                    result.TestResult = true;
                    LogTestSuccessful(vixServer, VixTestType.TOMCAT);                    
                }
                
            }
            catch(WebException ex)            
            {
                
                result.TestResult = false;
                result.TestResultMessage = ex.Message;
                LogFailedTestMsg(vixServer, VixTestType.TOMCAT.ToString(), ex.Message);
            }
            return result;
        }

        private static VixTestResult TestIDSQuery(VixServer vixServer)
        {
            VixTestResult result = new VixTestResult(VixTestType.IDS_QUERY,true);
            LogTestStarted(vixServer, VixTestType.IDS_QUERY);
            try
            {
                WebClient client = new WebClient();
                
                string vixApplicationPathUrl = "http://" + vixServer.VixHost + ":" + vixServer.VixPort + "/" +
                    ConfigurationManager.AppSettings["IDSWebApplicationPath"] +
                    ConfigurationManager.AppSettings["IDSServicePath"];

                WebRequest request = WebRequest.Create(vixApplicationPathUrl);
                request.Timeout = 1000 * 30; // 30 seconds
                Stream readStream = request.GetResponse().GetResponseStream();

                //Stream readStream = client.OpenRead(vixApplicationPathUrl);
                readStream.Close();
                result.TestResult = true;
                LogTestSuccessful(vixServer, VixTestType.IDS_QUERY);
            }
            catch (WebException ex)
            {
                result.TestResult = false;
                result.TestResultMessage = ex.Message;
                LogFailedTestMsg(vixServer, VixTestType.IDS_QUERY.ToString(), ex.Message);
            }
            return result;
        }

        /// <summary>
        /// This method is deprecated - no longer used in favor of other, better methods
        /// </summary>
        /// <param name="vixServer"></param>
        /// <returns></returns>
        private static VixTestResult TestFailedVixQuery(VixServer vixServer)
        {
            VixTestResult result = new VixTestResult(VixTestType.FAILED_QUERY, true);
            try
            {
                WebClient client = new WebClient();
                string vixApplicationPathUrl = "http://" + vixServer.VixHost + ":" + vixServer.VixPort + "/" + 
                    ConfigurationManager.AppSettings["ExchangeWebApplicationPath"] + 
                    ConfigurationManager.AppSettings["VixWebservicePath"];
                Stream readStream = client.OpenRead(vixApplicationPathUrl);
                result.TestResult = false;
                result.TestResultMessage = "Exception was not thrown";

            }
            catch (WebException ex)
            {
                if (ex.Status == WebExceptionStatus.ProtocolError)
                {
                    if (ex.Response is HttpWebResponse)
                    {
                        HttpWebResponse res = (HttpWebResponse)ex.Response;
                        if (res.StatusCode == HttpStatusCode.Unauthorized)
                        {
                            result.TestResult = true;
                        }
                        else
                        {
                            result.TestResult = false;
                            result.TestResultMessage = (int)res.StatusCode + ", " + res.StatusDescription;
                        }
                    }
                    else
                    {
                        result.TestResult = false;
                        result.TestResultMessage = ex.Message;
                    }                    
                }
                else
                {
                    result.TestResult = false;
                    result.TestResultMessage = ex.Message;
                }
            }
            return result;
        }


        private static VixTestResult TestVixPing(VixServer vixServer)
        {
            VixTestResult result = new VixTestResult(VixTestType.PING, true);
            LogTestStarted(vixServer, VixTestType.PING);
            try
            {
                Ping ping = new Ping();
                
                PingReply reply = ping.Send(vixServer.VixHost, 1000 * 30);                
                if (reply.Status == IPStatus.Success)
                {
                    result.TestResult = true;
                    LogTestSuccessful(vixServer, VixTestType.PING);
                }
                else
                {
                    result.TestResult = false;
                    result.TestResultMessage = reply.Status.ToString();
                    LogFailedTestMsg(vixServer, VixTestType.PING.ToString(), reply.Status.ToString());
                }
            }
            catch (Exception ex)
            {
                result.TestResult = false;
                result.TestResultMessage = ex.Message;
                LogFailedTestMsg(vixServer, VixTestType.PING.ToString(), ex.Message);
            }
            return result;
        }

        private static VixHealthTestResult TestVixLongRunningThreads(VixServer vixServer, int vixHealthTimeout)
        {
            VixHealthTestResult result = new VixHealthTestResult(VixTestType.VIX_HEALTH, false);
            LogTestStarted(vixServer, VixTestType.VIX_HEALTH);
            try
            {

                VixHealthMonitorCommon.VASite site = new VixHealthMonitorCommon.VASite(vixServer.SiteName, 
                    vixServer.SiteNumber, vixServer.SiteAbbreviation, vixServer.VistaHost, 
                    vixServer.VistaPort, vixServer.VixHost, vixServer.VixPort);

                site.vixServerHealth.UpdateSiteHealth(1000 * vixHealthTimeout);
                result.VixHealth = site.vixServerHealth;

                if (site.vixServerHealth.ErrorMessage != null)
                {
                    result.TestResult = false;
                    result.TestResultMessage = site.vixServerHealth.ErrorMessage;
                    LogFailedTestMsg(vixServer, VixTestType.VIX_HEALTH.ToString(), site.vixServerHealth.ErrorMessage);
                }
                else
                {
                    VixHealthMonitorCommon.VixHealthMonitorSettings settings = new VixMonitorVixHealthMonitorSettings("");
                    int threadProcessingCriticalLimitMinutes = int.Parse(ConfigurationManager.AppSettings["ThreadProcessingCriticalLimit"]);

                    settings.ThreadProcessingTimeCriticalLimit = 1000 * 60 * threadProcessingCriticalLimitMinutes; // 10 minutes
                    int busyThreadCount = site.vixServerHealth.GetThreadsProcessingAboveCriticalCount(settings);
                    if (busyThreadCount > 0)
                    {
                        result.TestResult = false;
                        string msg = busyThreadCount + " threads processing longer than " + threadProcessingCriticalLimitMinutes + " minutes";
                        result.TestResultMessage = msg;
                        LogFailedTestMsg(vixServer, VixTestType.VIX_HEALTH.ToString(), msg);
                    }
                    else
                    {
                        result.TestResult = true;
                        LogTestSuccessful(vixServer, VixTestType.VIX_HEALTH);
                    }
                }
                 
            }
            catch (Exception ex)
            {
                result.TestResult = false;
                result.TestResultMessage = ex.Message;
                LogFailedTestMsg(vixServer, VixTestType.VIX_HEALTH.ToString(), ex.Message);
            }
            return result;
        }

        private static void LogTestSuccessful(VixServer vixServer, VixTestType testType)
        {
            LogMsg("Test '" + testType.ToString() + "' on VIX Server '" + vixServer + "' completed successfully.");
        }

        private static void LogTestStarted(VixServer vixServer, VixTestType testType)
        {
            LogMsg("Starting test '" + testType.ToString() + "' on VIX Server '" + vixServer + "'.");
        }

        private static void LogFailedTestMsg(VixServer vixServer, string testName, string errMsg)
        {
            LogMsg("VIX Server '" + vixServer + "' failed test '" + testName + "' with error [" + errMsg + "]", true);
        }

        private static void LogMsg(string msg)
        {
            LogMsg(msg, false);
        }
        private static void LogMsg(string msg, bool isError)
        {
            if (logger != null)
            {
                logger.LogMsg(msg, isError);
            }
        }
    }
}
