/**
 * 
  Package: MAG - VistA Imaging
  WARNING: Per VHA Directive 2004-038, this routine should not be modified.
  Date Created:
  Site Name:  Washington OI Field Office, Silver Spring, MD
  Developer:
  Description: 

        ;; +--------------------------------------------------------------------+
        ;; Property of the US Government.
        ;; No permission to copy or redistribute this software is given.
        ;; Use of unreleased versions of this software requires the user
        ;;  to execute a written test agreement with the VistA Imaging
        ;;  Development Office of the Department of Veterans Affairs,
        ;;  telephone (301) 734-0100.
        ;;
        ;; The Food and Drug Administration classifies this software as
        ;; a Class II medical device.  As such, it may not be changed
        ;; in any way.  Modifications to this software may result in an
        ;; adulterated medical device under 21CFR820, the use of which
        ;; is considered to be a violation of US Federal Statutes.
        ;; +--------------------------------------------------------------------+

 */
using System;
using System.Collections.Generic;
using System.Text;
using VixMonitorBusiness.Vix.Monitor.Business;
using System.Net.Mail;
using System.Configuration;
using System.IO;

namespace VixMonitorConsole
{
    public class VixMonitorConsoleHelper
    {
        private readonly VixMonitorLogger logger = null;

        public VixMonitorConsoleHelper(VixMonitorLogger logger)
        {
            this.logger = logger;
        }

        private static string getXmlExportFilename()
        {
            return ConfigurationManager.AppSettings["XmlExportFilename"];
        }

        public void SendEmail(ReportType reportType, string htmlMsg, string plainTextMsg)
        {
            bool useIIS = false;
            useIIS = bool.Parse(ConfigurationManager.AppSettings["UseIISForEmail"]);
            string smtpServer = ConfigurationManager.AppSettings["SMTPServer"];

            SmtpClient client = null;
            if (useIIS)
            {
                client = new SmtpClient();
                client.DeliveryMethod = SmtpDeliveryMethod.PickupDirectoryFromIis;
            }
            else
            {
                client = new SmtpClient(smtpServer);
                client.DeliveryMethod = SmtpDeliveryMethod.Network;
            }

            MailAddress msgFrom = new MailAddress("");
            //MailAddress msgTo = new MailAddress(ConfigurationManager.AppSettings["EmailUsers"]);


            MailMessage message = new MailMessage();//(msgFrom, msgTo);
            message.From = msgFrom;
            string emailTo = ConfigurationManager.AppSettings["EmailUsers"];
            LogMsg("Sending email to [" + emailTo + "]");
            string[] sendTo = emailTo.Split(';');
            for (int i = 0; i < sendTo.Length; i++)
            {
                try
                {
                    message.To.Add(sendTo[i]);
                }
                catch (Exception ex)
                {
                    LogMsg("Error adding email addresss '" + sendTo[i] + "', " + ex.Message);
                }
            }
            if (reportType == ReportType.FAILED)
            {
                message.Priority = MailPriority.High;
                message.Subject = "ViX Monitor Console Report - FAILED VIX!";
            }
            else if (reportType == ReportType.PASSED)
            {
                message.Subject = "ViX Monitor Console Report - Online";
                message.Priority = MailPriority.Normal;
            }
            else
            {
                message.Priority = MailPriority.High;
                message.Subject = "ViX Monitor Console Report - VIX Status Changed";
            }
            bool hasPlainText = false;
            if ((plainTextMsg != null) && (plainTextMsg.Length > 0))
            {
                hasPlainText = true;                
            }

            if (hasPlainText)
            {
                AlternateView alternate = AlternateView.CreateAlternateViewFromString(htmlMsg, new System.Net.Mime.ContentType(System.Net.Mime.MediaTypeNames.Text.Html));
                message.AlternateViews.Add(alternate);
                message.IsBodyHtml = false;
                message.Body = plainTextMsg;
            }
            else
            {
                message.IsBodyHtml = true;
                message.Body = htmlMsg;
            }            
            client.Send(message);
            LogMsg("Email sent");
        }

        private void LogMsg(string msg)
        {
            logger.LogMsg(msg);
        }

        public static List<VixServer> LoadSitesFromSiteService()
        {
            SiteServiceHelper siteServicehelper = new SiteServiceHelper();
            List<VixServer> vixServers = siteServicehelper.GetVixServers();            
            return vixServers;
        }

        public static void UpdateVixServersFromExportFile(List<VixServer> vixServers)
        {
            string xmlFilename = getXmlExportFilename();
            List<VixServer> xmlServers = TestResultsOutputBuilder.LoadResultsFromXML(xmlFilename);
            if (xmlServers.Count > 0)
            {
                foreach (VixServer server in vixServers)
                {
                    bool serverFound = false;
                    foreach (VixServer xmlServer in xmlServers)
                    {
                        if (server.SiteNumber == xmlServer.SiteNumber)
                        {
                            xmlServer.AddTestResults(server.ServerTests);
                            serverFound = true;
                        }
                    }
                    // if a new vix was added, was not adding it to the xml list 
                    if (!serverFound)
                    {
                        xmlServers.Add(server);
                    }
                }
                TestResultsOutputBuilder.SaveResultsAsXml(xmlFilename, xmlServers);
            }
            else
            {
                TestResultsOutputBuilder.SaveResultsAsXml(xmlFilename, vixServers);
            }
        }

        public static void PruneXmlExportIfNeeded()
        {
            string xmlFilename = getXmlExportFilename();
            FileInfo file = new FileInfo(xmlFilename);
            if (file.Exists)
            {
                //int mb=1024*1024*1024; // too large
                int mb=1024*1024;
                int maxSize = 2 * mb;
                if (ConfigurationManager.AppSettings["MaxOutputFileSize"] != null)
                {
                    maxSize = int.Parse(ConfigurationManager.AppSettings["MaxOutputFileSize"]) * mb;
                }
                if (file.Length >= maxSize)
                {
                    TestResultsOutputBuilder.PruneXmlResults(xmlFilename);
                }
            }

        }

        /*
        public static void MergeVixServersFromExportFile(List<VixServer> vixServers)
        {
            string xmlFilename = ConfigurationManager.AppSettings["XmlExportFilename"];
            List<VixServer> xmlServers = TestResultsOutputBuilder.LoadResultsFromXML(xmlFilename);
            foreach (VixServer server in vixServers)
            {
                foreach (VixServer xmlServer in xmlServers)
                {
                    if (server.SiteNumber == xmlServer.SiteNumber)
                    {
                        server.AddTestResults(xmlServer.ServerTests);
                    }
                }
            }
        }
         */ 

        public static List<VixServer> RemoveSuccessfulVixServers(List<VixServer> vixServers)
        {
            List<VixServer> failedVix = new List<VixServer>();

            foreach (VixServer server in vixServers)
            {
                VixTestResultGroup currentStatus = server.GetMostRecentTest();
                if (currentStatus != null)
                {
                    if (!currentStatus.TestsPassed)
                    {
                        failedVix.Add(server);
                    }
                }
            }


            return failedVix;
        }

        public static List<VixServer> FindStatusChangedVix(List<VixServer> vixServers)
        {
            List<VixServer> statusChangedVix = new List<VixServer>();

            string xmlFilename = getXmlExportFilename();
            List<VixServer> xmlServers = TestResultsOutputBuilder.LoadResultsFromXML(xmlFilename);
            if (xmlServers.Count > 0)
            {
                foreach (VixServer server in vixServers)
                {
                    foreach (VixServer xmlServer in xmlServers)
                    {
                        if (server.SiteNumber == xmlServer.SiteNumber)
                        {
                            VixTestResultGroup currentStatus = server.GetMostRecentTest();
                            VixTestResultGroup previousStatus = xmlServer.GetMostRecentTest();

                            if ((currentStatus == null) && (previousStatus != null))
                            {
                                statusChangedVix.Add(xmlServer);
                            }
                            else if ((previousStatus == null) && (currentStatus != null))
                            {
                                statusChangedVix.Add(server);
                            }
                            else if (currentStatus.TestsPassed != previousStatus.TestsPassed)
                            {
                                statusChangedVix.Add(server);
                            }
                        }
                    }
                }
            }
            else
            {
                statusChangedVix.AddRange(vixServers);
            }
            return statusChangedVix;
        }
    }

    public enum ReportType
    {
        PASSED, FAILED, STATUS_CHANGED
    }
}
